/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int audio_debug;
module_param(audio_debug, int, 0644);
MODULE_PARM_DESC(audio_debug, "enable debug messages [audio]");

static unsigned int audio_alsa_during_irq = 1;
module_param(audio_alsa_during_irq, int, 0644);
MODULE_PARM_DESC(audio_alsa_during_irq, "feed alsa during the irq handler, not via a dpc [audio]");

#define dprintk(level, fmt, arg...)\
	do { if (audio_debug >= level)\
		printk(KERN_DEBUG "%s/0: " fmt, channel->dev->name, ## arg);\
	} while (0)

#define MIXER_RCA_JACKS 1

static int analyze_pcm16(struct vc8x0_dma_channel *channel, u16 *buf, int entries)
{
#if ENABLE_PCM_ANALYSIS
	int i;
#if ENABLE_PCM_FIFO_DEBUG
	u32 *vp;
#endif
	u32 v;
	u16 oldv;
	u8 *b = (u8 *)buf;

#if ENABLE_PCM_FIFO_DEBUG
	if ((channel->aa_buf == 0) || (channel->aa_size != (65536 * sizeof(u32)))) {
		return 0;
	}
#endif

	/* Array of u32's to represent counts of samples */
	for (i = 0; i < entries; i++) {
		v = *(buf + i);
#if ENABLE_PCM_FIFO_DEBUG
		vp = channel->aa_buf + v;
		(*vp)++;
#endif

		if (
			(v == (u32)0xac40) ||
			(v == (u32)0xacc0) ||
			(v == (u32)0xad80) ||
			(v == (u32)0xae00)
		) {

			dprintk(4,
				"%s(%d) %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
			__func__, i - 8,
			*(b + (i*2) - 8),  *(b + (i*2) - 7),  *(b + (i*2) - 6),
			*(b + (i*2) - 5),  *(b + (i*2) - 4),  *(b + (i*2) - 3),
			*(b + (i*2) - 2),  *(b + (i*2) - 1),  *(b + (i*2) - 0),
			*(b + (i*2) + 1),  *(b + (i*2) + 2),  *(b + (i*2) + 3),
			*(b + (i*2) + 4),  *(b + (i*2) + 5),  *(b + (i*2) + 6),
			*(b + (i*2) + 7)
			);

			if (i > 4) {
				*(buf + i) = *(buf + i - 2);
				*(buf + i - 1) = *(buf + i - 3);
			}
			dprintk(4, "a fixup at %d (used 0x%x)\n", i, oldv);

			dprintk(4,
				"%s(%d) %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
			__func__, i - 8,
			*(b + (i*2) - 8),  *(b + (i*2) - 7),  *(b + (i*2) - 6),
			*(b + (i*2) - 5),  *(b + (i*2) - 4),  *(b + (i*2) - 3),
			*(b + (i*2) - 2),  *(b + (i*2) - 1),  *(b + (i*2) - 0),
			*(b + (i*2) + 1),  *(b + (i*2) + 2),  *(b + (i*2) + 3),
			*(b + (i*2) + 4),  *(b + (i*2) + 5),  *(b + (i*2) + 6),
			*(b + (i*2) + 7)
			);

		} else
			oldv = v;
	}
#endif
	return 1;
}

/* Repack 24 bit audio samples (in 32bit alignment)
 * into 16bit samples within the same buffer, and
 * return the new buffer length in bytes.
 *
 * Input Sample:
 * LEFT         RIGHT
 * 00 B0 B1 B2  00 B1 B1 B2

 * Output Sample:
 * LEFT   RIGHT
 * B1 B2  B1 B2
 */
static int repack_24_to_16(u8 *buf, int len)
{
	int i;
	u8 *dst = buf;
	u8 *src = buf + 2;

	if (audio_debug > 2) {
		for (i = 0; i < 0x40; i += 0x10) {
			printk(KERN_ERR
				"%s() %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
			__func__,
			*(buf + i + 0), *(buf + i + 1), *(buf + i + 2),
			*(buf + i + 3), *(buf + i + 4), *(buf + i + 5),
			*(buf + i + 6), *(buf + i + 7), *(buf + i + 8),
			*(buf + i + 9), *(buf + i + 10), *(buf + i + 11),
			*(buf + i + 12), *(buf + i + 13), *(buf + i + 14),
			*(buf + i + 15)
	    );
		}
	}

	/* For each 24 bit sample */
	for (i = 0; i < (len / 4); i++) {
		*(dst) = *(src);
		*(dst + 1) = *(src + 1);
		dst += 2;
		src += 4;
	}

	return ((len / 4) * 2);
}

static void complete_handler_audio(struct vc8x0_dma_buffer *buf);

/* Take the payload held in this buffer and pass it to user space.
 * We don't touch lists, we only take a spinlock on the videobuf
 * list, dequeue a user buffer if necessary.
 */
static void vc8x0_audio_buffer_dequeue(struct vc8x0_dma_buffer *buf)
{
	struct vc8x0_dma_channel *channel = buf->channel;

	dprintk(4, "%s() buffer %p\n", __func__, buf);

	dprintk(4,
		"%s() %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
		__func__, *(buf->cpu + 0), *(buf->cpu + 1), *(buf->cpu + 2),
		*(buf->cpu + 3), *(buf->cpu + 4), *(buf->cpu + 5),
		*(buf->cpu + 6), *(buf->cpu + 7), *(buf->cpu + 8),
		*(buf->cpu + 9), *(buf->cpu + 10), *(buf->cpu + 11),
		*(buf->cpu + 12), *(buf->cpu + 13), *(buf->cpu + 14),
		*(buf->cpu + 15)
	    );

	complete_handler_audio(buf);

	memset(buf->cpu, 0x00, buf->cpu_size);
}

void vc8x0_audio_work_handler(struct work_struct *work)
{
	struct vc8x0_dev *dev;
	struct vc8x0_dma_buffer *buf;
	struct list_head *p, *q;
	unsigned long flags;

	struct vc8x0_dma_channel *channel =
	    container_of(work, struct vc8x0_dma_channel, workhandler);

	if ((channel == 0) || (channel->dev == 0)) {
		printk(KERN_ERR "%s() bug\n", __func__);
		return;
	}

	dev = channel->dev;

	dprintk(3, "%s()\n", __func__);

	/* The interrupt handler shuffles buffers from the busy list to the dpc list.
	 * The interrupt handler pulls empty buffers from the free list and places them
	 * on the hardware.
	 */

	/* Move any dpc buffers to the full list, we're doing this
	 * to keep the interrupt spinlock usage on dpc to a minimum.
	 * we'll then take a more leasurly dequeue approach when
	 * processing the full list.
	 */
	/* For each dpc buffer move it to the full list in the minimum amount of time */
	spin_lock_irqsave(&channel->dma_buffers_dpc_lock, flags);
	spin_lock(&channel->dma_buffers_full_lock);
	list_for_each_safe(p, q, &channel->dma_buffers_dpc) {

		buf = list_entry(p, struct vc8x0_dma_buffer, list);
		dprintk(2, "%s() dequeueing DPC buffer %p\n", __func__, buf);

		/* Last, put the buffer on the full list for leasurely processing */
		buf->state = STATE_FULL;
		list_move_tail(&buf->list, &channel->dma_buffers_full);

	}
	spin_unlock(&channel->dma_buffers_full_lock);
	spin_unlock_irqrestore(&channel->dma_buffers_dpc_lock, flags);

	/* BAM! The interrupt handler is now free to move on */
	/* BAM! The interrupt handler is now free to move on */
	/* BAM! The interrupt handler is now free to move on */
	/* BAM! The interrupt handler is now free to move on */

	/* Now let's dequeue the full buffers */
	/* For each full buffer, send it to user space */
	spin_lock(&channel->dma_buffers_full_lock);
	list_for_each_safe(p, q, &channel->dma_buffers_full) {

		buf = list_entry(p, struct vc8x0_dma_buffer, list);

		vc8x0_audio_buffer_dequeue(buf);

		/* Put the dequeue'd buffer back into the busy list */
		spin_lock_irqsave(&channel->dma_buffers_busy_lock, flags);
		buf->state = STATE_BUSY;
		list_move_tail(&buf->list, &channel->dma_buffers_busy);
		spin_unlock_irqrestore(&channel->dma_buffers_busy_lock, flags);

	}
	spin_unlock(&channel->dma_buffers_full_lock);
}

int vc8x0_audio_irqhandler(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_dma_buffer *buf;
	int handled = 0;

	u32 lastidx = 0, aud_pciaddr = 0;
	u32 buflen, reg;
	unsigned long flags;

	/* Process the interrupt */
	channel->buffers_processed++;
	buflen = MAX_USER_AUDIO_BUFFER_SIZE;

	/* Find the finished frame and locate it's pci address */
	lastidx = vc_read32(channel->regs.m_nLastFrameIndex);
	if ((lastidx < 1) || (lastidx > MAX_USER_VIDEO_BUFFERS)) {
		printk(KERN_ERR "%s() lastidx out of range, critical (%x).\n", __func__, lastidx);
		goto badaddr;
	}

	if (channel->lastidx == lastidx) {	
		/* We've already processed this frame, skip it,
		 * duplicate interrupt.
		 */
		goto out;
	}
	channel->lastidx = lastidx;

	if (lastidx == 9)
		lastidx = 1;
	else
		lastidx++;

	lastidx--;

	reg = channel->regs.m_nSP1DmaAddrL + (lastidx * 2 * sizeof(u32));
	aud_pciaddr = vc_read32(reg);

	dprintk(3, "%s() buf completed, reg=%x, physical 0x%x, lastidx=%d\n",
		__func__, reg, aud_pciaddr, lastidx);

	/* Find the buffer by address */
	buf = vc8x0_buffer_busy_get_by_address(channel, aud_pciaddr, TYPE_AUDIO);
	if (buf == 0) {
		printk(KERN_ERR "%s() No busy dma buffer for address 0x%x\n",
		       __func__, aud_pciaddr);
		channel->buffers_bad_address++;
		goto badaddr;
	}
	buf->used_len = buflen;
	buf->used_len = (0x1f40 * 8);
	buf->used_len = 19200;
	buf->used_len = 28800;
	buf->used_len = 3840; /* 480 samples per sec * 2 channels * 4 bytes per channel */
	buf->used_len = repack_24_to_16(buf->cpu, buf->used_len);
	dprintk(3, "%s() buf completed %p nr=%d used_len = %d\n",
		__func__, buf, buf->nr, buf->used_len);

	if (IS_845E(dev)) {
		/* Analyze the audio buffer and remove any spurious fifo tags */
		analyze_pcm16(channel, (u16 *)buf->cpu, buf->used_len / 2);
	}

	/* Put the buffer on the DPC list. This list is spinlock held very briefly
	 * so we're not going to be held up.
	 * The only person that holds the DPC spinlock is the deferred work queue.
	 * The only person that holds the busy spinlock is the interrupt handler.
	 * These spinlocks are also held briefly for reporting purposes in other
	 * parts of the driver but they are very timely.
	 */
	spin_lock_irqsave(&channel->dma_buffers_busy_lock, flags);

	/* Last, put the buffer on the DPC list for our deferred worker to process */
	spin_lock_irqsave(&channel->dma_buffers_dpc_lock, flags);
	buf->state = STATE_DPC;
	list_move_tail(&buf->list, &channel->dma_buffers_dpc);
	spin_unlock_irqrestore(&channel->dma_buffers_dpc_lock, flags);

	spin_unlock_irqrestore(&channel->dma_buffers_busy_lock, flags);

	handled++;
badaddr:
	if (handled) {
		if (audio_alsa_during_irq == 1) {
			/* Process audio now, zero latency */
			vc8x0_audio_work_handler(&channel->workhandler);
		} else {
			/* Process audio via a deferred queue, with latencies up to 1.999 ms or longer */
			schedule_work(&channel->workhandler);
		}
	}
out:
	return handled;
}

static struct snd_pcm_hardware snd_vc8x0_hw_capture = {
	.info = SNDRV_PCM_INFO_BLOCK_TRANSFER |
	    SNDRV_PCM_INFO_MMAP |
	    SNDRV_PCM_INFO_INTERLEAVED | SNDRV_PCM_INFO_MMAP_VALID,

	.formats = SNDRV_PCM_FMTBIT_S16_LE,
	.rates = SNDRV_PCM_RATE_48000,
	.rate_min = 48000,
	.rate_max = 48000,
	.channels_min = 2,
	.channels_max = 2,
	.buffer_bytes_max = 28800 * 2,
	.period_bytes_min = 64,
	.period_bytes_max = 0x1f40 * 20,
	.periods_min = 2,
	.periods_max = 98
};

void vc8x0_audio_timeout(unsigned long data)
{
	struct vc8x0_dma_channel *channel = (struct vc8x0_dma_channel *)data;

	dprintk(1, "%s()\n", __func__);
}

static int snd_vc8x0_capture_open(struct snd_pcm_substream *substream)
{
	struct vc8x0_audio_dev *chip;
	struct vc8x0_dma_channel *channel;
	struct snd_pcm_runtime *runtime;
	struct vc8x0_dev *dev;

	chip = snd_pcm_substream_chip(substream);

	if (!chip) {
		printk(KERN_ERR "%s() No chip\n", __func__);
		return -ENODEV;
	}
	channel = chip->channel;

	if (!channel) {
		printk(KERN_ERR "%s() No channel\n", __func__);
		return -ENODEV;
	}
	dev = channel->dev;

	runtime = substream->runtime;

	chip->users++;
	chip->card_close = 0;
	chip->capture_pcm_substream = substream;
	runtime->private_data = chip;

	runtime->hw = snd_vc8x0_hw_capture;
	snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS);
	return 0;
}

static int snd_vc8x0_pcm_close(struct snd_pcm_substream *substream)
{
	struct vc8x0_audio_dev *chip = snd_pcm_substream_chip(substream);

	chip->users--;

	if (chip->users == 0)
		chip->card_close = 1;

	if (substream->runtime->dma_area) {
		vfree(substream->runtime->dma_area);
		substream->runtime->dma_area = 0;
	}

	return 0;
}

static int snd_vc8x0_hw_capture_params(struct snd_pcm_substream *substream,
					struct snd_pcm_hw_params *hw_params)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned int size;

	size = params_buffer_bytes(hw_params);
	if (runtime->dma_area) {
		if (runtime->dma_bytes > size)
			return 0;
		vfree(runtime->dma_area);
	}
	runtime->dma_area = vmalloc(size);
	if (!runtime->dma_area)
		return -ENOMEM;
	else
		runtime->dma_bytes = size;
	return 0;
}

static int snd_vc8x0_hw_capture_free(struct snd_pcm_substream *substream)
{
	struct vc8x0_audio_dev *chip = snd_pcm_substream_chip(substream);
	struct vc8x0_dma_channel *channel = chip->channel;
	dprintk(1, "%s() rate = %d\n", __func__, substream->runtime->rate);
	return 0;
}

static int snd_vc8x0_prepare(struct snd_pcm_substream *substream)
{
	struct vc8x0_audio_dev *chip = snd_pcm_substream_chip(substream);
	struct vc8x0_dma_channel *channel = chip->channel;
	dprintk(1, "%s() rate = %d\n", __func__, substream->runtime->rate);

	return 0;
}

static inline void handle_audio_data(struct vc8x0_dma_buffer *buf,
	int *period_elapsed)
{
	struct vc8x0_dma_channel *channel = buf->channel;
	struct vc8x0_audio_dev *chip = channel->audio_dev;
	struct snd_pcm_runtime *runtime = chip->capture_pcm_substream->runtime;
	int stride;
	int len, rdb, cpsafe[3];
	unsigned char *cp;
	unsigned int oldptr;

	stride = runtime->frame_bits >> 3;
	if (stride == 0) {
		printk(KERN_ERR "%s() divbyzero BUG\n", __func__);
		stride = 4;
	}

	len = buf->used_len / stride;
	cp = buf->cpu;
	oldptr = chip->rcv_position;
	rdb = runtime->dma_bytes;

	*period_elapsed = 0;

	dprintk(2, "%s() used_len = 0x%x len = 0x%x stride = 0x%x\n", __func__,
		buf->used_len, len, stride);

	/* do the copy */
	if ((oldptr + len) >= runtime->buffer_size) {
		unsigned int cnt = runtime->buffer_size - oldptr;
		dprintk(2,
			"len=0x%x cnt=%d stride=%d oldptr=%d rbs=%lu rdb=%d\n",
			len, cnt, stride, oldptr, runtime->buffer_size,
			runtime->dma_bytes);

		/* safety, extensive bounds checking on memcpy's */
		cpsafe[0] = 1;
		if ((rdb == 0) || (runtime->dma_area == 0)) {
			printk(KERN_ERR "%s() critical0\n", __func__);
			cpsafe[0] = 0;
		}

		if ((cnt * stride) > buf->cpu_size) {
			printk(KERN_ERR "%s() critical1\n", __func__);
			cpsafe[0] = 0;
		}
		if (((cnt * stride) + (oldptr * stride)) > rdb) {
			printk(KERN_ERR "%s() critical2 %d > %d\n",
				__func__,
				(cnt * stride) + (oldptr * stride),
				rdb
				);
			cpsafe[0] = 0;
		}

		if (cpsafe[0]) {
			memcpy(runtime->dma_area + (oldptr * stride),
				cp, cnt * stride);
		}

		cpsafe[1] = 1;
		if ((((len * stride) - (cnt * stride)) + (cnt * stride)) >
			buf->cpu_size) {
			printk(KERN_ERR "%s() critical3\n", __func__);
			cpsafe[1] = 0;
		}

		if (((buf->used_len) - (cnt * stride)) > rdb) {
			printk(KERN_ERR "%s() critical4 %d > %d\n",
				__func__,
				(len * stride) - (cnt * stride), rdb);
			cpsafe[1] = 0;
		}

		if (cpsafe[1]) {
			memcpy(runtime->dma_area, (cp + (cnt * stride)),
				((len * stride) - (cnt * stride)));
		}
	} else {
		cpsafe[0] = 1;
		if ((rdb == 0) || (runtime->dma_area == 0)) {
			printk(KERN_ERR "%s() critical5\n", __func__);
			cpsafe[0] = 0;
		}
		if ((len * stride) > buf->cpu_size) {
			printk(KERN_ERR "%s() critical6 %d > %d\n",
				__func__,
				(len * stride), buf->cpu_size);
			cpsafe[0] = 0;
		}
		if (((len * stride) + (oldptr * stride)) > rdb) {
			printk(KERN_ERR "%s() critical7 %d > %d\n",
				__func__,
				(len * stride) + (oldptr * stride),
				buf->cpu_size);
			cpsafe[0] = 0;
		}
		
		if (cpsafe[0]) {
			memcpy(runtime->dma_area + (oldptr * stride),
				cp, len * stride);
		}
	}

	/* update the status */
	snd_pcm_stream_lock(chip->capture_pcm_substream);
	chip->rcv_position += len;
	if (chip->rcv_position >= runtime->buffer_size)
		chip->rcv_position -= runtime->buffer_size;

	chip->copied_position += (len);
	if (chip->copied_position >= runtime->period_size) {
		chip->copied_position -= runtime->period_size;
		*period_elapsed = 1;
	}
	snd_pcm_stream_unlock(chip->capture_pcm_substream);
}

static void complete_handler_audio(struct vc8x0_dma_buffer *buf)
{
	struct vc8x0_dma_channel *channel = buf->channel;
	struct vc8x0_audio_dev *chip = channel->audio_dev;

	struct snd_pcm_substream *substream = chip->capture_pcm_substream;
	int period_elapsed = 0;

	if ((chip->card_close == 1) || (chip->capture_stream != STREAM_ON))
		return;

	if (substream) {
		handle_audio_data(buf, &period_elapsed);
		if (period_elapsed) {
			dprintk(3, "%s() pulsing audio data\n", __func__);
			snd_pcm_period_elapsed(substream);
		}
	}

	return;
}

static int snd_vc8x0_capture_trigger(struct snd_pcm_substream *substream,
				      int cmd)
{
	struct vc8x0_audio_dev *chip = snd_pcm_substream_chip(substream);
	struct vc8x0_dma_channel *channel = chip->channel;

	dprintk(1, "%s() cmd %d, audio stat : %d\n", __func__,
		cmd, chip->capture_stream);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_RESUME:
	case SNDRV_PCM_TRIGGER_START:
		if (chip->capture_stream == STREAM_ON)
			return 0;

		chip->rcv_position = chip->copied_position = 0;
		chip->capture_stream = STREAM_ON;
		return 0;

	case SNDRV_PCM_TRIGGER_SUSPEND:
		chip->capture_stream = STREAM_SUSPEND;
		return 0;
	case SNDRV_PCM_TRIGGER_STOP:
		chip->capture_stream = STREAM_OFF;
		return 0;
	default:
		return -EINVAL;
	}
}

static snd_pcm_uframes_t snd_vc8x0_capture_pointer(struct snd_pcm_substream
						    *substream)
{
	struct vc8x0_audio_dev *chip = snd_pcm_substream_chip(substream);
	return chip->rcv_position;
}

static struct page *snd_pcm_pd_get_page(struct snd_pcm_substream *subs,
					unsigned long offset)
{
	void *pageptr = subs->runtime->dma_area + offset;
	return vmalloc_to_page(pageptr);
}

static struct snd_pcm_ops pcm_capture_ops = {
	.open = snd_vc8x0_capture_open,
	.close = snd_vc8x0_pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = snd_vc8x0_hw_capture_params,
	.hw_free = snd_vc8x0_hw_capture_free,
	.prepare = snd_vc8x0_prepare,
	.trigger = snd_vc8x0_capture_trigger,
	.pointer = snd_vc8x0_capture_pointer,
	.page = snd_pcm_pd_get_page,
};

#if ENABLE_ALSA_MIXER
int vc8x0_audio_mixer_volume_get(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ev)
{
	struct vc8x0_audio_dev *chip = snd_kcontrol_chip(kcontrol);
	struct vc8x0_dma_channel *channel = chip->channel;
	int v = 0;

	ev->value.integer.value[0] = v;

	dprintk(1, "%s() = %d\n", __func__, v);

	return 0;
}

int vc8x0_audio_mixer_volume_set(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ev)
{
	struct vc8x0_audio_dev *chip = snd_kcontrol_chip(kcontrol);
	struct vc8x0_dma_channel *channel = chip->channel;

	dprintk(1, "%s(%ld)\n", __func__, ev->value.integer.value[0]);

	return 0;
}

int vc8x0_audio_mixer_volume_info(struct snd_kcontrol *kcontrol,
				   struct snd_ctl_elem_info *ei)
{
	struct vc8x0_audio_dev *chip = snd_kcontrol_chip(kcontrol);
	struct vc8x0_dma_channel *channel = chip->channel;

	dprintk(1, "%s()\n", __func__);

	/* WM8739 supports 32 discrete volume levels from -34 to +12db */
	ei->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	ei->count = 1;
	ei->value.integer.min = -34;
	ei->value.integer.max = 12;

	return 0;
}

static DECLARE_TLV_DB_SCALE(vc8x0_mixer_tv_vol_db_scale, -9600, 100, 0);

#define vc8x0_VOLUME(xname, xindex, addr) \
{	.iface = SNDRV_CTL_ELEM_IFACE_MIXER, \
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE | SNDRV_CTL_ELEM_ACCESS_TLV_READ, \
	.name = xname, \
	.index = xindex, \
	.info = vc8x0_audio_mixer_volume_info, \
	.get = vc8x0_audio_mixer_volume_get, \
	.put = vc8x0_audio_mixer_volume_set, \
	.private_value = addr, \
	.tlv.p = vc8x0_mixer_tv_vol_db_scale, \
}

static struct snd_kcontrol_new vc8x0_audio_mixer_controls[] = {
	vc8x0_VOLUME("Component", 0, MIXER_RCA_JACKS),
};

static int snd_card_new_mixer(struct vc8x0_audio_dev *chip)
{
	struct snd_card *card = chip->card;
	int i, r;

	for (i = 0; i < ARRAY_SIZE(vc8x0_audio_mixer_controls); i++) {
		r = snd_ctl_add(card,
				snd_ctl_new1(&vc8x0_audio_mixer_controls[i],
					     chip));
		if (r < 0)
			return r;
	}

	return 0;
}
#endif

void vc8x0_audio_unregister(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_audio_dev *chip = channel->audio_dev;

	dprintk(1, "%s()\n", __func__);
	dprintk(0, "Unregistered ALSA audio device %p\n", chip);

	if (!chip) {
		printk(KERN_ERR "%s() no chip!\n", __func__);
		return;
	}

	snd_card_free(chip->card);
}

/*
 * create a PCM device
 */
static int snd_vc8x0_pcm(struct vc8x0_audio_dev *chip, int device, char *name)
{
	int err;
	struct snd_pcm *pcm;
#if ENABLE_ALSA_MIXER
	err = snd_card_new_mixer(chip);
	if (err < 0)
		return err;
#endif
	err = snd_pcm_new(chip->card, name, device, 0, 1, &pcm);
	if (err < 0)
		return err;

	pcm->private_data = chip;

	strcpy(pcm->name, name);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &pcm_capture_ops);

	return 0;
}

int vc8x0_audio_register(struct vc8x0_dma_channel *channel)
{
	struct snd_card *card;
	struct vc8x0_audio_dev *chip;
	struct vc8x0_dev *dev = channel->dev;
	int err;

	err = snd_card_create(SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1,
			      THIS_MODULE, sizeof(struct vc8x0_audio_dev),
			      &card);
	if (err < 0)
		goto error;

	chip = (struct vc8x0_audio_dev *)card->private_data;
	chip->card = card;
	chip->channel = channel;

	err = snd_vc8x0_pcm(chip, 0, "VC8X0 Digital");
	if (err < 0)
		goto error;

	strcpy(card->driver, "VC8X0");
	sprintf(card->shortname, "ViewCast VC8X0");
	sprintf(card->longname, "%s at %s", card->shortname, channel->dev->name);
	strcpy(card->mixername, "VC8X0");

	err = snd_card_register(card);
	if (err < 0)
		goto error;

	channel->audio_dev = chip;
	snd_card_set_dev(card, &dev->pci->dev);

	dprintk(0, "Registered ALSA audio device %p card %p\n",
		channel->audio_dev, channel->audio_dev->card);
	return 0;

error:
	snd_card_free(card);
	printk(KERN_ERR "%s(): Failed to register analog "
	       "audio adapter\n", __func__);

	return 0;
}
