/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
VC8X0 Buffering Model:

The 800 family will be capable of supporting up to 4 concurrent video
channels each with an embedded audio channel. It can also support an
additional 4 discrete audio channels with an off-board daughter card.

DMA buffers per channel must be contigious, reside only in 32bit
memory.

The PCIe bridge (GN4124) supports up to 18 'fifos', essentially
discrete DMA channels. The GN4124 uses a DMA Sequencer architecture
to control which dma buffers are targets for which channel. The sequencer
is a list of program instructions that effictivel handle the data modement
within the FPGA / PCIe bridge. The DMA sequence and fifo control is
configured completely using BAR0.

The sequence program is loaded incrementally into address BAR0+DPRAM, byte
by byte.

The hardware is fixed to operate with exactly 9 DMA buffers
for each video channel, and an additional 9 DMA buffers for each
audio channel. Additionally, each channel as a single DMA buffer
that is used for timestamp usage (primarily designed for the
windows directshow streaming model).

Video buffers delivered from the FPGA are in the following format:
	Y Plane VBI data	for field0
	Y Plane Active video	for field0
	U Plane VBI data	for field0
	U Plane Active video	for field0
	V Plane VBI data	for field0
	V Plane Active video	for field0
	Y Plane VBI data	for field1
	Y Plane Active video	for field1
	U Plane VBI data	for field1
	U Plane Active video	for field1
	V Plane VBI data	for field1
	V Plane Active video	for field1

Audio buffers delivered from the FPGA are in the following format:
	Stereo PCM L/R data	audio input 1
	Stereo PCM L/R data	audio input 2
	Stereo PCM L/R data	audio input 3
	Stereo PCM L/R data	audio input 4

Timestamp buffers delivered from the FPGA are in the following format:
	Not yet documented but is statically allocated
	as 1024 * 16 bytes in LoadDMAResources();

The largest resolution supported by the driver is 1920x1080p.

The 820e, via LoadDMAResources(), the driver will support:

VideoA (dma_channel[0])
	MAX_VIDEO_DSP_FRAMES = 9
	9 x DMA buffers in 32bit contigious memory only.
	each buffer (1920 * 1250 * 2) + 4096;
VideoB (dma_channel[1])
	9 x DMA buffers in 32bit contigious memory only.
	each buffer (1920 * 1250 * 2) + 4096;
AudioA (dma_channel[2])
	9 x DMA buffers in 32bit contigious memory only.
	4 = nChannelCount
	each buffer ((1024 * 16) * 4) + 4096;

*/
 
#include "vc8x0.h"

static unsigned int buffer_debug;
module_param(buffer_debug, int, 0644);
MODULE_PARM_DESC(buffer_debug, "enable debug messages [buffer]");

#define dprintk(level, fmt, arg...)\
	do { if (buffer_debug >= level)\
		printk(KERN_ERR "%s/0: " fmt, dev->name, ## arg);\
	} while (0)

void vc8x0_buffer_recycle(struct vc8x0_dma_buffer *buf)
{
	memset_io(buf->cpu, 0x0, buf->cpu_size);
}

void vc8x0_buffer_analyze(u8 *buf, int len)
{
	int i;
	u32 data[256];
	memset(data, 0, sizeof(data));

	for (i = 0; i < len; i++) {
		data[ *(buf + i) ]++;
	}

	for (i = 0; i < 256; i++) {
		if (data[i]) {
			printk(KERN_ERR "%02x %x\n", i, data[i]);
		}
	}
}

void vc8x0_buffer_display(struct vc8x0_dma_buffer *buf)
{
	struct vc8x0_dev *dev = buf->channel->dev;

	dprintk(1, "%s(%p)        cpu->virtual = %p, physical = 0x%x, type = %s, nr = %d\n",
		__func__, buf, buf->cpu, (u32)buf->cpu_dma,
		buf->type == TYPE_VIDEO ? "VIDEO" :
		buf->type == TYPE_AUDIO ? "AUDIO" :
		buf->type == TYPE_TIMESTAMP ? "TIMESTAMP" :
		"UNKNOWN",
		buf->nr
		);
}

/* Install this buffer into the running hardware.
 * The buffer will be placed into the busy list as a result
 * of this call.
 * Locks on lists are assumed to be held!!!!!!!!!
 */
void vc8x0_buffer_activate(struct vc8x0_dma_buffer *buf)
{
	struct vc8x0_dma_channel *channel;
	struct vc8x0_dev *dev;

	if ((buf == 0) || (buf->channel == 0) || (buf->channel->dev == 0)) {
		printk(KERN_ERR "%s() missing critical parameter, safety first\n", __func__);
		return;
	}

	channel = buf->channel;
	dev = channel->dev;
	buf->last_used = jiffies_to_msecs(jiffies);

	if ((buf->type == TYPE_VIDEO) || (buf->type == TYPE_AUDIO)) {
		buf->state = STATE_BUSY;
		list_add_tail(&buf->list, &channel->dma_buffers_busy);
		dprintk(1, "%s() Placed buffer %p type %d on hardware\n", __func__, buf, buf->type);
	} else
	if (buf->type == TYPE_TIMESTAMP) {
		buf->state = STATE_BUSY;
		list_add_tail(&buf->list, &channel->dma_timestamp_buffers_busy);
		dprintk(1, "%s() Placed TIMESTAMP buffer %p on hardware\n", __func__, buf);
	} else {
		printk(KERN_ERR "%s() avoid BUG\n", __func__);
	}

}

/* VC8X0 can occasionally return a pad buffer address on startup.
 * a helper function. Return busy buffer or 0 if not found.
 */
struct vc8x0_dma_buffer *vc8x0_buffer_busy_get_by_address(
	struct vc8x0_dma_channel *channel, u32 physaddr, enum buffer_payload_t type)
{
	struct vc8x0_dma_buffer *ret = 0, *buf;
	struct list_head *p, *q;
	unsigned long flags;

	if ((type == TYPE_VIDEO) || (type == TYPE_AUDIO)) {
		spin_lock_irqsave(&channel->dma_buffers_busy_lock, flags);
		list_for_each_safe(p, q, &channel->dma_buffers_busy) {
			buf = list_entry(p, struct vc8x0_dma_buffer, list);
			if (buf->cpu_dma == physaddr) {
				ret = buf;
				break;
			}
			buf = 0;
		}
		spin_unlock_irqrestore(&channel->dma_buffers_busy_lock, flags);

	} else
	if (type == TYPE_TIMESTAMP) {
		spin_lock_irqsave(&channel->dma_timestamp_buffers_busy_lock, flags);
		list_for_each_safe(p, q, &channel->dma_timestamp_buffers_busy) {
			buf = list_entry(p, struct vc8x0_dma_buffer, list);
			if (buf->cpu_dma == (physaddr & 0xfffff000)) {
				ret = buf;
				break;
			}
			buf = 0;
		}
		spin_unlock_irqrestore(&channel->dma_timestamp_buffers_busy_lock, flags);
		if (ret == 0) {
			/* Look on the free list */
			spin_lock_irqsave(&channel->dma_timestamp_buffers_free_lock, flags);
			list_for_each_safe(p, q, &channel->dma_timestamp_buffers_free) {
				buf = list_entry(p, struct vc8x0_dma_buffer, list);
				if (buf->cpu_dma == (physaddr & 0xfffff000)) {
					ret = buf;
					break;
				}
				buf = 0;
			}
			spin_unlock_irqrestore(&channel->dma_timestamp_buffers_free_lock, flags);
		}
	} else {
		printk(KERN_ERR "%s() BUG\n", __func__);
	}

	return ret;
}

struct vc8x0_dma_buffer *vc8x0_buffer_alloc(struct vc8x0_dma_channel *channel,
	enum buffer_payload_t type, u32 nr)
{
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_dma_buffer *buf;
	u32 cpu_size = 0;
	u8 *virt;

	dprintk(1, "%s(channel%d)\n", __func__, channel->nr);

	if (type == TYPE_VIDEO) {
		dprintk(1, "%s() MAX_USER_VIDEO_BUFFER_SIZE = 0x%x\n", __func__, MAX_USER_VIDEO_BUFFER_SIZE);
		cpu_size = MAX_USER_VIDEO_BUFFER_SIZE;
	} else
	if (type == TYPE_AUDIO) {
		dprintk(1, "%s() MAX_USER_AUDIO_BUFFER_SIZE = 0x%x\n", __func__, MAX_USER_AUDIO_BUFFER_SIZE);
		cpu_size = MAX_USER_AUDIO_BUFFER_SIZE;
	} else
	if (type == TYPE_TIMESTAMP) {
		dprintk(1, "%s() MAX_USER_TIMESTAMP_BUFFER_SIZE = 0x%x\n", __func__, MAX_USER_TIMESTAMP_BUFFER_SIZE);
		cpu_size = MAX_USER_TIMESTAMP_BUFFER_SIZE;
	} else {
		BUG();
	}

	buf = kzalloc(sizeof(*buf), GFP_KERNEL);
        if (buf == 0) {
		printk(KERN_ERR "%s() kzalloc -ENOMEM\n", __func__);
		goto err1;
	}

	buf->nr = nr;
	buf->channel = channel;
	buf->cpu_size = cpu_size;
	buf->state = STATE_FREE;
	buf->type = type;

	/* Allocate contiguous memory for the video/vbi/audio data */
	buf->cpu = pci_alloc_consistent(dev->pci, buf->cpu_size, &buf->cpu_dma);
	if (!buf->cpu) {
		printk(KERN_ERR "%s() pci_alloc_consistent -ENOMEM (%d bytes)\n", __func__, buf->cpu_size);
		goto err2;
	}

	/* Erase the buffer */
	vc8x0_buffer_recycle(buf);

	/* Create the address mappings list */
	virt = buf->cpu;
		
	/* Show some useful information on the sg configuration */
	vc8x0_buffer_display(buf);

	return buf;

err2:
	kfree(buf);
err1:
	return 0;
}

void vc8x0_buffer_dealloc(struct vc8x0_dma_buffer *buf)
{
	struct vc8x0_dev *dev;
	struct vc8x0_dma_channel *channel;

	if (buf == 0)
		BUG();
	if (buf->channel == 0)
		BUG();
	if (buf->channel->dev == 0)
		BUG();

	dprintk(1, "%s() Deallocating buffer %p nr = %d\n", __func__, buf, buf->nr);

	channel = buf->channel;
	dev = channel->dev;

	vc8x0_buffer_display(buf);

	if (buf->cpu) {
		pci_free_consistent(dev->pci, buf->cpu_size, buf->cpu, buf->cpu_dma);
		buf->cpu_size = 0;
		buf->cpu = 0;
		buf->cpu_dma = 0;
	}

	memset(buf, 0, sizeof(struct vc8x0_dma_buffer));
	kfree(buf);
}

