/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * 12 channels in total.
 *
 * Channels 01    DVI Video A       (VD_1) EVENT_ID_0
 * Channels 02    DVI Video B       (VD_2) EVENT_ID_2
 * Channels 03    DVI Video C       (VD_3) EVENT_ID_4
 * Channels 04    DVI Video D       (VD_4) EVENT_ID_6
 * Channels 05    DVI Audio A       (AD_1) EVENT_ID_1
 * Channels 06    DVI Audio B       (AD_2) EVENT_ID_3
 * Channels 07    DVI Audio C       (AD_3) EVENT_ID_5
 * Channels 08    DVI Audio D       (AD_4) EVENT_ID_7
 * Channels 09    Unbalanced Audio  (AD_5) EVENT_ID_11
 * Channels 10    Balanced Audio    (AD_6) EVENT_ID_12
 * Channels 11    AES Audio 0       (AD_7) EVENT_ID_13
 * Channels 12    AES Audio 1       (AD_8) EVENT_ID_14
 *
 */
#include "vc8x0.h"

static unsigned int channel_debug;
module_param(channel_debug, int, 0644);
MODULE_PARM_DESC(channel_debug, "enable debug messages [channel]");

#define dprintk(level, fmt, arg...)\
	do { if (channel_debug >= level)\
		printk(KERN_ERR "%s: " fmt, dev->name, ## arg);\
	} while (0)

int vc8x0_channel_register(struct vc8x0_dev *dev, struct vc8x0_dma_channel *channel, int nr)
{
	struct vc8x0_dma_buffer *buf;
	int i, ret = -ENODEV;
	unsigned long flags;

	dprintk(1, "%s(dev = %p) nr = %d\n", __func__, dev, nr);
	memset(channel, 0, sizeof(struct vc8x0_dma_channel));

	mutex_init(&channel->lock);
	channel->nr = nr;
	channel->dev = dev;
	channel->state = STATE_STOPPED;

	if ((nr >= 1) && (nr <= 4)) {
		channel->type = TYPE_VIDEO;
		channel->max_buffers = MAX_USER_VIDEO_BUFFERS;
	} else
	if ((nr >= 5) && (nr <= 16)) {
		channel->type = TYPE_AUDIO;
		channel->max_buffers = MAX_USER_AUDIO_BUFFERS;
	} else
	if (nr == 17) {
		channel->type = TYPE_TIMESTAMP;
		channel->max_buffers = MAX_USER_TIMESTAMP_BUFFERS;
	} else
		BUG();
	
	/* External Master Bus - Video */
	mutex_init(&channel->i2c_bus.lock);
	channel->i2c_bus.nr = nr;
	channel->i2c_bus.dev = dev;

	/* SPI Controller per channel */
	mutex_init(&channel->spi_bus.lock);
	channel->spi_bus.nr = nr;
	channel->spi_bus.dev = dev;

	if (channel->type == TYPE_VIDEO) {
		channel->i2c_bus.reg_base = (nr * 0x100);
		channel->spi_bus.reg_base = (nr * 0x100);

		/* V4L2 */
		channel->timeout.function = vc8x0_video_timeout;
		channel->timeout.data = (unsigned long)channel;
		init_timer(&channel->timeout);
		INIT_LIST_HEAD(&channel->v4l2_capture);
		spin_lock_init(&channel->v4l2_capture_lock);
		INIT_WORK(&channel->workhandler, vc8x0_video_work_handler);

		/* VBI */
		channel->vbi_timeout.function = vc8x0_vbi_timeout;
		channel->vbi_timeout.data = (unsigned long)channel;
		init_timer(&channel->vbi_timeout);
		INIT_LIST_HEAD(&channel->vbi_capture);
		spin_lock_init(&channel->vbi_capture_lock);

	} else
	if (channel->type == TYPE_AUDIO) {
		channel->i2c_bus.reg_base = 0x900 + ((nr - 9) * 0x80);
		INIT_WORK(&channel->workhandler, vc8x0_audio_work_handler);

		channel->timeout.function = vc8x0_audio_timeout;
		channel->timeout.data = (unsigned long)channel;
		init_timer(&channel->timeout);
#if ENABLE_PCM_FIFO_DEBUG
		channel->aa_size = 65536 * sizeof(u32);
		channel->aa_buf = kzalloc(channel->aa_size, GFP_KERNEL);
#endif
	} else
		/* DO I really mean this????? TODO BUG */
		channel->i2c_bus.reg_base = (nr * 0x100);

	/* DMA VIDEO buffer managerment */
	INIT_LIST_HEAD(&channel->dma_buffers_free);
	INIT_LIST_HEAD(&channel->dma_buffers_busy);
	INIT_LIST_HEAD(&channel->dma_buffers_dpc);
	INIT_LIST_HEAD(&channel->dma_buffers_full);
	spin_lock_init(&channel->dma_buffers_free_lock);
	spin_lock_init(&channel->dma_buffers_busy_lock);
	spin_lock_init(&channel->dma_buffers_dpc_lock);
	spin_lock_init(&channel->dma_buffers_full_lock);

	/* DMA TIMESTAMP buffer managerment */
	INIT_LIST_HEAD(&channel->dma_timestamp_buffers_free);
	INIT_LIST_HEAD(&channel->dma_timestamp_buffers_busy);
	INIT_LIST_HEAD(&channel->dma_timestamp_buffers_dpc);
	INIT_LIST_HEAD(&channel->dma_timestamp_buffers_full);
	spin_lock_init(&channel->dma_timestamp_buffers_free_lock);
	spin_lock_init(&channel->dma_timestamp_buffers_busy_lock);
	spin_lock_init(&channel->dma_timestamp_buffers_dpc_lock);
	spin_lock_init(&channel->dma_timestamp_buffers_full_lock);

	if (channel->type == TYPE_VIDEO) {
		/* Debugging OSD */
		vc8x0_display_init(channel, &channel->display_ctx);
	}

	/* Now we allocate any video hardware buffers */
	for (i = 0; i < channel->max_buffers; i++) {
		buf = vc8x0_buffer_alloc(channel, channel->type, i);
		if (buf) {
			spin_lock_irqsave(&channel->dma_buffers_free_lock, flags);

			buf->state = STATE_FREE;
			list_add_tail(&buf->list, &channel->dma_buffers_free);

			spin_unlock_irqrestore(&channel->dma_buffers_free_lock, flags);
			dprintk(2, "%s() Added buffer %p to type %d free list\n",
				__func__, buf, channel->type);
		}
	}

	/* Now we allocate any timestamp hardware buffers */
	for (i = 0; i < MAX_USER_TIMESTAMP_BUFFERS; i++) {
		buf = vc8x0_buffer_alloc(channel, TYPE_TIMESTAMP, i);
		if (buf) {
			spin_lock_irqsave(&channel->dma_timestamp_buffers_free_lock, flags);

			buf->state = STATE_FREE;
			list_add_tail(&buf->list, &channel->dma_timestamp_buffers_free);

			spin_unlock_irqrestore(&channel->dma_timestamp_buffers_free_lock, flags);
			dprintk(2, "%s() Added buffer %p to timestamp free list\n",
				__func__, buf);
		}
	}

	/* Configure channel specific registers */
	vc8x0_dma_configure_channel_registers(channel);

	do {
		if (vc8x0_i2c_register(&channel->i2c_bus) < 0) {
			printk(KERN_ERR "%s(nr=%d) Error registering vida i2c bus\n",
				__func__, channel->nr);
			break;
		}
		if (vc8x0_spi_register(&channel->spi_bus) < 0) {
			printk(KERN_ERR "%s(nr=%d) Error registering vida spi bus\n",
				__func__, channel->nr);
			break;
		}

		if (channel->type == TYPE_VIDEO) {
			if (IS_820E(dev)) {
				/* AD7441 */
				if (vc8x0_ad7441_init(channel, 0x42) < 0) {
					printk(KERN_ERR "%s(nr=%d) Error detecting the ad7441\n",
						__func__, channel->nr);
					break;
				}

				/* Flip the FPGA bit to inform it the EDID is loaded (enable hotplug) */
				vc_set32(channel->regs.m_nMisc, 0x01);

				if (vc8x0_ad7441_enable(&channel->ad7441_ctx) < 0) {
					printk(KERN_ERR "%s(nr=%d) Error initializing the ad7441\n",
						__func__, channel->nr);
					break;
				}
			}
			if (IS_845E(dev)) {
				/* GS2971 SDI Receiver */
				if (vc8x0_gs2971a_init(channel) < 0) {
					printk(KERN_ERR "%s(nr=%d) Error detecting the gs2971a\n",
						__func__, channel->nr);
					break;
				}

				if (vc8x0_gs2971a_enable(&channel->gs2971a_ctx) < 0) {
					printk(KERN_ERR "%s(nr=%d) Error initializing the gs2971a\n",
						__func__, channel->nr);
					break;
				}
			}
		}

		if (channel->type == TYPE_AUDIO) {
			if (IS_820E(dev)) {
				/* PCM3052 */
				if (vc8x0_pcm3052_init(channel, 0x8e) < 0) {
					printk(KERN_ERR "%s(nr=%d) Error detecting the pcm3052\n",
						__func__, channel->nr);
					break;
				}
				if (vc8x0_pcm3052_enable(&channel->pcm3052_ctx) < 0) {
					printk(KERN_ERR "%s(nr=%d) Error initializing the pcm3052\n",
						__func__, channel->nr);
					break;
				}
			}
		}

		if (channel->type == TYPE_VIDEO) {
			/* Video Subsystem per channel */
			if (vc8x0_video_register(channel) < 0) {
				printk(KERN_ERR "%s(nr=%d) Error initializing the video subsystem\n",
					__func__, channel->nr);
				break;
			}
		}

		if (channel->type == TYPE_AUDIO) {
#if ENABLE_ALSA
			/* Audio Subsystem per channel */
			if (vc8x0_audio_register(channel) < 0) {
				printk(KERN_ERR "%s(nr=%d) Error initializing the audio subsystem\n",
					__func__, channel->nr);
				break;
			}
#endif
		}

		/* Configure the DMA registers for the Flex, now that we have the PCI buffers,
		 * video registers and formats are all configure.
		 */
		vc8x0_dma_configure(channel);

		if (channel_debug >= 2)
			vc8x0_dma_dump_channel_registers(channel);

		ret = 0;

	} while (0);

	return ret;
}

/* Move one set of buffers to another list */
void vc8x0_channel_list_move(struct vc8x0_dma_channel *channel,
	struct list_head *srchead,
	spinlock_t *srclock,
	struct list_head *dsthead,
	spinlock_t *dstlock,
	enum state_t state);

/* I'm specifically taking the spinlocks without disabling
 * irq's else pci_free_consistent warns, and by this stage the
 * hardware should not be running.
 */
int vc8x0_channel_unregister(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev;
	struct vc8x0_dma_buffer *buf;
	struct list_head *p, *q;

	if (channel == 0)
		BUG();
	if (channel->dev == 0)
		BUG();

	dev = channel->dev;

	dprintk(1, "%s(nr=%d)\n", __func__, channel->nr);

	if (channel->type == TYPE_AUDIO) {
		if (IS_820E(dev)) {
			if (vc8x0_pcm3052_disable(&channel->pcm3052_ctx) < 0)
				printk(KERN_ERR "%s(nr=%d) Error initializing the pcm3052\n",
					__func__, channel->nr);
		}
#if ENABLE_PCM_FIFO_DEBUG
		if (channel->aa_buf) {
			kfree(channel->aa_buf);
			channel->aa_buf = 0;
		}
#endif
	}

	if (channel->type == TYPE_VIDEO) {
		if (IS_820E(dev)) {
			/* Flip the FPGA bit to inform it the EDID is unloaded (disable hotplug) */
			vc_clear32(channel->regs.m_nMisc, 0x01);

			if (vc8x0_ad7441_disable(&channel->ad7441_ctx) < 0)
				printk(KERN_ERR "%s(nr=%d) Error disabling the ad7441\n",
					__func__, channel->nr);
		}
		if (IS_845E(dev)) {
			if (vc8x0_gs2971a_disable(&channel->gs2971a_ctx) < 0)
				printk(KERN_ERR "%s(nr=%d) Error disabling the gs2971a\n",
					__func__, channel->nr);
		}
	}

	/* Move any non-free AV buffers to the free buffer list then dealloc
	 * everything on the free buffer list.
	 * Finish by destorying the entire channel object.
	 */

	/* Move any full video buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_full,
		&channel->dma_buffers_full_lock,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		STATE_FREE);

	/* Move any dpc video buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_dpc,
		&channel->dma_buffers_dpc_lock,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		STATE_FREE);

	/* Move any busy video buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_busy,
		&channel->dma_buffers_busy_lock,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		STATE_FREE);

	/* Delete any free video buffers */
	spin_lock(&channel->dma_buffers_free_lock);
	list_for_each_safe(p, q, &channel->dma_buffers_free) {
		buf = list_entry(p, struct vc8x0_dma_buffer, list);

		dprintk(2, "%s() Deallocating FREE buffer %p\n",
			__func__, buf);

		list_del(p);
		vc8x0_buffer_dealloc(buf);
	}
	spin_unlock(&channel->dma_buffers_free_lock);

	/* Move any non-free timestamp buffers to the free buffer list then dealloc
	 * everything on the free buffer list.
	 * Finish by destorying the entire channel object.
	 */

	/* Move any full timestamp buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_timestamp_buffers_full,
		&channel->dma_timestamp_buffers_full_lock,
		&channel->dma_timestamp_buffers_free,
		&channel->dma_timestamp_buffers_free_lock,
		STATE_FREE);

	/* Move any dpc timestamp buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_timestamp_buffers_dpc,
		&channel->dma_timestamp_buffers_dpc_lock,
		&channel->dma_timestamp_buffers_free,
		&channel->dma_timestamp_buffers_free_lock,
		STATE_FREE);

	/* Move any busy timestamp buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_timestamp_buffers_busy,
		&channel->dma_timestamp_buffers_busy_lock,
		&channel->dma_timestamp_buffers_free,
		&channel->dma_timestamp_buffers_free_lock,
		STATE_FREE);

	/* Delete any free timestamp buffers */
	spin_lock(&channel->dma_timestamp_buffers_free_lock);
	list_for_each_safe(p, q, &channel->dma_timestamp_buffers_free) {
		buf = list_entry(p, struct vc8x0_dma_buffer, list);

		dprintk(2, "%s() Deallocating FREE buffer %p\n",
			__func__, buf);

		list_del(p);
		vc8x0_buffer_dealloc(buf);
	}
	spin_unlock(&channel->dma_timestamp_buffers_free_lock);

	if (channel->type == TYPE_AUDIO) {
#if ENABLE_ALSA
		vc8x0_audio_unregister(channel);
#endif
	}
	if (channel->type == TYPE_VIDEO) {
		vc8x0_video_unregister(channel);
	}

	vc8x0_i2c_unregister(&channel->i2c_bus);
	vc8x0_spi_unregister(&channel->spi_bus);

	memset(channel, 0, sizeof(struct vc8x0_dma_channel));

	return 0;
}

/* Move one set of buffers to another list */
void vc8x0_channel_list_move(struct vc8x0_dma_channel *channel,
	struct list_head *srchead,
	spinlock_t *srclock,
	struct list_head *dsthead,
	spinlock_t *dstlock,
	enum state_t state)
{
	struct vc8x0_dma_buffer *buf;
	struct vc8x0_dev *dev = channel->dev;
	struct list_head *p, *q;
	unsigned long flags, flags2;

	/* Move any buffers from srclist to the free list */
	spin_lock_irqsave(srclock, flags);
        list_for_each_safe(p, q, srchead) {
                buf = list_entry(p, struct vc8x0_dma_buffer, list);

                dprintk(1, "%s() Moving buffer %p to the %d list\n",
                        __func__, buf, state);

                spin_lock_irqsave(dstlock, flags2);
                buf->state = state;
                list_move_tail(&buf->list, dsthead);
                spin_unlock_irqrestore(dstlock, flags2);
        }
        spin_unlock_irqrestore(srclock, flags);
}

/* Move any non-free buffers to the free buffer list. Call this
 * when streaming has stopped.
 */
void vc8x0_channel_free_all_buffers(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev;

	if (channel == 0)
		BUG();
	if (channel->dev == 0)
		BUG();

	dev = channel->dev;

	/* Move any full buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_full,
		&channel->dma_buffers_full_lock,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		STATE_FREE);

	/* Move any dpc buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_dpc,
		&channel->dma_buffers_dpc_lock,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		STATE_FREE);

	/* Move any busy buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_busy,
		&channel->dma_buffers_busy_lock,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		STATE_FREE);

	/* Move any full buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_timestamp_buffers_full,
		&channel->dma_timestamp_buffers_full_lock,
		&channel->dma_timestamp_buffers_free,
		&channel->dma_timestamp_buffers_free_lock,
		STATE_FREE);

	/* Move any dpc buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_timestamp_buffers_dpc,
		&channel->dma_timestamp_buffers_dpc_lock,
		&channel->dma_timestamp_buffers_free,
		&channel->dma_timestamp_buffers_free_lock,
		STATE_FREE);

	/* Move any busy buffers to the free list */
	vc8x0_channel_list_move(channel,
		&channel->dma_timestamp_buffers_busy,
		&channel->dma_timestamp_buffers_busy_lock,
		&channel->dma_timestamp_buffers_free,
		&channel->dma_timestamp_buffers_free_lock,
		STATE_FREE);

}

void vc8x0_channel_statistics_reset(struct vc8x0_dma_channel *channel)
{
	channel->buffers_dropped = 0;
	channel->buffers_processed = 0;
	channel->buffers_bad_address = 0;
	channel->alignment_errors = 0;
}

void vc8x0_channel_statistics_procfs_buffer_list(struct seq_file *m,
	struct vc8x0_dma_channel *channel,
	char *title,
	spinlock_t *lock,
	struct list_head *head)
{
	struct vc8x0_dma_buffer *buf;
	struct list_head *p, *q;
	int i = 0;
	unsigned long flags;

	seq_printf(m, " ->%s\n", title);

	spin_lock_irqsave(lock, flags);
	list_for_each_safe(p, q, head) {
		buf = list_entry(p, struct vc8x0_dma_buffer, list);

		seq_printf(m, "    buffer %p state %s type %s va %p size %d last_used %lld phys addr %x\n",
			buf,
			buf->state == STATE_FREE ? "FREE" :
			buf->state == STATE_BUSY ? "BUSY" :
			buf->state == STATE_DPC  ? "DPC" :
			buf->state == STATE_FULL ? "FULL" :
			"UNKNOWN",
			buf->type == TYPE_VIDEO ? "VIDEO" :
			buf->type == TYPE_AUDIO ? "AUDIO" :
			buf->type == TYPE_TIMESTAMP ? "TIMESTAMP" :
			"UNKNOWN",
			buf->cpu,
			buf->cpu_size,
			buf->last_used,
			(u32)buf->cpu_dma
			);

		i++;
	}
	spin_unlock_irqrestore(lock, flags);
	seq_printf(m, "    %s = %d\n", title, i);
}

void vc8x0_channel_statistics_procfs(struct seq_file *m, struct vc8x0_dma_channel *channel)
{
	seq_printf(m, "Channel%d(%s) Statistics:\n",
		channel->nr,
		channel->type == TYPE_VIDEO ? "VIDEO" :
		channel->type == TYPE_AUDIO ? "AUDIO" :
		channel->type == TYPE_TIMESTAMP ? "TIMESTAMP" :
		"UNKNOWN"
		);

	seq_printf(m, "    running_status = %s\n",
		channel->state == STATE_RUNNING ? "RUNNING" :
		channel->state == STATE_STOPPED ? "STOPPED" :
		"UNDEFINED");

	seq_printf(m, " buffers_dropped     = %llu\n", channel->buffers_dropped);
	seq_printf(m, " buffers_processed   = %llu\n", channel->buffers_processed);
	seq_printf(m, " buffers_bad_address = %llu\n", channel->buffers_bad_address);
	seq_printf(m, " alignment_errors    = %llu\n", channel->alignment_errors);
	seq_printf(m, " vbi_enabled         = %d\n", channel->vbi_enabled);

	if (IS_820E(channel->dev))
		vc8x0_ad7441_procfs_detection_show(m, &channel->ad7441_ctx);

	/* VIDEO buffers */
	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_buffers_free",
		&channel->dma_buffers_free_lock,
		&channel->dma_buffers_free);

	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_buffers_busy",
		&channel->dma_buffers_busy_lock,
		&channel->dma_buffers_busy);

	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_buffers_dpc",
		&channel->dma_buffers_dpc_lock,
		&channel->dma_buffers_dpc);

	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_buffers_full",
		&channel->dma_buffers_full_lock,
		&channel->dma_buffers_full);

	/* TIMESTAMP buffers */
	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_timestamp_buffers_free",
		&channel->dma_timestamp_buffers_free_lock,
		&channel->dma_timestamp_buffers_free);

	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_timestamp_buffers_busy",
		&channel->dma_timestamp_buffers_busy_lock,
		&channel->dma_timestamp_buffers_busy);

	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_timestamp_buffers_dpc",
		&channel->dma_timestamp_buffers_dpc_lock,
		&channel->dma_timestamp_buffers_dpc);

	vc8x0_channel_statistics_procfs_buffer_list(m,
		channel,
		"dma_timestamp_buffers_full",
		&channel->dma_timestamp_buffers_full_lock,
		&channel->dma_timestamp_buffers_full);
}

int vc8x0_channel_state_procfs(struct seq_file *m, struct vc8x0_dma_channel *channel)
{
	seq_printf(m, "Channel%d Statistics:\n", channel->nr);
	if (channel->type == TYPE_VIDEO) {
		if (IS_820E(channel->dev))
			vc8x0_ad7441_procfs_show(m, &channel->ad7441_ctx);
		if (IS_845E(channel->dev))
			vc8x0_gs2971a_procfs_show(m, &channel->gs2971a_ctx);
	}

	if (channel->type == TYPE_AUDIO) {
		if (IS_820E(channel->dev))
			vc8x0_pcm3052_procfs_show(m, &channel->pcm3052_ctx);
	}

	return 0;
}

int vc8x0_channel_event(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	int handled = 0;

	if (channel->state != STATE_RUNNING)
		return 0;

	if (channel->m_arrLastFrameIndex == 0)
		channel->m_arrLastFrameIndex = MAX_USER_VIDEO_BUFFERS - 1;
	else
		channel->m_arrLastFrameIndex--;

	if (channel->type == TYPE_VIDEO) {

		handled = vc8x0_video_irqhandler(channel);

		if (channel->nr == 1) {
			vc_set32(VDMA_EVENT_CLR, 0x00000001);
		} else
		if (channel->nr == 2) {
			vc_set32(VDMA_EVENT_CLR, 0x00000004);
		} else
		if (channel->nr == 3) {
			vc_set32(VDMA_EVENT_CLR, 0x00000010);
		} else
		if (channel->nr == 4) {
			vc_set32(VDMA_EVENT_CLR, 0x00000040);
		}
	} else
	if (channel->type == TYPE_AUDIO) {

		handled = vc8x0_audio_irqhandler(channel);

		if (channel->nr == 5)
			vc_set32(VDMA_EVENT_CLR, 0x00000002);

		/* TODO: Review this, why is channel 9 clearing event 2, not 11? */
		if (channel->nr == 9)
			vc_set32(VDMA_EVENT_CLR, 0x00000002);
	}

	return handled;
}

int vc8x0_channel_irq(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	int handled = 0;

	dprintk(1, "%s(%d)\n", __func__, channel->nr);

	if (channel->state != STATE_RUNNING)
		return 0;

	return handled;
}

int vc8x0_channel_thread(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;

	if (channel->type == TYPE_VIDEO) {
		if (IS_820E(channel->dev)) {
			if (vc8x0_ad7441_is_locked(&channel->ad7441_ctx)) {
				/* Do nothing */
			}
		}
		if (IS_845E(channel->dev)) {
			if (vc8x0_gs2971a_is_locked(&channel->gs2971a_ctx)) {
				/* Do nothing */
			}
		}
	}

	if (channel->thread_actions & THREAD_ACTION_FLEX_RESET) {
		if (time_after(jiffies, (long)channel->last_timestamp_err +
			msecs_to_jiffies(1000))) {
			dprintk(1, "%s() Fixing alignment issue channel->nr = %d\n",
				__func__, channel->nr);
			vc8x0_dma_reset(channel);
                        channel->last_timestamp_err = jiffies;
                }
		channel->thread_actions &= ~THREAD_ACTION_FLEX_RESET;
	}

	return 0;
}

static void vc8x0_channel_enable_interrupts(struct vc8x0_dma_channel *channel, int enable)
{
	struct vc8x0_dev *dev = channel->dev;
	dprintk(1, "%s(%d, %d)\n", __func__, channel->nr, enable);

	if (enable) {

		if (channel->nr == 1)
			dev->m_nInterruptMask1 |= 0x10001; /* event 0 */
		if (channel->nr == 2)
			dev->m_nInterruptMask1 |= 0x20004; /* event 2 */
		if (channel->nr == 3)
			dev->m_nInterruptMask1 |= 0x40010; /* event 4 */
		if (channel->nr == 4)
			dev->m_nInterruptMask1 |= 0x80040; /* event 6 */
		if (channel->nr == 5)
			dev->m_nInterruptMask1 |= 0x00002; /* event 1 */
		if (channel->nr == 6)
			dev->m_nInterruptMask1 |= 0x00008; /* event 3 */
		if (channel->nr == 7)
			dev->m_nInterruptMask1 |= 0x00020; /* event 5 */
		if (channel->nr == 8)
			dev->m_nInterruptMask1 |= 0x00080; /* event 7 */
		if (channel->nr == 9)
			dev->m_nInterruptMask1 |= 0x00800; /* event 11 */

		vc_write32(0xe08, dev->m_nInterruptMask1);
		vc_write32(0xe0c, dev->m_nInterruptMask2);

	} else {

		if (channel->nr == 1)
			dev->m_nInterruptMask1 &= ~0x10001;
		if (channel->nr == 2)
			dev->m_nInterruptMask1 &= ~0x20004;
		if (channel->nr == 3)
			dev->m_nInterruptMask1 &= ~0x40010;
		if (channel->nr == 4)
			dev->m_nInterruptMask1 &= ~0x80040;
		if (channel->nr == 5)
			dev->m_nInterruptMask1 &= ~0x00002;
		if (channel->nr == 6)
			dev->m_nInterruptMask1 &= ~0x00008;
		if (channel->nr == 7)
			dev->m_nInterruptMask1 &= ~0x00020;
		if (channel->nr == 8)
			dev->m_nInterruptMask1 &= ~0x00080;
		if (channel->nr == 9)
			dev->m_nInterruptMask1 &= ~0x00800;

		vc_write32(0xe08, dev->m_nInterruptMask1);
		vc_write32(0xe0c, dev->m_nInterruptMask2);
	}
}

int vc8x0_channel_start(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	u32 val;
	dprintk(1, "%s(%d)\n", __func__, channel->nr);

	if (dev->risc_running == 0) {
		dprintk(1, "%s(%d) Risc is not running\n", __func__, channel->nr);
		return 0;
	}

	if (channel->state == STATE_RUNNING) {
		dprintk(1, "%s(%d) Channel already running\n", __func__, channel->nr);
		return 0;
	}

	/* Move any free video buffers to the busy list */
	vc8x0_channel_list_move(channel,
		&channel->dma_buffers_free,
		&channel->dma_buffers_free_lock,
		&channel->dma_buffers_busy,
		&channel->dma_buffers_busy_lock,
		STATE_BUSY);

	if (channel->nr == 1) {
		dprintk(1, "%s(%d) Starting the engine\n", __func__, channel->nr);

		/* Re-position the video fifo */
		vc8x0_dma_reset(channel);

		/* VBI, bit 3 low = disabled */
		val = vc_read32(channel->regs.m_nMisc);

		if (channel->vbi_enabled)
			vc_write32(channel->regs.m_nMisc, val | 0x08);
		else
			vc_write32(channel->regs.m_nMisc, val & 0xF7);

		if (IS_820E(dev)) {
			vc_set32(VDMA_EVENT_CLR, 0x00000801);
			vc_set32(VDMA_EVENT_EN, 0x00000801);
		}
		if (IS_845E(dev)) {
			vc_set32(VDMA_EVENT_CLR, 0x00000003);
			vc_set32(VDMA_EVENT_EN, 0x00000003);
		}

		vc8x0_channel_enable_interrupts(channel, 0);

		vc_write32(channel->regs.m_nStreamDisabled, 0); /* Video flex on */
		vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Video Fifo on */
#if ENABLE_ALSA
		if (IS_820E(dev))
			vc8x0_channel_start(&dev->dma_channel[DMA_CHANNEL9]);
		if (IS_845E(dev)) {
			vc8x0_channel_start(&dev->dma_channel[DMA_CHANNEL5]);
		}
#endif
		vc8x0_channel_enable_interrupts(channel, 1);

	} else
	if (channel->nr == 2) {
		dprintk(1, "%s(%d) Starting the engine\n", __func__, channel->nr);

		/* Re-position the video fifo */
		vc8x0_dma_reset(channel);

		/* VBI, bit 3 low = disabled */
		val = vc_read32(channel->regs.m_nMisc);

		if (channel->vbi_enabled)
			vc_write32(channel->regs.m_nMisc, val | 0x08);
		else
			vc_write32(channel->regs.m_nMisc, val & 0xF7);

		if (IS_820E(dev)) {
			vc_set32(VDMA_EVENT_CLR, 0x00000004);
			vc_set32(VDMA_EVENT_EN, 0x00000004);
		}

		vc8x0_channel_enable_interrupts(channel, 0);

		vc_write32(channel->regs.m_nStreamDisabled, 0); /* Video flex on */
		vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Video Fifo on */

#if ENABLE_ALSA
		if (IS_820E(dev)) {
			/* Audio Slaved to Video A currently only */
		}
		if (IS_845E(dev)) {
			vc8x0_channel_start(&dev->dma_channel[DMA_CHANNEL6]);
		}
#endif
		vc8x0_channel_enable_interrupts(channel, 1);
	} else
	if (channel->nr == 3) {
		dprintk(1, "%s(%d) Starting the engine\n", __func__, channel->nr);

		/* Re-position the video fifo */
		vc8x0_dma_reset(channel);

		/* VBI, bit 3 low = disabled */
		val = vc_read32(channel->regs.m_nMisc);

		if (channel->vbi_enabled)
			vc_write32(channel->regs.m_nMisc, val | 0x08);
		else
			vc_write32(channel->regs.m_nMisc, val & 0xF7);

		if (IS_845E(dev)) {
			vc_set32(VDMA_EVENT_CLR, 0x00000010);
			vc_set32(VDMA_EVENT_EN, 0x00000010);
		}

		vc8x0_channel_enable_interrupts(channel, 0);

		vc_write32(channel->regs.m_nStreamDisabled, 0); /* Video flex on */
		vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Video Fifo on */
#if ENABLE_ALSA
		if (IS_845E(dev)) {
			vc8x0_channel_start(&dev->dma_channel[DMA_CHANNEL7]);
		}
#endif
		vc8x0_channel_enable_interrupts(channel, 1);

	} else
	if (channel->nr == 4) {
		dprintk(1, "%s(%d) Starting the engine\n", __func__, channel->nr);

		/* Re-position the video fifo */
		vc8x0_dma_reset(channel);

		/* VBI, bit 3 low = disabled */
		val = vc_read32(channel->regs.m_nMisc);

		if (channel->vbi_enabled)
			vc_write32(channel->regs.m_nMisc, val | 0x08);
		else
			vc_write32(channel->regs.m_nMisc, val & 0xF7);

		if (IS_845E(dev)) {
			vc_set32(VDMA_EVENT_CLR, 0x00000040);
			vc_set32(VDMA_EVENT_EN, 0x00000040);
		}

		vc8x0_channel_enable_interrupts(channel, 0);

		vc_write32(channel->regs.m_nStreamDisabled, 0); /* Video flex on */
		vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Video Fifo on */
#if ENABLE_ALSA
		if (IS_845E(dev)) {
			vc8x0_channel_start(&dev->dma_channel[DMA_CHANNEL8]);
		}
#endif
		vc8x0_channel_enable_interrupts(channel, 1);

	} else
	if ((channel->nr >= 5) || (channel->nr <= 9)) {
		dprintk(1, "%s(%d) Starting the audio engine\n", __func__, channel->nr);
#if ENABLE_PCM_FIFO_DEBUG
		memset(channel->aa_buf, 0, channel->aa_size);
#endif
		vc8x0_channel_enable_interrupts(channel, 0);

		/* Re-position the audio fifo */
		vc8x0_dma_reset(channel);

		/* Fix sample rate to 48KHz */
		vc_write32(0x940, 0x00000002);
		vc_write32(0x9c0, 0x00000002);
		vc_write32(0xa40, 0x00000002);
		vc_write32(0xac0, 0x00000002);

		vc_write32(channel->regs.m_nStreamDisabled, 0); /* Audio Flex on */
		vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Audio Fifo on */

		vc8x0_channel_enable_interrupts(channel, 1);
	}

	channel->prev_timestamp = 0;
	channel->curr_timestamp = 0;
	channel->diff_timestamp = 0;
	channel->last_timestamp_err = jiffies;
	channel->state = STATE_RUNNING;
	mod_timer(&channel->timeout, jiffies + (HZ / 2));
	if (channel->type == TYPE_VIDEO) {
		mod_timer(&channel->vbi_timeout, jiffies + (HZ / 2));
	}
	return 0;
}

int vc8x0_channel_stop(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
#if ENABLE_PCM_FIFO_DEBUG
	int i;
	u32 v;
#endif
	dprintk(1, "%s(%d)\n", __func__, channel->nr);

	if (dev->risc_running == 0) {
		dprintk(1, "%s(%d) Risc is not running\n", __func__, channel->nr);
		return 0;
	}

	if (channel->state == STATE_STOPPED) {
		dprintk(1, "%s(%d) Channel already stopped\n", __func__, channel->nr);
		return 0;
	}

	channel->state = STATE_STOPPED;
	del_timer(&channel->timeout);

	if (channel->nr == 1) {
		dprintk(1, "%s(%d) Stopping the engine\n", __func__, channel->nr);

		/* Stop the FLEX program */
		vc_write32(channel->regs.m_nStreamDisabled, 1);

		/* Stop the FPGA FIFO */
		vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Video Fifo off */

		/* Reset the Audio Flex registers */
		msleep(200);

#if ENABLE_ALSA
		if (IS_820E(dev))
			vc8x0_channel_stop(&dev->dma_channel[DMA_CHANNEL9]);
		if (IS_845E(dev)) {
			vc8x0_channel_stop(&dev->dma_channel[DMA_CHANNEL5]);
		}
#endif
	} else
	if (channel->nr == 2) {
		dprintk(1, "%s(%d) Stopping the engine\n", __func__, channel->nr);

		/* Stop the FLEX program */
		vc_write32(channel->regs.m_nStreamDisabled, 1);

		/* Stop the FPGA FIFO */
		vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Video Fifo off */

		/* Reset the Audio Flex registers */
		msleep(200);

#if ENABLE_ALSA
		if (IS_845E(dev)) {
			vc8x0_channel_stop(&dev->dma_channel[DMA_CHANNEL6]);
		}
#endif
	} else
	if (channel->nr == 3) {
		dprintk(1, "%s(%d) Stopping the engine\n", __func__, channel->nr);

		/* Stop the FLEX program */
		vc_write32(channel->regs.m_nStreamDisabled, 1);

		/* Stop the FPGA FIFO */
		vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Video Fifo off */

		/* Reset the Audio Flex registers */
		msleep(200);

#if ENABLE_ALSA
		if (IS_845E(dev)) {
			vc8x0_channel_stop(&dev->dma_channel[DMA_CHANNEL7]);
		}
#endif
	} else
	if (channel->nr == 4) {
		dprintk(1, "%s(%d) Stopping the engine\n", __func__, channel->nr);

		/* Stop the FLEX program */
		vc_write32(channel->regs.m_nStreamDisabled, 1);

		/* Stop the FPGA FIFO */
		vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Video Fifo off */

		/* Reset the Audio Flex registers */
		msleep(200);

#if ENABLE_ALSA
		if (IS_845E(dev)) {
			vc8x0_channel_stop(&dev->dma_channel[DMA_CHANNEL8]);
		}
#endif
	} else
	if ((channel->nr >= 5) && (channel->nr <= 9)) {
		dprintk(1, "%s(%d) Stopping the engine\n", __func__, channel->nr);

		/* Stop the FLEX program */
		vc_write32(channel->regs.m_nStreamDisabled, 1);

		/* Stop the FPGA FIFO */
		vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Audio Fifo off */

#if ENABLE_PCM_FIFO_DEBUG
		if (channel->aa_buf) {
			for (i = 0; i < 65536; i++) {
				v = channel->aa_buf[i];
				if (v) {
					printk(KERN_ERR "(a)0x%04x = %d\n", i, v);
				}
			}
		}
#endif
	}

	dprintk(1, "%s() free all buffers\n", __func__);
	vc8x0_channel_free_all_buffers(channel);
	dprintk(1, "%s() freed all buffers\n", __func__);
	return 0;
}

