/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

MODULE_DESCRIPTION("Driver for the family of ViewCast 8x0 cards");
MODULE_AUTHOR("Steven Toth <stoth@kernellabs.com>");
MODULE_LICENSE("GPL");

/* 1 = Basic device statistics
 * 2 = PCIe register dump for entire device
 * 4 = AD9985 register dump
 * 8 = SIL9013 register dump
 */
unsigned int vc8x0_procfs_verbosity = 3;
module_param(vc8x0_procfs_verbosity, int, 0644);
MODULE_PARM_DESC(vc8x0_procfs_verbosity, "enable procfs debugging via /proc/vc8x0");

unsigned int vc8x0_dual_1080p60 = 0;
module_param(vc8x0_dual_1080p60, int, 0644);
MODULE_PARM_DESC(vc8x0_dual_1080p60, "enable dual channel 1080p60 only (def: 0)");

unsigned int vc8x0_thread_active = 1;
module_param(vc8x0_thread_active, int, 0644);
MODULE_PARM_DESC(vc8x0_thread_active, "should keep alive thread run");

static unsigned int debug;
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "enable debug messages");

static unsigned int pcie_serialize;
module_param(pcie_serialize, int, 0644);
MODULE_PARM_DESC(pcie_serialize, "enable pcie_serialize read/writes");

static unsigned int irq_debug;
module_param(irq_debug, int, 0644);
MODULE_PARM_DESC(irq_debug, "enable debug messages [irq]");

static unsigned int msi_enable;
module_param_named(msi_enable, msi_enable, int, 0644);
MODULE_PARM_DESC(msi_enable, "use msi interrupts (def: 0)");

static unsigned int card[]  = {[0 ... (VC8X0_MAXBOARDS - 1)] = UNSET };
module_param_array(card,  int, NULL, 0444);
MODULE_PARM_DESC(card, "card type");

#define dprintk(level, fmt, arg...)\
	do { if (debug >= level)\
		printk(KERN_DEBUG "%s: " fmt, dev->name, ## arg);\
	} while (0)

static unsigned int vc8x0_devcount;
static DEFINE_MUTEX(devlist);
LIST_HEAD(vc8x0_devlist);

__inline__ u32 vc8x0_read32(struct vc8x0_dev *dev, u32 reg)
{
	u32 val;
	unsigned long flags = 0;

	if (pcie_serialize)
		spin_lock_irqsave(&dev->reglock, flags);

        val = readl(dev->lmmio032 + ((reg)>>2));

	if (pcie_serialize)
		spin_unlock_irqrestore(&dev->reglock, flags);

//	printk(KERN_ERR "%s() %x = %x\n", __func__, reg, val);
	return val;
}

__inline__ void vc8x0_write32(struct vc8x0_dev *dev, u32 reg, u32 value)
{
	unsigned long flags = 0;

	if (pcie_serialize)
		spin_lock_irqsave(&dev->reglock, flags);

//	printk(KERN_ERR "%s() %x = %x\n", __func__, reg, value);
	writel((value), dev->lmmio032 + ((reg)>>2));

	if (pcie_serialize)
		spin_unlock_irqrestore(&dev->reglock, flags);
}

__inline__ void vc8x0_andor32(struct vc8x0_dev *dev, u32 reg, u32 mask, u32 value)
{
	writel((value), dev->lmmio032 + ((reg)>>2));
}

__inline__ u64 vc8x0_read64(struct vc8x0_dev *dev, u32 reg)
{
	u64 val;

	unsigned long flags = 0;

	if (pcie_serialize)
		spin_lock_irqsave(&dev->reglock, flags);

        val = readl(dev->lmmio032 + ((reg)>>2));

	if (pcie_serialize)
		spin_unlock_irqrestore(&dev->reglock, flags);

	return val;
}

__inline__ void vc8x0_write64(struct vc8x0_dev *dev, u32 reg, u32 value)
{
	unsigned long flags = 0;

	if (pcie_serialize)
		spin_lock_irqsave(&dev->reglock, flags);

	writel((value), dev->lmmio032 + ((reg)>>2));

	if (pcie_serialize)
		spin_unlock_irqrestore(&dev->reglock, flags);
}

static void vc8x0_shutdown(struct vc8x0_dev *dev)
{
	/* Disable all interrupts */
	/* Power down all function blocks */
	vc_write32(0xe00, 0);
	vc_write32(0xe04, 0);
	vc_write32(0xe08, 0);
	vc_write32(0xe0c, 0);
}

static int get_resources(struct vc8x0_dev *dev)
{
	/* The documentation states we should ignore BAR1 and BAR2 */
	if (!request_mem_region(pci_resource_start(dev->pci, BAR0),
			       pci_resource_len(dev->pci, BAR0),
			       dev->name)) {
		printk(KERN_ERR "%s: can't get BAR0 MMIO memory @ 0x%llx\n",
			dev->name,
			(unsigned long long)pci_resource_start(dev->pci, BAR0));
	} else {
		if (!request_mem_region(pci_resource_start(dev->pci, BAR4),
			       pci_resource_len(dev->pci, BAR4),
			       dev->name)) {
			printk(KERN_ERR "%s: can't get BAR4 MMIO memory @ 0x%llx\n",
				dev->name,
				(unsigned long long)pci_resource_start(dev->pci, BAR4));
		} else
			return 0;
	}

	return -EBUSY;
}

static int vc8x0_dev_setup(struct vc8x0_dev *dev)
{
	int i;

	mutex_init(&dev->lock);
	spin_lock_init(&dev->reglock);

	atomic_inc(&dev->refcount);

	dev->nr = vc8x0_devcount++;
	sprintf(dev->name, "vc8x0[%d]", dev->nr);

	/* board config */
	dev->board = UNSET;
	if (card[dev->nr] < vc8x0_bcount)
		dev->board = card[dev->nr];
	for (i = 0; UNSET == dev->board  &&  i < vc8x0_idcount; i++)
		if (dev->pci->subsystem_vendor == vc8x0_subids[i].subvendor &&
		    dev->pci->subsystem_device == vc8x0_subids[i].subdevice)
			dev->board = vc8x0_subids[i].card;
	if (UNSET == dev->board) {
		dev->board = VC8X0_BOARD_UNKNOWN;
		vc8x0_card_list(dev);
	}

	/* The keepalive thread needs a mutex */
	mutex_init(&dev->kthread_lock);

	/* Main Master 0 Bus incl. eeprom */
	mutex_init(&dev->i2c_bus.lock);
	dev->i2c_bus.nr = 0;
	dev->i2c_bus.dev = dev;
	dev->i2c_bus.reg_base = 0xd80;

	if (get_resources(dev) < 0) {
		printk(KERN_ERR "CORE %s No more PCIe resources for "
		       "subsystem: %04x:%04x\n",
		       dev->name, dev->pci->subsystem_vendor,
		       dev->pci->subsystem_device);

		vc8x0_devcount--;
		return -ENODEV;
	}

	/* PCIe stuff */
	dev->lmmio032 = ioremap(pci_resource_start(dev->pci, BAR0),
			     pci_resource_len(dev->pci, BAR0));
	dev->lmmio064 = (u64 *)dev->lmmio032;
	dev->bmmio = (u8 *)dev->lmmio032;
	dev->lmmio4 = ioremap(pci_resource_start(dev->pci, BAR4),
			     pci_resource_len(dev->pci, BAR4));

	dev->m_nInterruptMask1 = 0;
	dev->m_nInterruptMask2 = 0;

	printk(KERN_INFO "CORE %s: subsystem: %04x:%04x, board: %s [card=%d,%s]\n",
	       dev->name, dev->pci->subsystem_vendor,
	       dev->pci->subsystem_device, vc8x0_boards[dev->board].name,
	       dev->board, card[dev->nr] == dev->board ?
	       "insmod option" : "autodetected");

	return 0;
}

static void vc8x0_dev_unregister(struct vc8x0_dev *dev)
{
	release_mem_region(pci_resource_start(dev->pci, BAR0),
			   pci_resource_len(dev->pci, BAR0));
	release_mem_region(pci_resource_start(dev->pci, BAR4),
			   pci_resource_len(dev->pci, BAR4));

	if (!atomic_dec_and_test(&dev->refcount))
		return;

	iounmap(dev->lmmio032);
	iounmap(dev->lmmio4);
}

static irqreturn_t vc8x0_irq(int irq, void *dev_id)
{
	struct vc8x0_dev *dev = dev_id;
	int handled = 1;
	u32 irq_status = 0, maskedval1 = 0, maskedval2 = 0, hwint, i, v, v2, gis;
	u32 irq_statusH;
	u32 devmask = 0x00010000;

	if (!(dev->interrupts_enabled)) {
		goto out;
	}

	gn_write(0xa04, 0);

	/* TODO: make sure interrupts are flagged as enabled else return */
	/* DeepBlueProtocol0800::Isr0800 */

	/* Get the primary interrupt */
	irq_status = vc_read32(0xe00);
	if (irq_status == 0xffffffff) {
		printk(KERN_ERR
			"%s() irq_status is probably bad 0x%x, aborting\n",
			__func__, irq_status);
		return IRQ_RETVAL(0);
	}

	irq_statusH = vc_read32(0xe04);
	if (irq_statusH != dev->last0e04) {
		dprintk(1, "%s() status went from %08x to %08x\n", __func__, dev->last0e04, irq_statusH);
		dev->last0e04 = irq_statusH;
	}

	/* Interrupts ever 10micro seconds */
	maskedval1 = irq_status & dev->m_nInterruptMask1;

	hwint = maskedval1 & 0xefff0000;

	gis = gn_read(GN4124_INT_STAT);

	if (irq_debug)
		printk(KERN_ERR "%s() irq_status = %x, gis = %x, hwint = %x, maskedval1 = %x\n", __func__,
			irq_status, gis, hwint, maskedval1);

	/* Check the device interrupts. 00XX0000. If any of the interrupts
	 * were raised then disable them if they were considered spurious.
	 */
	if (hwint) {
		/* Determine which channel raised the interrupt */
		for (i = 0; i < 8; i++) {
			/* Check each specific channel */
			if (hwint & devmask) {
				/* This channel signalled.... but was it's interrupt mask enabled? */
				if (dev->m_nInterruptMask1 & devmask) {
					/* Disable channel */
					dev->m_nInterruptMask1 &= ~devmask;
				}
			}
			devmask <<= 1;
		}
		vc_write32(0xe08, dev->m_nInterruptMask1);
	}
	vc_write32(0xe00, irq_status);

	/* Secondary interrupt */
	if (maskedval1 & 0x80000000) {
		v = vc_read32(0xe04);
		dev->m_nInterruptMask2 &= ~v;
		if (dev->m_nInterruptMask2) {
			vc_write32(0xe0c, dev->m_nInterruptMask2);
		}
		if (v) {
			vc_write32(0xe04, v);
		}
		maskedval2 = v;
	}

	/* Deal with the GN4124 interrupts */
	v2 = (gis & GN4124_INTERRPT_MASK) ? 1 : 0;
	for (i = 0; i < 7; i++) {
		if (gis & (INT_STAT_ALI0 < i))
			dev->m_nALInterruptCounts[i]++;
	}
	gis = gn_read(GN4124_GPIO_INT_STATUS);

	/* See if the GN4124 has fignalled that one or more FLEX risc processor
	 * events have occured. We need to service each event.
	 */
	if (maskedval1 & VDMA_EVENT_0) {
		/* Video Channel A */
		handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL1 ]);
	}
	if (maskedval1 & VDMA_EVENT_1) {
		/* Audio Channel - SDI Audio AD1 */
		if (IS_845E(dev))
			handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL5 ]);
	}
	if (maskedval1 & VDMA_EVENT_11) {
#if ENABLE_ALSA
		if (IS_820E(dev)) {
			/* Audio Channel - Unbalanced Audio AD5 */
			handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL9 ]);
		}
#else
		printk(KERN_ERR "%s() spurious AD5/DMA9 interrupt, ignoring\n",
			__func__);
#endif
	}

	/* Video / Audio Channel B */
	if (maskedval1 & VDMA_EVENT_2) {
		/* Video Channel B */
		handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL2 ]);
	}
	if (maskedval1 & VDMA_EVENT_3) {
		/* Audio Channel - SDI Audio AD2 */
		if (IS_845E(dev))
			handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL6 ]);
	}

	/* Video / Audio Channel C */
	if ((maskedval1 & VDMA_EVENT_4) && (!vc8x0_dual_1080p60)) {
		/* Video Channel C */
		handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL3 ]);
	}
	if ((maskedval1 & VDMA_EVENT_5) && (!vc8x0_dual_1080p60)) {
		/* Audio Channel - SDI Audio AD3 */
		if (IS_845E(dev))
			handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL7 ]);
	}

	/* Video / Audio Channel D */
	if ((maskedval1 & VDMA_EVENT_6) && (!vc8x0_dual_1080p60)) {
		/* Video Channel D */
		handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL4 ]);
	}
	if ((maskedval1 & VDMA_EVENT_7) && (!vc8x0_dual_1080p60)) {
		/* Audio Channel - SDI Audio AD4 */
		if (IS_845E(dev))
			handled += vc8x0_channel_event(&dev->dma_channel[ DMA_CHANNEL8 ]);
	}

#if ENABLE_MONITOR_REGISTERS
	for (i = 0, v = 0; i < MAX_MONITOR_REGISTERS; i++, v += 4) {
		dev->monitor_regs[i] = vc_read32(0xb8 + v);
	}
#endif

out:
	gn_write(0xa04, 1);
	handled = 1;
	return IRQ_RETVAL(handled);
}

#ifdef CONFIG_PROC_FS
static int vc8x0_proc_state_show(struct seq_file *m, void *v)
{
	struct vc8x0_dev *dev;
	struct list_head *list;
	int i;

	if (vc8x0_devcount == 0)
		return 0;

	/* For each vc8x0 in the system */
	list_for_each(list, &vc8x0_devlist) {
		dev = list_entry(list, struct vc8x0_dev, devlist);

		/* Show channel metrics */
		if (IS_820E(dev)) {
			for (i = 0; i < MAX_DMA_CHANNELS; i++)
				vc8x0_channel_state_procfs(m, &dev->dma_channel[ i ]);
		}

		if (IS_845E(dev)) {
			vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL1 ]);
			vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL2 ]);
			if (!vc8x0_dual_1080p60) {
				vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL3 ]);
				vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL4 ]);
			}
			vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL5 ]);
			vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL6 ]);
			if (!vc8x0_dual_1080p60) {
				vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL7 ]);
				vc8x0_channel_state_procfs(m, &dev->dma_channel[ DMA_CHANNEL8 ]);
			}
		}

	}

	return 0;
}

static int vc8x0_proc_show(struct seq_file *m, void *v)
{
	struct vc8x0_dev *dev;
	struct list_head *list;
	int i = 0;
	u32 frmmax_a = 1, frmmax_b = 1, irqmax; /* Never div by zero */

	if (vc8x0_devcount == 0)
		return 0;

	/* For each vc8x0 in the system */
	list_for_each(list, &vc8x0_devlist) {
		dev = list_entry(list, struct vc8x0_dev, devlist);
		seq_printf(m, "%s = %p\n", dev->name, dev);

		if (vc8x0_procfs_verbosity & 0x02) {

			seq_printf(m, "dev->m_nInterruptMask1 = %08x\n", dev->m_nInterruptMask1);
			seq_printf(m, "dev->m_nInterruptMask2 = %08x\n", dev->m_nInterruptMask2);

			seq_printf(m, "GN4124: Full PCI Register Dump:\n");
			for (i = 0x0000; i <= 0x0004; i += 4)
				seq_printf(m, " 0x%04x = %08x\n", i, gn_read(i));
			for (i = 0x0800; i <= 0x0880; i += 4)
				seq_printf(m, " 0x%04x = %08x\n", i, gn_read(i));
			for (i = 0x0900; i <= 0x0928; i += 4)
				seq_printf(m, " 0x%04x = %08x\n", i, gn_read(i));
			for (i = 0x0a00; i <= 0x0a2c; i += 4)
				seq_printf(m, " 0x%04x = %08x\n", i, gn_read(i));
			for (i = 0x0b00; i <= 0x0b30; i += 4)
				seq_printf(m, " 0x%04x = %08x\n", i, gn_read(i));
			for (i = 0x0e00; i <= 0x0e04; i += 4)
				seq_printf(m, " 0x%04x = %08x\n", i, gn_read(i));

			seq_printf(m, "FLEX: Control:\n");
			for (i = 0x0; i <= 0x60; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 0 control:\n");
			for (i = 0x0100; i < 0x180; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 1 control:\n");
			for (i = 0x0200; i < 0x280; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 2 control:\n");
			for (i = 0x0300; i < 0x380; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 3 control:\n");
			for (i = 0x0400; i < 0x480; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Audio 0 control:\n");
			for (i = 0x0900; i < 0x97f; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 0 Dump:\n");
			for (i = 0x4c80; i < 0x4ea0; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 1 Dump:\n");
			for (i = 0x4ea0; i < 0x5040; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 2 Dump:\n");
			for (i = 0x5040; i < 0x51e0; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Video 3 Dump:\n");
			for (i = 0x51e0; i < 0x5380; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Audio 5 Dump (SDI):\n");
			for (i = 0x5380; i < 0x5470; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Audio 6 Dump (SDI):\n");
			for (i = 0x5740; i < 0x5560; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Audio 7 Dump (SDI):\n");
			for (i = 0x5560; i < 0x5650; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Audio 8 Dump (SDI):\n");
			for (i = 0x5650; i < 0x5740; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Audio 5 Dump (unbalanced):\n");
			for (i = 0x5740; i <= 0x5814; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FPGA: Audio Clocks:\n");
			for (i = 0x0d00; i <= 0x0d20; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FLEX: Interrupts:\n");
			for (i = 0x0e00; i <= 0x0e10; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			seq_printf(m, "FPGA: Full Register Dump:\n");
			for (i = 0x00a0; i <= 0x00a8; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

			/* Interrupts */
			for (i = 0x0e00; i <= 0x0e10; i += 4)
				seq_printf(m, " 0x%04x = %08x\n",
					i, vc_read32(i));

#if ENABLE_MONITOR_REGISTERS
			/* Monitor Registers */
			seq_printf(m, "FPGA: Monitor Registers:\n");
			for (i = 0; i < MAX_MONITOR_REGISTERS; i++)
				seq_printf(m, " %02d = %08x\n",
					i, dev->monitor_regs[i]);
#endif
		}

		if (IS_820E(dev)) {
			/* Check which channel is running when calculate the maximum amount of irq latency
			 * given that each channel could have a different frame rate.
			 */
			if (dev->dma_channel[ DMA_CHANNEL1 ].ad7441_ctx.fmt)
				frmmax_a = dev->dma_channel[ DMA_CHANNEL1 ].ad7441_ctx.fmt->frate;

			if (dev->dma_channel[ DMA_CHANNEL2 ].ad7441_ctx.fmt)
				frmmax_b = dev->dma_channel[ DMA_CHANNEL2 ].ad7441_ctx.fmt->frate;

			if (frmmax_a > frmmax_b)
				irqmax = (1000 / frmmax_a) + 1;
			else
				irqmax = (1000 / frmmax_b) + 1;

			/* Show Channel Statistics */
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL1 ]);
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL2 ]);
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL9 ]);
		}
		if (IS_845E(dev)) {
			/* Show Channel Statistics */
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL1 ]);
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL2 ]);
			if (!vc8x0_dual_1080p60) {
				vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL3 ]);
				vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL4 ]);
			}
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL5 ]);
			vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL6 ]);
			if (!vc8x0_dual_1080p60) {
				vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL7 ]);
				vc8x0_channel_statistics_procfs(m, &dev->dma_channel[ DMA_CHANNEL8 ]);
			}
		}
	}

	return 0;
}

static int vc8x0_proc_open(struct inode *inode, struct file *filp)
{
	return single_open(filp, vc8x0_proc_show, NULL);
}

static int vc8x0_proc_state_open(struct inode *inode, struct file *filp)
{
	return single_open(filp, vc8x0_proc_state_show, NULL);
}

static struct file_operations vc8x0_proc_fops = {
	.open		= vc8x0_proc_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

static struct file_operations vc8x0_proc_state_fops = {
	.open		= vc8x0_proc_state_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

static int vc8x0_proc_create(void)
{
	struct proc_dir_entry *pe;

	pe = proc_create("vc8x0", S_IRUGO, NULL, &vc8x0_proc_fops);
	if (!pe)
		return -ENOMEM;

	pe = proc_create("vc8x0-state", S_IRUGO, NULL, &vc8x0_proc_state_fops);
	if (!pe)
		return -ENOMEM;

	return 0;
}
#endif

static int vc8x0_thread_function(void *data)
{
	struct vc8x0_dev *dev = data;
	struct vc8x0_dma_channel *channel;
	struct vc8x0_dma_channel *aud_channel = 0;
#if ENABLE_AUDIO_KEEPALIVE && ENABLE_ALSA
	struct vc8x0_audio_dev *chip;
	u64 snd_time = 0;
#endif

//	u64 last_led_strobe = 0;

	dprintk(1, "%s() Started\n", __func__);

	channel = &dev->dma_channel[ DMA_CHANNEL1 ];

	if (IS_820E(dev))
		aud_channel = &dev->dma_channel[ DMA_CHANNEL9 ];
	if (IS_845E(dev))
		aud_channel = &dev->dma_channel[ DMA_CHANNEL5 ];

	msleep(2000);

	set_freezable();

	while (1) {
		msleep_interruptible(100);

		if (kthread_should_stop())
			break;

		try_to_freeze();

		if (vc8x0_thread_active == 0)
			continue;

		/* Let the channels do any thread processing */
		if (IS_820E(dev)) {
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL1 ]);
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL2 ]);
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL9 ]);
		}

		if (IS_845E(dev)) {
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL1 ]);
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL2 ]);
			if (!vc8x0_dual_1080p60) {
				vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL3 ]);
				vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL4 ]);
			}
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL5 ]);
			vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL6 ]);
			if (!vc8x0_dual_1080p60) {
				vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL7 ]);
				vc8x0_channel_thread(&dev->dma_channel[ DMA_CHANNEL8 ]);
			}
		}

#if ENABLE_ALSA && ENABLE_AUDIO_KEEPALIVE
		/* The PCM audio subsystem throws this messages:
		 * ALSA sound/core/pcm_lib.c:1765: capture write error (DMA or IRQ trouble?)
		 * when no audio is delivered for 10 seconds. It basically means it's worker thread
		 * didn't receive a notification with 10 seconds. The message is poor.
		 * In terms of the vc8x0 driver this message will appear by default if the HDMI
		 * cable is disconnected for > 10 seconds, and it will appear every 10 seconds.
		 * If you don't want this IRQ message to appear then set ENABLE_AUDIO_KEEPLIVE=1
		 */

		/* keep the audio channel alive by sending a zero length packet every 5 seconds.
		 * Technically we don't ajust the pcm read/write buffers so the effect is we're
		 * telling the PCM subsystem to process a buffer which has nothing to process.
		 * It's a reasonable workaround.
		 */
		if ((channel->nr == 1) && (channel->state == STATE_RUNNING)) {
			
			chip = aud_channel->audio_dev;
			if (chip->users > 0) {
				/* We only pulse the audio system if the device is open, else
				 * it throws a backtrace and a warning.
				 */
				if (time_after(jiffies, (long)snd_time + msecs_to_jiffies(500))) {
					snd_pcm_period_elapsed(chip->capture_pcm_substream);
					snd_time = jiffies;
				}
			}

		}
#endif
		/* Other parts of the driver need to guarantee that
		 * various 'keep alives' aren't happening. We'll
		 * prevent race conditions by allowing the
		 * rest of the driver to dictate when
		 * this keepalives can occur.
		 */
		mutex_lock(&dev->kthread_lock);

		mutex_unlock(&dev->kthread_lock);
	}

	vc8x0_thread_active = 0;
	dprintk(1, "%s() Stopped\n", __func__);
	return 0;
}

#if ENABLE_BAD_READS
void bad_read_work_handler(struct work_struct *work)
{
        struct vc8x0_dev *dev =
            container_of(work, struct vc8x0_dev, bad_read_workhandler);
	u32 val;

	if (dev->interrupts_enabled == 0)
		return;

	val = vc_read32(0x4d00);
	if (val != dev->bad_read_prior_val) {
		printk(KERN_ERR "%s() read %x expected %x\n", __func__, val, dev->bad_read_prior_val);
	}
	dev->bad_read_prior_val = val;

	schedule_work(&dev->bad_read_workhandler);
}
#endif

#define RESET_HOLD 1
#define RESET_STROBE 2
static void vc8x0_reset(struct vc8x0_dev *dev, int action)
{
	u32 v;

	/* Warm reset the FPGA? */
	/* Assert RSTOUT */
	v = gn_read(GN4124_PCI_SYS_CFG_SYSTEM);
	v &= ~0x4000;
	gn_write(GN4124_PCI_SYS_CFG_SYSTEM, v);

	if (action == RESET_STROBE) {
		/* De-assert RSTOUT */
		v |= 0x4000;
		gn_write(GN4124_PCI_SYS_CFG_SYSTEM, v);
	}
}

static int __devinit vc8x0_initdev(struct pci_dev *pci_dev,
	const struct pci_device_id *pci_id)
{
	struct vc8x0_dev *dev;
	u64 v = 0xf0e0d0c0b0a09080;
	int err;

	dev = kzalloc(sizeof(*dev), GFP_KERNEL);
	if (NULL == dev)
		return -ENOMEM;

	/* pci init */
	dev->pci = pci_dev;
	if (pci_enable_device(pci_dev)) {
		err = -EIO;
		goto fail_unreg;
	}

	/* print pci info */
	pci_read_config_byte(pci_dev, PCI_CLASS_REVISION, &dev->pci_rev);
	pci_read_config_byte(pci_dev, PCI_LATENCY_TIMER,  &dev->pci_lat);
	printk(KERN_INFO "vc8x0 device found at %s, rev: %d, irq: %d, latency: %d\n",
		pci_name(pci_dev), dev->pci_rev, pci_dev->irq,
		dev->pci_lat);

	printk(KERN_INFO "vc8x0 BAR0 mmio: 0x%llx [0x%x bytes]\n",
		(unsigned long long)pci_resource_start(pci_dev, BAR0),
		(unsigned int)pci_resource_len(pci_dev, BAR0));

	printk(KERN_INFO "vc8x0 BAR4 mmio: 0x%llx [0x%x bytes]\n",
		(unsigned long long)pci_resource_start(pci_dev, BAR4),
		(unsigned int)pci_resource_len(pci_dev, BAR4));

	pci_set_master(pci_dev);
	if (!pci_dma_supported(pci_dev, 0xffffffff)) {
		printk("%s/0: Oops: no 32bit PCI DMA ???\n", dev->name);
		err = -EIO;
		goto fail_irq;
	}

	/* MAP the PCIe space, register i2c, program any PCIe quirks. */
	if (vc8x0_dev_setup(dev) < 0) {
		err = -EINVAL;
		goto fail_unreg;
	}

	if ((msi_enable) && (!pci_enable_msi(pci_dev))) {
		err = request_irq(pci_dev->irq, vc8x0_irq,
			IRQF_DISABLED, dev->name, dev);
	} else {
		err = request_irq(pci_dev->irq, vc8x0_irq,
			IRQF_SHARED | IRQF_DISABLED, dev->name, dev);
	}
	if (err < 0) {
		printk(KERN_ERR "%s: can't get IRQ %d\n",
		       dev->name, pci_dev->irq);
		goto fail_irq;
	}

	/* Card specific tweaks with subsystems etc */
	vc8x0_card_setup(dev);

	pci_set_drvdata(pci_dev, dev);

	/* Warm reset the FPGA? */
	vc8x0_reset(dev, RESET_STROBE);
	
	/* Load the FPGA firmware */
	if (vc8x0_fw_load(dev) < 0) {
		printk(KERN_ERR "%s() FPGA failed to load\n", __func__);
		err = -EIO;
		goto fail_irq;
	}
	msleep(50);

	printk(KERN_ERR "%s() About to read the firmware version register...\n", __func__);

	/* After power on set all the register defaults */
	/* O820 = 0x000000a0 */
	/* 0845 = 0x00010018 */
	v = vc_read32(0x00a0);
	printk(KERN_ERR "%s() Release Version Value: 0x%llx\n", __func__, v);
	if (v == 0xffffffff) {
		printk(KERN_ERR "%s() FPGA failed to initialize\n", __func__);
		err = -EIO;
		goto fail_irq;
	}

	if (vc8x0_i2c_register(&dev->i2c_bus) < 0)
		printk(KERN_ERR "%s() Error registering the primary i2c bus\n", __func__);

	if (IS_820E(dev)) {
		/* DMA Channels */
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL1 ], 1) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 1\n", __func__);
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL2 ], 2) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 2\n", __func__);
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL9 ], 9) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 9\n", __func__);
	}

	if (IS_845E(dev)) {
		/* Video DMA Channels */
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL1 ], 1) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 1\n", __func__);
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL2 ], 2) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 2\n", __func__);

		if (!vc8x0_dual_1080p60) {
			if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL3 ], 3) < 0)
				printk(KERN_ERR "%s() Failed to initialize DMA channel 3\n", __func__);
			if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL4 ], 4) < 0)
				printk(KERN_ERR "%s() Failed to initialize DMA channel 4\n", __func__);
		}

		/* Audio DMA Channels */
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL5 ], 5) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 5\n", __func__);
		if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL6 ], 6) < 0)
			printk(KERN_ERR "%s() Failed to initialize DMA channel 6\n", __func__);
		if (!vc8x0_dual_1080p60) {
			if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL7 ], 7) < 0)
				printk(KERN_ERR "%s() Failed to initialize DMA channel 7\n", __func__);
			if (vc8x0_channel_register(dev, &dev->dma_channel[ DMA_CHANNEL8 ], 8) < 0)
				printk(KERN_ERR "%s() Failed to initialize DMA channel 8\n", __func__);
		}
	}

	/* System level I2C devices */
	if (vc8x0_eeprom_init(&dev->i2c_bus, &dev->eeprom_ctx, 0xa0) < 0)
		printk(KERN_ERR "%s() Error detecting the eeprom\n", __func__);
	if (vc8x0_eeprom_enable(&dev->eeprom_ctx) < 0)
		printk(KERN_ERR "%s() Error initializing the eeprom\n", __func__);
	if (vc8x0_eeprom_unlockFPGA(&dev->eeprom_ctx) < 0)
		printk(KERN_ERR "%s() Error initializing the eeprom/FPGA\n", __func__);

	if (vc8x0_dma_loadsequencer(dev) == 0) {
		if (vc8x0_dma_sequencer_start(dev) < 0) {
			printk(KERN_ERR "%s() Failed to start DMA sequencer\n", __func__);
		}
	} else
		printk(KERN_ERR "%s() Failed to initialize DMA sequencer\n", __func__);

	/* Enable GN4124 GPIO Interrupts */
	/* set GPIO 0 as an interrupt source input (GPIO 0 - 9 inputs GPIO 10 - 15 outputs) */
	gn_write(GN4124_GPIO_DIRECTION_MODE, 0x0000001);
        /* disable GPIO as outputs - GPIOs enable */
        gn_write(GN4124_GPIO_OUTPUT_ENABLE, 0x00000001);
        /* set logic analyzer signals to zero, turn off the LEDs in the GPIO register */
        gn_write(GN4124_GPIO_OUTPUT_VALUE, 0x00000001);
	/* End - Enable GN4124 GPIO Interrupts */

	/* Put this in a global list so we can track multiple boards */
	mutex_lock(&devlist);
	list_add_tail(&dev->devlist, &vc8x0_devlist);
	mutex_unlock(&devlist);

	dev->kthread = kthread_run(vc8x0_thread_function, dev, "vc8x0 housekeeping");
	if (!dev->kthread) {
		printk(KERN_ERR "%s() Failed to create "
			"kernel thread\n", __func__);
	} else
		dprintk(1, "%s() Created the housekeeping thread\n", __func__);

	dev->interrupts_enabled = 1;

	vc_write32(0x0d00, 0);
	vc_write32(0x0d08, 0);
	vc_write32(0x0d10, 3);
	vc_write32(0x0d18, 3);

#if ENABLE_BAD_READS
	INIT_WORK(&dev->bad_read_workhandler, bad_read_work_handler);
	schedule_work(&dev->bad_read_workhandler);
#endif
	return 0;

fail_irq:
	/* Assert RSTOUT  - Put the FPGA in Reset? */
	vc8x0_reset(dev, RESET_HOLD);
	vc8x0_dev_unregister(dev);
fail_unreg:
	kfree(dev);
	return err;
}

static void __devexit vc8x0_finidev(struct pci_dev *pci_dev)
{
	struct vc8x0_dev *dev = pci_get_drvdata(pci_dev);
	int i = 0;

	dev->interrupts_enabled = 0;

	if (dev->kthread) {
		kthread_stop(dev->kthread);
		dev->kthread = NULL;

		while (vc8x0_thread_active) {
			msleep(500);
			if (i++ > 8)
				break;
		}
	}

	if (vc8x0_eeprom_disable(&dev->eeprom_ctx) < 0)
		printk(KERN_ERR "%s() Error disabling the eeprom\n",
			__func__);

	gn_write(GN4124_GPIO_INT_DISABLE, 0x00000001);
	vc8x0_dma_sequencer_stop(dev);
	vc8x0_shutdown(dev);

	if (vc8x0_i2c_unregister(&dev->i2c_bus) < 0)
		printk(KERN_ERR "%s() Error unregistering the primary i2c bus\n", __func__);

	if (IS_820E(dev)) {
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL1 ]);
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL2 ]);
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL9 ]);
	}

	if (IS_845E(dev)) {
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL1 ]);
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL2 ]);
		if (!vc8x0_dual_1080p60) {
			vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL3 ]);
			vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL4 ]);
		}
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL5 ]);
		vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL6 ]);
		if (!vc8x0_dual_1080p60) {
			vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL7 ]);
			vc8x0_channel_unregister(&dev->dma_channel[ DMA_CHANNEL8 ]);
		}
	}

	pci_disable_device(pci_dev);

	/* unregister stuff */
	free_irq(pci_dev->irq, dev);
	if (msi_enable)
		pci_disable_msi(pci_dev);

	mutex_lock(&devlist);
	list_del(&dev->devlist);
	mutex_unlock(&devlist);

	/* Assert RSTOUT  - Put the FPGA in Reset? */
	vc8x0_reset(dev, RESET_HOLD);
	
	vc8x0_dev_unregister(dev);

	kfree(dev);
}

static struct pci_device_id vc8x0_pci_tbl[] = {
	{
		/* VC8X0 */
		.vendor       = 0x1a39,
		.device       = 0x0004,
		.subvendor    = PCI_ANY_ID,
		.subdevice    = PCI_ANY_ID,
	}, {
		/* --- end of list --- */
	}
};
MODULE_DEVICE_TABLE(pci, vc8x0_pci_tbl);

static struct pci_driver vc8x0_pci_driver = {
	.name     = "vc8x0",
	.id_table = vc8x0_pci_tbl,
	.probe    = vc8x0_initdev,
	.remove   = __devexit_p(vc8x0_finidev),
	/* TODO */
	.suspend  = NULL,
	.resume   = NULL,
};

static int __init vc8x0_init(void)
{
	printk(KERN_INFO "vc8x0 driver version %d.%d.%d loaded\n",
	       (VC8X0_VERSION_CODE >> 16) & 0xff,
	       (VC8X0_VERSION_CODE >>  8) & 0xff,
	       VC8X0_VERSION_CODE & 0xff);
#ifdef SNAPSHOT
	printk(KERN_INFO "vc8x0: snapshot date %04d-%02d-%02d\n",
	       SNAPSHOT/10000, (SNAPSHOT/100)%100, SNAPSHOT%100);
#endif
#ifdef CONFIG_PROC_FS
	vc8x0_proc_create();
#endif
	return pci_register_driver(&vc8x0_pci_driver);
}

static void __exit vc8x0_fini(void)
{
#ifdef CONFIG_PROC_FS
	remove_proc_entry("vc8x0", NULL);
	remove_proc_entry("vc8x0-state", NULL);
#endif
	pci_unregister_driver(&vc8x0_pci_driver);
}

module_init(vc8x0_init);
module_exit(vc8x0_fini);

