/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int eeprom_debug;
module_param(eeprom_debug, int, 0644);
MODULE_PARM_DESC(eeprom_debug, "enable debug messages [eeprom]");

#define dprintk(level, fmt, arg...)\
	do { if (eeprom_debug >= level)\
		printk(KERN_DEBUG "%s/0: " fmt, dev->name, ## arg);\
	} while (0)

#define DS28CN01_SUBADDR_EEPROM_0 0x00
#define DS28CN01_SUBADDR_UID      0xA0
#define DS28CN01_SUBADDR_MODE     0xA8
#define DS28CN01_SUBADDR_COMMAND  0xA9
#define DS28CN01_SUBADDR_MAC      0xB0

static int GetUid(struct vc8x0_eeprom_context *ctx, u8 *pucUid, int len);
static int configureFPGA(struct vc8x0_eeprom_context *ctx);
static int ReadPage0(struct vc8x0_eeprom_context *ctx);

static int vc8x0_eeprom_i2c_read8(struct vc8x0_eeprom_context *ctx, u8 reg, u8 *val)
{
	int ret;

	ret = vc8x0_I2cReadRegister(ctx->bus, ctx->addr, reg, val);
	return ret;
}

static int vc8x0_eeprom_i2c_write8(struct vc8x0_eeprom_context *ctx, u8 reg, u8 val)
{
	int ret;

	ret = vc8x0_I2cWriteRegister(ctx->bus, ctx->addr, reg, val);
	return ret;
}

static int vc8x0_eeprom_i2c_readblock(struct vc8x0_eeprom_context *ctx, u8 reg, u8 *buf, int len)
{
	int ret;

	ret = vc8x0_I2cReadRegisterBlock(ctx->bus, ctx->addr, reg, buf, len);

	return ret;
}

static int vc8x0_eeprom_i2c_writeblock(struct vc8x0_eeprom_context *ctx, u8 reg, u8 *buf, int len)
{
	int ret;

	ret = vc8x0_I2cWriteRegisterBlock(ctx->bus, ctx->addr, reg, buf, len);

	return ret;
}

int vc8x0_eeprom_init(struct vc8x0_i2c *bus, struct vc8x0_eeprom_context *ctx, u8 i2caddr)
{
	struct vc8x0_dev *dev = bus->dev;
	dprintk(1, "%s(addr = 0x%02x)\n", __func__, i2caddr);

	memset(ctx, 0, sizeof(*ctx));
	ctx->bus = bus;
	ctx->addr = i2caddr;

	return 0;
}

int vc8x0_eeprom_enable(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	int i;
	u8 mode;

	dprintk(1, "%s(base = %04x)\n", __func__, ctx->bus->reg_base);

	vc8x0_eeprom_i2c_read8(ctx, DS28CN01_SUBADDR_MODE, &mode);
        dprintk(1, "Power-on communications mode is 0x%02x", mode);
        mode = 0x00;
        vc8x0_eeprom_i2c_write8(ctx, DS28CN01_SUBADDR_MODE, mode);
	vc8x0_eeprom_i2c_read8(ctx, DS28CN01_SUBADDR_MODE, &mode);
        dprintk(1, "New communications mode is 0x%02x", mode);

	for (i = 0; i < sizeof(ctx->sn); i++) {
		vc8x0_eeprom_i2c_read8(ctx, 0x20 + i, &ctx->sn[i]);
	}

	/* Read the entire page0 of the eeprom for later use */
	if (ReadPage0(ctx) < 0) {
		printk(KERN_ERR "%s() Unable to query eeprom page 0\n", __func__);
		return -1;
	}

	printk(KERN_INFO "%s/0: ViewCast O8x0 card present, serial# %s\n", dev->name, ctx->sn);

	return 0;
}

int vc8x0_eeprom_disable(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	dprintk(1, "%s()\n", __func__);
	return 0;
}

int vc8x0_eeprom_unlockFPGA(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;

	dprintk(1, "%s(%p)\n", __func__, ctx);

	if (GetUid(ctx, &ctx->uid[0], sizeof(ctx->uid)) < 0) {
		printk(KERN_ERR "%s() UID is invalid, unable to configure FPGA.\n", __func__);
		return -1;
	}

	if (configureFPGA(ctx) < 0) {
		printk(KERN_ERR "%s() Unable to configure the FPGA.\n", __func__);
		return -1;
	}

	return 0;
}

static int ReadPage0(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	int i;
	dprintk(1, "%s()\n", __func__);

	for (i = 0; i < sizeof(ctx->page0); i++)
		vc8x0_eeprom_i2c_read8(ctx, DS28CN01_SUBADDR_EEPROM_0 + i, &ctx->page0[i]);

	return 0;
}

#define DS_BUS_BUSY         0x02
#define DS_COMMAND_COMPLETE 0x04

static int WaitForReady(struct vc8x0_eeprom_context *ctx, u32 delay, u32 tries)
{
	u32 c = 0;
	u8 mode = 0xff;

	while (++c < tries) {
		msleep(delay);
		vc8x0_eeprom_i2c_read8(ctx, DS28CN01_SUBADDR_MODE, &mode);
		if ((mode & DS_BUS_BUSY) == 0)
			return 0;
	}

        return -1;
}

static int GetMACFromChip(struct vc8x0_eeprom_context *ctx,
	u8* pucChallenge, int clen, u8* pucMac, int mlen)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	u8 cmd[9];
	u8 response[20];

	/* Must use I2C block read/writes here, not reg read writes,
	 * or the SHA1 crypto block in the DS28 gets confused and the
	 * results are inaccurate.
	 */

	dprintk(1, "%s()\n", __func__);

	if ((clen != 7) || (mlen != 20))
		return -1;

	memset(pucMac, 0, mlen);
	memset(response, 0, sizeof(response));

	cmd[0] = 0xD0;
	memcpy(&cmd[1], pucChallenge, clen);

	dprintk(1, "%s() CMD# %02x %02x %02x %02x %02x %02x %02x %02x %02x\n", __func__,
		cmd[0], cmd[1], cmd[2], cmd[3], cmd[4], cmd[5], cmd[6], cmd[7], cmd[8]);
 
	vc8x0_eeprom_i2c_writeblock(ctx, DS28CN01_SUBADDR_COMMAND, &cmd[0], 9);

	/* Expected compute time is 2 msec. */
	if (WaitForReady(ctx, 100, 10) != 0) {
		return -1;
	}

	vc8x0_eeprom_i2c_readblock(ctx, DS28CN01_SUBADDR_MAC, &response[0], sizeof(response));

	dprintk(1, "%s() MAC# %02x %02x %02x %02x %02x %02x %02x %02x\n", __func__,
		response[0x00], response[0x01], response[0x02], response[0x03],
		response[0x04], response[0x05], response[0x06], response[0x07]);
	dprintk(1, "%s()      %02x %02x %02x %02x %02x %02x %02x %02x\n", __func__,
		response[0x08], response[0x09], response[0x0a], response[0x0b],
		response[0x0c], response[0x0d], response[0x0e], response[0x0f]);
	dprintk(1, "%s()      %02x %02x %02x %02x\n", __func__,
		response[0x10], response[0x11], response[0x12], response[0x13]);

	memcpy(pucMac, response, mlen);

	return 0;
}

static int WaitForShaChallenge(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	u32 status, c = 0;
	int ret = 0;

	/* Wait for a minimum of 2ms to begin with */
	msleep(2);

	while (1) {
		status = vc_read32(FPGA_SHA_STATUS_REG);

		if ((status & FPGA_SHA_RAND_CHLNG_DONE) == FPGA_SHA_RAND_CHLNG_DONE) {
			break;
		}

		msleep(20);
		if (c++ > 10) {
			ret = -1;
			break;
		}
	}
	return ret;
}

static int WaitForShaComplete(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	u32 status, c = 0;
	int ret = 0;

	dprintk(1, "%s()\n", __func__);

	/* Wait for a minimum of 2ms to begin with */
	msleep(2);

	while (1) {
		status = vc_read32(FPGA_SHA_STATUS_REG);
		if ((status & FPGA_SHA_HASH_COMPARE_CMD) == FPGA_SHA_HASH_COMPARE_CMD) {
			break;
		}

		msleep(20);
		if (c++ > 10) {
			ret = -1;
			break;
		}
	}
	return ret;
}

static int configureFPGA(struct vc8x0_eeprom_context *ctx)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	u32 paddeduid[2];
	u32 page0[8];
	u32 challenge[2];
	u8 challengersp[8];
	u8 mac[20];
	u32 paddedmac[5];
	int i;

	dprintk(1, "%s()\n", __func__);

	memset(mac, 0, sizeof(mac));

	/* Convert bytes into dwords */
	for (i = 0; i < 2; i++) {
		paddeduid[i] =
			(ctx->uid[(i * 4) + 3] << 24) |
			(ctx->uid[(i * 4) + 2] << 16) |
			(ctx->uid[(i * 4) + 1] <<  8) |
			(ctx->uid[(i * 4) + 0] <<  0);
		dprintk(1, "%s() paddeduid[%d] = 0x%08x\n", __func__, i, paddeduid[i]);
	}
	/* on dword 1 shave off byte 7 (which is a crc) */
	paddeduid[1] &= 0x00FFFFFF;

	for (i = 0; i < 8; i++) {
		page0[i] =
			(ctx->page0[(i * 4) + 3] << 24) |
			(ctx->page0[(i * 4) + 2] << 16) |
			(ctx->page0[(i * 4) + 1] <<  8) |
			(ctx->page0[(i * 4) + 0] <<  0);
		dprintk(1, "%s() page0[%d] = 0x%08x\n", __func__, i, page0[i]);
	}

	/* Reset random challenge */
	vc_write32(FPGA_SHA_COMMAND_REG, FPGA_SHA_CLR_RAND_CHLNG_CMD);

	/* Write out the security chip UID to FPGA */
	for (i = 0; i < 2; i++) {
		vc_write32(FPGA_SHA_SECURITY_UID_REG_1 + (i * 4), paddeduid[i]);
		dprintk(1, "%s() 0x%08x <- 0x%08x\n", __func__, FPGA_SHA_SECURITY_UID_REG_1 + (i * 4), paddeduid[i]);
	}

	/* Write out eeprom page 0 to FPGA */
	for (i = 0; i < 8; i++) {
		vc_write32(FPGA_SHA_EEPROM_DATA_REG_0 + (i * 4), page0[i]);
		dprintk(1, "%s() 0x%08x <- 0x%08x\n", __func__, FPGA_SHA_EEPROM_DATA_REG_0 + (i * 4), page0[i]);
	}

	/* Write command to request random challenge */
	vc_write32(FPGA_SHA_COMMAND_REG, FPGA_SHA_RAND_CHLNG_CMD);

	/* Wait for it to finish */
	if (WaitForShaChallenge(ctx) == 0) {
		/* And now read the challenge generated by the FPGA */
		for (i = 0; i < 2; i++) {
			challenge[i] = vc_read32(FPGA_SHA_CHALLENGE_REG_1 + (i * 4));
			challengersp[(i * 4) + 0] = challenge[i] & 0xff;
			challengersp[(i * 4) + 1] = (challenge[i] >> 8) & 0xff;
			challengersp[(i * 4) + 2] = (challenge[i] >> 16) & 0xff;
			challengersp[(i * 4) + 3] = (challenge[i] >> 24) & 0xff;

			dprintk(1, "%s() challenge[%d] 0x%08x = 0x%08x\n", __func__, i, FPGA_SHA_CHALLENGE_REG_1 + (i * 4), challenge[i]);
		}
		if (GetMACFromChip(ctx, &challengersp[0], 7, &mac[0], sizeof(mac)) != 0) {
			printk(KERN_ERR "%s() Unable to acquire MAC\n", __func__);
			return -1;
		}

		/* Write it out to the FPGA */
		for (i = 0; i < 5; i++) {
			paddedmac[i] =
				(mac[(i * 4) + 3] << 24) |
				(mac[(i * 4) + 2] << 16) |
				(mac[(i * 4) + 1] <<  8) |
				(mac[(i * 4) + 0] <<  0);
			vc_write32(FPGA_SHA_HASH_REG_1 + (i * 4), paddedmac[i]);
			dprintk(1, "%s() 0x%08x <- 0x%08x\n", __func__, FPGA_SHA_HASH_REG_1 + (i * 4), paddedmac[i]);
		}

		/* Start the hash compare processing in the FPGA */
		vc_write32(FPGA_SHA_COMMAND_REG, FPGA_SHA_HASH_COMPARE_CMD);

		/* Wait for completion */
		if (WaitForShaComplete(ctx) == 0) {
			dprintk(1, "%s() FPGA SHA compare SUCCEEDED\n", __func__);
		} else {
			dprintk(1, "%s() FPGA SHA compare FAILED\n", __func__);
		}
#if 0
                CreateOptionsFile();
#endif
	}

	return 0;
}

/* DS2Ds28cn01460 UID crc - x8 + x5 + x4 + 1 */
static u8 CalcCrc(u8 *puc, int len)
{
	u64 ull = *(u64*)puc;
	u8 ucCrc = 0;
	u8 ucXor;
	int i;

	if (len != 8)
		return -1;

	for (i = 0; i < 56; i++) {
		ucXor = (((u8)(ull & 1)) ^ ucCrc) << 7;
		if (ucXor)
			ucCrc ^= 0x18;
		ucCrc = (ucCrc >> 1) | ucXor;
		ull >>= 1;
	}

	return ucCrc;
}

static int CheckCrc(u8 *puc, int len)
{
	if (len != 8)
		return -1;

        return (CalcCrc(puc, len) == puc[7]) ? 0 : -1;
}

static int GetUid(struct vc8x0_eeprom_context *ctx, u8 *pucUid, int len)
{
	struct vc8x0_dev *dev = ctx->bus->dev;
	int i;

	if (len != 8)
		return -1;

	for (i = 0; i < len; i++)
		vc8x0_eeprom_i2c_read8(ctx, DS28CN01_SUBADDR_UID + i, pucUid + i);

	dprintk(1, "%s() UID# %02x %02x %02x %02x %02x %02x %02x %02x\n", __func__,
		*(pucUid + 0), *(pucUid + 1), *(pucUid + 2), *(pucUid + 3),
		*(pucUid + 4), *(pucUid + 5), *(pucUid + 6), *(pucUid + 7));

	return CheckCrc(pucUid, len);
}

