/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

#define CLK_DIV		(5)
#define FW_RETRY	(5)
#define TIME_OUT	(1000000)
#define DEV_FW_NAME	"S8_Rel_0_Ver_11.fpga"
#define DEV_FW_SIZE	(1286284)
#define PROD_FW_NAME	"v4l-osprey-820e-firmware-v1.5.0.rom"
#define PROD_FW_SIZE	(1932760)
#define SDI_FW_NAME	"v4l-osprey-845e-firmware-v1.0.18.rom"
#define SDI_FW_SIZE	(3104728)

static unsigned int fw_debug;
module_param(fw_debug, int, 0644);
MODULE_PARM_DESC(channel_debug, "enable debug messages [fw]");

unsigned int vc8x0_clk_div = CLK_DIV;
module_param(vc8x0_clk_div, int, 0644);
MODULE_PARM_DESC(vc8x0_clk_div, "Firmware clocking speed");

#define dprintk(level, fmt, arg...)\
	do { if (fw_debug >= level)\
		printk(KERN_DEBUG "%s: " fmt, dev->name, ## arg);\
	} while (0)

/* GN4124AppLayerInit - Application Layer and interrupt routing initialization */
void vc8x0_fw_GN4124AppLayerInit(struct vc8x0_dev *dev)
{
	u32 reg, i, targetFreq;

	/* Check PCI_SUB_VENDOR and PCI_SUB_ID to identify GN4124 / GN4121 */
	reg = gn_read(0x2C);
	if (reg == 0x00041A39) {
		targetFreq = 0x1F03C; /* 200 MHz */
	} else {
		/* 100MHz for GN4121 */
		targetFreq = 0x1F07C;
	}
	targetFreq = 0x1F03C; /* 200 MHz */
	
	dprintk(1, "%s() Setting the GN4124 Target Frequency to 0x%08x (%sMHz)", __func__,
		targetFreq,
		targetFreq == 0x1f03c ? "200" : "100");

	/* Check to see if PLL is setup correctly and locked */
	reg = gn_read(GN4124_CLK_CSR);
	if (reg != (0xE0000000 | targetFreq)) {

		/* Initialize local bus clock to 160Mhz (oscillator as reference) and enable the localbus */
		reg = gn_read(GN4124_PCI_SYS_CFG_SYSTEM);
		reg |= 0x00001000;
		reg &= ~0x0000C000;
		gn_write(GN4124_PCI_SYS_CFG_SYSTEM, reg);

		/* Setup PLL */
		gn_write(GN4124_CLK_CSR, (0x8000000 | targetFreq));
		reg = gn_read(GN4124_CLK_CSR);
		gn_write(GN4124_CLK_CSR, targetFreq);

		/* Wait until PLL & DLLs are locked or timeout occurs. */
		i = 0;
		do
		{
			reg = gn_read(GN4124_CLK_CSR);
			msleep(100);
		} while ((reg & 0xE0000000) == 0xE0000000 && (i++ < 20));
		/* Timeout occurs */
		if (i >= 20) {
			printk(KERN_ERR "%s() PLL & DLLs failed to lock.", __func__);
			return;
		}

		/* De-assert RSTOUT */
		reg = gn_read(GN4124_PCI_SYS_CFG_SYSTEM);
		reg |= 0x4000;
		gn_write(GN4124_PCI_SYS_CFG_SYSTEM, reg);

		/* Wait for 100us for RSTOUT to complete the state change and be ready for 
		 * local bus accessing
		 */
		mdelay(100);
	}

	/* Turn on virtual channel option */
//	reg = gn_read(GN4124_PCI_SYS_CFG_SYSTEM);
//	reg |= 0x00080000; 
//	gn_write(GN4124_PCI_SYS_CFG_SYSTEM, reg);

	vc_write32(DMA_STATUS_MASK, 0x0);
	vc_write32(APP_STATUS_MASK, 0x0);

	/* Set video formatter mode of operation */
//	reg = APP_CFG_YUY2 | APP_CFG_CAPTURE;
//	reg = APP_CFG_UYVY | APP_CFG_CAPTURE;
	reg = 0;

	vc_write32(APP_CFG, reg);

	/* Set max payload size in DMA */
	reg = gn_read(GN4124_PCIE_DCR);

	/* Units are number of DWORDS hence the >> 2 */
	dev->m_nMaxReadSize = (128 << (( reg >> 12) & 0x7)) >> 2;
	if (dev->m_nMaxReadSize == 128) {
		dev->m_nVideoBlockSize = 512;
//		m_nVideoBlockSize = 1024;
//		m_nAudioBlockSize = 256;
		dev->m_nAudioBlockSize = 64;
//		m_nAudioBlockSize = 32;
	} else {
		dev->m_nVideoBlockSize = 1024;
		dev->m_nAudioBlockSize = 32;
	}
		
	dev->m_nPCIePayloadSize = (128 << (( reg >> 5) & 0x7)) >> 2;
	reg = (dev->m_nMaxReadSize & 0x3ff) << 10 | (dev->m_nPCIePayloadSize & 0x3ff);
	dprintk(1, "%s() Max read size is %d, Max payload size is %d, Video block size is %d, Audio block size is %d",
		__func__,
		dev->m_nMaxReadSize,
		dev->m_nPCIePayloadSize,
		dev->m_nVideoBlockSize,
		dev->m_nAudioBlockSize);
	vc_write32(DMA_PAYLOAD_SIZE, reg);

	/* Enable DMA status */
//	vc_write32(DMA_STATUS_MASK, DMA_STATUS_MASK_LDM_IDLE);
	vc_write32(DMA_STATUS_MASK, DMA_STATUS_MASK_LDM_IDLE | DMA_STATUS_MASK_PD_ERR);

	/* Enable application status mask */
//	vc_write32(APP_STATUS_MASK, APP_STATUS_MASK_FIFO_FULL | APP_STATUS_MASK_FIFO_EMPTY | APP_STATUS_MASK_DVB);

	/* Clear DMA status */
	vc_write32(DMA_STATUS, DMA_STATUS_MASK_PDM_IDLE |
		DMA_STATUS_MASK_LDM_IDLE | DMA_STATUS_MASK_PD_ERR);

	/* Clear APP status */
//	vc_write32(APP_STATUS, APP_STATUS_MASK_FIFO_FULL | APP_STATUS_MASK_FIFO_EMPTY | APP_STATUS_MASK_DVB);

	/* Reset the PDM and LDM */
	vc_write32(DMA_CFG, DMA_CFG_PDM_FLUSH |
		DMA_CFG_LDM_FLUSH | DMA_CFG_CSR_ENABLE | DMA_CFG_PDM_ENABLE | DMA_CFG_LDM_ENABLE);
	mdelay(50);

	/* DMA_CFG_LDM_WORD_BUFFER_EN is set required if buffers are not multiptles of 64 bits */
	vc_write32(DMA_CFG, DMA_CFG_LDM_WORD_BUFFER_EN | DMA_CFG_CSR_ENABLE | DMA_CFG_PDM_ENABLE | DMA_CFG_LDM_ENABLE); 

	/* Assume INT0 will have the GPIO input as a interrupt source set edge sensitive inputs */
	vc_write32(VDMA_EVENT_CLR, VDMA_EVENT_ALL);
	gn_write(GN4124_INT_CTRL, 0x0);
//	gn_write(GN4124_INT_CTRL, 0x0000000F);

	/* Set GPIO 0 interrupt type to "Hi" level trigger (GPIO 0 - 8 "Hi level") */
	gn_write(GN4124_GPIO_INT_TYPE, 0x00000001);
	gn_write(GN4124_GPIO_INT_VALUE, 0x00000001);
//	gn_write(GN4124_GPIO_INT_TYPE, 0x01ff);
//	gn_write(GN4124_GPIO_INT_VALUE, 0x01ff);

	/* Mask all interrupt sources except gpio 0 */
	gn_write(GN4124_GPIO_INT_MASK, 0x0000fffe);

	/* Read current interrupt status to clear any pending interrupts */
	reg = gn_read(GN4124_GPIO_INT_STATUS);

	/* Clear the INT_CFG registers */
	gn_write(GN4124_INT_CFG1, 0x0);
	gn_write(GN4124_INT_CFG2, 0x0);
	gn_write(GN4124_INT_CFG3, 0x0);
	gn_write(GN4124_INT_CFG4, 0x0);
	gn_write(GN4124_INT_CFG5, 0x0);
	gn_write(GN4124_INT_CFG6, 0x0);
	gn_write(GN4124_INT_CFG7, 0x0);

	/* Clear all the interrupt sources */
	gn_write(GN4124_INT_STAT, ~(INT_STAT_SWI0 | INT_STAT_SWI0));

	/* Allow GPIO to generate interrupt to the MSI block */
//	gn_write(GN4124_INT_CFG0, INT_STAT_GPIO);
	gn_write(GN4124_INT_CFG0, GN4124_INTERRPT_MASK);
}

/* Bit-banging through FCL method */
int vc8x0_fw_FCLBitBangTransfer(struct vc8x0_dev *dev, u8 *buffer, u32 len, u32 nClkDiv)
{
	u32 i, j, k;
	u8 bitMask;
	u8 *pu8Ptr = buffer;
	int ret = 0;

	/* Set FCL_CLK_DIV */
	gn_write(GN4124_FCL_CLK_DIV, nClkDiv);

	/*  Reset FCL */
	i = 0x40;
	gn_write(GN4124_FCL_CTRL, i);

	/* Read back to ensure FCL block is operational */
	i = gn_read(GN4124_FCL_CTRL);
	if (i != 0x40) {
		printk(KERN_ERR "%s() fcl block failed\n", __func__);
		ret = -EINVAL;
		goto out;
	}

	i = 0x0;
	gn_write(GN4124_FCL_CTRL, i);

	/* Clear IRQ */
	i = gn_read(GN4124_FCL_IRQ);

	/*  Set FCL output enable */
	i = 0x17;
	gn_write(GN4124_FCL_EN, i);

	/*  Assert SPRI_CONFIG */
	i = 0x4;
	gn_write(GN4124_FCL_IODATA_OUT, i);
	i = 0x0;
	gn_write(GN4124_FCL_IODATA_OUT, i);

	/* Wait until both SPRI_STATUS & SPRI_DONE signals go low */
	j = 0;
	do {
		i = gn_read(GN4124_FCL_STATUS);
		if (j++ > TIME_OUT) {
			printk(KERN_ERR "%s() SPRI_STATUS and/or SPRI_DONE are stuck high!\n", __func__);
			ret = -EINVAL;
			goto out;
		}
	} while ((i & 0x28) != 0);

	/*  De-assert SPRI_CONFIG */
	i = 0x4;
	gn_write(GN4124_FCL_IODATA_OUT, i);

	/* Wait until SPRI_STATUS signal goes high */
	j = 0;
	do {
		i = gn_read(GN4124_FCL_STATUS);
		if (j++ > TIME_OUT) {
			printk(KERN_ERR "%s() SPRI_STATUS stuck low!\n", __func__);
			ret = -EINVAL;
			goto out;
		}
	} while ((i & 0x20) == 0);

	/* Clock out the first 6 bytes of the input buffer (Encription header)*/
	for (i = 0; i < 6; i++) {
		for (j = 0, bitMask = 1; j < 8; j++) {
			/* Clock out data bit.  FPGA latches data bit on rising edge. */
			if (pu8Ptr[i] & bitMask) {
				gn_write(GN4124_FCL_IODATA_OUT, 0x6);
				gn_write(GN4124_FCL_IODATA_OUT, 0x7);
			} else {
				gn_write(GN4124_FCL_IODATA_OUT, 0x4);
				gn_write(GN4124_FCL_IODATA_OUT, 0x5);
			}
			bitMask <<= 1;
		}
	}
	gn_write(GN4124_FCL_IODATA_OUT, 0x04);
	gn_write(GN4124_FCL_IODATA_OUT, 0x05);
	gn_write(GN4124_FCL_IODATA_OUT, 0x04);
	gn_write(GN4124_FCL_IODATA_OUT, 0x05);
	gn_write(GN4124_FCL_IODATA_OUT, 0x04);
	msleep(1);

	/* Clock out all bits from the input buffer */
	for (; i < len; i++) {
		for (j = 0, bitMask = 1; j < 8; j++) {
			/* Clock out data bit.  FPGA latches data bit on rising edge. */
			if (pu8Ptr[i] & bitMask) {
				gn_write(GN4124_FCL_IODATA_OUT, 0x6);
				gn_write(GN4124_FCL_IODATA_OUT, 0x7);
			} else {
				gn_write(GN4124_FCL_IODATA_OUT, 0x4);
				gn_write(GN4124_FCL_IODATA_OUT, 0x5);
			}
			bitMask <<= 1;
		}
		/* Check programming status on every 128th bytes */
		if (i % 128 == 0) {
			k = gn_read(GN4124_FCL_STATUS);
			if ((k & 0x28) == 0) {
				printk(KERN_ERR "%s() SPRI_STATUS & SPRI_DONE stuck low!\n", __func__);
				ret = -EINVAL;
				goto out;
			}
		}
	}

	/* Generate extra 16 clocks after all data has been pushed out */
	for (i = 0; i < 16; i++) {
		gn_write(GN4124_FCL_IODATA_OUT, 0x4);
		gn_write(GN4124_FCL_IODATA_OUT, 0x5);
	}
	gn_write(GN4124_FCL_IODATA_OUT, 0x4);

	/* Check for SPRI_DONE signal */
	i = gn_read(GN4124_FCL_STATUS);
	if ((i & 0x8) == 0) {
		printk(KERN_ERR "%s() SPRI_DONE is low!\n", __func__);
		ret = -EINVAL;
		goto out;
	}

	ret = 0;

out:
	if (ret == 0) {
		dprintk(0, "%s() Firmware loaded successfully.\n", __func__);
		printk(KERN_ERR "%s() firmware bit-banged, now starting it\n", __func__);
		vc8x0_fw_GN4124AppLayerInit(dev);
		printk(KERN_ERR "%s() started\n", __func__);
	} else
		printk(KERN_ERR "%s() Firmware failed to load.\n", __func__);
	return ret;
}

int vc8x0_fw_format(struct vc8x0_dev *dev, u8 *buffer, u32 len)
{
	int i;
	u8 k;

	dprintk(1, "%s(%p, %d)\n", __func__, buffer, len);

	/* Bit reorder file content */
	for (i = 0; i < len; i++) {
		k = *(buffer + i);

		k = (	((k & 0x80) >> 7) |
			((k & 0x40) >> 5) |
			((k & 0x20) >> 3) |
			((k & 0x10) >> 1) |
			((k & 0x08) << 1) |
			((k & 0x04) << 3) |
			((k & 0x02) << 5) |
			((k & 0x01) << 7));

		*(buffer + i) = k;
	}
	return 0;
}

int vc8x0_fw_load(struct vc8x0_dev *dev)
{
	const struct firmware *fw = 0;
	int i, ret, clkdiv = vc8x0_clk_div;
	u8 *buf;
	char *fwname;
	u32 fwsize;

	dprintk(1, "%s()\n", __func__);

	if (vc8x0_boards[dev->board].fwtype == FW_DEVELOPMENT) {
		fwname = DEV_FW_NAME;
		fwsize = DEV_FW_SIZE;
	} else if (vc8x0_boards[dev->board].fwtype == FW_SDI) {
		fwname = SDI_FW_NAME;
		fwsize = SDI_FW_SIZE;
	} else {
		fwname = PROD_FW_NAME;
		fwsize = PROD_FW_SIZE;
	}

	printk(KERN_INFO "%s() Loading firmware file %s.\n", __func__, fwname);
	ret = request_firmware(&fw, fwname, &dev->pci->dev);
	if (ret < 0) {
                printk(KERN_ERR "%s() Could not load firmware file %s.\n", __func__, fwname);
                printk(KERN_INFO "%s() Copy %s to your hotplug directory!\n", __func__, fwname);
                ret = -EINVAL;
		goto no_fw;
        }

	if (fw->size != fwsize) {
		printk(KERN_ERR "%s() Firmware %s has invalid size!", __func__, fwname);
		ret = -EINVAL;
        } else {
		dprintk(1, "%s() Loaded Firmware %d bytes in ram!", __func__, fw->size);

		buf = kzalloc(fw->size, GFP_KERNEL);
		if (!buf) {
			printk(KERN_ERR "%s() -ENOMEM\n", __func__);
			ret = -ENOMEM;
		} else {
			/* Format the data, we can't modify fw->data directly */
			memcpy(buf, fw->data, fw->size);
			ret = vc8x0_fw_format(dev, buf, fw->size);

			for (i = 0; i < FW_RETRY; i++) {
				printk(KERN_ERR "%s() clkdiv = %d\n", __func__,
					clkdiv);
				if (vc8x0_fw_FCLBitBangTransfer(dev, buf,
					fw->size, clkdiv) == 0)
					break;
				clkdiv++;
			}
			if (i == FW_RETRY) {
				printk(KERN_ERR "%s() Failed to bitbang\n",
					__func__);
				ret = -EIO;
			} else {
				ret = 0;
			}

			kfree(buf);
		}
	}

	release_firmware(fw);
no_fw:
	return ret;
}

int vc8x0_fw_register(struct vc8x0_dev *dev)
{
	dprintk(1, "%s()\n", __func__);
	return 0;
}

int vc8x0_fw_unregister(struct vc8x0_dev *dev)
{
	dprintk(1, "%s()\n", __func__);
	return 0;
}

