/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int gs2971a_debug;
module_param(gs2971a_debug, int, 0644);
MODULE_PARM_DESC(gs2971a_debug, "enable debug messages [gs2971a]");

#define IOPROC_1		0x00
#define IOPROC_2		0x01
#define ERROR_STAT_1		0x02
#define ERROR_STAT_2		0x03
#define EDH_FLAG_IN		0x04
#define EDH_FLAG_OUT		0x05
#define DATA_FORMAT_DS1		0x06
#define DATA_FORMAT_DS2		0x07
#define IO_CONFIG		0x08
#define IO_CONFIG_2		0x09
#define ANC_CONTROL		0x0A
#define ANC_LINE_A		0x0B
#define ANC_LINE_B		0x0C
#define ANC_TYPE1_AP2		0x0F
#define ANC_TYPE2_AP2		0x10
#define ANC_TYPE_3_AP1		0x11
#define ANC_TYPE_4_AP1		0x12
#define ANC_TYPE_5_AP1		0x13
#define ANC_TYPE_1_AP2		0x14
#define ANC_TYPE_2_AP2		0x15
#define ANC_TYPE_3_AP2		0x16
#define ANC_TYPE_4_AP2		0x17
#define ANC_TYPE_5_AP2		0x18
#define VIDEO_FORMAT_352_A_1	0x19
#define VIDEO_FORMAT_352_B_1	0x1A
#define VIDEO_FORMAT_352_A_2	0x1B
#define VIDEO_FORMAT_352_B_3	0x1C
#define VIDEO_FORMAT_352_INS_A	0x1D
#define VIDEO_FORMAT_352_INS_B	0x1E
#define RASTER_STRUCT_1		0x1F
#define RASTER_STRUCT_2		0x20
#define RASTER_STRUCT_3		0x21
#define RASTER_STRUCT_4		0x22
#define FLYWHEEL_STATUS		0x23
#define RATE_SEL		0x24
#define TIM_861_FORMAT		0x25
#define TIM_861_CFG		0x26
#define ERROR_MASK_1		0x37
#define ERROR_MASK_2		0x38
#define AGC_CTRL		0x39
#define CLK_GEN			0x6C
#define IO_DRIVE_STRENGTH	0x6D

#define HD_CFG_AUD			0x200
#define HD_ACS_DET			0x201
#define HD_AUD_DET1			0x202
#define HD_CH_MUTE			0x205

#define HD_AUD_RATEA			0x221
#define HD_AUD_ACTA			0x222
#define HD_AUD_RATEB			0x231
#define HD_AUD_ACTB			0x232

#define HD_ACSR1_2A_BYTE_0_1		0x240
#define HD_ACSR1_2A_BYTE_2_3		0x241
#define HD_ACSR1_2A_BYTE_4_5		0x242
#define HD_ACSR3_4A_BYTE_0_1		0x250
#define HD_ACSR3_4A_BYTE_2_3		0x251
#define HD_ACSR3_4A_BYTE_4_5		0x252
#define HD_ACSR1_2B_BYTE_0_1		0x260
#define HD_ACSR1_2B_BYTE_2_3		0x261
#define HD_ACSR1_2B_BYTE_4_5		0x262
#define HD_ACSR3_4B_BYTE_0_1		0x270
#define HD_ACSR3_4B_BYTE_2_3		0x271
#define HD_ACSR3_4B_BYTE_4_5		0x272

#define TIMING_861_MASK				0x0400
#define ACTLINE_PER_FIELD_MASK			0x07FF
#define INT_PROG_MASK				0x0800
#define STD_LOCK_MASK				0x1000
#define M_MASK					0x2000
#define DATA_RATE_MASK				0xC000
#define DATA_RATE_SHIFT				14

#define HD_AUD_CFG_PRI_GRP_MASK		0x0003
#define HD_AUD_CFG_PRI_GRP_SHIFT	0x0000
#define HD_AUD_CFG_SEC_GRP_MASK		0x000C
#define HD_AUD_CFG_SEC_GRP_SHIFT	0x0002

#define HD_AUD_CFG_PRI_FMT_MASK		0x0030
#define HD_AUD_CFG_PRI_FMT_SHIFT	0x0004
#define HD_AUD_CFG_SEC_FMT_MASK		0x00C0
#define HD_AUD_CFG_SEC_FMT_SHIFT	0x0006

#define HD_AUD_CFG_PRI_OWL_MASK		0x0300
#define HD_AUD_CFG_PRI_OWL_SHIFT	0x0008
#define HD_AUD_CFG_SEC_OWL_MASK		0x0C00
#define HD_AUD_CFG_SEC_OWL_SHIFT	0x000A

#define HD_AUD_CFG_MUTE_MASK		0x2000
#define HD_AUD_CFG_MUTE_SHIFT		0x000D

#define HD_AUD_PG_DET_1_MASK		0x0002
#define HD_AUD_PG_DET_2_MASK		0x0004
#define HD_AUD_PG_DET_3_MASK		0x0008
#define HD_AUD_PG_DET_4_MASK		0x0010

#define HD_ACS_DET1_2A_MASK		0x0001
#define HD_ACS_DET3_4A_MASK		0x0002
#define HD_ACS_DET1_2B_MASK		0x0004
#define HD_ACS_DET3_4B_MASK		0x0008

#define HD_CTRA_DET_MASK		0x0010
#define HD_CTRB_DET_MASK		0x0020

#define HD_ADPG1_DET			0x0002
#define HD_ADPG2_DET			0x0004
#define HD_ADPG3_DET			0x0008
#define HD_ADPG4_DET			0x0010

#define SD_CFG_AUD			0x400
#define SD_ACS_DET			0x401
#define SD_AUD_DET1			0x403

#define SD_AUD_RATEA			0x422
#define SD_AUD_ACTA			0x423
#define SD_AUD_RATEB			0x432
#define SD_AUD_ACTB			0x433


#define SD_ACSR1_2A_BYTE_0_1		0x440
#define SD_ACSR1_2A_BYTE_2_3		0x441
#define SD_ACSR1_2A_BYTE_4_5		0x442
#define SD_ACSR3_4A_BYTE_0_1		0x450
#define SD_ACSR3_4A_BYTE_2_3		0x451
#define SD_ACSR3_4A_BYTE_4_5		0x452
#define SD_ACSR1_2B_BYTE_0_1		0x460
#define SD_ACSR1_2B_BYTE_2_3		0x461
#define SD_ACSR1_2B_BYTE_4_5		0x462
#define SD_ACSR3_4B_BYTE_0_1		0x470
#define SD_ACSR3_4B_BYTE_2_3		0x471
#define SD_ACSR3_4B_BYTE_4_5		0x472

#define SD_CTRA_DET_MASK		0x0010
#define SD_CTRB_DET_MASK		0x0020
#define SD_ADPG1_DET			0x0010
#define SD_ADPG2_DET			0x0020
#define SD_ADPG3_DET			0x0040
#define SD_ADPG4_DET			0x0080

#define SD_ACS_DET1_2A_MASK		0x0001
#define SD_ACS_DET3_4A_MASK		0x0002
#define SD_ACS_DET1_2B_MASK		0x0004
#define SD_ACS_DET3_4B_MASK		0x0008

#define SD_ACS_DET_PRIMARY_MASK		0x0003
#define SD_ACS_DET_SECONDARY_MASK	0x000C

#define dprintk(level, fmt, arg...)\
	do {\
		if (gs2971a_debug >= level) {\
			printk(KERN_DEBUG "%s: " fmt, channel->dev->name, ## arg);\
		}\
	} while (0)

/* Register read / write */
static int vc8x0_gs2971a_read(struct vc8x0_gs2971a_context *ctx, u32 reg, u32 *val)
{
	int ret;
	ret = vc8x0_SPIRead(ctx->bus, reg, val);
	return ret;
}

static int vc8x0_gs2971a_write(struct vc8x0_gs2971a_context *ctx, u32 reg, u32 val)
{
	int ret;
	ret = vc8x0_SPIWrite(ctx->bus, reg, val);
	return ret;
}

int vc8x0_gs2971a_init(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_gs2971a_context *ctx = &channel->gs2971a_ctx;

	dprintk(1, "%s(%p)\n", __func__, channel);
	memset(ctx, 0, sizeof(*ctx));

	ctx->dma_channel = channel;
	ctx->bus = &channel->spi_bus;
	ctx->mute = 0;
	ctx->gain = 0;
	ctx->audio_inputnr = 0;
	ctx->last_locked = jiffies;

	/* Force a small delay so the driver initializes correctly,
	 * else we only get colorbars when gs2971a_debug=0.
	 */
	msleep(50);
	dprintk(1, "%s() bus = %p\n", __func__, ctx->bus);

	return 0;
}

int vc8x0_gs2971a_enable(struct vc8x0_gs2971a_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	struct vc8x0_dev *dev = channel->dev;

	dprintk(1, "%s(base = %04x)\n", __func__, ctx->bus->reg_base);

	/* Power up and assume reasonable defaults */
	/* TODO: Hardcoded channel 1 defaults here */
	vc_write32(0x120, 0x00000001);
	vc_write32(0x130, 0x00000000);
	vc_write32(0x220, 0x00000001);
	vc_write32(0x230, 0x00000000);
	vc_write32(0x320, 0x00000001);
	vc_write32(0x330, 0x00000000);
	vc_write32(0x420, 0x00000001);
	vc_write32(0x430, 0x00000000);
	msleep(50);

	vc8x0_gs2971a_write(ctx, ERROR_MASK_1, 0x03FF);
	vc8x0_gs2971a_write(ctx, ERROR_MASK_2, 0x003F);

	/* Set Audio Master Clock select to 256fs */
	vc8x0_gs2971a_write(ctx, AGC_CTRL, 1);
	
	return 0;
}

int vc8x0_gs2971a_disable(struct vc8x0_gs2971a_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;

	dprintk(1, "%s()\n", __func__);

	/* Power Down */

	return 0;
}

int vc8x0_gs2971a_set_input(struct vc8x0_gs2971a_context *ctx, int nr)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	dprintk(1, "%s(%d)\n", __func__, nr);
	return 0;
}

int vc8x0_gs2971a_set_mute(struct vc8x0_gs2971a_context *ctx, u32 enable)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;

	if (!ctx->dma_channel)
		return 0;

	dprintk(1, "%s(nr=%d, %d)\n", __func__, channel->nr, enable);

	ctx->mute = enable;
#if 0
	if (enable) {
		vc8x0_gs2971a_i2c_write8(ctx, 72, 0x00);
	} else {
		vc8x0_gs2971a_i2c_write8(ctx, 72, ctx->gain);
	}
#endif
	return 0;
}

int vc8x0_gs2971a_get_mute(struct vc8x0_gs2971a_context *ctx, u32 *enabled)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	*enabled = ctx->mute;
	dprintk(1, "%s(nr=%d, %d)\n", __func__, channel->nr, *enabled);
	return 0;
}

int vc8x0_gs2971a_set_volume(struct vc8x0_gs2971a_context *ctx, u32 val)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	u8 v = 8 + val;

	if (!ctx->dma_channel)
		return 0;

	ctx->gain = v;

	dprintk(1, "%s(nr=%d, %d, %02x)\n", __func__, channel->nr, val, ctx->gain);
#if 0
	vc8x0_gs2971a_i2c_write8(ctx, 72, ctx->gain);
#endif
	return 0;
}

int vc8x0_gs2971a_get_volume(struct vc8x0_gs2971a_context *ctx, u32 *val)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	*val = ctx->gain;
	dprintk(1, "%s(nr=%d, %d)\n", __func__, channel->nr, *val);
	return 0;
}

#define FRAMERATE_UNKNOWN       0x00
#define FRAMERATE_23            (23)
#define FRAMERATE_24            (24)
#define FRAMERATE_25            (25)
#define FRAMERATE_29            (29)
#define FRAMERATE_30            (30)
#define FRAMERATE_50            (50)
#define FRAMERATE_59            (59)
#define FRAMERATE_60            (60)
        int frate;

#define	UNDEFINED_ENTRY	{ FORMAT_UNKNOWN,	1,	1,	1,	1,	{ 1,	0},	{ 0,	0 },	{ 0,	0 },	0,	0,	1,	0, }
static struct gs2970_format_t
{
	u32 id;
	u32 nTotalPixels;
	u32 nTotalLines;
	u32 width;
	u32 nBlankingPixels;
	u32 height[2];
	u32 nCapturedVBILines[2];
	u32 nCountedVBILines[2];
	u32 nPixelRepitition;
	u32 interlaced;
	u32 frate;
	u32 nFrameRateDF;
} gs2970_formats[0x40] =
	/* Page 58 in the 2970 datasheet */
	/* Video Standard	T Pxl	T Lns	A Pxl	B Pxl	A Ln1	A Ln2	CapV1	CapV2	CntV1	CntV2	PMode	ilace	frate	Fr099 */
{
/* 0x00 - 0x0f */
	{ FORMAT_1280x720p60,	1650,	750,	1280,	370,	{ 720,	0},	{ 30,	0 },	{ 30,	0 },	0,	0,	FRAMERATE_60,	5994, },	// 1280x720/60
	UNDEFINED_ENTRY,
	{ FORMAT_1280x720p30,	3300,	750,	1280,	2020,	{ 720,	0},	{ 30,	0 },	{ 30,	0 },	0,	0,	FRAMERATE_30,	2997, },	// 1280x720/30
	UNDEFINED_ENTRY,
	{ FORMAT_1280x720p50,	1980,	750,	1280,	700,	{ 720,	0},	{ 30,	0 },	{ 30,	0 },	0,	0,	FRAMERATE_50,	5000, },	// 1280x720/50
	UNDEFINED_ENTRY,
	{ FORMAT_1280x720p25,	3960,	750,	1280,	2680,	{ 720,	0},	{ 30,	0 },	{ 30,	0 },	0,	0,	FRAMERATE_25,	2500, },	// 1280x720/25
	UNDEFINED_ENTRY,
	{ FORMAT_1280x720p24,	4125,	750,	1280,	2845,	{ 720,	0},	{ 30,	0 },	{ 30,	0 },	0,	0,	FRAMERATE_24,	2400, },	// 1280x720/24
	UNDEFINED_ENTRY,
	{ FORMAT_1920x1080i60,	2200,	1125,	1920,	280,	{ 540,	540},	{ 23,	22 },	{ 23,	22 },	0,	1,	FRAMERATE_60,	5994, },	// 1920x1080i60 1920x1080/30PsF
	{ FORMAT_1920x1080p30,	2200,	1125,	1920,	280,	{ 1080,	0},	{ 45,	0 },	{ 45,	0 },	0,	0,	FRAMERATE_30,	2997, },	// 1920x1080p30
	{ FORMAT_1920x1080i50,	2200,	1125,	1920,	280,	{ 540,	540},	{ 23,	22 },	{ 23,	22 },	0,	1,	FRAMERATE_50,	5000, },	// 1920x1080i50 1920x1080/25PsF
	{ FORMAT_1920x1080p25,	2640,	1125,	1920,	720,	{ 1080,	0},	{ 45,	0 },	{ 45,	0 },	0,	0,	FRAMERATE_25,	2500, },	// 1920x1080p25
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
/* 0x10 - 0x1f */
	{ FORMAT_1920x1080p24,	2750,	1125,	1920,	830,	{ 1080,	0},	{ 45,	0 },	{ 45,	0 },	0,	0,	FRAMERATE_24,	2398, },	// 1920x1080/24
	{ FORMAT_1920x1080p24,	2750,	1125,	1920,	830,	{ 540,	540},	{ 23,	22 },	{ 23,	22 },	0,	1,	FRAMERATE_24,	2398, },	// 1920x1080/24PsF
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	{ FORMAT_1920x1080i50,	2376,	1250,	1920,	456,	{ 540,	540},	{ 23,	22 },	{ 23,	22 },	0,	1,	FRAMERATE_50,	5000, },	// 1920x1080/50
	{ FORMAT_1920x1080i60,	2200,	1125,	1920,	280,	{ 518,	517},	{ 45,	45 },	{ 45,	45 },	0,	1,	FRAMERATE_60,	5994, },	// 1920x1035/60
	{ FORMAT_720x480i60,	1716,	525,	720,	996,	{ 244,	243},	{ 19,	19 },	{ 19,	19 },	0,	1,	FRAMERATE_29,	2997, },	// 720x487/30
	{ FORMAT_720x480i60,	1716,	525,	720,	996,	{ 254,	253},	{ 9,	9 },	{ 9,	9 },	0,	1,	FRAMERATE_29,	2997, },	// 720x507/30
	{ FORMAT_720x576i50,	1728,	625,	720,	1008,	{ 288,	288},	{ 25,	24 },	{ 25,	24 },	0,	1,	FRAMERATE_25,	2500, },	// 720x576/25
	{ FORMAT_720x480i60,	1716,	525,	720,	996,	{ 244,	243},	{ 19,	19 },	{ 19,	19 },	0,	1,	FRAMERATE_29,	2997, },	// 720x487/30
	{ FORMAT_720x576i50,	1728,	625,	720,	1008,	{ 288,	288},	{ 25,	24 },	{ 25,	24 },	0,	1,	FRAMERATE_25,	2500, },	// 720x576/25
	{ FORMAT_720x480i60,	1716,	525,	720,	996,	{ 254,	253},	{ 9,	9 },	{ 9,	9 },	0,	0,	FRAMERATE_29,	2997, },	// 720x507/30
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
/* 0x20 - 0x2f */
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	{ FORMAT_1920x1080p60,	2200,	1125,	1920,	280,	{ 1080,	0},	{ 45,	0 },	{ 45,	0 },	0,	0,	FRAMERATE_60,	5994, },	// 1920x1080/60 3G
	UNDEFINED_ENTRY,
	{ FORMAT_1920x1080p60,	2640,	1125,	1920,	720,	{ 1080,	0},	{ 45,	0 },	{ 45,	0 },	0,	0,	FRAMERATE_50,	5000, },	// 1920x1080/50 3G
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
/* 0x30 - 0x3f */
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
	UNDEFINED_ENTRY,
};

static struct gs2970_format_t *gs2970_find_format(u32 vdstd)
{
	struct gs2970_format_t *ret = 0;
	int max = (sizeof(gs2970_formats) / sizeof(struct gs2970_format_t));

	if (vdstd < max)
		ret = &gs2970_formats[vdstd];

	return ret;
}

/* Return 1 on lock else 0 = no lock */
static int vc8x0_gs2971a_is_sdi_locked(struct vc8x0_gs2971a_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	int locked;

	vc8x0_gs2971a_read(ctx, RASTER_STRUCT_4, &ctx->regs[RASTER_STRUCT_4]);

	locked = ctx->regs[RASTER_STRUCT_4] & STD_LOCK_MASK ? 1 : 0;

	dprintk(2, "%s(nr=%d) locked = %d\n", __func__, channel->nr, locked);

	return locked;
}

/* We don't need to check the locked register in here, but since
 * we're reading it for other registers, we'll do it anyway.
 * Detect the incoming signal, compare against the requested
 * signal and if we don't match, or if no supported signal
 * is found, enable colorbars.
 * If the detected signal matches the user requested signal,
 * disable colorbar generation.
 */
static void vc8x0_gs2971a_detect(struct vc8x0_gs2971a_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	struct gs2970_format_t *gsfmt;
	int locked, interlaced, colorbars_reqd = 1;

	vc8x0_gs2971a_read(ctx, RASTER_STRUCT_4, &ctx->regs[RASTER_STRUCT_4]);

	locked = ctx->regs[RASTER_STRUCT_4] & STD_LOCK_MASK ? 1 : 0;
	interlaced = ctx->regs[RASTER_STRUCT_4] & INT_PROG_MASK ? 1 : 0;

	if (locked) {
		/* Get the detected input signal */
		vc8x0_gs2971a_read(ctx, DATA_FORMAT_DS1, &ctx->regs[DATA_FORMAT_DS1]);
		ctx->detected_vd_std = (ctx->regs[DATA_FORMAT_DS1] & 0x3F00) >> 8;

		gsfmt = gs2970_find_format(ctx->detected_vd_std);
		if (gsfmt) {
			dprintk(2, "%s() ctx->detected_vd_std = 0x%x, size = %d/%d interlaced = %d rate = %d\n",
				__func__,
				ctx->detected_vd_std,
				gsfmt->width, gsfmt->height[0], gsfmt->interlaced,
				gsfmt->frate);

			dprintk(2, "%s() VD_STD = 0x%02x (fmtid = 0x%x)\n", __func__,
				ctx->detected_vd_std,
				gsfmt->id);

			if (gsfmt->id != FORMAT_UNKNOWN)
				ctx->detected_fmt = vc8x0_video_find_formatById(channel, gsfmt->id);
			else if (gsfmt->interlaced)
				ctx->detected_fmt = vc8x0_video_find_format(ctx->dma_channel, V4L2_PIX_FMT_YUYV,
					gsfmt->width, gsfmt->height[0] * 2,
					FORMAT_INTERLACED, gsfmt->frate);
			else
				ctx->detected_fmt = vc8x0_video_find_format(ctx->dma_channel, V4L2_PIX_FMT_YUYV,
					gsfmt->width, gsfmt->height[0],
					FORMAT_PROGRESSIVE, gsfmt->frate);

			if (ctx->detected_fmt && (ctx->requested_fmt == 0)) {
				dprintk(2, "%s() Detected format->id = %d [%s], user has not yet requested a format\n",
					__func__, ctx->detected_fmt->id, ctx->detected_fmt->name);
			}

			if (ctx->detected_fmt && ctx->requested_fmt) {
				if (ctx->detected_fmt != ctx->requested_fmt) {
					dprintk(2, "%s() Detected incorrect format->id = %d [%s], need %d\n",
						__func__,
						ctx->detected_fmt->id,
						ctx->detected_fmt->name,
						ctx->requested_fmt->id);
				} else {
					dprintk(2, "%s() Detected format->id = %d [%s] correctly.\n",
						__func__,
						ctx->detected_fmt->id,
						ctx->detected_fmt->name);
					colorbars_reqd = 0;
				}
			} else {
				dprintk(2, "%s() detected size = unknown\n", __func__);
			}
		} else {
			dprintk(1, "%s() VD_STD = 0x%02x\n", __func__, ctx->detected_vd_std);
		}
	} else {
		dprintk(2, "%s() not locked\n", __func__);
	}

	if (colorbars_reqd == 1) {
		/* We didn't detect the video format correctly, throw the colorbars up. */
		vc8x0_channel_loss_of_sync(ctx->dma_channel, 1);
	} else {
		/* format is good. Remove colorbars if they were present. */
		vc8x0_channel_loss_of_sync(ctx->dma_channel, 0);
	}

	dprintk(1, "%s() colorbars = %d locked = %d interlaced = %d\n", __func__, colorbars_reqd, locked, interlaced);
}

int vc8x0_gs2971a_procfs_show(struct seq_file *m, struct vc8x0_gs2971a_context *ctx)
{
	struct arr_t {
		u32 reg;
		const char *name;
	} arr[] = {
		{ RASTER_STRUCT_1,	"RASTER_STRUCT_1"	},
		{ RASTER_STRUCT_2,	"RASTER_STRUCT_2"	},
		{ RASTER_STRUCT_3,	"RASTER_STRUCT_3"	},
		{ RASTER_STRUCT_4,	"RASTER_STRUCT_4"	},
		{ IO_CONFIG,		"IO_CONFIG"		},
		{ IO_CONFIG_2,		"IO_CONFIG_2"		},
		{ DATA_FORMAT_DS1,	"DATA_FORMAT_DS1"	},
		{ RATE_SEL,		"RATE_SEL"		},
		{ SD_CFG_AUD,		"SD_CFG_AUD"		},
		{ HD_CFG_AUD,		"HD_CFG_AUD"		},
		{ HD_ACS_DET,		"HD_ACS_DET"		},
		{ HD_AUD_DET1,		"HD_AUD_DET1"		},
		{ HD_AUD_ACTA,		"HD_AUD_ACTA"		},
		{ HD_AUD_ACTB,		"HD_AUD_ACTB"		},
		{ HD_AUD_RATEA,		"HD_AUD_RATEA"		},
		{ HD_AUD_RATEB,		"HD_AUD_RATEB"		},
		{ HD_CH_MUTE,		"HD_CH_MUTE"		},
		{ HD_ACSR1_2A_BYTE_0_1, "HD_ACSR1_2A_BYTE_0_1"	},
		{ HD_ACSR1_2A_BYTE_2_3, "HD_ACSR1_2A_BYTE_2_3"	},
		{ HD_ACSR1_2A_BYTE_4_5, "HD_ACSR1_2A_BYTE_4_5"	},
		{ HD_ACSR3_4A_BYTE_0_1, "HD_ACSR3_4A_BYTE_0_1"	},
		{ HD_ACSR3_4A_BYTE_2_3, "HD_ACSR3_4A_BYTE_2_3"	},
		{ HD_ACSR3_4A_BYTE_4_5, "HD_ACSR3_4A_BYTE_4_5"	},
		{ HD_ACSR1_2B_BYTE_0_1, "HD_ACSR1_2B_BYTE_0_1"	},
		{ HD_ACSR1_2B_BYTE_2_3, "HD_ACSR1_2B_BYTE_2_3"	},
		{ HD_ACSR1_2B_BYTE_4_5, "HD_ACSR1_2B_BYTE_4_5"	},
		{ HD_ACSR3_4B_BYTE_0_1, "HD_ACSR3_2B_BYTE_0_1"	},
		{ HD_ACSR3_4B_BYTE_2_3, "HD_ACSR3_2B_BYTE_2_3"	},
		{ HD_ACSR3_4B_BYTE_4_5, "HD_ACSR3_2B_BYTE_4_5"	},

	};
	u32 r, v;
	int i;

	seq_printf(m, "GS2971A:\n");

	for (i = 0; i < (sizeof(arr) / sizeof(struct arr_t)); i++) {
		r = arr[i].reg;
		vc8x0_gs2971a_read(ctx, r, &ctx->regs[r]);
		seq_printf(m, "%20s: 0x%04x = 0x%04x", arr[i].name, r, ctx->regs[r]);
		if (arr[i].reg == RASTER_STRUCT_4) {
			v = ctx->regs[r];
			seq_printf(m, " - Rate: %s",
				(v & 0xc000) == 0xc000 ? "SD" :
				(v & 0xc000) == 0x4000 ? "SD" :
				(v & 0xc000) == 0x0000 ? "HD" :
				(v & 0xc000) == 0x8000 ? "3G" : "Unknown");
			seq_printf(m, " M: %s",
				(v & 0x2000) == 0x2000 ? "1.001" : "1.000");
			seq_printf(m, " STD_Lock: %d", v & 0x1000 ? 1 : 0);
			seq_printf(m, " ScanMode: %s", v & 0x0800 ? "Interlaced" : "Progressive");
			
		}
		if (arr[i].reg == HD_AUD_DET1) {
			v = ctx->regs[r];
			seq_printf(m, " - Audio packets detected in groups 1:%s 2:%s 3:%s 4:%s",
				(v & HD_ADPG1_DET) ? "YES" : "NO",
				(v & HD_ADPG2_DET) ? "YES" : "NO",
				(v & HD_ADPG3_DET) ? "YES" : "NO",
				(v & HD_ADPG4_DET) ? "YES" : "NO");
		}
		if (arr[i].reg == HD_ACS_DET) {
			v = ctx->regs[r];
			seq_printf(m, "\n");
			seq_printf(m, " - Primary group audio status detected for channels 1/2:%s 3/4:%s\n",
				(v & HD_ACS_DET1_2A_MASK) ? "YES" : "NO",
				(v & HD_ACS_DET3_4A_MASK) ? "YES" : "NO");
			seq_printf(m, " - Primary group audio control packeted detected: %s\n",
				(v & HD_CTRA_DET_MASK) ? "YES" : "NO");

			seq_printf(m, " - Secondary group audio status detected for channels 1/2:%s 3/4:%s",
				(v & HD_ACS_DET1_2B_MASK) ? "YES" : "NO",
				(v & HD_ACS_DET3_4B_MASK) ? "YES" : "NO");
			seq_printf(m, " - Secondary group audio control packeted detected: %s",
				(v & HD_CTRB_DET_MASK) ? "YES" : "NO");
		}
		seq_printf(m, "\n");
	}

	if (ctx->detected_fmt) {
		seq_printf(m, "Detected video format->id = %d [%s].\n",
			ctx->detected_fmt->id,
			ctx->detected_fmt->name);
	}

	/* Audio settings, based on the detected signal data rate (r) */
	r = ((ctx->regs[RASTER_STRUCT_4] & DATA_RATE_MASK) >> DATA_RATE_SHIFT);

	/* Is this HD Audio? */
	if ((r == 0) || (r == 2)) {

		/* Is the channel status for primary group detected? */
		if ((ctx->regs[HD_ACS_DET] & HD_CTRA_DET_MASK) == HD_CTRA_DET_MASK) {
			/* Yes.  Go get the pertinant channels status bytes */
			seq_printf(m, "Audio: HD - Detected\n");

			vc8x0_gs2971a_read(ctx, HD_ACSR1_2A_BYTE_0_1, &ctx->regs[HD_ACSR1_2A_BYTE_0_1]);
			vc8x0_gs2971a_read(ctx, HD_ACSR1_2A_BYTE_2_3, &ctx->regs[HD_ACSR1_2A_BYTE_2_3]);
			vc8x0_gs2971a_read(ctx, HD_ACSR1_2A_BYTE_4_5, &ctx->regs[HD_ACSR1_2A_BYTE_4_5]);
			vc8x0_gs2971a_read(ctx, HD_ACSR3_4A_BYTE_0_1, &ctx->regs[HD_ACSR3_4A_BYTE_0_1]);
			vc8x0_gs2971a_read(ctx, HD_ACSR3_4A_BYTE_2_3, &ctx->regs[HD_ACSR3_4A_BYTE_2_3]);
			vc8x0_gs2971a_read(ctx, HD_ACSR3_4A_BYTE_4_5, &ctx->regs[HD_ACSR3_4A_BYTE_4_5]);
			vc8x0_gs2971a_read(ctx, HD_AUD_RATEA, &ctx->regs[HD_AUD_RATEA]);
			vc8x0_gs2971a_read(ctx, HD_AUD_ACTA, &ctx->regs[HD_AUD_ACTA]);
		} else {
			seq_printf(m, "Audio: HD - Not detected\n");
			ctx->regs[HD_ACSR1_2A_BYTE_0_1] = 0;
			ctx->regs[HD_ACSR1_2A_BYTE_2_3] = 0;
			ctx->regs[HD_ACSR1_2A_BYTE_4_5] = 0;
			ctx->regs[HD_ACSR3_4A_BYTE_0_1] = 0;
			ctx->regs[HD_ACSR3_4A_BYTE_2_3] = 0;
			ctx->regs[HD_ACSR3_4A_BYTE_4_5] = 0;
			ctx->regs[HD_AUD_RATEA] = 0;
			ctx->regs[HD_AUD_ACTA] = 0;
		}

		/* Is the channel status for secondary group detected? */
		if ((ctx->regs[HD_ACS_DET] & HD_CTRB_DET_MASK) == HD_CTRB_DET_MASK) {
			/* Yes.  Go get the pertinant channels status bytes */
			vc8x0_gs2971a_read(ctx, HD_ACSR1_2B_BYTE_0_1, &ctx->regs[HD_ACSR1_2B_BYTE_0_1]);
			vc8x0_gs2971a_read(ctx, HD_ACSR1_2B_BYTE_2_3, &ctx->regs[HD_ACSR1_2B_BYTE_2_3]);
			vc8x0_gs2971a_read(ctx, HD_ACSR1_2B_BYTE_4_5, &ctx->regs[HD_ACSR1_2B_BYTE_4_5]);
			vc8x0_gs2971a_read(ctx, HD_ACSR3_4B_BYTE_0_1, &ctx->regs[HD_ACSR3_4B_BYTE_0_1]);
			vc8x0_gs2971a_read(ctx, HD_ACSR3_4B_BYTE_2_3, &ctx->regs[HD_ACSR3_4B_BYTE_2_3]);
			vc8x0_gs2971a_read(ctx, HD_ACSR3_4B_BYTE_4_5, &ctx->regs[HD_ACSR3_4B_BYTE_4_5]);
			vc8x0_gs2971a_read(ctx, HD_AUD_RATEB, &ctx->regs[HD_AUD_RATEB]);
			vc8x0_gs2971a_read(ctx, HD_AUD_ACTB, &ctx->regs[HD_AUD_ACTB]);
		} else {
			ctx->regs[HD_ACSR1_2B_BYTE_0_1] = 0;
			ctx->regs[HD_ACSR1_2B_BYTE_2_3] = 0;
			ctx->regs[HD_ACSR1_2B_BYTE_4_5] = 0;
			ctx->regs[HD_ACSR3_4B_BYTE_0_1] = 0;
			ctx->regs[HD_ACSR3_4B_BYTE_2_3] = 0;
			ctx->regs[HD_ACSR3_4B_BYTE_4_5] = 0;
			ctx->regs[HD_AUD_RATEB] = 0;
			ctx->regs[HD_AUD_ACTB] = 0;
		}

		/* Update our status */
		seq_printf(m, "   AudioPresent0: %d (Primary group active channels)\n", ((ctx->regs[HD_AUD_ACTA] & 0x0003) && (ctx->regs[HD_AUD_DET1] & HD_ADPG1_DET)) ? 1 : 0);
		seq_printf(m, "   AudioPresent1: %d (Primary group active channels)\n", ((ctx->regs[HD_AUD_ACTA] & 0x000c) && (ctx->regs[HD_AUD_DET1] & HD_ADPG1_DET)) ? 1 : 0);
		seq_printf(m, "   AudioPresent2: %d (Primary group active channels)\n", ((ctx->regs[HD_AUD_ACTB] & 0x0003) && (ctx->regs[HD_AUD_DET1] & HD_ADPG2_DET)) ? 1 : 0);
		seq_printf(m, "   AudioPresent3: %d (Primary group active channels)\n", ((ctx->regs[HD_AUD_ACTB] & 0x000c) && (ctx->regs[HD_AUD_DET1] & HD_ADPG2_DET)) ? 1 : 0);
#if 0
		seq_printf(m, "AudioSampleRate0: %d (Primary group sampling frequency for channels 1 and 2)\n", (ctx->regs[HD_AUD_RATEA] &0x000e) >> 1);
		seq_printf(m, "AudioSampleRate1: %d (Primary group sampling frequency for channels 1 and 2)\n", (ctx->regs[HD_AUD_RATEA] &0x000e) >> 1);
		seq_printf(m, "  AudioDataType0: %d (Audio Group B status for channels 1 and 2)\n", (ctx->regs[HD_ACSR1_2B_BYTE_0_1] &0x0002) >> 1);
		seq_printf(m, "  AudioDataType1: %d (Audio Group B status for channels 1 and 2)\n", (ctx->regs[HD_ACSR1_2B_BYTE_0_1] &0x0002) >> 1);
#endif
#if 0
		seq_printf(m, "AudioSampleRate2: %d (Secondary group sampling frequency for channels 1 and 2)\n", (ctx->regs[HD_AUD_RATEB] &0x000e) >> 1);
		seq_printf(m, "AudioSampleRate3: %d (Secondary group sampling frequency for channels 1 and 2)\n", (ctx->regs[HD_AUD_RATEB] &0x000e) >> 1);
		seq_printf(m, "  AudioDataType2: %d (Audio Group B status for channels 1 and 2)\n", (ctx->regs[HD_ACSR1_2B_BYTE_0_1] &0x0002) >> 1);
		seq_printf(m, "  AudioDataType3: %d (Audio Group B status for channels 1 and 2)\n", (ctx->regs[HD_ACSR1_2B_BYTE_0_1] &0x0002) >> 1);
#endif

	} else {
		seq_printf(m, "Audio: SD - TODO\n");
	}

	return 0;
}

int vc8x0_gs2971a_set_audio_input(struct vc8x0_gs2971a_context *ctx, int inputnr)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;

	if (!ctx)
		return -ENODEV;

	if ((inputnr < 0) || (inputnr > 3))
		return -EINVAL;

	dprintk(1, "%s(%p, %d)\n", __func__, ctx, inputnr);
	ctx->audio_inputnr = inputnr;

	return 0;
}

int vc8x0_gs2971a_get_audio_input(struct vc8x0_gs2971a_context *ctx, int *inputnr)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;

	if ((!ctx) || (!inputnr))
		return -ENODEV;

	dprintk(1, "%s(%p, %d)\n", __func__, ctx, ctx->audio_inputnr);
	*inputnr = ctx->audio_inputnr;

	return 0;
}

int vc8x0_gs2971a_set_format(struct vc8x0_gs2971a_context *ctx, struct vc8x0_format *fmt)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	int ret = -ENODEV;

	if ((!ctx) || (!fmt))
		goto out;

	dprintk(1, "%s(%p, %d)\n", __func__, ctx, fmt->id);

	/* Poke the hardware */
	/* Set any format settings */
//	vc8x0_ad7441_sw_reset(ctx);

	if ((fmt->id == FORMAT_720x480i60) || (fmt->id == FORMAT_720x480i59))
		ctx->dma_channel->vbi_enabled = 1;
	else
		ctx->dma_channel->vbi_enabled = 0;

	ctx->requested_fmt = fmt;
	ret = 0;

out:
	return ret;	
}

int vc8x0_gs2971a_get_format(struct vc8x0_gs2971a_context *ctx, struct vc8x0_format **fmt)
{
	*fmt = ctx->detected_fmt;
	return 0;
}

/* Driver Lock detect and format recognition works in the following way.
 * We check the SDI registers directly. It auto-detects the incoming
 * signal. If we don't detect anything, or what we detected doesn't match
 * the user requested formed, we request colorbar generation.
 */
/* Return 0 on not locked, 1 on locked */
int vc8x0_gs2971a_is_locked(struct vc8x0_gs2971a_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	struct vc8x0_dev *dev = channel->dev;
	int ret;

	if (!time_after(jiffies, (long)ctx->last_locked + msecs_to_jiffies(500)))
	{
		/* Skip the detect, because it hasn't been 500ms since the last attempt */
		return 1;
	}
	ctx->last_locked = jiffies;

	switch (vc8x0_boards[dev->board].input[channel->video_inputnr].type) {
	case VC8X0_VMUX_SDI:
		ret = vc8x0_gs2971a_is_sdi_locked(ctx);
		break;
	default:
		ret = 0;
	}

	if (ret == 0) {
		/* We didn't detect the video lock correctly, throw the colorbars up. */
		vc8x0_channel_loss_of_sync(ctx->dma_channel, 1);
	} else {
		vc8x0_gs2971a_detect(ctx);
	}

	return ret;
}
