/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int i2c_debug;
module_param(i2c_debug, int, 0644);
MODULE_PARM_DESC(i2c_debug, "enable debug messages [i2c]");

static unsigned int i2c_scan;
module_param(i2c_scan, int, 0444);
MODULE_PARM_DESC(i2c_scan, "scan i2c bus at insmod time");

#define dprintk(level, fmt, arg...)\
	do { if (i2c_debug >= level)\
		printk(KERN_DEBUG "%s/0: " fmt, dev->name, ## arg);\
	} while (0)

#define I2C_PRE_LO_DATA	0x31
#define I2C_PRE_HI_DATA	0x00
#define I2C_RETRY_COUNT 250

/* Read an I2C byte and set the next command byte. */
static int vc8x0_I2cReadStep(struct vc8x0_i2c *bus, u8 *buf, u8 cmd)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret = 0, c = 0;
	u64 status, reg;

	*buf = 0x00;

	dprintk(2, "%s(cmd = %02x)\n", __func__, cmd);

	while (1)
	{
		status = vc_read64(bus->reg_base + I2C_STATUS_REGISTER);

		if ((status & I2C_IN_PROGRESS) == 0)
		{
			/* Read byte is ready. */
			reg = vc_read64(bus->reg_base + I2C_RECEIVE_REGISTER);
			reg &= 0xFF;
			*buf = (u8)reg;

			/* Will be ACK, NACK, or STOP. */
			vc_write64(bus->reg_base + I2C_COMMAND_REGISTER, cmd);
			ret = 1;
			break;
		}
		udelay(100);
		if (++c >= I2C_RETRY_COUNT)
		{
			printk(KERN_ERR "%s() Attempt to read I2c value failed. status = 0x%llx", __func__, status);
			break;
		}
	}
	return ret;
}

/* return 1 on success else 0 */
static int vc8x0_I2cWriteStep(struct vc8x0_i2c *bus, u8 buf, u8 cmd)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret = 0, c = 0;
	u64 status;

	dprintk(2, "%s(buf=%02x, cmd = %02x)\n", __func__, buf, cmd);

	while (1)
	{
		status = vc_read64(bus->reg_base + I2C_STATUS_REGISTER);
		if ((status & I2C_IN_PROGRESS) == 0)
		{
			/* Next byte can be written now.  There is only data byte in a WRITE command. */
			if (cmd & I2C_WRITE)
				vc_write64(bus->reg_base + I2C_TRANSMIT_REGISTER, buf);
			vc_write64(bus->reg_base + I2C_COMMAND_REGISTER, cmd);
			ret = 1;
			break;
		}
		udelay(100);
		if (++c >= I2C_RETRY_COUNT)
		{
			printk(KERN_ERR "%s() Attempt to read I2c value failed. status = 0x%llx", __func__, status);
			break;
		}
	}
	return ret;
}

/* return 1 on success else 0 */
static int vc8x0_I2cRead(struct vc8x0_i2c *bus, u8 addr, u8 reg, u8 *buf)
{
	int ret = 0;

	*buf = 0;

	ret = vc8x0_I2cWriteStep(bus, addr, I2C_START | I2C_WRITE);
	if (ret)
	{
		ret = vc8x0_I2cWriteStep(bus, reg, I2C_WRITE);
		if (ret)
		{
			ret = vc8x0_I2cWriteStep(bus, addr | 1, I2C_START | I2C_WRITE);
			if (ret)
			{
				ret = vc8x0_I2cWriteStep(bus, addr | 1, I2C_READ | I2C_NACK);
				if (ret)
					return vc8x0_I2cReadStep(bus, buf, I2C_STOP);
			}
		}
	}
	vc8x0_I2cWriteStep(bus, addr, I2C_STOP);
	return ret;
}

/* return 1 on success else 0 */
static int vc8x0_I2cWrite(struct vc8x0_i2c *bus, u8 addr, u8 reg, u8 buf)
{
	int ret;

	ret = vc8x0_I2cWriteStep(bus, addr, I2C_START | I2C_WRITE);
	if (ret)
	{
		ret = vc8x0_I2cWriteStep(bus, reg, I2C_WRITE);
		if (ret)
//			ret = vc8x0_I2cWriteStep(bus, buf, I2C_WRITE | I2C_STOP);
			return vc8x0_I2cWriteStep(bus, buf, I2C_WRITE | I2C_STOP);
	}
	vc8x0_I2cWriteStep(bus, addr, I2C_STOP);
	return ret;
}

/*
 * Read a block of registers using device's subaddress autoincrement.
 * Lock must already be held.  At least two bytes must be read.
 */
static int vc8x0_I2cReadBlock(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 *pucBlock, u32 len)
{
	int ret, i;

	if (len < 2)
		return 0;

	ret = vc8x0_I2cWriteStep(bus, device, I2C_START | I2C_WRITE);
	if (ret)
	{
		ret = vc8x0_I2cWriteStep(bus, addr, I2C_WRITE);
		if (ret)
		{
			ret = vc8x0_I2cWriteStep(bus, device | 1, I2C_START | I2C_WRITE);
			if (ret)
			{
				ret = vc8x0_I2cWriteStep(bus, device | 1, I2C_READ);
				if (ret)
				{
					for (i = 0; i < len - 2; i++)
					{
						ret = vc8x0_I2cReadStep(bus, &pucBlock[i], I2C_READ);
						if (!ret)
							break;
					}

					if (ret)
					{
						ret = vc8x0_I2cReadStep(bus, &pucBlock[len - 2], I2C_READ | I2C_NACK);
						if (ret)
						{
							ret = vc8x0_I2cReadStep(bus, &pucBlock[len - 1], I2C_STOP);
							return ret;
						}
					}
				}
			}
		}
	}

	vc8x0_I2cWriteStep(bus, device, I2C_STOP);
	return ret;
}

/*
 * One I2C write or masked read-modify-write.
 * I2c lock must already have been acquired by caller.
 */
static int vc8x0_I2cWriteRegisterLocked(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 buf, u8 ucMask)
{
	int ret = 1;
	u8 v;

	if (ucMask != 0xFF)
	{
		ret = vc8x0_I2cRead(bus, device, addr, &v);
		buf = (buf & ucMask) | (v & ~ucMask);
	}

	if (ret)
		ret = vc8x0_I2cWrite(bus, device, addr, buf);

	return ret;
}

int vc8x0_I2cWriteRegister(struct vc8x0_i2c *bus, u8 device, u8 reg, u8 val)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret;

	mutex_lock(&bus->lock);
	ret = vc8x0_I2cWriteRegisterLocked(bus, device, reg, val, 0xff);
	if (ret) {
		dprintk(1, "%s() device=0x%02X, reg=0x%02X, val=0x%02X", __func__,
			device, reg, val);
	}
	mutex_unlock(&bus->lock);
	msleep(2);
	return ret;
}


/* One I2C read.
 * This routine acquires i2c lock.
 */
int vc8x0_I2cReadRegister(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 *buf)
{
	struct vc8x0_dev *dev = 0;
	int ret;

	BUG_ON(bus == 0);
	BUG_ON(buf == 0);

	dev = bus->dev;
	mutex_lock(&bus->lock);
	ret = vc8x0_I2cRead(bus, device, addr, buf);
	if (ret) {
		dprintk(1, "%s() device=0x%02X, addr=0x%02X, buf=0x%02X",
			__func__, device, addr, *buf);
	}
	mutex_unlock(&bus->lock);
	msleep(2);
	return ret;
}

/* Write a block of registers using device's subaddress autoincrement.
 * Lock must already be held.
 */
static int vc8x0_I2cWriteBlock(struct vc8x0_i2c *bus, u8 ucAddr, u8 ucSubAddr, u8 *pucBlock, u32 ulNRegs)
{
	int ret, i;

	ret = vc8x0_I2cWriteStep(bus, ucAddr, I2C_START | I2C_WRITE);
	ret = vc8x0_I2cWriteStep(bus, ucSubAddr, I2C_WRITE);

	for (i = 0; i < ulNRegs - 1; i++) {
		ret = vc8x0_I2cWriteStep(bus, pucBlock[i], I2C_WRITE);
	}

	ret = vc8x0_I2cWriteStep(bus, pucBlock[ulNRegs - 1], I2C_WRITE | I2C_STOP);
#if 0
	I2cWriteStep(ucAddr, I2C_STOP);
#endif
	return ret;
}

int vc8x0_I2cWriteRegisterBlock(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 *buf, u32 len)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret;

	mutex_lock(&bus->lock);
        ret = vc8x0_I2cWriteBlock(bus, device, addr, buf, len);
	if (ret) {
		dprintk(1, "%s() device=0x%02X, addr=0x%02X, buf=?",
			__func__, device, addr);
	}
	mutex_unlock(&bus->lock);

        return ret;
}

int vc8x0_I2cReadRegisterBlock(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 *buf, u32 len)
{
	int ret;
	mutex_lock(&bus->lock);
	ret = vc8x0_I2cReadBlock(bus, device, addr, buf, len);
	mutex_unlock(&bus->lock);
	return ret;
}

int vc8x0_i2c_register(struct vc8x0_i2c *bus)
{
	struct vc8x0_dev *dev = bus->dev;

	dprintk(1, "%s(bus = %d, reg_base = %x)\n", __func__,
		bus->nr, bus->reg_base);

	vc_write64(bus->reg_base + I2C_PRE_LO_REGISTER, I2C_PRE_LO_DATA);
	vc_write64(bus->reg_base + I2C_PRE_HI_REGISTER, I2C_PRE_HI_DATA);
	vc_write64(bus->reg_base + I2C_CONTROL_REGISTER, I2C_BLOCK_ENABLE);

	return 0;
}

int vc8x0_i2c_unregister(struct vc8x0_i2c *bus)
{
	return 0;
}

