/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int pcm3052_debug;
module_param(pcm3052_debug, int, 0644);
MODULE_PARM_DESC(pcm3052_debug, "enable debug messages [pcm3052]");

#define dprintk(level, fmt, arg...)\
	do { if (pcm3052_debug >= level)\
		printk(KERN_DEBUG "%s/0: " fmt, channel->dev->name, ## arg);\
	} while (0)

static int vc8x0_pcm3052_i2c_read8(struct vc8x0_pcm3052_context *ctx, u8 reg, u8 *val)
{
	int ret;

	ret = vc8x0_I2cReadRegister(ctx->bus, ctx->addr, reg, val);
	return ret;
}

static int vc8x0_pcm3052_i2c_write8(struct vc8x0_pcm3052_context *ctx, u8 reg, u8 val)
{
	int ret;

	ret = vc8x0_I2cWriteRegister(ctx->bus, ctx->addr, reg, val);
	return ret;
}

int vc8x0_pcm3052_init(struct vc8x0_dma_channel *channel, u8 i2caddr)
{
	struct vc8x0_pcm3052_context *ctx = &channel->pcm3052_ctx;

	dprintk(1, "%s(%p, 0x%x)\n", __func__, channel, i2caddr);
	memset(ctx, 0, sizeof(*ctx));

	ctx->dma_channel = channel;
	ctx->bus = &channel->i2c_bus;
	ctx->addr = i2caddr;
	ctx->mute = 0;
	ctx->gain = 0x0d;

	dprintk(1, "%s() bus = %p\n", __func__, ctx->bus);

	return 0;
}

int vc8x0_pcm3052_enable(struct vc8x0_pcm3052_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	struct vc8x0_dev *dev = channel->dev;

	dprintk(1, "%s(base = %04x)\n", __func__, ctx->bus->reg_base);

	/* Power Up the PCM3052 */
	vc_write32(0x930, 0x00000000);
	msleep(50);

	/* Assume reasonable defaults */
	vc8x0_pcm3052_i2c_write8(ctx, 72, ctx->gain);
	
	return 0;
}

int vc8x0_pcm3052_disable(struct vc8x0_pcm3052_context *ctx)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	struct vc8x0_dev *dev = channel->dev;

	dprintk(1, "%s()\n", __func__);

	/* Power Down the PCM3052 */
	vc_write32(0x930, 0x00000002);

	return 0;
}

int vc8x0_pcm3052_procfs_show(struct seq_file *m, struct vc8x0_pcm3052_context *ctx)
{
	int i, c = 0;
	unsigned char buf[0x100];
	seq_printf(m, "PCM3052:\n");

	for (i = 0x41; i <= 0x50; i++)
		vc8x0_pcm3052_i2c_read8(ctx, i, &buf[i]);

	seq_printf(m, "     00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f\n");
	for (i = 0; i < sizeof(buf); i++) {
		if (c == 0)
			seq_printf(m, " %02x:", i);

		if ((i >= 0x41) && (i <= 0x50))
			seq_printf(m, " %02x", buf[i]);
		else
			seq_printf(m, "   ");

		if (++c == 16) {
			seq_printf(m, "\n");
			c = 0;
		}
	}

	return 0;
}

int vc8x0_pcm3052_set_input(struct vc8x0_pcm3052_context *ctx, int nr)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	dprintk(1, "%s(%d)\n", __func__, nr);
	return 0;
}

int vc8x0_pcm3052_set_mute(struct vc8x0_pcm3052_context *ctx, u32 enable)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;

	if (!ctx->dma_channel)
		return 0;

	dprintk(1, "%s(nr=%d, %d)\n", __func__, channel->nr, enable);

	ctx->mute = enable;

	if (enable) {
		vc8x0_pcm3052_i2c_write8(ctx, 72, 0x00);
	} else {
		vc8x0_pcm3052_i2c_write8(ctx, 72, ctx->gain);
	}

	return 0;
}

int vc8x0_pcm3052_get_mute(struct vc8x0_pcm3052_context *ctx, u32 *enabled)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	*enabled = ctx->mute;
	dprintk(1, "%s(nr=%d, %d)\n", __func__, channel->nr, *enabled);
	return 0;
}

int vc8x0_pcm3052_set_volume(struct vc8x0_pcm3052_context *ctx, u32 val)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	u8 v = 8 + val;

	if (!ctx->dma_channel)
		return 0;

	ctx->gain = v;

	dprintk(1, "%s(nr=%d, %d, %02x)\n", __func__, channel->nr, val, ctx->gain);

	vc8x0_pcm3052_i2c_write8(ctx, 72, ctx->gain);

	return 0;
}

int vc8x0_pcm3052_get_volume(struct vc8x0_pcm3052_context *ctx, u32 *val)
{
	struct vc8x0_dma_channel *channel = ctx->dma_channel;
	*val = ctx->gain;
	dprintk(1, "%s(nr=%d, %d)\n", __func__, channel->nr, *val);
	return 0;
}

