/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int spi_debug;
module_param(spi_debug, int, 0644);
MODULE_PARM_DESC(spi_debug, "enable debug messages [spi]");

#define dprintk(level, fmt, arg...)\
	do { if (spi_debug >= level)\
		printk(KERN_DEBUG "%s/0: " fmt, dev->name, ## arg);\
	} while (0)

static int GSPIWrite(struct vc8x0_spi *bus, u32 offset, u32 val)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret = -1, nOuterLoop, nInnerLoop;
	u32 nGSPIStatus, rval;
	u32 nGSPICommand = GSPI_WRITE_COMMAND | ((offset & 0x7FFF) << 16) | (val & 0xFFFF);
	int bComplete = 0;

	dprintk(2, "%s(%d) offset=0x%08x value=0x%08x\n", __func__, bus->nr, offset, val);

	/* Hardware is installed so do it to it... */
	for (nOuterLoop = 0; nOuterLoop < GSPI_RETRY_COUNT; nOuterLoop++)
	{
		nGSPIStatus = vc_read32(bus->reg_base + GSPI_STATUS_REGISTER);
		if (nGSPIStatus == GSPI_DATA_READY_TO_WRITE)
		{
			vc_write32(bus->reg_base + GSPI_WRITE_REGISTER, nGSPICommand);
			for (nInnerLoop = 0; nInnerLoop < GSPI_RETRY_COUNT; nInnerLoop++)
			{
				nGSPIStatus = vc_read32(bus->reg_base + GSPI_STATUS_REGISTER);
				if (nGSPIStatus == GSPI_DATA_READY_TO_READ)
				{
					rval = vc_read32(bus->reg_base + GSPI_READ_REGISTER);
					nInnerLoop = nOuterLoop = GSPI_RETRY_COUNT;
					ret = 0;
					bComplete = 1;
					break;
				}
				msleep(1);
			}
		}
		if (!bComplete)
			msleep(1);
	}
	return ret;
}

static int GSPIRead(struct vc8x0_spi *bus, u32 offset, u32 *val)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret = -1, nOuterLoop, nInnerLoop;
	u32 nGSPIStatus, rval;
	u32 nGSPICommand = GSPI_READ_COMMAND | ((offset & 0x7FFF) << 16);
	int bComplete = 0;

	dprintk(2, "%s(%d) offset=0x%08x SPICommand = 0x%x\n", __func__, bus->nr, offset, nGSPICommand);

	/* Hardware is installed so do it to it... */
	for (nOuterLoop = 0; nOuterLoop < GSPI_RETRY_COUNT; nOuterLoop++)
	{
		nGSPIStatus = vc_read32(bus->reg_base + GSPI_STATUS_REGISTER);
		if (nGSPIStatus == GSPI_DATA_READY_TO_WRITE)
		{
			vc_write32(bus->reg_base + GSPI_WRITE_REGISTER, nGSPICommand);
			for (nInnerLoop = 0; nInnerLoop < GSPI_RETRY_COUNT; nInnerLoop++)
			{
				nGSPIStatus = vc_read32(bus->reg_base + GSPI_STATUS_REGISTER);
				if (nGSPIStatus == GSPI_DATA_READY_TO_READ)
				{
					rval = vc_read32(bus->reg_base + GSPI_READ_REGISTER);
					*val = rval & 0x0000FFFF;
					nInnerLoop = nOuterLoop = GSPI_RETRY_COUNT;
					ret = 0;
					bComplete = 1;
					break;
				}
				msleep(1);
			}
		}
#if 0
		else
			printk(KERN_ERR "%s() not ready_to_write\n", __func__);
#endif

		if (!bComplete)
			msleep(1);
	}
	return ret;
}

int vc8x0_SPIRead(struct vc8x0_spi *bus, u32 offset, u32 *val)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret;

	mutex_lock(&bus->lock);
	ret = GSPIRead(bus, offset, val);
	if (ret) {
		dprintk(1, "%s() offset=0x%08X val=0x%08X", __func__, offset, *val);
	}
	mutex_unlock(&bus->lock);
	msleep(2);
	return ret;
}

int vc8x0_SPIWrite(struct vc8x0_spi *bus, u32 offset, u32 val)
{
	struct vc8x0_dev *dev = bus->dev;
	int ret;

	mutex_lock(&bus->lock);
	ret = GSPIWrite(bus, offset, val);
	if (ret) {
		dprintk(1, "%s() offset=0x%082X val=0x%08X", __func__, offset, val);
	}
	mutex_unlock(&bus->lock);
	msleep(2);
	return ret;
}

int vc8x0_spi_register(struct vc8x0_spi *bus)
{
	struct vc8x0_dev *dev = bus->dev;
	u32 v;

	dprintk(1, "%s(bus = %d, reg_base = %x)\n", __func__,
		bus->nr, bus->reg_base);


	/* Issue a few reads from the rx pipe else the bus seems to be stalled
	 * on startup.
	 */
	v = vc_read32(bus->reg_base + GSPI_READ_REGISTER);
//	v = vc_read32(bus->reg_base + GSPI_READ_REGISTER);
//	v = vc_read32(bus->reg_base + GSPI_READ_REGISTER);
	return 0;
}

int vc8x0_spi_unregister(struct vc8x0_spi *bus)
{
	struct vc8x0_dev *dev = bus->dev;

	dprintk(1, "%s(bus = %d, reg_base = %x)\n", __func__,
		bus->nr, bus->reg_base);

	return 0;
}

