/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * The FPGA generates timestamp records into DMA buffers. The timestamps are in 100ns units
 * since the FPGA fifo begun, stored in a u64. (52 bits are cycled, 12 bits are fixed).
 * Measuring the difference between two timestamps for two 1920x1080i video frames
 * would thus return 333670 or 333671 (100ns units)
 * or 33.37ms.
 * Timestamp frames are 0x20 long and held in the DMA buffers at N*0x20;
 * So each video fifo has 9 video buffers and thus 9 TS buffers.
 * If we're monitoring the timestamps for videoA (AC0) then we look in
 * ts buffer (m_nrLastFrameIdx - as read from the h.w), find the TS buffer
 * then the offset for the record is buf->cpu + (lastframeindex * 0x20)
 * struct {
 *  u64 fifochannel[4];
 * }
 * The upper 12 bits of a timestamp values contains the following fixed markers.
 * vc8x0[0]: vc8x0_timestamp_dump() ac000000020da438 ac100000020da438 ac300000020da438 ac200000020da438
 * 0xAC0 - 1st SDI_0
 * 0xAC1 - 2nd SDI_0
 * 0xAC2 - SDI_0 Cb
 * 0xAC3 - SDI_0 Cr
 * 0xAC4 - SDI_0 I2S0
 * 0xAC5 - SDI_0 I2S1
 * 0xAC6 - SDI_0 I2S2
 * 0xAC7 - SDI_0 I2S3
 * 0xAC8 - 1st DVI_0
 * 0xAC9 - 2nd DVI_0
 * 0xACA - DVI_0 Cb
 * 0xACB - DVI_0 Cr
 * 0xACC - DVI_0 I2S0
 * 0xACD - DVI_0 I2S1
 * 0xACE - DVI_0 I2S2
 * 0xACF - DVI_0 I2S3
 * 0xAD0 - Audio_Ch_0
 * 0xAD1 - Audio_Ch_1
 * 0xAD2 - Audio_Ch_2
 * 0xAD3 - Audio_Ch_3
 *
 * Each timestamp contains the frame time of the last sampled frame for the fifochannel.
 * (one per video channel A-D)
 *
 */

#include "vc8x0.h"

static unsigned int timestamp_debug;
module_param(timestamp_debug, int, 0644);
MODULE_PARM_DESC(timestamp_debug, "enable debug messages [timestamp]");

#define dprintk(level, fmt, arg...)\
	do { if (timestamp_debug >= level)\
		printk(KERN_ERR "%s: " fmt, dev->name, ## arg);\
	} while (0)

void vc8x0_timestamp_dump(struct vc8x0_dma_buffer *buf, int idx)
{
	struct vc8x0_dma_channel *channel = buf->channel;
	struct vc8x0_dev *dev = channel->dev;
	u64 *ts;

	if (buf->type != TYPE_TIMESTAMP) {
		printk(KERN_ERR "%s() invalid buffer %p type %d\n", __func__, buf, buf->type);
		return;
	}

	if ((idx < 0) && (idx > 16)) {
		printk(KERN_ERR "%s() invalid idx %d\n", __func__, idx);
		return;
	}

	channel = buf->channel;
	dev = channel->dev;
	ts = (u64 *)(buf->cpu + (idx * 0x20));

	dprintk(1, "%s() %08llx %08llx %08llx %08llx\n", __func__,
		*(ts + 0), *(ts + 1), *(ts + 2), *(ts + 3)
		);
}

int vc8x0_timestamp_validate(struct vc8x0_dma_buffer *buf, int idx)
{
	struct vc8x0_dma_channel *channel = buf->channel;
	struct vc8x0_dev *dev = channel->dev;
	u64 *ts = 0;
	int restart_fifo = 0;

	if ((idx < 0) && (idx > 16)) {
		printk(KERN_ERR "%s() invalid idx %d\n", __func__, idx);
		return -EINVAL;
	}

	ts = (u64 *)(buf->cpu + (idx * 0x20));

	dprintk(2, "%s() %08llx %08llx %08llx %08llx\n", __func__,
		*(ts + 0), *(ts + 1), *(ts + 2), *(ts + 3)
		);

	channel->prev_timestamp = channel->curr_timestamp;
	channel->curr_timestamp = *ts;
	channel->diff_timestamp = channel->curr_timestamp - channel->prev_timestamp;

	if (channel->type == TYPE_VIDEO) {
		/* 24p is one frame every 41ms, allow 10% headroom then consdier the stream broken and out of sync */
		if ((channel->diff_timestamp == 0) || (channel->diff_timestamp > 450000 /* 45ms */)) {
			restart_fifo++;
		}

		if (restart_fifo) {
			dprintk(1, "%s() %08llx (%08llx) timestamp discontinuity, possible alignment issue detected\n", __func__,
				channel->curr_timestamp,
				channel->diff_timestamp);

			dprintk(1, "%s() Requesting repair of alignment issue (restarting the fifo)\n", __func__);
			channel->alignment_errors++;
			channel->thread_actions |= THREAD_ACTION_FLEX_RESET;
		}
	}
	
	return 0;
}
