/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int colorbars_on_los = 1;
module_param(colorbars_on_los, int, 0644);
MODULE_PARM_DESC(colorbars_on_los, "enable colorbar generation [don't show malformed video]");

static unsigned int video_osd_debug = 0;
module_param(video_osd_debug, int, 0644);
MODULE_PARM_DESC(video_osd_debug, "enable OSD messages [video]");

static int errors_on_colorsbars = 0;

static unsigned int video_debug;
module_param(video_debug, int, 0644);
MODULE_PARM_DESC(video_debug, "enable debug messages [video]");

static unsigned int video_during_irq = 1;
module_param(video_during_irq, int, 0644);
MODULE_PARM_DESC(video_during_irq, "enable during_irq handling [video]");

static unsigned int video_1080p = 1;
module_param(video_1080p, int, 0644);
MODULE_PARM_DESC(video_1080p, "enable 1080i50/60 (0) or 1080p50/60 (1) handling [default 1, 1080p]");

static unsigned int video_nr[] = {[0 ... (VC8X0_MAXBOARDS - 1)] = UNSET };
module_param_array(video_nr, int, NULL, 0444);
MODULE_PARM_DESC(video_nr, "video device numbers");

static unsigned int vbi_nr[] = {[0 ... (VC8X0_MAXBOARDS - 1)] = UNSET };
module_param_array(vbi_nr, int, NULL, 0444);
MODULE_PARM_DESC(vbi_nr, "vbi device numbers");

extern unsigned int vc8x0_dual_1080p60;

#define dprintk(level, fmt, arg...)\
	do { if (video_debug >= level)\
		printk(KERN_ERR "%s/0: " fmt, channel->dev->name, ## arg);\
	} while (0)

#define VBI_LINE_LENGTH 1440
#define VBI_LINE_COUNT 22

static struct vc8x0_format formats[] = {
	{
		.name     = "422packed,YUYV,640x480p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 640,
		.height   = 480,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_640x480p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
/* 720 x 480 */
	{
		.name     = "422packed,YUYV,720x480i59",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 720,
		.height   = 480,
		.depth    = 16,
		.flags    = FORMAT_INTERLACED,
		.id       = FORMAT_720x480i59,
		.frate    = FRAMERATE_59,
		.vbi_total_lines = 45,
		.vbi_field0_lines = 22,
		.vbi_field1_lines = 23,
		.input    = FMT_INPUT_DVI |
			FMT_INPUT_YPRPB |
			FMT_INPUT_SVIDEO |
			FMT_INPUT_COMPOSITE1 |
			FMT_INPUT_COMPOSITE2 |
			FMT_INPUT_COMPOSITE3,
	},
	{
		.name     = "422packed,YUYV,720x480i60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 720,
		.height   = 480,
		.depth    = 16,
		.flags    = FORMAT_INTERLACED,
		.id       = FORMAT_720x480i60,
		.frate    = FRAMERATE_60,
		.vbi_total_lines = 45,
		.vbi_field0_lines = 22,
		.vbi_field1_lines = 23,
		.input    = FMT_INPUT_DVI |
			FMT_INPUT_YPRPB |
			FMT_INPUT_SVIDEO |
			FMT_INPUT_COMPOSITE1 |
			FMT_INPUT_COMPOSITE2 |
			FMT_INPUT_COMPOSITE3,
	},
/* 720 x 576 */
	{
		.name     = "422packed,YUYV,720x576p25",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 720,
		.height   = 576,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_720x576p25,
		.frate    = FRAMERATE_25,
		.input    = FMT_INPUT_DVI |
			FMT_INPUT_YPRPB |
			FMT_INPUT_SVIDEO |
			FMT_INPUT_COMPOSITE1 |
			FMT_INPUT_COMPOSITE2 |
			FMT_INPUT_COMPOSITE3,
	},
	{
		.name     = "422packed,YUYV,720x576i50",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 720,
		.height   = 576,
		.depth    = 16,
		.flags    = FORMAT_INTERLACED,
		.id       = FORMAT_720x576i50,
		.frate    = FRAMERATE_50,
		.input    = FMT_INPUT_DVI |
			FMT_INPUT_YPRPB |
			FMT_INPUT_SVIDEO |
			FMT_INPUT_COMPOSITE1 |
			FMT_INPUT_COMPOSITE2 |
			FMT_INPUT_COMPOSITE3,
	},
/* 800 x 600 */
	{
		.name     = "422packed,YUYV,800x600p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 800,
		.height   = 600,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_800x600p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
/* 1024 x 768 */
	{
		.name     = "422packed,YUYV,1024x768p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1024,
		.height   = 768,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1024x768p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
/* 1280 x 960 */
	{
		.name     = "422packed,YUYV,1280x960p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 960,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x960p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
	{
		.name     = "422packed,YUYV,1280x1024p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 1024,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x1024p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
/* 1280 x 720 */
	{
		.name     = "422packed,YUYV,1280x720p23",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p23,
		.frate    = FRAMERATE_23,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1280x720p24",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p24,
		.frate    = FRAMERATE_24,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1280x720p25",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p25,
		.frate    = FRAMERATE_25,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1280x720p29",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p29,
		.frate    = FRAMERATE_29,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1280x720p30",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p30,
		.frate    = FRAMERATE_30,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1280x720p50",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p50,
		.frate    = FRAMERATE_50,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB | FMT_INPUT_SDI,
	}, {
		.name     = "422packed,YUYV,1280x720p59",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p59,
		.frate    = FRAMERATE_59,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1280x720p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1280,
		.height   = 720,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1280x720p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB | FMT_INPUT_VGA | FMT_INPUT_SDI,
	},
/* 1400 x 1050 */
	{
		.name     = "422packed,YUYV,1400x1050p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1400,
		.height   = 1050,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1400x1050p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
/* 1600 x 1200 */
	{
		.name     = "422packed,YUYV,1600x1200p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1600,
		.height   = 1200,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1600x1200p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
/* 1920 x 1080 */
	{
		.name     = "422packed,YUYV,1920x1080i50",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_INTERLACED,
		.id       = FORMAT_1920x1080i50,
		.frate    = FRAMERATE_50,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	},
	{
		.name     = "422packed,YUYV,1920x1080i59",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_INTERLACED,
		.id       = FORMAT_1920x1080i59,
		.frate    = FRAMERATE_59,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1920x1080i60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_INTERLACED,
		.id       = FORMAT_1920x1080i60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	},
	{
		.name     = "422packed,YUYV,1920x1080p23",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1920x1080p23,
		.frate    = FRAMERATE_23,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1920x1080p24",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1920x1080p24,
		.frate    = FRAMERATE_24,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1920x1080p25",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1920x1080p25,
		.frate    = FRAMERATE_25,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1920x1080p30",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1920x1080p30,
		.frate    = FRAMERATE_30,
		.input    = FMT_INPUT_DVI | FMT_INPUT_YPRPB,
	}, {
		.name     = "422packed,YUYV,1920x1080p50",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1920x1080p50,
		.frate    = FRAMERATE_50,
		.input    = FMT_INPUT_DVI,
	},
	{
		.name     = "422packed,YUYV,1920x1080p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_1920x1080p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_DVI,
	},
	{
		.name     = "422packed,YUYV,1600x1200p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1600,
		.height   = 1200,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_UXGA_1600x1200p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
	{
		.name     = "422packed,YUYV,1440x900p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1440,
		.height   = 900,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_WXGA_1440x900p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
	{
		.name     = "422packed,YUYV,1680x1050p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1680,
		.height   = 1050,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_WSXGA_1680x1050p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
	{
		.name     = "422packed,YUYV,1920x1080p60",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.width    = 1920,
		.height   = 1080,
		.depth    = 16,
		.flags    = FORMAT_PROGRESSIVE,
		.id       = FORMAT_VGAHD_1920x1080p60,
		.frate    = FRAMERATE_60,
		.input    = FMT_INPUT_VGA,
	},
};

/* 75% IRE colorbars */
unsigned char colorbars[7][4] = {{ 0xc0, 0x80, 0xc0, 0x80 },
				 { 0xaa, 0x20, 0xaa, 0x8f },
				 { 0x86, 0xa0, 0x86, 0x20 },
				 { 0x70, 0x40, 0x70, 0x2f },
				 { 0x4f, 0xbf, 0x4f, 0xd0 },
				 { 0x39, 0x5f, 0x39, 0xe0 },
				 { 0x15, 0xe0, 0x15, 0x70 } };
unsigned char blackscreen[4] = { 0x00, 0x80, 0x00, 0x80 };
unsigned char bluescreen[4] = { 0x1d, 0xff, 0x1d, 0x6b };

#define FILL_MODE_COLORBARS 0
#define FILL_MODE_GREENSCREEN 1
#define FILL_MODE_BLUESCREEN 2
#define FILL_MODE_BLACKSCREEN 3
static void fill_frame(struct vc8x0_dma_channel *channel,
	unsigned char *dest_frame, unsigned int width,
	unsigned int height, unsigned int fillmode)
{
	unsigned int width_bytes = width * 2;
	unsigned int i, divider;

	if (fillmode > 3)
		fillmode = 0;

	switch (fillmode) {
	case FILL_MODE_COLORBARS:
		divider = (width_bytes / 7) + 1;
		for (i = 0; i < width_bytes; i += 4)
			memcpy(&dest_frame[i], &colorbars[i / divider], 4);
		break;
	case FILL_MODE_GREENSCREEN:
		memset(dest_frame, 0, width_bytes);
		break;
	case FILL_MODE_BLUESCREEN:
		for (i = 0; i < width_bytes; i += 4)
			memcpy(&dest_frame[i], bluescreen, 4);
		break;
	case FILL_MODE_BLACKSCREEN:
		for (i = 0; i < width_bytes; i += 4)
			memcpy(&dest_frame[i], blackscreen, 4);
	}

	for (i = 1; i < height; i++) {
		memcpy(dest_frame + (width_bytes * i), dest_frame,
		       width_bytes);
	}
}

static int vc8x0_video_generate_osd(struct vc8x0_dma_channel *channel, u8 *dst)
{
	/* Do some text rendering */
	struct vc8x0_format *fmt = 0;
	unsigned char tmp[256];
	int ret;

	if (IS_820E(channel->dev))
		fmt = channel->ad7441_ctx.detected_fmt;

	ret = vc8x0_display_render_reset(&channel->display_ctx, dst, channel->fmt->width);
	if (ret < 0)
		return ret;

	sprintf(tmp, "Uptime %d", jiffies_to_msecs(jiffies));
	ret |= vc8x0_display_render_string(&channel->display_ctx, tmp, strlen(tmp), 0, 0);

	if (fmt) {
		sprintf(tmp, "%dx%dx%d%s",
			fmt->width, fmt->height, fmt->frate,
			fmt->flags == FORMAT_INTERLACED ? "i" : "p");
		ret |= vc8x0_display_render_string(&channel->display_ctx, tmp, strlen(tmp), 0, 1);
	}

	return ret;
}

void vc8x0_channel_loss_of_sync(struct vc8x0_dma_channel *channel, int enable)
{
	/* Option, if the user wants to see malform or incorrectly formatted video
	 * then they can override the default and disable colorbar generation.
	 */
	if (colorbars_on_los == 0)
		return;

	/* If we're asked to disable colorbars and they were already disabled... */
	if ((enable == 0) && (channel->colorbars_active == 0)) {
		/* Do nothing */
		return;
	}

	/* If we're asked to disable colorbars and they were active... */
	if ((enable == 0) && channel->colorbars_active) {
		channel->colorbars_active = 0;

		/* Restart the DMA if we come out of video sync loss */
		channel->thread_actions |= THREAD_ACTION_FLEX_RESET;

		return;
	}

	/* If we're asked to enable colorbars and they were inactive... */
	if (enable && (!channel->colorbars_active)) {
		channel->colorbars_active = 1;
		channel->loss_of_sync++;
		return;
	}

	/* If we're asked to enable colorbars and they were active... */
	if (enable && channel->colorbars_active)
	{
		/* Do nothing, we're already out of sync */
	}
}

void vc8x0_video_timeout(unsigned long data)
{
	struct vc8x0_dma_channel *channel = (struct vc8x0_dma_channel *)data;
	struct vc8x0_buffer *buf;
	u8 *dst;
	unsigned long flags;

	dprintk(1, "%s()\n", __func__);

	if (channel->state != STATE_RUNNING) {
		/* Return without processing the buffer or restarting
		 * the timer
		 */
		printk(KERN_ERR "%s() channel stopped, aborting\n", __func__);
		return;
	}

	/* Return all of the buffers in error state, so the vbi/vid inode
	 * can return from blocking.
	 */
	spin_lock_irqsave(&channel->v4l2_capture_lock, flags);
	while (!list_empty(&channel->v4l2_capture)) {
		buf = list_entry(channel->v4l2_capture.next,
			struct vc8x0_buffer, vb.queue);

		/* See the notes in the video dequeue section related to generating colorbars */
		dst = videobuf_to_vmalloc(&buf->vb);
		if (dst) {
			fill_frame(channel, dst, buf->vb.width, buf->vb.height, FILL_MODE_COLORBARS);
			vc8x0_channel_loss_of_sync(channel, 1);
		}

		if (video_osd_debug & 1)
			vc8x0_video_generate_osd(channel, dst);

		/* re-set the buffer timeout */
		mod_timer(&channel->timeout, jiffies + (HZ / 2));

		if (errors_on_colorsbars == 1)
			buf->vb.state = VIDEOBUF_ERROR;
		else
			buf->vb.state = VIDEOBUF_DONE;
		do_gettimeofday(&buf->vb.ts);
		list_del(&buf->vb.queue);
		wake_up(&buf->vb.done);

	}
	spin_unlock_irqrestore(&channel->v4l2_capture_lock, flags);
}

void vc8x0_vbi_timeout(unsigned long data)
{
	struct vc8x0_dma_channel *channel = (struct vc8x0_dma_channel *)data;
	struct vc8x0_buffer *buf;
	unsigned long flags;

	dprintk(1, "%s()\n", __func__);

	if (channel->state != STATE_RUNNING) {
		/* Return without processing the buffer or restarting
		 * the timer
		 */
		printk(KERN_ERR "%s() channel stopped, aborting\n", __func__);
		return;
	}

	/* Return all of the buffers in error state, so the vbi/vid inode
	 * can return from blocking.
	 */
	spin_lock_irqsave(&channel->vbi_capture_lock, flags);
	while (!list_empty(&channel->vbi_capture)) {
		buf = list_entry(channel->vbi_capture.next,
			struct vc8x0_buffer, vb.queue);

		/* re-set the buffer timeout */
		mod_timer(&channel->vbi_timeout, jiffies + (HZ / 2));

		list_del(&buf->vb.queue);
		buf->vb.state = VIDEOBUF_ERROR;
		wake_up(&buf->vb.done);
		printk(KERN_ERR "%s() Removed a buffer from the queue\n",
			__func__);
	}
	spin_unlock_irqrestore(&channel->vbi_capture_lock, flags);
}

/* Linux VBI handling wants lines 5-21 in a single videobuf buffer in YUY2 format.
 * We'll skip the first 4 lines of the FPGA buffer, convert to 422 and place the resulting
 * pixeldata into a short VBI buffer. Unlikely video, once the single field of VBI data
 * is processed, we'll hand it off to the user sometime after this function has created
 * the content.
 */
void vc8x0_process_vbi_field(u8 *py, struct vc8x0_buffer *vb_buf, int nr,
	struct vc8x0_format *fmt)
{
	u8 *dst = 0;
	u8 *y;
	int i;

	dst = videobuf_to_vmalloc(&vb_buf->vb);

	/* VBI collected buy the PGA starts at line 2, so we need to put this in line 2 in the dest
	 * buffer.
	 */
//	dst += (vb_buf->vb.width);
	if (nr == 1) {
		dst += (vb_buf->vb.width);
		dst += (vb_buf->vb.width) * 21;
	}

	y  = (py + ((fmt->width * (fmt->height / 2)) * nr));
	if (nr == 1) {
		y += (fmt->width * fmt->vbi_field0_lines);
		//y += (fmt->width * 2);
	}

	/* We're going to deinterlace in dwords */
	/* We're going to deinterlace 4 dwords at a time, 8 pixels per cycle */

	/* Process a single field of vbi data, 17 lines max */
	for (i = 0; i < 22; i++) {
		memcpy(dst + (i * 1440), y + (i * fmt->width), 720);
	}
}

void vc8x0_process_video_field(u8 *py, u8 *pu, u8 *pv, struct vc8x0_buffer *vb_buf, int nr,
	u32 vbi_enabled, struct vc8x0_format *fmt)
{
	u32 px[4];
	u32 yp, up, vp;
	u8 *dst = 0;
	u32 *ddst = 0;
	u32 *y, *u, *v;
	u32 ymax;
	int i, pixelcount;

	dst = videobuf_to_vmalloc(&vb_buf->vb);
	dst += ((vb_buf->vb.width * 2) * nr);
	ddst = (u32 *)dst;

	y  = (u32 *)(py + ((vb_buf->vb.width * (vb_buf->vb.height / 2)) * nr));
	u  = (u32 *)(pu + (((vb_buf->vb.width / 2) * (vb_buf->vb.height / 2)) * nr));
	v  = (u32 *)(pv + (((vb_buf->vb.width / 2) * (vb_buf->vb.height / 2)) * nr));

	if (vbi_enabled) {
		/* VBI */
		if (nr == 0) {
			y += ((fmt->width * fmt->vbi_field0_lines) / sizeof(u32));
			u += (((fmt->width / 2) * fmt->vbi_field0_lines) / sizeof(u32));
			v += (((fmt->width / 2) * fmt->vbi_field0_lines) / sizeof(u32));
		} else {
			y += ((fmt->width * fmt->vbi_field0_lines) / sizeof(u32));
			u += (((fmt->width / 2) * fmt->vbi_field0_lines) / sizeof(u32));
			v += (((fmt->width / 2) * fmt->vbi_field0_lines) / sizeof(u32));

			y += ((fmt->width * fmt->vbi_field1_lines) / sizeof(u32));
			u += (((fmt->width / 2) * fmt->vbi_field1_lines) / sizeof(u32));
			v += (((fmt->width / 2) * fmt->vbi_field1_lines) / sizeof(u32));
		}
	}

	/* We're going to deinterlace in dwords */
	/* We're going to deinterlace 4 dwords at a time, 8 pixels per cycle */

	/* Process a single field (height / 2) of width ant 8 pixels per time */
	ymax = ((vb_buf->vb.height / 2) * vb_buf->vb.width) / 8;
	for (i = 0, pixelcount = 0; i < ymax; i++, pixelcount += 8) {

		if (pixelcount == (vb_buf->vb.width)) {
			ddst += (vb_buf->vb.width / 2);
			pixelcount = 0;
		}

		/* highly optimized for intel i686 little endian, which is what the SOW calls for. */
		/* Input data in ram looks like:
		 * y_plane: Y0 Y1 Y2 Y3 Y4 Y5 Y6 Y7 Y8
		 * u_plane: U0 U1 U2 U3
		 * v_plane: V0 V1 V2 V3
		 *
		 * We shuffle these bytes into dwords ...
		 * px[0] = 0xV0Y1U0Y0
		 * px[1] = 0xV1Y3U1Y2
		 * px[2] = 0xV2Y5U2Y4
		 * px[3] = 0xV3Y7U3Y6
		 *
		 * and the dwords in little ending are stored back into a byte buffer,
		 * which then looks like:
		 * bytes 0 - 15 ->
		 * Y0 U0 Y1 V0  Y2 U1 Y3 V1  Y3 U2 Y4 V2  Y5 U3 Y6 V3
		 *
		 */
		
		yp = *(y++);
		up = *(u++);
		vp = *(v++);

		/* Pixels 1, 2 */
		/* Y0 xx Y1 xx */
		px[0] = (yp & 0x000000ff) | ((yp & 0x0000ff00) << 8);

		/* xx U0 xx xx */
		px[0] |= ((up & 0x000000ff) << 8);

		/* xx xx xx V0 */
		px[0] |= ((vp & 0x000000ff) << 24);

		/* Pixels 3, 4 */
		/* Y0 xx Y1 xx */
		px[1] = ((yp & 0x00ff0000) >> 16) | ((yp & 0xff000000) >> 8);

		/* xx U0 xx xx */
		px[1] |= (up & 0x0000ff00);

		/* xx xx xx V0 */
		px[1] |= ((vp & 0x0000ff00) << 16);


		yp = *(y++);

		/* Pixels 5, 6 */
		/* Y0 xx Y1 xx */
		px[2] = (yp & 0x000000ff) | ((yp & 0x0000ff00) << 8);

		/* xx U0 xx xx */
		px[2] |= ((up & 0x00ff0000) >> 8);

		/* xx xx xx V0 */
		px[2] |= ((vp & 0x00ff0000) << 8);


		/* Pixels 7, 8 */
		/* Y0 xx Y1 xx */
		px[3] = ((yp & 0x00ff0000) >> 16) | ((yp & 0xff000000) >> 8);

		/* xx U0 xx xx */
		px[3] |= ((up & 0xff000000) >> 16);

		/* xx xx xx V0 */
		px[3] |= (vp & 0xff000000);

		/* Clock the 8 pixels (4 dwords) into the videobuf image, now as YUYV422 */

		*(ddst++) = px[0];
		*(ddst++) = px[1];
		*(ddst++) = px[2];
		*(ddst++) = px[3];
	}
}

/* Take the payload held in this buffer and pass it to user space.
 * We don't touch lists, we only take a spinlock on the videobuf
 * list, dequeue a user buffer if necessary.
 */
void vc8x0_video_buffer_dequeue(struct vc8x0_dma_buffer *buf)
{
	struct vc8x0_dma_channel *channel = buf->channel;
	struct vc8x0_format *fmt = channel->fmt;
	struct vc8x0_buffer *vb_buf, *vbi_buf;
	unsigned int i, j, do_colorbars = 0;
	u32 shipframe = 0;
	u32 crc = 0;
	u8 *dst;
	u8 *pdst;
#define MEASURE_YUYV 0
#if MEASURE_YUYV
	u64 yuyv[2];
#endif

	/* Local variables to precalculate to performance enhance the deinterlacing. */
	u32 stride, plane_offset, ymax, umax, tmp;

	dprintk(3, "%s() buffer %p\n", __func__, buf);

	/* Process current buffer */
	spin_lock(&channel->v4l2_capture_lock);
	do
	{
		if (list_empty(&channel->v4l2_capture))
			break;

		vb_buf = list_entry(channel->v4l2_capture.next, struct vc8x0_buffer, vb.queue);

		dst = videobuf_to_vmalloc(&vb_buf->vb);
		if (!dst)
			break;

		if (vb_buf->vb.state != VIDEOBUF_QUEUED) {
			printk(KERN_ERR "%s() vb was not QUEUED, is 0x%x\n", __func__, vb_buf->vb.state);
			break;
		}

		if (channel->colorbars_active)
			do_colorbars = 1;

		do
		{
			/* Video resolution input doesn't match the users requested size */
			if (vb_buf->vb.width != fmt->width) {
				dprintk(2, "%s() Resolution miss-match %d != %d, enabling colorbars\n",
					__func__, vb_buf->vb.width, fmt->width);
				do_colorbars = 1;
			}

		} while (0);

		if (do_colorbars == 1) {

			/* Colorbars can happen in one of two ways.
			 * 1) In 1280 mode, when user switches video rf from 1280 to 1920
			 * then we get partial frames and interrupts, we detect this and display colorbars.
			 * 2) In 1920 mode, when user switches video rf from 1920 to 1280
			 * then interrupts don't trigger, buffers timeout and we show color
			 * bars in the vc8x0_video_timeout function.
			 */
			fill_frame(channel, dst, vb_buf->vb.width, vb_buf->vb.height, FILL_MODE_COLORBARS);
			vc8x0_channel_loss_of_sync(channel, 1);
			shipframe = 1;

		} else {
			dprintk(2, "%s() Copied video data into user buffer width = %d height = %d, "
				"size = %ld/%d (dst=%p src=%p) fmt w/h=%d/%d\n",
				__func__,
				vb_buf->vb.width,
				vb_buf->vb.height,
				vb_buf->vb.size,
				(u32)vb_buf->vb.bsize,
				dst, buf->cpu, channel->fmt->width, channel->fmt->height
				);

			vc8x0_channel_loss_of_sync(channel, 0);
			pdst = dst;

			stride = (vb_buf->vb.width * 2);
			plane_offset = vb_buf->vb.width * (vb_buf->vb.height / 2);
			ymax = (((vb_buf->vb.height / 2) - 1) * vb_buf->vb.width);
			umax = ((((vb_buf->vb.height / 2) - 2) * vb_buf->vb.width) / 2);
			tmp = 0;

#if MEASURE_YUYV
			yuyv[0] = jiffies;
#endif
			if (fmt->flags == FORMAT_INTERLACED) {

				vc8x0_process_video_field(
					buf->cpu_y_plane,
					buf->cpu_u_plane,
					buf->cpu_v_plane,
					vb_buf,
					0,
					channel->vbi_enabled,
					fmt);
				vc8x0_process_video_field(
					buf->cpu_y_plane,
					buf->cpu_u_plane,
					buf->cpu_v_plane,
					vb_buf,
					1,
					channel->vbi_enabled,
					fmt);

				if (channel->vbi_enabled) {
					spin_lock(&channel->vbi_capture_lock);
					do {
						if (list_empty(&channel->vbi_capture)) {
							break;
						}

						vbi_buf = list_entry(channel->vbi_capture.next, struct vc8x0_buffer, vb.queue);
						dst = videobuf_to_vmalloc(&vbi_buf->vb);
						if (!dst)
							break;
						vc8x0_process_vbi_field(buf->cpu_y_plane, vbi_buf, 0, fmt);
						vc8x0_process_vbi_field(buf->cpu_y_plane, vbi_buf, 1, fmt);

						do_gettimeofday(&vbi_buf->vb.ts);
						list_del(&vbi_buf->vb.queue);

						vbi_buf->vb.state = VIDEOBUF_DONE;
						wake_up(&vbi_buf->vb.done);

						/* re-set the buffer timeout */
						mod_timer(&channel->vbi_timeout, jiffies + (HZ / 2));

					} while (0);
					spin_unlock(&channel->vbi_capture_lock);
				}

			} else
			if (fmt->flags == FORMAT_PROGRESSIVE) {
				for (i = 0, j = 0; i < (vb_buf->vb.height * vb_buf->vb.width); i++, j += 2) {
					crc += *(buf->cpu_y_plane + i);
					*(pdst + j + 0) = *(buf->cpu_y_plane + i);
				}
				for (i = 0, j = 1; i < ((vb_buf->vb.height * vb_buf->vb.width) / 2); i++, j += 4) {
					*(pdst + j + 0) = *(buf->cpu_u_plane + i);
					*(pdst + j + 2) = *(buf->cpu_v_plane + i);
				}
			}
#if MEASURE_YUYV
			yuyv[1] = jiffies;
			printk(KERN_ERR "%s() %lld jiffies to convert frame\n", __func__, yuyv[1] - yuyv[0]);
#endif

			shipframe = 1;

			if (crc == channel->crc) {
//				printk(KERN_ERR "%s() dup frame %x\n", __func__, crc);
			}
			channel->crc = crc;
		}

		if (shipframe) {

			if (video_osd_debug & 1)
				vc8x0_video_generate_osd(channel, dst);

			do_gettimeofday(&vb_buf->vb.ts);
			list_del(&vb_buf->vb.queue);

			if (do_colorbars && (errors_on_colorsbars == 1))
				vb_buf->vb.state = VIDEOBUF_ERROR;
			else
				vb_buf->vb.state = VIDEOBUF_DONE;
			wake_up(&vb_buf->vb.done);

			/* re-set the buffer timeout */
			mod_timer(&channel->timeout, jiffies + (HZ / 2));
		}

	} while (0);

	spin_unlock(&channel->v4l2_capture_lock);
}

void vc8x0_video_work_handler(struct work_struct *work)
{
        struct vc8x0_dev *dev;
        struct vc8x0_dma_buffer *buf;
        struct list_head *p, *q;
	unsigned long flags;

	struct vc8x0_dma_channel *channel =
		container_of(work, struct vc8x0_dma_channel, workhandler);

	if ((channel == 0) || (channel->dev == 0)) {
		printk(KERN_ERR "%s() bug\n", __func__);
		return;
	}

	dev = channel->dev;

	dprintk(3, "%s()\n", __func__);

	/* The interrupt handler shuffles buffers from the busy list to the dpc list.
	 * The interrupt handler pulls empty buffers from the free list and places them
	 * on the hardware.
	 */

	/* Move any dpc buffers to the full list, we're doing this
	 * to keep the interrupt spinlock usage on dpc to a minimum.
	 * we'll then take a more leasurly dequeue approach when
	 * processing the full list.
	 */
	/* For each dpc buffer move it to the full list in the minimum amount of time */
        spin_lock_irqsave(&channel->dma_buffers_dpc_lock, flags);
        list_for_each_safe(p, q, &channel->dma_buffers_dpc) {

		buf = list_entry(p, struct vc8x0_dma_buffer, list);
                dprintk(3, "%s() dequeueing DPC buffer %p\n", __func__, buf);

		/* Last, put the buffer on the full list for leasurely processing */
		buf->state = STATE_FULL;
		spin_lock(&channel->dma_buffers_full_lock);
                list_move_tail(&buf->list, &channel->dma_buffers_full);
		spin_unlock(&channel->dma_buffers_full_lock);

	}
	spin_unlock_irqrestore(&channel->dma_buffers_dpc_lock, flags);

	/* BAM! The interrupt handler is now free to move on */
	/* BAM! The interrupt handler is now free to move on */
	/* BAM! The interrupt handler is now free to move on */
	/* BAM! The interrupt handler is now free to move on */

	/* Now let's dequeue the full buffers */
	/* For each full buffer, send it to user space */
	spin_lock(&channel->dma_buffers_full_lock);
        list_for_each_safe(p, q, &channel->dma_buffers_full) {

		buf = list_entry(p, struct vc8x0_dma_buffer, list);

		vc8x0_video_buffer_dequeue(buf);

		spin_lock_irqsave(&channel->dma_buffers_busy_lock, flags);
		buf->state = STATE_BUSY;
                list_move_tail(&buf->list, &channel->dma_buffers_busy);
		spin_unlock_irqrestore(&channel->dma_buffers_busy_lock, flags);

	}
	spin_unlock(&channel->dma_buffers_full_lock);
}

int vc8x0_video_irqhandler(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_dma_buffer *buf, *ts_buf;
	int handled = 0;

	u32 lastidx = 0, vid_pciaddr = 0, ts_pciaddr = 0;
	u32 buflen, reg;
	unsigned long flags;

	/* Process the interrupt */
	
	channel->buffers_processed++;
	buflen = channel->fmt->width * channel->fmt->height * 2;
	if (buflen > MAX_USER_VIDEO_BUFFER_SIZE) {
		printk(KERN_ERR "%s() buffer length is corrupt (%x).\n",
			__func__, buflen);
		goto badaddr; 
	}

	/* Find the finished frame and locate it's pci address */
	lastidx = vc_read32(channel->regs.m_nLastFrameIndex);
	if ((lastidx < 1) || (lastidx > MAX_USER_VIDEO_BUFFERS)) {
		printk(KERN_ERR "%s() lastidx out of range, critical (%x).\n", __func__, lastidx);
		goto badaddr;
	}

	if (channel->lastidx == lastidx) {	
		/* We've already processed this frame, skip it,
		 * duplicate interrupt.
		 */
		goto out;
	}
	channel->lastidx = lastidx;

	if (lastidx == 9)
		lastidx = 1;
	else
		lastidx++;

	lastidx--;

	reg = channel->regs.m_nYDmaAddrL + (lastidx * 6 * sizeof(u32));
	vid_pciaddr = vc_read32(reg);

	reg = channel->regs.m_nYTSDmaAddrL + (lastidx * 2 * sizeof(u32));
	ts_pciaddr = vc_read32(reg);

	dprintk(2, "%s() buf completed, reg=%x, vid_physical 0x%x, ts_physical 0x%x, lastidx=%d\n",
		__func__, reg, vid_pciaddr, ts_pciaddr, lastidx);

	/* Find the video buffer by address */
	buf = vc8x0_buffer_busy_get_by_address(channel, vid_pciaddr, TYPE_VIDEO);
	if (buf == 0) {
		dprintk(1, "%s() No busy video dma buffer for address 0x%x\n", __func__, vid_pciaddr);
		channel->buffers_bad_address++;
		goto badaddr;
	}
	buf->used_len = buflen;

	/* Find the timestamp buffer by address */
	ts_buf = vc8x0_buffer_busy_get_by_address(channel, ts_pciaddr, TYPE_TIMESTAMP);
	if (ts_buf == 0) {
		dprintk(1, "%s() No busy timestamp dma buffer for address 0x%x\n", __func__, ts_pciaddr);
		channel->buffers_bad_address++;
		goto badaddr;
	}
	ts_buf->used_len = 0;
	dprintk(2, "%s() bufs completed vid %p nr=%d, ts %p nr=%d\n",
		__func__, buf, buf->nr, ts_buf, ts_buf->nr);

	vc8x0_timestamp_validate(ts_buf, lastidx);

	/* Put the buffer on the DPC list. This list is spinlock held very briefly
	 * so we're not going to be held up.
	 * The only person that holds the DPC spinlock is the deferred work queue.
	 * The only person that holds the busy spinlock is the interrupt handler.
	 * These spinlocks are also held briefly for reporting purposes in other
	 * parts of the driver but they are very timely.
	 */
	spin_lock_irqsave(&channel->dma_buffers_busy_lock, flags);

	/* Last, put the buffer on the DPC list for our deferred worker to process */
	spin_lock_irqsave(&channel->dma_buffers_dpc_lock, flags);
	buf->state = STATE_DPC;
	list_move_tail(&buf->list, &channel->dma_buffers_dpc);
	spin_unlock_irqrestore(&channel->dma_buffers_dpc_lock, flags);

	spin_unlock_irqrestore(&channel->dma_buffers_busy_lock, flags);

	handled++;
badaddr:
	if (handled) {
                if (video_during_irq == 1) {
                        /* Process video now, zero latency */
                        vc8x0_video_work_handler(&channel->workhandler);
               } else {
                        /* Process video via a deferred queue, with possible latencies */
                        schedule_work(&channel->workhandler);
               }
        }
out:
	return handled;
}

static void vc8x0_video_dump_format(struct vc8x0_dma_channel *channel, struct vc8x0_format *fmt)
{
	if (!fmt)
		return;

	dprintk(1, "%s()\n", __func__);
	dprintk(1, " .name = %s\n", fmt->name);
	dprintk(1, " .fourcc = 0x%x\n", fmt->fourcc);
	dprintk(1, " .width = %d\n", fmt->width);
	dprintk(1, " .height = %d\n", fmt->height);
	dprintk(1, " .depth = %d\n", fmt->depth);
	dprintk(1, " .flags = %d\n", fmt->flags);
	dprintk(1, " .id = %d\n", fmt->id);
	dprintk(1, " .frate = %d\n", fmt->frate);
	dprintk(1, " .input = 0x%02x\n", fmt->input);
	if (fmt->input & FMT_INPUT_DVI)
		dprintk(1, "   = DVI\n");
	if (fmt->input & FMT_INPUT_YPRPB)
		dprintk(1, "   = YPRPB\n");
	if (fmt->input & FMT_INPUT_VGA)
		dprintk(1, "   = VGA\n");
	if (fmt->input & FMT_INPUT_SVIDEO)
		dprintk(1, "   = SVIDEO\n");
	if (fmt->input & FMT_INPUT_COMPOSITE1)
		dprintk(1, "   = COMPOSITE1\n");
	if (fmt->input & FMT_INPUT_COMPOSITE2)
		dprintk(1, "   = COMPOSITE2\n");
	if (fmt->input & FMT_INPUT_COMPOSITE3)
		dprintk(1, "   = COMPOSITE3\n");
}

/* Check to see if a given cable input type is compatible with the video format. */
static int vc8x0_video_is_input_compatible(enum video_input_type input, struct vc8x0_format *fmt)
{
	int ret = 0;

//	printk(KERN_ERR "%s(%d, %d [%s])\n", __func__, input, fmt->id, fmt->name);

	switch (input) {
	case VID_INPUT_DVI:
		if (fmt->input & FMT_INPUT_DVI)
			ret = 1;
		break;
	case VID_INPUT_YPRPB:
		if (fmt->input & FMT_INPUT_YPRPB)
			ret = 1;
		break;
	case VID_INPUT_SVIDEO:
		if (fmt->input & FMT_INPUT_SVIDEO)
			ret = 1;
		break;
	case VID_INPUT_COMPOSITE1:
		if (fmt->input & FMT_INPUT_COMPOSITE1)
			ret = 1;
		break;
	case VID_INPUT_COMPOSITE2:
		if (fmt->input & FMT_INPUT_COMPOSITE2)
			ret = 1;
		break;
	case VID_INPUT_COMPOSITE3:
		if (fmt->input & FMT_INPUT_COMPOSITE3)
			ret = 1;
		break;
	case VID_INPUT_VGA:
		if (fmt->input & FMT_INPUT_VGA)
			ret = 1;
		break;
	default:
		ret = 0;
	}

//	printk(KERN_ERR "%s() ret = %d\n", __func__, ret);
	return ret;
}

struct vc8x0_format *vc8x0_video_find_formatById(struct vc8x0_dma_channel *channel, int id)
{
	unsigned int i;
	dprintk(1, "%s(%x)\n", __func__, id);

	for (i = 0; i < ARRAY_SIZE(formats); i++) {
		if (formats[i].id== id)
			return formats + i;
	}

	return 0;
}

/* for the given input params, try to find a matching supported driver rate */
struct vc8x0_format *vc8x0_video_find_format(struct vc8x0_dma_channel *channel,
	u32 pixelformat, u32 width, u32 height, u32 flags, u32 rate)
{
	unsigned int i;
	dprintk(1, "%s(%x, %d, %d, %d, rate=%d)\n", __func__, pixelformat, width, height, flags, rate);

	for (i = 0; i < ARRAY_SIZE(formats); i++) {
		if (video_1080p && (formats[i].id == FORMAT_1920x1080i60))
			continue;
		if (video_1080p && (formats[i].id == FORMAT_1920x1080i50))
			continue;
		if ((!vc8x0_dual_1080p60) && (formats[i].id == FORMAT_1920x1080p50))
			continue;
		if ((!vc8x0_dual_1080p60) && (formats[i].id == FORMAT_1920x1080p60))
			continue;

		/* Skip any formats not comatible with the users selected cable input type.
		 * This will prevent us from selecting an HDMI format for 1920x1080 when the
		 * caller is selecting VGA for example.
		 */
		if (vc8x0_video_is_input_compatible(channel->video_inputnr, &formats[i]) == 0)
			continue;

		if ((formats[i].fourcc == pixelformat) &&
			(formats[i].width == width) &&
			(formats[i].height == height) &&
			(formats[i].flags == flags) &&
			(formats[i].frate == rate)
		)
		{
			return formats + i;
		}
	}

	return 0;
}

static struct vc8x0_format *find_format_by_rate(struct vc8x0_dma_channel *channel,
	struct vc8x0_format *fmt, u32 frate)
{
	unsigned int i;
	dprintk(1, "%s(%p, %d, input=%d)\n", __func__, fmt, frate, channel->video_inputnr);

	if (fmt) {

		for (i = 0; i < ARRAY_SIZE(formats); i++) {
			if (video_1080p && (formats[i].id == FORMAT_1920x1080i60))
				continue;
			if (video_1080p && (formats[i].id == FORMAT_1920x1080i50))
				continue;
			if ((!vc8x0_dual_1080p60) && (formats[i].id == FORMAT_1920x1080p50))
				continue;
			if ((!vc8x0_dual_1080p60) && (formats[i].id == FORMAT_1920x1080p60))
				continue;

			/* Skip any formats not comatible with the users selected cable input type.
			 * This will prevent us from selecting an HDMI format for 1920x1080 when the
			 * caller is selecting VGA for example.
			 */
			if (vc8x0_video_is_input_compatible(channel->video_inputnr, &formats[i]) == 0)
				continue;

			if ((formats[i].fourcc == fmt->fourcc) &&
				(formats[i].width == fmt->width) &&
				(formats[i].height == fmt->height) &&
				(formats[i].frate == frate))
			{
				return formats + i;
			}
		}
	}

	return 0;
}

static struct vc8x0_format *find_format(struct vc8x0_dma_channel *channel, struct v4l2_format *f, u32 flags)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(formats); i++) {
		if (video_1080p && (formats[i].id == FORMAT_1920x1080i60))
			continue;
		if (video_1080p && (formats[i].id == FORMAT_1920x1080i50))
			continue;
		if ((!vc8x0_dual_1080p60) && (formats[i].id == FORMAT_1920x1080p50))
			continue;
		if ((!vc8x0_dual_1080p60) && (formats[i].id == FORMAT_1920x1080p60))
			continue;

		/* Skip any formats not comatible with the users selected cable input type.
		 * This will prevent us from selecting an HDMI format for 1920x1080 when the
		 * caller is selecting VGA for example.
		 */
		if (vc8x0_video_is_input_compatible(channel->video_inputnr, &formats[i]) == 0)
			continue;

		if ((formats[i].fourcc == f->fmt.pix.pixelformat) &&
			(formats[i].width == f->fmt.pix.width) &&
			(formats[i].height == f->fmt.pix.height))
		{
			if ((flags == FORMAT_UNKNOWN) || (flags == formats[i].flags))
				return formats + i;
		}
	}
	return 0;
}
/* ------------------------------------------------------------------- */

static const struct v4l2_queryctrl no_ctl = {
	.name  = "42",
	.flags = V4L2_CTRL_FLAG_DISABLED,
};

static struct vc8x0_ctrl vc8x0_ctls[] = {
	{
		.v = {
			.id            = V4L2_CID_BRIGHTNESS,
			.name          = "Brightness",
			.minimum       = -127,
			.maximum       = 128,
			.step          = 1,
			.default_value = 128,
			.type          = V4L2_CTRL_TYPE_INTEGER,
		},
	}, {
		.v = {
			.id            = V4L2_CID_CONTRAST,
			.name          = "Contrast",
			.minimum       = 0,
			.maximum       = 255,
			.step          = 1,
			.default_value = 128,
			.type          = V4L2_CTRL_TYPE_INTEGER,
		},
	}, {
		.v = {
			.id            = V4L2_CID_SATURATION,
			.name          = "Saturation",
			.minimum       = 0,
			.maximum       = 255,
			.step          = 1,
			.default_value = 128,
			.type          = V4L2_CTRL_TYPE_INTEGER,
		},
	}, {
		.v = {
			.id            = V4L2_CID_HUE,
			.name          = "Hue",
			.minimum       = 0,
			.maximum       = 255,
			.step          = 1,
			.default_value = 0,
			.type          = V4L2_CTRL_TYPE_INTEGER,
		},
	}, {
		.v = {
			.id            = V4L2_CID_AUDIO_MUTE,
			.name          = "Mute",
			.minimum       = 0,
			.maximum       = 1,
			.default_value = 0,
			.type          = V4L2_CTRL_TYPE_BOOLEAN,
		},
	}, {
		.v = {
			.id            = V4L2_CID_AUDIO_VOLUME,
			.name          = "Volume",
			.minimum       = -4,
			.maximum       = 20,
			.step          = 1,
			.default_value = 5,
			.type          = V4L2_CTRL_TYPE_INTEGER,
		},
	}
};
static const int VC8X0_CTLS = ARRAY_SIZE(vc8x0_ctls);

/* Must be sorted from low to high control ID! */
static const u32 vc8x0_user_ctrls[] = {
	V4L2_CID_USER_CLASS,
	V4L2_CID_BRIGHTNESS,
	V4L2_CID_CONTRAST,
	V4L2_CID_SATURATION,
	V4L2_CID_HUE,
	V4L2_CID_AUDIO_VOLUME,
	V4L2_CID_AUDIO_MUTE,
	0
};

static const u32 *ctrl_classes[] = {
	vc8x0_user_ctrls,
	NULL
};

static int vc8x0_set_tvnorm(struct vc8x0_dma_channel *channel, v4l2_std_id norm)
{
	dprintk(1, "%s(norm = 0x%08x) name: [%s]\n",
		__func__,
		(unsigned int)norm,
		v4l2_norm_to_name(norm));

	return 0;
}

static struct video_device *vc8x0_vdev_init(struct vc8x0_dma_channel *channel,
	struct pci_dev *pci,
	struct video_device *template,
	char *type)
{
	struct video_device *vfd;
	dprintk(1, "%s()\n", __func__);

	vfd = video_device_alloc();
	if (NULL == vfd)
		return NULL;
	*vfd = *template;
	vfd->parent = &pci->dev;
	vfd->release = video_device_release;

	snprintf(vfd->name, sizeof(vfd->name), "%s (%s)",
		 vc8x0_boards[channel->dev->board].name, type);

	video_set_drvdata(vfd, channel);

	return vfd;
}

static int vc8x0_ctrl_query(struct v4l2_queryctrl *qctrl)
{
	int i;

	if (qctrl->id < V4L2_CID_BASE ||
	    qctrl->id >= V4L2_CID_LASTP1)
		return -EINVAL;
	for (i = 0; i < VC8X0_CTLS; i++)
		if (vc8x0_ctls[i].v.id == qctrl->id)
			break;
	if (i == VC8X0_CTLS) {
		*qctrl = no_ctl;
		return 0;
	}
	*qctrl = vc8x0_ctls[i].v;
	return 0;
}

/* Video VideoBuf */
static int video_buffer_setup(struct videobuf_queue *q, unsigned int *count,
	unsigned int *size)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_dma_channel *channel = fh->channel;
	struct vc8x0_format *fmt = fh->channel->fmt;

	dprintk(1, "%s() width = %d, height = %d, count = %d\n", __func__,
		fmt->width, fmt->height, *count);

	*size = (fmt->depth * fmt->width * fmt->height) >> 3;

	if (*count == 0)
		*count = 16;

	dprintk(1, "%s() *size = %d *count = %d\n", __func__, *size, *count);

	return 0;
}

void vc8x0_dma_free(struct videobuf_queue *q,
	struct vc8x0_dma_channel *channel, struct vc8x0_buffer *buf)
{
	videobuf_waiton(q, &buf->vb, 0, 0);
	videobuf_vmalloc_free(&buf->vb);
	buf->vb.state = VIDEOBUF_NEEDS_INIT;
}

static int vc8x0_prepare_buffer(struct videobuf_queue *q,
	struct vc8x0_dma_channel *channel,
	struct vc8x0_buffer *buf,
	enum v4l2_field field)
{
	struct vc8x0_format *fmt = channel->fmt;
	int rc = 0;

	/* check settings */
	if (fmt == 0)
		return -EINVAL;

	buf->vb.size = (fmt->width * fmt->height * (fmt->depth / 8));

	dprintk(3, "%s() Resolution: %dx%d\n", __func__, fmt->width, fmt->height);
	dprintk(3, "%s() vb.width = %d\n", __func__, buf->vb.width);
	dprintk(3, "%s() vb.height = %d\n", __func__, buf->vb.height);
	dprintk(3, "%s() vb.size = %ld\n", __func__, buf->vb.size);
	dprintk(3, "%s() vb.bsize = %d\n", __func__, buf->vb.bsize);
	dprintk(3, "%s() vb.baddr = %lx\n", __func__, buf->vb.baddr);

	if ((buf->vb.baddr != 0)  && (buf->vb.bsize < buf->vb.size)) {
		return -EINVAL;
	}

	/* alloc + fill struct (if changed) */
	if (buf->vb.width != fmt->width || buf->vb.height != fmt->height ||
	    buf->vb.field != field || buf->fmt != fmt)
	{
		buf->vb.width  = fmt->width;
		buf->vb.height = fmt->height;
		buf->vb.field  = field;
		buf->fmt       = fmt;

		vc8x0_dma_free(q, channel, buf);
	}

	if (0 != buf->vb.baddr  &&  buf->vb.bsize < buf->vb.size) {
		return -EINVAL;
	}

	if (buf->vb.field == 0 )
		buf->vb.field = V4L2_FIELD_INTERLACED;

	if (VIDEOBUF_NEEDS_INIT == buf->vb.state) {
		buf->vb.width  = fmt->width;
		buf->vb.height = fmt->height;
		buf->vb.field  = field;
		buf->fmt       = fmt;

		if (0 != (rc = videobuf_iolock(q, &buf->vb, 0)))
			goto fail;
	}

	buf->vb.state = VIDEOBUF_PREPARED;
	return 0;

fail:
	vc8x0_dma_free(q, channel, buf);
	return rc;

}

static int video_buffer_prepare(struct videobuf_queue *q, struct videobuf_buffer *vb,
	enum v4l2_field field)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_buffer *buf = container_of(vb, struct vc8x0_buffer, vb);

	return vc8x0_prepare_buffer(q, fh->channel, buf, field);
}

static void video_buffer_queue(struct videobuf_queue *q, struct videobuf_buffer *vb)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_buffer *buf = container_of(vb, struct vc8x0_buffer, vb);

	buf->vb.state = VIDEOBUF_QUEUED;
	list_add_tail(&buf->vb.queue, &fh->channel->v4l2_capture);
}

static void video_buffer_release(struct videobuf_queue *q,
	struct videobuf_buffer *vb)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_buffer *buf = container_of(vb, struct vc8x0_buffer, vb);

	return vc8x0_dma_free(q, fh->channel, buf);
}

static struct videobuf_queue_ops vc8x0_video_qops = {
	.buf_setup    = video_buffer_setup,
	.buf_prepare  = video_buffer_prepare,
	.buf_queue    = video_buffer_queue,
	.buf_release  = video_buffer_release,
};

/* VBI VideoBuf */
static int vbi_buffer_setup(struct videobuf_queue *q, unsigned int *count,
	unsigned int *size)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_dma_channel *channel = fh->channel;
	struct vc8x0_format *fmt = fh->channel->fmt;

	dprintk(1, "%s()\n", __func__);

	if (fmt) {
		dprintk(1, "%s() channel width = %d, height = %d, count = %d\n", __func__,
			fmt->width, fmt->height, *count);
	}

	/* We have to deliver VBI in YUYV 422 */
	*size = VBI_LINE_LENGTH * VBI_LINE_COUNT * 2;
        if (0 == *count)
                *count = 4;
        if (*count < 2)
                *count = 2;
        if (*count > 32)
                *count = 4;

	dprintk(1, "%s() *size = %d *count = %d\n", __func__, *size, *count);

	return 0;
}

static int vc8x0_prepare_vbi_buffer(struct videobuf_queue *q,
	struct vc8x0_dma_channel *channel,
	struct vc8x0_buffer *buf,
	enum v4l2_field field)
{
	struct vc8x0_format *fmt = channel->fmt;
	int rc = 0, size;

	/* check settings */
	if (fmt == 0)
		return -EINVAL;

	size = VBI_LINE_LENGTH * VBI_LINE_COUNT * 2;

	if ((buf->vb.baddr != 0)  && (buf->vb.bsize < buf->vb.size)) {
		return -EINVAL;
	}

	if (buf->vb.field == 0 )
		buf->vb.field = V4L2_FIELD_INTERLACED;

	if (VIDEOBUF_NEEDS_INIT == buf->vb.state) {
		buf->vb.width  = VBI_LINE_LENGTH;
		buf->vb.height = VBI_LINE_COUNT * 2;
		buf->vb.size   = size;
		buf->vb.field  = V4L2_FIELD_SEQ_TB;

		if (0 != (rc = videobuf_iolock(q, &buf->vb, 0)))
			goto fail;
	}

	dprintk(3, "%s() Resolution: %dx%d\n", __func__, fmt->width, fmt->height);
	dprintk(3, "%s() vb.width = %d\n", __func__, buf->vb.width);
	dprintk(3, "%s() vb.height = %d\n", __func__, buf->vb.height);
	dprintk(3, "%s() vb.size = %ld\n", __func__, buf->vb.size);
	dprintk(3, "%s() vb.bsize = %d\n", __func__, buf->vb.bsize);
	dprintk(3, "%s() vb.baddr = %lx\n", __func__, buf->vb.baddr);

	buf->vb.state = VIDEOBUF_PREPARED;
	return 0;

fail:
	vc8x0_dma_free(q, channel, buf);
	return rc;

}
static int vbi_buffer_prepare(struct videobuf_queue *q, struct videobuf_buffer *vb,
	enum v4l2_field field)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_buffer *buf = container_of(vb, struct vc8x0_buffer, vb);

	return vc8x0_prepare_vbi_buffer(q, fh->channel, buf, field);
}

static void vbi_buffer_queue(struct videobuf_queue *q, struct videobuf_buffer *vb)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_buffer *buf = container_of(vb, struct vc8x0_buffer, vb);

	buf->vb.state = VIDEOBUF_QUEUED;
	list_add_tail(&buf->vb.queue, &fh->channel->vbi_capture);
}

static void vbi_buffer_release(struct videobuf_queue *q,
	struct videobuf_buffer *vb)
{
	struct vc8x0_fh *fh = q->priv_data;
	struct vc8x0_buffer *buf = container_of(vb, struct vc8x0_buffer, vb);

	return vc8x0_dma_free(q, fh->channel, buf);
}

static struct videobuf_queue_ops vc8x0_vbi_qops = {
	.buf_setup    = vbi_buffer_setup,
	.buf_prepare  = vbi_buffer_prepare,
	.buf_queue    = vbi_buffer_queue,
	.buf_release  = vbi_buffer_release,
};

static struct videobuf_queue *get_queue(struct vc8x0_fh *fh)
{
	switch (fh->type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		return &fh->vidq;
	case V4L2_BUF_TYPE_VBI_CAPTURE:
		return &fh->vbiq;
	default:
		return 0;
	}
}

static int vc8x0_video_open(struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct vc8x0_dma_channel *channel = video_drvdata(file);
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_fh *fh;
	enum v4l2_buf_type type = 0;

	switch (vdev->vfl_type) {
	case VFL_TYPE_GRABBER:
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		break;
	case VFL_TYPE_VBI:
		vdev->tvnorms = V4L2_STD_NTSC;
		type = V4L2_BUF_TYPE_VBI_CAPTURE;
		break;
	}

	dprintk(1, "%s() dev=%s type=%s channel=%p\n",
		__func__,
		video_device_node_name(vdev),
		v4l2_type_names[type], channel);

	/* allocate + initialize per filehandle data */
	fh = kzalloc(sizeof(*fh), GFP_KERNEL);
	if (NULL == fh)
		return -ENOMEM;

	file->private_data = fh;
	fh->dev      = dev;
	fh->channel  = channel;
	fh->type     = type;

	videobuf_queue_vmalloc_init(&fh->vidq, &vc8x0_video_qops,
		&dev->pci->dev, &channel->v4l2_capture_lock,
		V4L2_BUF_TYPE_VIDEO_CAPTURE,
		V4L2_FIELD_INTERLACED,
		sizeof(struct vc8x0_buffer),
		fh,
		NULL);

	videobuf_queue_vmalloc_init(&fh->vbiq, &vc8x0_vbi_qops,
		&dev->pci->dev, &channel->vbi_capture_lock,
		V4L2_BUF_TYPE_VBI_CAPTURE,
		V4L2_FIELD_INTERLACED,
		sizeof(struct vc8x0_buffer),
		fh,
		NULL);

	mutex_lock(&channel->lock);
	channel->videousers++;
	mutex_unlock(&channel->lock);

	return 0;
}

static ssize_t vc8x0_video_read(struct file *file, char __user *data,
	size_t count, loff_t *ppos)
{
	struct vc8x0_fh *fh = file->private_data;
	struct vc8x0_dma_channel *channel = fh->channel;

	dprintk(1, "%s(%d)\n", __func__, count);

	switch (fh->type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		return videobuf_read_one(&fh->vidq, data, count, ppos,
			file->f_flags & O_NONBLOCK);
	case V4L2_BUF_TYPE_VBI_CAPTURE:
		return videobuf_read_stream(&fh->vbiq, data, count, ppos, 0,
			file->f_flags & O_NONBLOCK);
	default:
		return 0;
	}
}

static unsigned int vc8x0_video_poll(struct file *file,
	struct poll_table_struct *wait)
{
	struct vc8x0_fh *fh = file->private_data;
	struct vc8x0_dma_channel *channel = fh->channel;

	dprintk(3, "%s()\n", __func__);

	switch (fh->type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		return videobuf_poll_stream(file, &fh->vidq, wait);
	case V4L2_BUF_TYPE_VBI_CAPTURE:
		return videobuf_poll_stream(file, &fh->vbiq, wait);
	default:
		return 0;
	}
}

static int vc8x0_video_close(struct file *file)
{
	struct vc8x0_fh *fh = file->private_data;
	struct vc8x0_dma_channel *channel = fh->channel;

	dprintk(1, "%s()\n", __func__);

	mutex_lock(&fh->channel->lock);
	fh->channel->videousers--;

	if (fh->channel->videousers == 0) {
		vc8x0_channel_stop(fh->channel);
	}

	mutex_unlock(&fh->channel->lock);

	switch (fh->type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		videobuf_queue_cancel(&fh->vidq);
		videobuf_mmap_free(&fh->vidq);
		break;
	case V4L2_BUF_TYPE_VBI_CAPTURE:
		break;
	default:
		BUG();
	}

	file->private_data = NULL;
	kfree(fh);
	return 0;
}

static int vc8x0_video_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct vc8x0_fh *fh = file->private_data;
	struct vc8x0_dma_channel *channel = fh->channel;

	dprintk(1, "%s()\n", __func__);

	return videobuf_mmap_mapper(get_queue(fh), vma);
}

/* ------------------------------------------------------------------ */
/* VIDEO CTRL IOCTLS                                                  */

static int vc8x0_get_control(struct vc8x0_dma_channel *channel,
	struct v4l2_control *ctl)
{
	struct vc8x0_dev *dev = channel->dev;
	dprintk(1, "%s()\n", __func__);

	switch (ctl->id) {
	case V4L2_CID_AUDIO_MUTE:
		if (IS_820E(dev))
			vc8x0_pcm3052_get_mute(&dev->dma_channel[ DMA_CHANNEL9 ].pcm3052_ctx, &ctl->value);
		break;
	case V4L2_CID_AUDIO_VOLUME:
		if (IS_820E(dev))
			vc8x0_pcm3052_get_volume(&dev->dma_channel[ DMA_CHANNEL9 ].pcm3052_ctx, &ctl->value);
		break;
	case V4L2_CID_BRIGHTNESS:
		if (IS_820E(dev))
			vc8x0_ad7441_get_brightness(&channel->ad7441_ctx, &ctl->value);
		break;
	case V4L2_CID_CONTRAST:
		if (IS_820E(dev))
			vc8x0_ad7441_get_contrast(&channel->ad7441_ctx, &ctl->value);
		break;
	case V4L2_CID_SATURATION:
		if (IS_820E(dev))
			vc8x0_ad7441_get_saturation(&channel->ad7441_ctx, &ctl->value);
		break;
	case V4L2_CID_HUE:
		if (IS_820E(dev))
			vc8x0_ad7441_get_hue(&channel->ad7441_ctx, &ctl->value);
		break;
	}
	return 0;
}

static int vc8x0_set_control(struct vc8x0_dma_channel *channel, struct v4l2_control *ctl)
{
	struct vc8x0_dev *dev = channel->dev;
	int ret = 0;

	dprintk(1, "%s()\n", __func__);
	switch (ctl->id) {
	case V4L2_CID_AUDIO_MUTE:
		if ((ctl->value >= 0) && (ctl->value <= 1)) {
			if (IS_820E(dev))
				vc8x0_pcm3052_set_mute(&dev->dma_channel[ DMA_CHANNEL9 ].pcm3052_ctx, ctl->value);
		} else
			ret = -EINVAL;
		break;
	case V4L2_CID_AUDIO_VOLUME:
		if ((ctl->value >= -4) && (ctl->value <= 20)) {
			if (IS_820E(dev))
				vc8x0_pcm3052_set_volume(&dev->dma_channel[ DMA_CHANNEL9 ].pcm3052_ctx, ctl->value);
		} else
			ret = -EINVAL;
		break;
	case V4L2_CID_BRIGHTNESS:
		if ((ctl->value >= 0) && (ctl->value <= 255)) {
			if (IS_820E(dev))
				vc8x0_ad7441_set_brightness(&channel->ad7441_ctx, ctl->value);
		} else
			ret = -EINVAL;
		break;
	case V4L2_CID_CONTRAST:
		if ((ctl->value >= 0) && (ctl->value <= 255)) {
			if (IS_820E(dev))
				vc8x0_ad7441_set_contrast(&channel->ad7441_ctx, ctl->value);
		} else
			ret = -EINVAL;
		break;
	case V4L2_CID_SATURATION:
		if ((ctl->value >= 0) && (ctl->value <= 255)) {
			if (IS_820E(dev))
				vc8x0_ad7441_set_saturation(&channel->ad7441_ctx, ctl->value);
		} else
			ret = -EINVAL;
		break;
	case V4L2_CID_HUE:
		if ((ctl->value >= 0) && (ctl->value <= 255)) {
			if (IS_820E(dev))
				vc8x0_ad7441_set_hue(&channel->ad7441_ctx, ctl->value);
		} else
			ret = -EINVAL;
		break;
	}
	return ret;
}

static void init_controls(struct vc8x0_dma_channel *channel)
{
	struct v4l2_control ctrl;
	int i;

	for (i = 0; i < VC8X0_CTLS; i++) {
		ctrl.id = vc8x0_ctls[i].v.id;
		ctrl.value = vc8x0_ctls[i].v.default_value;

		vc8x0_set_control(channel, &ctrl);
	}
}

/* ------------------------------------------------------------------ */
/* VIDEO IOCTLS                                                       */

static int vidioc_g_fmt_vid_cap(struct file *file, void *priv,
	struct v4l2_format *f)
{
	struct vc8x0_fh *fh   = priv;
	struct vc8x0_format *fmt = fh->channel->fmt;
	struct vc8x0_dma_channel *channel = fh->channel;

	dprintk(1, "%s() returns format[%d] %dx%d flags=%d\n", __func__,
		fmt->id, fmt->width, fmt->height, fmt->flags);

	f->fmt.pix.width        = fmt->width;
	f->fmt.pix.height       = fmt->height;

	if (fmt->flags & FORMAT_PROGRESSIVE)
		f->fmt.pix.field        = V4L2_FIELD_NONE;
	else
		f->fmt.pix.field        = V4L2_FIELD_INTERLACED;

	f->fmt.pix.pixelformat  = fmt->fourcc;
	f->fmt.pix.bytesperline =
		(f->fmt.pix.width * fmt->depth) >> 3;
	f->fmt.pix.sizeimage =
		f->fmt.pix.height * f->fmt.pix.bytesperline;

	/* ... according to the spec ... */
	f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;

	return 0;
}

static int vidioc_enum_frameintervals(struct file *file, void *priv, struct v4l2_frmivalenum *f)
{
	struct vc8x0_fh *fh = priv;
	struct vc8x0_dma_channel *channel  = fh->channel;
	int framerates[4] = { 25, 30, 50, 60 };

	dprintk(1, "%s(index=%d)\n", __func__, f->index);

	if ((f->index < 0) || (f->index >= 4))
		return -EINVAL;

	f->type = V4L2_FRMSIZE_TYPE_DISCRETE;
	f->discrete.numerator = 1;
	f->discrete.denominator = framerates[ f->index ];

	dprintk(1, "%s(index=%d) rate = %d/%d\n", __func__, f->index,
		f->discrete.numerator, f->discrete.denominator);

	return 0;
}

/* Check the callers format requirements. If the the hardware is busy we refuse.
 * If the hardware is idle we configure it (we only need to configure the
 * analog frontends. If COMPONENT is the current input the reconfigure, else
 * just remember their chosen format and leave the hardware alone.
 *
 * We need to take into consideration the selected input type before we refuse
 * or accept the input format.
 */
static int vidioc_try_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	struct vc8x0_fh *fh = priv;
	struct vc8x0_dma_channel *channel  = fh->channel;
	struct vc8x0_format *fmt;
	int ret = 0;

	dprintk(1, "%s() width = %d, height = %d\n", __func__, f->fmt.pix.width, f->fmt.pix.height);

	fmt = find_format(channel, f, FORMAT_UNKNOWN);
	if (fmt == 0)
		return -EINVAL;

	dprintk(1, "%s() pixel format and resolution is good %s %dx%d\n", __func__,
		fmt->name, fmt->width, fmt->height);
	
	if (fh->channel->state == STATE_STOPPED) {
		/* Other that during system initialization, this is the only place
		 * we explicitly allow the video format to be modified.
		 */
		fh->channel->fmt = fmt;

		ret = vc8x0_dma_configure(channel);
		if (IS_820E(channel->dev))
			ret |= vc8x0_ad7441_set_format(&channel->ad7441_ctx, fmt);
		if (IS_845E(channel->dev))
			ret |= vc8x0_gs2971a_set_format(&channel->gs2971a_ctx, fmt);

	} else {
		printk(KERN_WARNING
			"%s() someone tried to set the video format while the hardware was running. avoided\n", __func__);
		ret = -EBUSY;
	}

	return ret;
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv,
	struct v4l2_format *f)
{
	return vidioc_try_fmt_vid_cap(file, priv, f);
}

static int vidioc_querycap(struct file *file, void  *priv,
	struct v4l2_capability *cap)
{
	struct vc8x0_dev *dev  = ((struct vc8x0_fh *)priv)->dev;

	strcpy(cap->driver, "vc8x0");
	strlcpy(cap->card, vc8x0_boards[dev->board].name,
		sizeof(cap->card));
	sprintf(cap->bus_info, "PCIe:%s", pci_name(dev->pci));
	cap->version = VC8X0_VERSION_CODE;
	cap->capabilities =
		V4L2_CAP_VIDEO_CAPTURE |
		V4L2_CAP_READWRITE     |
		V4L2_CAP_STREAMING     |
		V4L2_CAP_VBI_CAPTURE;
	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv,
	struct v4l2_fmtdesc *f)
{
	if (unlikely(f->index >= ARRAY_SIZE(formats)))
		return -EINVAL;

	f->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	f->flags = 0;
	strlcpy(f->description, formats[f->index].name,
		sizeof(f->description));
	f->pixelformat = formats[f->index].fourcc;
	f->reserved[0] = 0;
	f->reserved[1] = 0;
	f->reserved[2] = 0;
	f->reserved[3] = 0;

	return 0;
}

static int vidioc_reqbufs(struct file *file, void *priv,
	struct v4l2_requestbuffers *p)
{
	struct vc8x0_fh *fh = priv;
	return videobuf_reqbufs(get_queue(fh), p);
}

static int vidioc_querybuf(struct file *file, void *priv,
	struct v4l2_buffer *p)
{
	struct vc8x0_fh *fh = priv;
	return videobuf_querybuf(get_queue(fh), p);
}

static int vidioc_qbuf(struct file *file, void *priv,
	struct v4l2_buffer *p)
{
	struct vc8x0_fh *fh = priv;
	return videobuf_qbuf(get_queue(fh), p);
}

static int vidioc_dqbuf(struct file *file, void *priv,
	struct v4l2_buffer *p)
{
	struct vc8x0_fh *fh = priv;
	return videobuf_dqbuf(get_queue(fh), p, file->f_flags & O_NONBLOCK);
}

static int vidioc_streamon(struct file *file, void *priv,
	enum v4l2_buf_type i)
{
	int ret;
	struct vc8x0_fh *fh = priv;
	struct vc8x0_dma_channel *channel = fh->channel;
	
	dprintk(1, "%s()\n", __func__);
	dprintk(1, "%s() channel=%d\n", __func__, channel->nr);

	if (fh->type == V4L2_BUF_TYPE_VBI_CAPTURE) {
		dprintk(1, "%s() vbi\n", __func__);
		return videobuf_streamon(get_queue(fh));
	}
	if (unlikely(fh->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)) {
		printk(KERN_ERR "%s() err 1\n", __func__);
		return -EINVAL;
	}
	if (unlikely(i != fh->type)) {
		printk(KERN_ERR "%s() err 2\n", __func__);
		return -EINVAL;
	}

	if (vc8x0_channel_start(fh->channel) < 0) {
		printk(KERN_ERR "%s() channel_start fails\n", __func__);
		return -EINVAL;
	}

	ret = videobuf_streamon(get_queue(fh));

	return ret;
}

static int vidioc_streamoff(struct file *file, void *priv, enum v4l2_buf_type i)
{
	struct vc8x0_fh *fh = priv;
	struct vc8x0_dma_channel *channel = fh->channel;
	struct vc8x0_audio_dev *chip = channel->audio_dev;
	struct snd_pcm_runtime *runtime = 0;
	int err = 0;

	dprintk(1, "%s()\n", __func__);

	if (fh->type == V4L2_BUF_TYPE_VBI_CAPTURE) {
		dprintk(1, "%s() vbi\n", __func__);
		return videobuf_streamoff(get_queue(fh));
	}

	if (fh->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (i != fh->type)
		return -EINVAL;

	if (chip && chip->capture_pcm_substream && chip->capture_pcm_substream->runtime) {
		runtime = chip->capture_pcm_substream->runtime;
		if (runtime->status) {
			runtime->status->state = SNDRV_PCM_STATE_DISCONNECTED;
			wake_up(&runtime->tsleep);
		}
	}

	if (vc8x0_channel_stop(fh->channel) < 0)
		return -EINVAL;

	err = videobuf_streamoff(get_queue(fh));

	return err;
}

static int vidioc_g_fmt_vbi_cap(struct file *file, void *priv,
	struct v4l2_format *format)
{
	format->fmt.vbi.sampling_rate = 28636300;
	format->fmt.vbi.sampling_rate = 28636300 / 2;
	format->fmt.vbi.sampling_rate = 13600000;
	format->fmt.vbi.samples_per_line = VBI_LINE_LENGTH;
	format->fmt.vbi.sample_format = V4L2_PIX_FMT_GREY;
	format->fmt.vbi.offset = 0;

	format->fmt.vbi.start[0] = 1;
	format->fmt.vbi.count[0] = VBI_LINE_COUNT;
	format->fmt.vbi.start[1] = 264;
	format->fmt.vbi.count[1] = VBI_LINE_COUNT;

	return 0;
}

static int vidioc_s_std(struct file *file, void *priv, v4l2_std_id *tvnorms)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	dprintk(1, "%s()\n", __func__);

	mutex_lock(&channel->lock);
	vc8x0_set_tvnorm(channel, *tvnorms);
	mutex_unlock(&channel->lock);

	return 0;
}

static int vidioc_g_std(struct file *file, void *priv, v4l2_std_id *norm)
{
	*norm = V4L2_STD_NTSC_M;
	return 0;
}

static int vc8x0_enum_input(struct vc8x0_dma_channel *channel, struct v4l2_input *i)
{
	struct vc8x0_dev *dev = channel->dev;
	static const char *iname[] = {
		[VC8X0_VMUX_YPRPB] = "YPRPB",
		[VC8X0_VMUX_DVI] = "DVI",
		[VC8X0_VMUX_VGA] = "VGA",
		[VC8X0_VMUX_SVIDEO] = "SVIDEO",
		[VC8X0_VMUX_COMPOSITE1] = "COMPOSITE1",
		[VC8X0_VMUX_COMPOSITE2] = "COMPOSITE2",
		[VC8X0_VMUX_COMPOSITE3] = "COMPOSITE3",
		[VC8X0_VMUX_SDI] = "SDI",
	};
	unsigned int n;
	dprintk(1, "%s(%d)\n", __func__, i->index);

	n = i->index;
	if (n >= 7)
		return -EINVAL;

	if (INPUT(n)->type == 0)
		return -EINVAL;

	memset(i, 0, sizeof(*i));
	i->index = n;
	i->type  = V4L2_INPUT_TYPE_CAMERA;
	strcpy(i->name, iname[INPUT(n)->type]);

	return 0;
}

static int vidioc_enum_input(struct file *file, void *priv,
				struct v4l2_input *i)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	dprintk(1, "%s()\n", __func__);
	return vc8x0_enum_input(channel, i);
}

static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;

	*i = channel->video_inputnr;
	dprintk(1, "%s() returns %d\n", __func__, *i);
	return 0;
}

int vc8x0_video_set_input(struct vc8x0_dma_channel *channel, unsigned int i)
{
	struct vc8x0_dev *dev = channel->dev;
	int ret = 0;

	dprintk(1, "%s(%d)\n", __func__, i);

	if (i >= 7) {
		dprintk(1, "%s() -EINVAL\n", __func__);
		return -EINVAL;
	}

	if (channel->state == STATE_RUNNING) {
		dprintk(1, "%s() -EBUSY\n", __func__);
		return -EBUSY;
	}

	/* We need to stall the keepalive thread while we reconfigure any hardware */
	mutex_lock(&dev->kthread_lock);

	channel->video_inputnr = i;
	channel->vbi_enabled = 0;

	switch (vc8x0_boards[channel->dev->board].input[i].type) {
	case VC8X0_VMUX_DVI:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_DVI);
		break;
	case VC8X0_VMUX_YPRPB:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_YPRPB);
		break;
	case VC8X0_VMUX_VGA:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_VGA);
		break;
	case VC8X0_VMUX_COMPOSITE1:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_COMPOSITE1);
		break;
	case VC8X0_VMUX_COMPOSITE2:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_COMPOSITE2);
		break;
	case VC8X0_VMUX_COMPOSITE3:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_COMPOSITE3);
		break;
	case VC8X0_VMUX_SVIDEO:
		ret = vc8x0_ad7441_set_input(&channel->ad7441_ctx, VID_INPUT_SVIDEO);
		break;
	default:
		ret = -ENODEV;
	}

	mutex_unlock(&dev->kthread_lock);

	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
	struct vc8x0_fh *fh = (struct vc8x0_fh *)priv;
	struct vc8x0_dma_channel *channel = fh->channel;

	/* Do nothing if the input selection didn't change */
	if (channel->video_inputnr == i)
		return 0;

	return vc8x0_video_set_input(channel, i);
}

static int vidioc_queryctrl(struct file *file, void *priv,
				struct v4l2_queryctrl *qctrl)
{
	qctrl->id = v4l2_ctrl_next(ctrl_classes, qctrl->id);
	if (unlikely(qctrl->id == 0))
		return -EINVAL;
	return vc8x0_ctrl_query(qctrl);
}

static int vidioc_g_ctrl(struct file *file, void *priv,
				struct v4l2_control *ctl)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;

	return vc8x0_get_control(channel, ctl);
}

static int vidioc_s_ctrl(struct file *file, void *priv,
				struct v4l2_control *ctl)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;

	return vc8x0_set_control(channel, ctl);
}

static u32 vc8x0_video_get_framerate(struct vc8x0_dma_channel *channel)
{
	return 0;
}

static int vc8x0_video_g_parm(struct file *file, void *priv, struct v4l2_streamparm *s)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	struct v4l2_captureparm *c = &s->parm.capture;
	u32 ff;

	ff = vc8x0_video_get_framerate(channel);

	c->capability = V4L2_CAP_TIMEPERFRAME;
	c->timeperframe.numerator = 1;
	c->timeperframe.denominator = 0;

	dprintk(1, "%s() %d/%d/%d\n", 
		__func__,
		c->capability,
		c->timeperframe.numerator,
		c->timeperframe.denominator);

	return 0;
}

static int vc8x0_video_s_parm(struct file *file, void *priv, struct v4l2_streamparm *s)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	struct v4l2_captureparm *c = &s->parm.capture;
	struct vc8x0_format *fmt;
	int ret = -EINVAL;
	u32 fr = c->timeperframe.denominator;
//	u32 ff;

	dprintk(1, "%s() %d/%d\n", 
		__func__,
		c->timeperframe.numerator,
		c->timeperframe.denominator);

	ret = 0;

	/* Gstreamer will call this function prior to streaming because we told it
	 * via g_parm() that the framerate was 0, it thinks it needs to establish it.
	 */
	fmt = find_format_by_rate(channel, channel->fmt, fr);
	if (fmt) {
		/* Re-assign a new format based on existing established resolution
		 * and framerate.
		 */
		if (channel->state == STATE_STOPPED) {
			dprintk(1, "%s() Changing format based on frate, was:\n", __func__); 
			vc8x0_video_dump_format(channel, channel->fmt);

			/* Other that during system initialization, this is the only place
			 * we explicitly allow the video format to be modified.
			 */
			channel->fmt = fmt;

			dprintk(1, "%s() Changing format based on frate, now:\n", __func__); 
			vc8x0_video_dump_format(channel, channel->fmt);

			ret = vc8x0_dma_configure(channel);
			if (IS_820E(channel->dev))
				ret |= vc8x0_ad7441_set_format(&channel->ad7441_ctx, fmt);
			if (IS_845E(channel->dev))
				ret |= vc8x0_gs2971a_set_format(&channel->gs2971a_ctx, fmt);
		}
	} else
		ret = -EINVAL;

	return ret;
}

static const char *audio_inputs[] = {
	[0] = "HDMI/SDI ch1/2",
	[1] = "HDMI/SDI ch3/4",
	[2] = "HDMI/SDI ch5/6",
	[3] = "HDMI/SDI ch7/8",
};

static int vidioc_enumaudio(struct file *file, void *priv, struct v4l2_audio *i)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	int n = i->index, maxi;

	dprintk(1, "%s(%d)\n", __func__, i->index);

	if (IS_820E(channel->dev)) {
		/* A single audio input type from HDMI */
		maxi = 1; 
	} else {
		/* SDI Audio subchannels 1-4 */
		maxi = 4; 
	}

	if ((i->index < 0) || (i->index >= maxi))
		return -EINVAL;

	memset(i, 0, sizeof(*i));

	i->index = n;
	strcpy(i->name, audio_inputs[n]);
	i->capability = V4L2_AUDCAP_STEREO;
	i->mode = 0;

	return 0;
}

static int vidioc_g_audio(struct file *file, void *priv, struct v4l2_audio *i)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	int inputnr = 0;

	dprintk(1, "%s(%d)\n", __func__, i->index);

	memset(i, 0, sizeof(*i));

	if (IS_820E(channel->dev)) 
		inputnr = 0;
	else
	if (IS_845E(channel->dev)) {
		vc8x0_gs2971a_get_audio_input(&channel->gs2971a_ctx, &inputnr);
	}

	strcpy(i->name, audio_inputs[ inputnr ]);
	i->index = inputnr;
	i->capability = V4L2_AUDCAP_STEREO;

	return 0;
}

static int vidioc_s_audio(struct file *file, void *priv, struct v4l2_audio *i)
{
	struct vc8x0_dma_channel *channel = ((struct vc8x0_fh *)priv)->channel;
	int ret, maxi;

	dprintk(1, "%s(%d)\n", __func__, i->index);

	if (IS_820E(channel->dev)) {
		/* A single audio input type from HDMI */
		maxi = 1; 
	} else {
		/* SDI Audio subchannels 1-4 */
		maxi = 4; 
	}

	if ((i->index < 0) || (i->index >= maxi))
		return -EINVAL;

	if (IS_845E(channel->dev))
		ret = vc8x0_gs2971a_set_audio_input(&channel->gs2971a_ctx, i->index);
	else
		ret = 0;

	return ret;
}

static const struct v4l2_file_operations video_fops = {
	.owner	       = THIS_MODULE,
	.open	       = vc8x0_video_open,
	.release       = vc8x0_video_close,
	.read	       = vc8x0_video_read,
	.poll          = vc8x0_video_poll,
	.mmap	       = vc8x0_video_mmap,
	.ioctl	       = video_ioctl2,
};

static const struct v4l2_ioctl_ops video_ioctl_ops = {
	.vidioc_querycap      = vidioc_querycap,
	.vidioc_enum_fmt_vid_cap  = vidioc_enum_fmt_vid_cap,
	.vidioc_enum_frameintervals  = vidioc_enum_frameintervals,
	.vidioc_g_fmt_vid_cap     = vidioc_g_fmt_vid_cap,
	.vidioc_try_fmt_vid_cap   = vidioc_try_fmt_vid_cap,
	.vidioc_s_fmt_vid_cap     = vidioc_s_fmt_vid_cap,
	.vidioc_reqbufs       = vidioc_reqbufs,
	.vidioc_querybuf      = vidioc_querybuf,
	.vidioc_qbuf          = vidioc_qbuf,
	.vidioc_dqbuf         = vidioc_dqbuf,
	.vidioc_s_std         = vidioc_s_std,
	.vidioc_enum_input    = vidioc_enum_input,
	.vidioc_g_input       = vidioc_g_input,
	.vidioc_s_input       = vidioc_s_input,
	.vidioc_queryctrl     = vidioc_queryctrl,
	.vidioc_g_ctrl        = vidioc_g_ctrl,
	.vidioc_s_ctrl        = vidioc_s_ctrl,
	.vidioc_streamon      = vidioc_streamon,
	.vidioc_streamoff     = vidioc_streamoff,
#ifdef CONFIG_VIDEO_ADV_DEBUG
	.vidioc_g_register    = vc8x0_g_register,
	.vidioc_s_register    = vc8x0_s_register,
#endif
	.vidioc_s_parm        = vc8x0_video_s_parm,
	.vidioc_g_parm        = vc8x0_video_g_parm,
/* Need this for VBI */
	.vidioc_g_fmt_vbi_cap   = vidioc_g_fmt_vbi_cap,
//	.vidioc_s_fmt_vbi_cap   = vidioc_g_fmt_vbi_cap,
//	.vidioc_try_fmt_vbi_cap = vidioc_g_fmt_vbi_cap,
	.vidioc_g_std           = vidioc_g_std,
/* SDI audio input selection */
	.vidioc_enumaudio     = vidioc_enumaudio,
	.vidioc_g_audio       = vidioc_g_audio,
	.vidioc_s_audio       = vidioc_s_audio,
};

static struct video_device vc8x0_vbi_template;
static struct video_device vc8x0_video_template = {
	.name                 = "vc8x0-video",
	.fops                 = &video_fops,
	.ioctl_ops 	      = &video_ioctl_ops,
	.current_norm         = V4L2_STD_NTSC_M,
//	.tvnorms              = V4L2_STD_NTSC_M,
};

void vc8x0_video_unregister(struct vc8x0_dma_channel *channel)
{
	dprintk(1, "%s()\n", __func__);

	if (channel->vbi_dev) {
		if (video_is_registered(channel->vbi_dev))
			video_unregister_device(channel->vbi_dev);
		else
			video_device_release(channel->vbi_dev);
		channel->vbi_dev = NULL;
	}
	if (channel->video_dev) {
		if (video_is_registered(channel->video_dev))
			video_unregister_device(channel->video_dev);
		else
			video_device_release(channel->video_dev);
		channel->video_dev = NULL;
	}
}

int vc8x0_video_register(struct vc8x0_dma_channel *channel)
{
	int err;

	dprintk(1, "%s()\n", __func__);
	spin_lock_init(&channel->slock);

	/* init video dma queues */
	INIT_LIST_HEAD(&channel->vidq.active);
	INIT_LIST_HEAD(&channel->vidq.queued);

	/* Assume a reasonable default video format */
	channel->fmt = &formats[8];

	/* register v4l devices */
	channel->video_dev = vc8x0_vdev_init(channel,
		channel->dev->pci,
		&vc8x0_video_template, "video");
	if (channel->video_dev) {
		err = video_register_device(channel->video_dev,
			VFL_TYPE_GRABBER, video_nr[channel->nr]);
		if (err < 0) {
			printk(KERN_INFO "%s: can't register video device\n",
				channel->dev->name);
			goto fail_unreg;
		}
		printk(KERN_INFO "%s/0: registered device %s [v4l2]\n",
		       channel->dev->name, video_device_node_name(channel->video_dev));

		init_controls(channel);
	}

	/* Initialize VBI and template */
	INIT_LIST_HEAD(&channel->vbiq.active);
	INIT_LIST_HEAD(&channel->vbiq.queued);

	memcpy(&vc8x0_vbi_template, &vc8x0_video_template, sizeof(vc8x0_vbi_template));
	strcpy(vc8x0_vbi_template.name, "vc8x0-vbi");

	channel->vbi_dev = vc8x0_vdev_init(channel,
		channel->dev->pci,
		&vc8x0_vbi_template, "vbi");
	if (channel->vbi_dev) {
		err = video_register_device(channel->vbi_dev,
			VFL_TYPE_VBI, vbi_nr[channel->nr]);
		if (err < 0) {
			printk(KERN_INFO "%s: can't register vbi device\n",
				channel->dev->name);
			goto fail_unreg;
		}
		printk(KERN_INFO "%s/0: registered device %s [v4l2]\n",
		       channel->dev->name, video_device_node_name(channel->vbi_dev));
	}

	return 0;

fail_unreg:
	vc8x0_video_unregister(channel);
	return err;
}

