/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/init.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kmod.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/kdev_t.h>
#include <linux/version.h>
#include <linux/mutex.h>
#include <linux/kthread.h>
#include <linux/freezer.h>
#include <linux/firmware.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#endif
#include <asm/div64.h>
#include <asm/io.h>
#include <media/v4l2-device.h>
#include <media/tuner.h>
#include <media/tveeprom.h>
#include <media/videobuf-dma-sg.h>
#include <media/videobuf-vmalloc.h>
#include <media/v4l2-common.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-chip-ident.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/control.h>
#include <sound/initval.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
#include <sound/tlv.h>
#endif

#include "vc8x0-reg.h"

#define VC8X0_VERSION_CODE KERNEL_VERSION(0, 0, 1)

#define UNSET (-1U)

#define VC8X0_MAXBOARDS 8

/* Max number of inputs by card */
#define MAX_VC8X0_INPUT 8
#define INPUT(nr) (&vc8x0_boards[dev->board].input[nr])

#define VC8X0_BOARD_NOAUTO	UNSET
#define VC8X0_BOARD_UNKNOWN	0
#define VC8X0_BOARD_820E	1
#define VC8X0_BOARD_820E_DEV	2
#define VC8X0_BOARD_845E	3
#define IS_820E(dev)		((dev->board == VC8X0_BOARD_820E) || (dev->board == VC8X0_BOARD_820E_DEV))
#define IS_845E(dev)		(dev->board == VC8X0_BOARD_845E)
#define IS_CHANNEL_CONFIGURED(channel)		(channel->nr)

#define BAR0 0
#define BAR4 4

#define VID_FLEX_STOP	(0x01)
#define VID_FLEX_START	(0x02)
#define VID_FLEX_RESET	(0x04)

/* Enable this to enable bad register reads and verification */
#define ENABLE_BAD_READS 0

/* Read 12 registers from the FPGA during interrupt handling */
#define ENABLE_MONITOR_REGISTERS 1

/* ALSA adjustments */
#define ENABLE_ALSA 1
#define ENABLE_AUDIO_KEEPALIVE 1
#define ENABLE_AUDIO_DMA 1
#define ENABLE_ALSA_MIXER 1
#define ENABLE_ALSA_WORKAROUNDS 1
#define ENABLE_PCM_ANALYSIS 1
#define ENABLE_PCM_FIFO_DEBUG 0

struct vc8x0_dma_channel;

struct vc8x0_ctrl {
	struct v4l2_queryctrl v;
	/* This is a pretty empty structure.... refactor */
};

struct vga_size_t
{
	u32 lcf_min;
	u32 lcf_max;
	u32 width;
	u32 height[2];
	u32 interlaced;
};

struct vga_rate_t
{
	u32 fcl_min;
	u32 fcl_max;
	u32 ulInterval10M;
	u32 ucFps100;
	u32 ucFps099;
	u32 rate;
};

/* Selectable application inputs, consecutive numbered for --set-input
 * visible to the application.
 */
enum video_input_type {
	VID_INPUT_DVI = 0,
	VID_INPUT_YPRPB,
	VID_INPUT_VGA,
	VID_INPUT_SVIDEO,
	VID_INPUT_COMPOSITE1,
	VID_INPUT_COMPOSITE2,
	VID_INPUT_COMPOSITE3,
};

/* We cross reference video_input_type into out formats array, allowing the
 * driver to determine whether a video format is appropriate for a given
 * input. The following enum is used internally only for cross referencing.
 */
enum format_input_type {
	FMT_INPUT_UNKNOWN	= 0x00,
	FMT_INPUT_DVI		= 0x01,
	FMT_INPUT_YPRPB		= 0x02,
	FMT_INPUT_VGA		= 0x04,
	FMT_INPUT_SVIDEO	= 0x08,
	FMT_INPUT_COMPOSITE1	= 0x10,
	FMT_INPUT_COMPOSITE2	= 0x20,
	FMT_INPUT_COMPOSITE3	= 0x40,
	FMT_INPUT_SDI		= 0x80,
};

struct vc8x0_format {
	char  *name;
	u32   fourcc;          /* v4l2 format id */
	int   depth;
	int   width;
	int   height;
#define FORMAT_UNKNOWN          0x00
#define FORMAT_INTERLACED       0x01
#define FORMAT_PROGRESSIVE      0x02
	int   flags;
#define FORMAT_720x480i59       0x02
#define FORMAT_720x480i60       0x03
#define FORMAT_720x576i50       0x04
#define FORMAT_720x576p25       0x05
#define FORMAT_1280x720p23      0x06
#define FORMAT_1280x720p24      0x07
#define FORMAT_1280x720p25      0x08
#define FORMAT_1280x720p29      0x09
#define FORMAT_1280x720p30      0x0A
#define FORMAT_1280x720p50      0x0B
#define FORMAT_1280x720p59      0x0C
#define FORMAT_1280x720p60      0x0D
#define FORMAT_1920x1080i50     0x0E
#define FORMAT_1920x1080i59     0x0F
#define FORMAT_1920x1080i60     0x10
#define FORMAT_1920x1080p23     0x11
#define FORMAT_1920x1080p24     0x12
#define FORMAT_1920x1080p25     0x13
#define FORMAT_1920x1080p29     0x14
#define FORMAT_1920x1080p30     0x15
#define FORMAT_1920x1080p50     0x16
#define FORMAT_1920x1080p59     0x17
#define FORMAT_1920x1080p60     0x18
/* VGA */
#define FORMAT_640x480p60       0x30
#define FORMAT_800x600p60       0x31
#define FORMAT_1024x768p60      0x32
#define FORMAT_1280x960p60      0x33
#define FORMAT_1280x1024p60     0x34
#define FORMAT_1400x1050p60     0x35
#define FORMAT_1600x1200p60     0x36
#define FORMAT_UXGA_1600x1200p60	0x43
#define FORMAT_WXGA_1440x900p60		0x46
#define FORMAT_WSXGA_1680x1050p60	0x48
#define FORMAT_VGAHD_1920x1080p60	0x49
	int   id;
#define FRAMERATE_UNKNOWN	0x00
#define FRAMERATE_23		(23)
#define FRAMERATE_24		(24)
#define FRAMERATE_25		(25)
#define FRAMERATE_29		(29)
#define FRAMERATE_30		(30)
#define FRAMERATE_50		(50)
#define FRAMERATE_59		(59)
#define FRAMERATE_60		(60)
	int frate;

	/* VBI Related */
	u32 vbi_total_lines;
	u32 vbi_field0_lines;
	u32 vbi_field1_lines;

	/* The types of cable inputs this vodeo format is compatible with */
	enum format_input_type input;
};

struct vc8x0_display_context
{
	struct vc8x0_dma_channel *channel;

	u8 *frame, *ptr; /* top left of render image and a working ptr */
	int stride;

	int plotwidth, plotheight;

	int currx, curry;

	/* Rendered font fg and bg colors */
	u8 bg[2], fg[2];
};

struct vc8x0_pcm3052_context {
	struct vc8x0_dma_channel *dma_channel;

	/* I2C Addresses */
	struct vc8x0_i2c *bus;
	u8 addr;

	u32 mute;
	u32 gain;
};

struct vc8x0_gs2971a_context {
	struct vc8x0_dma_channel *dma_channel;
	struct vc8x0_spi *bus;

	u32 mute;
	u32 gain;
	int audio_inputnr;

	u32 regs[0x800];

	/* Detected resolution and rate */
	struct vc8x0_format *requested_fmt;

	struct vc8x0_format *detected_fmt;
	u32 detected_vd_std;

	u64 last_detected; /* jiffy last time in ms, when we attempt a format detect */
	u64 last_locked; /* jiffy last time in ms, when we attempt a lock detect */
};

struct vga_timing_t {
	u32 timing_mode;
	u32 spec_height;
	u32 spec_width;
	u32 active_pixels;
	u32 active_lines;
	u32 total_pixels;
	u32 total_lines;
	u32 fcl;
	u32 bl;
	double spec_frequency;
	double pixel_clock;
	double h_freq;
	double v_freq;
	double real_field_rate;
	double real_frame_rate;
	u32 h_blanking;
	u32 h_sync;
	u32 h_front_porch;
	u32 h_back_porch;
	u32 h_start;
	u32 h_end;
	u32 v_blaking;
	u32 v_sync;
	u32 v_front_porch;
	u32 v_back_porch;
	u32 v_start;
	u32 v_end;
	u32 left_margin;
	u32 right_margin;
	u32 top_margin;
	u32 bottom_margin;
	u32 reduced_blanking;
	u32 interlaced;
};

struct vc8x0_ad7441_context {
	struct vc8x0_dma_channel *dma_channel;

	/* I2C Addresses */
	struct vc8x0_i2c *bus;
	u8 usermap_addr;
	u8 user1map_addr;
	u8 user2map_addr;
	u8 user3map_addr;
	u8 reservedmap_addr;
	u8 hdmimap_addr;
	u8 rksvmap_addr;
	u8 edidmap_addr;

	u8 channel;
	struct vc8x0_format *fmt;

	/* Controls */
	u32 brightness;
	u32 contrast;
	u32 hue;
	u32 saturation;

	/* Detected settings */
	u64 last_detected; /* jiffy last time in ms, when we attempt a format detect */
	u64 last_locked; /* jiffy last time in ms, when we attempt a lock detect */
	u32 detected_width;
	u32 detected_height;
	u32 detected_interlaced;
	struct vc8x0_format *detected_fmt;
	struct vga_timing_t *vga_timing_mode;

	/* physical input */
	enum video_input_type video_input;

	struct detection_t {
		/* false(0) / true(1) */
		u32 valid;
		u32 locked;
		u32 interlaced;

		/* 00 Invalid
		 * 01 Separate HS and VS sync on pins
		 * 10 External CS sync on HS IN pin 
		 * 11 Embedded SOG/SOY
		 */
		u32 sync_mask;

		/* 0 - VS IN pin positiv
		 * 1 - HS IN pin positive
		 */
		/* 00 Invalid
		 * 01 GTF (in vga timings as 00)
		 * 10 CVT  (in vga timings as 01)
		 * 11 DMT (in vga timings as 02)
		 */
		u32 sync_polarity;

		u32 bl;

		/* Number of lines in the field */
		u32 lcf;

		/* Number of lines in a vsync period */
		u32 lcvs;

		/* Field Length Count - Number of clock cycles between successive vsyncs */
		u32 fcl28;

		/* Field Rate Hz * 100, eg 5994 */
		u32 field_rate;

		/* Raw register data */
		u8 stat_10[4], stdi_b1[5], fcl[2];
	
	} detection;
};

struct vc8x0_eeprom_context
{
	struct vc8x0_i2c *bus;
	u8 sn[10];
	u8 uid[8];
	u8 page0[32];

	u8 addr;
};

enum vc8x0_itype {
	VC8X0_VMUX_YPRPB = 1,
	VC8X0_VMUX_DVI,
	VC8X0_VMUX_VGA,
	VC8X0_VMUX_SVIDEO,
	VC8X0_VMUX_COMPOSITE1,
	VC8X0_VMUX_COMPOSITE2,
	VC8X0_VMUX_COMPOSITE3,
	VC8X0_VMUX_SDI,
};
#define IS_INPUT_YPRPB(channel)		(channel->video_inputnr == VC8X0_VMUX_YPRPB)
#define IS_INPUT_DVI(channel)		(channel->video_inputnr == VC8X0_VMUX_DVI)
#define IS_INPUT_VGA(channel)		(channel->video_inputnr == VC8X0_VMUX_VGA)
#define IS_INPUT_SVIDEO(channel)	(channel->video_inputnr == VC8X0_VMUX_SVIDEO)
#define IS_INPUT_COMPOSITE1(channel)	(channel->video_inputnr == VC8X0_VMUX_COMPOSITE1)
#define IS_INPUT_COMPOSITE2(channel)	(channel->video_inputnr == VC8X0_VMUX_COMPOSITE2)
#define IS_INPUT_COMPOSITE3(channel)	(channel->video_inputnr == VC8X0_VMUX_COMPOSITE3)
#define IS_INPUT_SDI(channel)		(channel->video_inputnr == VC8X0_VMUX_SDI)

#define MAX_VIDEO_DSP_FRAMES		(9)
#define MAX_AUDIO_DSP_FRAMES		(9)
#define MAX_USER_VIDEO_BUFFERS		MAX_VIDEO_DSP_FRAMES
#define MAX_USER_AUDIO_BUFFERS		MAX_AUDIO_DSP_FRAMES
#define MAX_USER_TIMESTAMP_BUFFERS	(1)

#define MAX_USER_VIDEO_BUFFER_SIZE	(((1920 * 2 ) * 1090) + 4096)
#define MAX_USER_AUDIO_BUFFER_SIZE	((1024 * 16 * 4) + 4096)
#define MAX_USER_TIMESTAMP_BUFFER_SIZE	((1024 * 16 * 1) + 4096)

enum buffer_payload_t {
	TYPE_UNKNOWN = 0,
	TYPE_VIDEO,
	TYPE_AUDIO,
	TYPE_TIMESTAMP
};

struct vc8x0_audio_dev {
	struct vc8x0_dma_channel *channel;

	unsigned int            copied_position;
        struct snd_pcm_substream   *capture_pcm_substream;

        unsigned int            rcv_position;
        struct  snd_card        *card;
        int                     card_close;

        int                     users;
        int                     pm_state;

	enum pcm_stream_state {
		STREAM_OFF,
		STREAM_ON,
		STREAM_SUSPEND,
	} capture_stream;
};

struct vc8x0_dma_buffer {
	struct list_head list;

	u32 nr;	/* 0..X */

	enum buffer_payload_t type;

	enum state_t {
		STATE_UNKNOWN = 0, STATE_FREE, STATE_BUSY, STATE_DPC, STATE_FULL
	} state;

	struct vc8x0_dma_channel *channel;

	/* How much of the buffer at cpu is really used */
	u32 used_len;

	/* User allocated buffer where the video/audio/vbi payload will
	 * be delivered to, by the hardware.
	 */
	u8 		*cpu;		/* Virtual address */
	u32		cpu_size;	/* Allocation size */
	dma_addr_t	cpu_dma;	/* Physical DMA address */

	/* Some in kernel pointers to VM memory so we can deal with the YUV
	 * planes after the fact during buffer dequeue.
	 */
	dma_addr_t	dma_y_plane;		/* Physical DMA address */
	u8 		*cpu_y_plane;		/* Virtual address */
	u32 		cpu_y_plane_size;	/* Bytes */

	dma_addr_t	dma_u_plane;		/* Physical DMA address */
	u8 		*cpu_u_plane;		/* Virtual address */
	u32 		cpu_u_plane_size;	/* Bytes */

	dma_addr_t	dma_v_plane;		/* Physical DMA address */
	u8 		*cpu_v_plane;		/* Virtual address */
	u32 		cpu_v_plane_size;	/* Bytes */

	u64 last_used; /* jiffy last time in ms, re received this via dma */
};

struct vc8x0_buffer {

	/* common v4l buffer stuff -- must be first */
	struct videobuf_buffer vb;

	unsigned int		bpl;
	struct vc8x0_format	*fmt;
	u32			count;
};

struct vc8x0_i2c {
	struct vc8x0_dev *dev;

	/* A lock to protect the I2C bus from concurrent access */
	struct mutex lock;
	int nr;

	/* Offset with the PCIe MMIO for base address of the I2C block */
	u32 reg_base;
};

struct vc8x0_spi {
	struct vc8x0_dev *dev;

	/* A lock to protect the I2C bus from concurrent access */
	struct mutex lock;
	int nr;

	/* Offset with the PCIe MMIO for base address of the I2C block */
	u32 reg_base;
};

struct vc8x0_fh {
	struct vc8x0_dev *dev;
	struct vc8x0_dma_channel *channel;
	enum v4l2_buf_type type;
	u32 resources;

	struct videobuf_queue vidq;
	struct videobuf_queue vbiq;

	atomic_t v4l_reading;
};

struct vc8x0_dmaqueue {
	struct list_head active;
	struct list_head queued;
	struct timer_list timeout;
	u32 count;
};

struct vc8x0_dma_channel {
	struct vc8x0_dev *dev;
	struct mutex lock;
	int nr;

	/* Thread actions. The irq handler can defer an action into the
	 * card specific thread, such as asking for the flex processor to
	 * be reset to clear up and video alignment issues.
	 */
#define THREAD_ACTION_FLEX_RESET	(1)
	u32 thread_actions;

	/* Type of channel AUDIO, VIDEO, TIMESTAMP, etc */
	enum buffer_payload_t type;
	u32 max_buffers;

	/* Each channels gets it's own I2C bus */
	struct vc8x0_i2c i2c_bus;

	/* Each channels gets it's own SPI controller */
	struct vc8x0_spi spi_bus;

	/* We're using a deferred work queue to hander IRQ related buffers */
	struct work_struct workhandler;

	/* Statistics */
	u64 buffers_dropped;
	u64 buffers_processed;
	u64 buffers_bad_address;
	u64 alignment_errors;

	/* AD7441 device */
	struct vc8x0_ad7441_context ad7441_ctx;

	/* PCM3052 device */
	struct vc8x0_pcm3052_context pcm3052_ctx;

	/* Gennum 2971A SDI A/V Recevier */
	struct vc8x0_gs2971a_context gs2971a_ctx;

	/* Debugging OSD */
	struct vc8x0_display_context display_ctx;

	enum channel_running_state {
		STATE_UNDEFINED = 0,
		STATE_STOPPED,
		STATE_RUNNING
	} state;

	/* DMA registers and programming */
	/* Register addresses specific to a channel */
	struct {

		/* Common Registers */
		u32 m_nXferCnt;
		u32 m_nDestXferCnt;
		u32 m_nFrameCnt;
		u32 m_nStreamDisabled;
		u32 m_nLastFrameIndex;
		u32 m_nLastAddrPtr;
		u32 m_nDestAddrPtr;
		u32 m_nLastTSAddrPtr;
		u32 m_nDestTSAddrPtr;
		u32 m_nFrameXferCnt;

		/* Video Registers */
		u32 m_nYXferSizeFull;
		u32 m_nUVXferSizeFull;
		u32 m_nYXferCtrlFull;
		u32 m_nUXferCtrlFull;
		u32 m_nVXferCtrlFull;
		u32 m_nYXferSizePart;
		u32 m_nUVXferSizePart;
		u32 m_nYXferCtrlPart;
		u32 m_nUXferCtrlPart;
		u32 m_nVXferCtrlPart;
		u32 m_nYTSXferCtrl;
		u32 m_nUTSXferCtrl;
		u32 m_nVTSXferCtrl;
		u32 m_nYDestAddrL;
		u32 m_nYDestAddrH;
		u32 m_nUDestAddrL;
		u32 m_nUDestAddrH;
		u32 m_nVDestAddrL;
		u32 m_nVDestAddrH;
		u32 m_nYTSDestAddrL;
		u32 m_nYTSDestAddrH;
		u32 m_nYDmaAddrL;
		u32 m_nYDmaAddrH;
		u32 m_nUDmaAddrL;
		u32 m_nUDmaAddrH;
		u32 m_nVDmaAddrL;
		u32 m_nVDmaAddrH;
		u32 m_nYTSDmaAddrL;
		u32 m_nYTSDmaAddrH;

		/* Audio registers */
		u32 m_nSP1XferStepSize;
		u32 m_nSP1XferCtrl;
		u32 m_nSP2XferCtrl;
		u32 m_nSP3XferCtrl;
		u32 m_nSP4XferCtrl;
		u32 m_nSP1TSXferCtrl;
		u32 m_nSP2TSXferCtrl;
		u32 m_nSP3TSXferCtrl;
		u32 m_nSP4TSXferCtrl;
		u32 m_nSP1DestAddrL;
		u32 m_nSP1DestAddrH;
		u32 m_nSP1TSDestAddrL;
		u32 m_nSP1TSDestAddrH;
		u32 m_nSP1DmaAddrL;
		u32 m_nSP1DmaAddrH;
		u32 m_nSP1TSDmaAddrL;
		u32 m_nSP1TSDmaAddrH;
		u32 m_nFifoGo;
		u32 m_nMisc;

	} regs, offsets, defaults;

	u32 m_nBlockSize;
	u32 m_nDMABuffers;
	u32 m_arrLastFrameIndexId;
	u32 m_arrLastFrameIndex;
	u32 lastidx;

	/* End - DMA registers and programming */

	/* Audio */
        struct vc8x0_audio_dev *audio_dev;

	/* Video */
	/* The currently selected video format, cached for query purposes to multiple
	 * reads can detect what another application is using.
	 */
	spinlock_t		v4l2_capture_lock;
	struct list_head	v4l2_capture;
	struct timer_list	timeout;
	u32			videousers;
	struct vc8x0_format	*fmt;
        spinlock_t		slock;
        struct video_device	*video_dev;
        struct vc8x0_dmaqueue	vidq;
        int			video_inputnr;

	/* VBI */
	spinlock_t		vbi_capture_lock;
	struct list_head	vbi_capture;
	struct video_device	*vbi_dev;
	struct vc8x0_dmaqueue	vbiq;
	struct timer_list	vbi_timeout;

	/* BUFFERING */

	/* DMA VIDEO buffers that we push to the hardware */
	/* Buffers that are free for allocations */
	spinlock_t dma_buffers_free_lock;
	struct list_head dma_buffers_free;

	/* Buffers that are busy on the hardware */
	spinlock_t dma_buffers_busy_lock;
	struct list_head dma_buffers_busy;

	/* Buffers that were dequeueed by the IRQ handler and
	 * are ready for dpc processing.  This is an unusual
	 * list but it exists inbetweem busy and full so that
	 * the interrupt handler isn't blocked while
	 * the full list is being processed, the spinlock for the dpc
	 * list is guaranteed by the driver only to be held
	 * for tiny amounts of time. */
	spinlock_t dma_buffers_dpc_lock;
	struct list_head dma_buffers_dpc;

	/* buffers that the DPC has placed into it's work list
	 * and are scheduled for pushing to user space. */
	spinlock_t dma_buffers_full_lock;
	struct list_head dma_buffers_full;
	/* End of DMA VIDEO buffers that we push to the hardware */

	/* DMA TIMESTAMP buffers that we push to the hardware */
	/* Buffers that are free for allocations */
	spinlock_t dma_timestamp_buffers_free_lock;
	struct list_head dma_timestamp_buffers_free;

	/* Buffers that are busy on the hardware */
	spinlock_t dma_timestamp_buffers_busy_lock;
	struct list_head dma_timestamp_buffers_busy;

	/* Buffers that were dequeueed by the IRQ handler and
	 * are ready for dpc processing.  This is an unusual
	 * list but it exists inbetweem busy and full so that
	 * the interrupt handler isn't blocked while
	 * the full list is being processed, the spinlock for the dpc
	 * list is guaranteed by the driver only to be held
	 * for tiny amounts of time. */
	spinlock_t dma_timestamp_buffers_dpc_lock;
	struct list_head dma_timestamp_buffers_dpc;

	/* buffers that the DPC has placed into it's work list
	 * and are scheduled for pushing to user space. */
	spinlock_t dma_timestamp_buffers_full_lock;
	struct list_head dma_timestamp_buffers_full;
	/* End of DMA TIMESTAMP buffers that we push to the hardware */

	u32 colorbars_active;
	u32 loss_of_sync;
	u32 crc;

	/* Timestamp, let's monitor incoming frames so we can detect frame loss
	 * or video frame alignment issues of the FPGA fails to detect AD7441 DE
	 * correctly.
	 */
	u64 prev_timestamp, curr_timestamp, diff_timestamp;
	u64 last_timestamp_err;

	/* VBI is available on some inputs */
	u32 vbi_enabled;

#if ENABLE_PCM_FIFO_DEBUG
	/* Audio analysis */
	int aa_size;
	u32 *aa_buf;
#endif
};

struct vc8x0_input {
	enum vc8x0_itype type;
};

struct vc8x0_board {
	char *name;
	enum {
		FW_UNKNOWN = 0,
		FW_DEVELOPMENT,
		FW_PRODUCTION,
		FW_SDI,
	} fwtype;

	/* Vendors can and do run the PCIe bridge at different
	 * clock rates, driven physically by crystals on the PCBs.
	 * The core has to accomodate this. This allows the user
	 * to add new boards with new frequencys. The value is
	 * expressed in Hz.
	 *
	 * The core framework will default this value based on
	 * current designs, but it can vary.
	 */
	struct vc8x0_input input[MAX_VC8X0_INPUT];
};

struct vc8x0_subid {
	u16     subvendor;
	u16     subdevice;
	u32     card;
};

struct vc8x0_dev {
	struct list_head devlist;

	atomic_t                   refcount;
	spinlock_t		   reglock;

	/* board details */
	int                        nr;
	struct mutex               lock;
	unsigned int               board;
	char                       name[32];

	/* pci stuff */
	struct pci_dev             *pci;
	unsigned char              pci_rev, pci_lat;
	u64                        __iomem *lmmio064;
	u32                        __iomem *lmmio032;
	u32                        __iomem *lmmio4;
	u8                         __iomem *bmmio;
	u32                        interrupts_enabled;

	/* I2C Bus for the eeprom */
	struct vc8x0_i2c i2c_bus;

	/* A kernel thread to keep the HDMI video frontend alive. */
 	struct task_struct	*kthread;
	struct mutex 		kthread_lock;

	/* DMA Channel related */
#define DMA_CHANNEL1 (0)
#define DMA_CHANNEL2 (1)
#define DMA_CHANNEL3 (2)
#define DMA_CHANNEL4 (3)
#define DMA_CHANNEL5 (4)
#define DMA_CHANNEL6 (5)
#define DMA_CHANNEL7 (6)
#define DMA_CHANNEL8 (7)
#define DMA_CHANNEL9 (8)
#define MAX_DMA_CHANNELS (9)
	struct vc8x0_dma_channel dma_channel[ MAX_DMA_CHANNELS ];

	/* EEPROM device */
	struct vc8x0_eeprom_context eeprom_ctx;

	/* GN4124 Application Layer */
	u32 m_nMaxReadSize;
	u32 m_nAudioBlockSize;
	u32 m_nVideoBlockSize;
	u32 m_nPCIePayloadSize;

	/* RISC Core */
	int risc_running;

	/* Interrupt handling */
	u32 m_nInterruptMask1;
	u32 m_nInterruptMask2;
	u64 m_nALInterruptCounts[7];

#if ENABLE_BAD_READS
	u32 bad_read_prior_val;
	struct work_struct bad_read_workhandler;
#endif
	u32 last0e04;

#if ENABLE_MONITOR_REGISTERS
	/* 12 new monitor registers as of v3.0.101 */
#define MAX_MONITOR_REGISTERS (12)
	u32 monitor_regs[MAX_MONITOR_REGISTERS];
#endif
};

/* PCIe register read/write for the GN4124 layer BAR4--------- */
#define gn_read(reg)             readl(dev->lmmio4 + ((reg)>>2))
#define gn_write(reg, value)     { writel((value), dev->lmmio4 + ((reg)>>2)); /* printk(KERN_ERR "wgn:%p(%x)\n", dev->lmmio4 + ((reg) >> 2), value ); */ } 
#define gn_andor(reg, mask, value)\
{\
	u32 newval = (readl(dev->lmmio4+((reg)>>2)) & ~(mask)) | ((value) & (mask));\
	writel(newval, dev->lmmio4+((reg)>>2));\
}
#define gn_set(reg, bit)	gn_andor((reg), (bit), (bit))
#define gn_clear(reg, bit)	gn_andor((reg), (bit), 0)

/* PCIe register read/write for the Application layer BAR0 (qword)---- */
/* Complication, my 32bit dev box is truncing 64bit writes via writeq,
 * compensate with a split write. oddly, reads appear to be OK.
 */
#define vc_read64(reg)		vc8x0_read64(dev, reg)
#define vc_write64(reg, value)	vc8x0_write64(dev, reg, value)

/* 32BIT REGISTER WRITES */
#define vc_read32(reg)			vc8x0_read32(dev, reg)
#define vc_write32(reg, value)		vc8x0_write32(dev, reg, value)
#define vc_andor32(reg, mask, value)	vc8x0_andor32(dev, reg, mask, value)
#define vc_set32(reg, bit)		vc_andor32((reg), (bit), (bit))
#define vc_clear32(reg, bit)		vc_andor32((reg), (bit), 0)


/* ----------------------------------------------------------- */
/* vc8x0-core.c                                              */

/* ----------------------------------------------------------- */
/* vc8x0-cards.c                                             */
extern struct vc8x0_board vc8x0_boards[];
extern const unsigned int vc8x0_bcount;

extern struct vc8x0_subid vc8x0_subids[];
extern const unsigned int vc8x0_idcount;

extern void vc8x0_card_list(struct vc8x0_dev *dev);
extern void vc8x0_gpio_setup(struct vc8x0_dev *dev);
extern void vc8x0_card_setup(struct vc8x0_dev *dev);
extern void vc8x0_card_setup_pre_i2c(struct vc8x0_dev *dev);
u32 vc8x0_read32(struct vc8x0_dev *dev, u32 reg);
void vc8x0_write32(struct vc8x0_dev *dev, u32 reg, u32 value);
void vc8x0_andor32(struct vc8x0_dev *dev, u32 reg, u32 mask, u32 value);

u64 vc8x0_read64(struct vc8x0_dev *dev, u32 reg);
void vc8x0_write64(struct vc8x0_dev *dev, u32 reg, u32 value);

/* ----------------------------------------------------------- */
/* vc8x0-i2c.c                                                */
extern int vc8x0_i2c_register(struct vc8x0_i2c *bus);
extern int vc8x0_i2c_unregister(struct vc8x0_i2c *bus);
int vc8x0_I2cReadRegister(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 *buf);
int vc8x0_I2cReadRegisterBlock(struct vc8x0_i2c *bus, u8 device, u8 reg, u8 *buf, u32 len);
int vc8x0_I2cWriteRegister(struct vc8x0_i2c *bus, u8 device, u8 reg, u8 val);
int vc8x0_I2cWriteRegisterBlock(struct vc8x0_i2c *bus, u8 device, u8 addr, u8 *buf, u32 len);

/* ----------------------------------------------------------- */
/* vc8x0-spi.c                                                */
extern int vc8x0_spi_register(struct vc8x0_spi *bus);
extern int vc8x0_spi_unregister(struct vc8x0_spi *bus);
int vc8x0_SPIRead(struct vc8x0_spi *bus, u32 offset, u32 *val);
int vc8x0_SPIWrite(struct vc8x0_spi *bus, u32 offset, u32 val);

/* ----------------------------------------------------------- */
/* vc8x0-eeprom.c                                            */
int vc8x0_eeprom_init(struct vc8x0_i2c *bus, struct vc8x0_eeprom_context *ctx, u8 i2caddr);
int vc8x0_eeprom_enable(struct vc8x0_eeprom_context *ctx);
int vc8x0_eeprom_disable(struct vc8x0_eeprom_context *ctx);
int vc8x0_eeprom_unlockFPGA(struct vc8x0_eeprom_context *ctx);

/* ----------------------------------------------------------- */
/* vc8x0-ad7441.c                                              */
int vc8x0_ad7441_init(struct vc8x0_dma_channel *dma_channel, u8 i2caddr);
int vc8x0_ad7441_enable(struct vc8x0_ad7441_context *ctx);
int vc8x0_ad7441_disable(struct vc8x0_ad7441_context *ctx);
int vc8x0_ad7441_set_format(struct vc8x0_ad7441_context *ctx, struct vc8x0_format *fmt);
int vc8x0_ad7441_get_format(struct vc8x0_ad7441_context *ctx, struct vc8x0_format **fmt);
int vc8x0_ad7441_is_locked(struct vc8x0_ad7441_context *ctx);
int vc8x0_ad7441_get_format_detected(struct vc8x0_ad7441_context *ctx, struct vc8x0_format **fmt);
void vc8x0_ad7441_procfs_detection_show(struct seq_file *m, struct vc8x0_ad7441_context *ctx);
int vc8x0_ad7441_detect(struct vc8x0_ad7441_context *ctx);

/* ----------------------------------------------------------- */
/* vc8x0-pcm3052.c                                             */
int vc8x0_pcm3052_init(struct vc8x0_dma_channel *dma_channel, u8 i2caddr);
int vc8x0_pcm3052_enable(struct vc8x0_pcm3052_context *ctx);
int vc8x0_pcm3052_disable(struct vc8x0_pcm3052_context *ctx);
int vc8x0_pcm3052_procfs_show(struct seq_file *m, struct vc8x0_pcm3052_context *ctx);
int vc8x0_pcm3052_set_mute(struct vc8x0_pcm3052_context *ctx, u32 enable);
int vc8x0_pcm3052_get_mute(struct vc8x0_pcm3052_context *ctx, u32 *enabled);
int vc8x0_pcm3052_set_volume(struct vc8x0_pcm3052_context *ctx, u32 val);
int vc8x0_pcm3052_get_volume(struct vc8x0_pcm3052_context *ctx, u32 *val);

/* ----------------------------------------------------------- */
/* vc8x0-gs2971a.c                                             */
int vc8x0_gs2971a_init(struct vc8x0_dma_channel *dma_channel);
int vc8x0_gs2971a_enable(struct vc8x0_gs2971a_context *ctx);
int vc8x0_gs2971a_disable(struct vc8x0_gs2971a_context *ctx);
int vc8x0_gs2971a_procfs_show(struct seq_file *m, struct vc8x0_gs2971a_context *ctx);
int vc8x0_gs2971a_set_mute(struct vc8x0_gs2971a_context *ctx, u32 enable);
int vc8x0_gs2971a_get_mute(struct vc8x0_gs2971a_context *ctx, u32 *enabled);
int vc8x0_gs2971a_set_volume(struct vc8x0_gs2971a_context *ctx, u32 val);
int vc8x0_gs2971a_get_volume(struct vc8x0_gs2971a_context *ctx, u32 *val);
int vc8x0_gs2971a_get_format(struct vc8x0_gs2971a_context *ctx, struct vc8x0_format **fmt);
int vc8x0_gs2971a_set_format(struct vc8x0_gs2971a_context *ctx, struct vc8x0_format *fmt);
int vc8x0_gs2971a_is_locked(struct vc8x0_gs2971a_context *ctx);
int vc8x0_gs2971a_set_audio_input(struct vc8x0_gs2971a_context *ctx, int inputnr);
int vc8x0_gs2971a_get_audio_input(struct vc8x0_gs2971a_context *ctx, int *inputnr);

/* ----------------------------------------------------------- */
/* vc8x0-channel.c                                            */
int vc8x0_channel_register(struct vc8x0_dev *dev, struct vc8x0_dma_channel *channel, int nr);
int vc8x0_channel_unregister(struct vc8x0_dma_channel *channel);
void vc8x0_channel_statistics_reset(struct vc8x0_dma_channel *channel);
void vc8x0_channel_statistics_procfs(struct seq_file *m, struct vc8x0_dma_channel *channel);
int vc8x0_channel_state_procfs(struct seq_file *m, struct vc8x0_dma_channel *channel);
int vc8x0_channel_event(struct vc8x0_dma_channel *channel);
int vc8x0_channel_irq(struct vc8x0_dma_channel *channel);
int vc8x0_channel_start(struct vc8x0_dma_channel *channel);
int vc8x0_channel_stop(struct vc8x0_dma_channel *channel);
int vc8x0_channel_thread(struct vc8x0_dma_channel *channel);

/* ----------------------------------------------------- */
/* vc8x0-fw.c                                            */
int vc8x0_fw_load(struct vc8x0_dev *dev);

/* -------------------------------------------------------- */
/* vc8x0-video.c                                            */
void vc8x0_vbi_timeout(unsigned long data);
void vc8x0_video_timeout(unsigned long data);
int vc8x0_video_register(struct vc8x0_dma_channel *channel);
void vc8x0_video_unregister(struct vc8x0_dma_channel *channel);
void vc8x0_video_work_handler(struct work_struct *work);
int vc8x0_video_irqhandler(struct vc8x0_dma_channel *channel);
struct vc8x0_format *vc8x0_video_find_format(struct vc8x0_dma_channel *channel,
	u32 pixelformat, u32 width, u32 height, u32 flags, u32 rate);
void vc8x0_channel_loss_of_sync(struct vc8x0_dma_channel *channel, int enable);
struct vc8x0_format *vc8x0_video_find_formatById(struct vc8x0_dma_channel *channel, int id);

/* ----------------------------------------------------------- */
/* vc8x0-ad7441.c                                              */
int vc8x0_ad7441_alloc(struct vc8x0_dma_channel *channel, u8 i2caddr);
int vc8x0_ad7441_dealloc(struct vc8x0_ad7441_context *chip);
int vc8x0_ad7441_procfs_show(struct seq_file *m, struct vc8x0_ad7441_context *ctx);
int vc8x0_ad7441_set_input(struct vc8x0_ad7441_context *ctx, enum video_input_type input);
int vc8x0_ad7441_set_brightness(struct vc8x0_ad7441_context *ctx, u32 v);
int vc8x0_ad7441_get_brightness(struct vc8x0_ad7441_context *ctx, u32 *v);
int vc8x0_ad7441_set_contrast(struct vc8x0_ad7441_context *ctx, u32 v);
int vc8x0_ad7441_get_contrast(struct vc8x0_ad7441_context *ctx, u32 *v);
int vc8x0_ad7441_set_hue(struct vc8x0_ad7441_context *ctx, u32 v);
int vc8x0_ad7441_get_hue(struct vc8x0_ad7441_context *ctx, u32 *v);
int vc8x0_ad7441_set_saturation(struct vc8x0_ad7441_context *ctx, u32 v);
int vc8x0_ad7441_get_saturation(struct vc8x0_ad7441_context *ctx, u32 *v);
int vc8x0_ad7441_keep_alive(struct vc8x0_ad7441_context *ctx);

/* ----------------------------------------------------------- */
/* vc8x0-buffer.c                                              */
void vc8x0_buffer_dealloc(struct vc8x0_dma_buffer *buf);
struct vc8x0_dma_buffer *vc8x0_buffer_alloc(struct vc8x0_dma_channel *channel,
	enum buffer_payload_t type, u32 nr);
void vc8x0_buffer_recycle(struct vc8x0_dma_buffer *buf);
void vc8x0_buffer_display(struct vc8x0_dma_buffer *buf);
void vc8x0_buffer_activate(struct vc8x0_dma_buffer *buf);
struct vc8x0_dma_buffer *vc8x0_buffer_busy_get_by_address(
        struct vc8x0_dma_channel *channel, u32 physaddr, enum buffer_payload_t type);

/* ----------------------------------------------------------- */
/* vc8x0-dma.c                                                 */
int vc8x0_dma_loadsequencer(struct vc8x0_dev *dev);
void vc8x0_dma_configure_channel_registers(struct vc8x0_dma_channel *channel);
void vc8x0_dma_dump_channel_registers(struct vc8x0_dma_channel *channel);
int vc8x0_dma_sequencer_start(struct vc8x0_dev *dev);
int vc8x0_dma_sequencer_stop(struct vc8x0_dev *dev);
int vc8x0_dma_configure(struct vc8x0_dma_channel *channel);
int vc8x0_dma_reset(struct vc8x0_dma_channel *channel);

/* ----------------------------------------------------------- */
/* vc8x0-display.c                                            */
int vc8x0_display_init(struct vc8x0_dma_channel *channel, struct vc8x0_display_context *ctx);
int vc8x0_display_render_reset(struct vc8x0_display_context *ctx, u8 *ptr, int stride);
int vc8x0_display_render_string(struct vc8x0_display_context *ctx, u8 *s, int len, int x, int y);

/* ----------------------------------------------------------- */
/* vc8x0-audio.c                                              */
int  vc8x0_audio_register(struct vc8x0_dma_channel *channel);
void vc8x0_audio_unregister(struct vc8x0_dma_channel *channel);
extern int vc8x0_audio_irqhandler(struct vc8x0_dma_channel *channel);
void vc8x0_audio_work_handler(struct work_struct *work);
void vc8x0_audio_timeout(unsigned long data);

/* ----------------------------------------------------------- */
/* vc8x0-timestamp.c                                           */
void vc8x0_timestamp_dump(struct vc8x0_dma_buffer *buf, int idx);
int vc8x0_timestamp_validate(struct vc8x0_dma_buffer *buf, int idx);

/* ----------------------------------------------------------- */
/* vc8x0-vga.c                                                 */
struct vga_timing_t *vc8x0_vga_lookup(int mode, u32 width, u32 height, u32 interlaced);
struct vga_rate_t *vc8x0_vga_HdGetRate(u32 fcl28);
struct vga_size_t *vc8x0_vga_HdGetSize(u32 lcf);
void vc8x0_vga_dump(struct vga_timing_t *mode);

