/*
 *  Driver for the ViewCast 845e PCIe Capture Card
 *
 *  Copyright 2011/2012, Kernel Labs Inc. www.kernellabs.com.
 *   - Steven Toth <stoth@kernellabs.com>
 *
 *  Project sponsored by ViewCast Texas, US.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "vc8x0.h"

static unsigned int dma_debug;
module_param(dma_debug, int, 0644);
MODULE_PARM_DESC(dma_debug, "enable debug messages [dma]");

#define dprintk(level, fmt, arg...)\
        do { if (dma_debug >= level)\
                printk(KERN_DEBUG "%s: " fmt, dev->name, ## arg);\
        } while (0)

#if 0
/*
LP1_VD_1
	VD_1_DEST_XFER_CNT	// Get the remaining XFER count (starts at X then decrements to 0)
	VD_1_Y_DEST_ADDR_L	// Y Plane destination address
	VD_1_U_DEST_ADDR_L	// U Plane destination address
	VD_1_V_DEST_ADDR_L	// V Plane destination address

	VD_1_LAST_TS_ADDR_PTR	// Load last DMA time stamp pointer for this FIFO
	VD_1_DEST_TS_ADDR_PTR	// Save it as the current DMA time stamp pointer
	VD_1_LAST_ADDR_PTR	// Load last DMA frame pointer for this FIFO
	VD_1_DEST_ADDR_PTR	// Save it as the current DMA frame pointer
	VD_1_FRAME_CNT		// Load DMA frames to xfer count
	VD_1_FRAME_XFER_CNT	// Save it 

LP1_VD_2
	
LP1_VD_3
*/
#endif

/* The number of buckets to count Map Registers with */
#define MAX_MAP_REGISTER_BUCKETS 256 * 4

/* VDMA sequencer code Definitions */
#define VDMA_IM (0)
#define VDMA_RA (2)
#define VDMA_RB (3)

/* Condition code for JMP instruction */
#define VDMA_RA_EQZ (0x8)
#define VDMA_RA_NEQZ (0)
#define VDMA_RB_EQZ (0x9)
#define VDMA_RB_NEQZ (1)
#define VDMA_ALWAYS (0xA)
#define VDMA_NEVER (0x2)
#define VDMA_C_HI (0xB)
#define VDMA_C_LO (0x3)
#define VDMA_PDM_CMD_QUEUE_FULL_HI (0xC)
#define VDMA_PDM_CMD_QUEUE_FULL_LO (0x4)
#define VDMA_LDM_CMD_QUEUE_FULL_HI (0xD)
#define VDMA_LDM_CMD_QUEUE_FULL_LO (0x5)
#define VDMA_EXT_COND_HI (0xF)
#define VDMA_EXT_COND_LO (0x7)

/* External condition select code for JMP instruction */
#define CAP_VSYNC (5)
#define PDM_IDLE (32)
#define LDM_IDLE (33)
#define EXT_COND_0 (34)
#define EXT_COND_1 (35)
#define EXT_COND_2 (36)
#define EXT_COND_3 (37)
#define EXT_COND_4 (38)
#define EXT_COND_5 (39)
#define EXT_COND_6 (40)
#define EXT_COND_7 (41)
#define EXT_COND_8 (42)
#define EXT_COND_9 (43)
#define EXT_COND_10 (44)
#define EXT_COND_11 (45)
#define EXT_COND_12 (46)
#define EXT_COND_13 (47)
#define EXT_COND_14 (48)
#define EXT_COND_15 (49)

/* VDMA instructions */
#define VDMA_NOP() \
    ((u32)0x0)

#define VDMA_LOAD_SYS_ADDR(R, ADDR) \
    ((u32)0x40000000 |    \
     ((u32)(R & 0x3) << 24) | \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_STORE_SYS_ADDR(R, ADDR) \
    ((u32)0x50000000 |    \
     ((u32)(R & 0x3) << 24) | \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_ADD_SYS_ADDR(DATA) \
    ((u32)0x60000000 |    \
     ((u32)(DATA & 0xFFFF)) \
    )

#define VDMA_ADD_SYS_ADDR_I(ADDR) \
    ((u32)0xE0000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_LOAD_XFER_CTL(R, ADDR) \
    ((u32)0xF0000000 |    \
     ((u32)(R & 0x3) << 24) | \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_LOAD_RA(ADDR) \
    ((u32)0x20000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_ADD_RA(ADDR) \
    ((u32)0x21000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_LOAD_RB(ADDR) \
    ((u32)0x24000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_ADD_RB(ADDR) \
    ((u32)0x25000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_STORE_RA(ADDR) \
    ((u32)0xA2000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_STORE_RB(ADDR) \
    ((u32)0xA3000000 |    \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_JMP(C, EXT_COND, ADDR) \
    ((u32)0x10000000 |    \
     (u32)((C & 0xF) << 24) | \
     (u32) ((EXT_COND & 0xFF) << 16) | \
     ((u32)(ADDR & 0xFFFF)) \
    )

#define VDMA_SIG_EVENT(S, A, EVENT_EN) \
    ((u32)0x80000000 |    \
     ((u32)(S & 0x1) << 27) | \
     ((u32)(A & 0x1) << 26) | \
     ((u32)(EVENT_EN & 0xFFFF)) \
    )

#define VDMA_WAIT_EVENT(EVENT_EN, EVENT_STATE) \
    ((u32)0x90000000 |    \
     ((u32)((EVENT_EN & 0xFFF) << 12)) | \
     ((u32)(EVENT_STATE & 0xFFF)) \
    )

/* Registers */
#define START				0x0000
#define MAIN				0x0000

#define LP1_VD_1_READY			0x0002
#define LP1_VD_1_Y_WAIT_FULL		0x0007
#define LP1_VD_1_U_WAIT_FULL		0x000C
#define LP1_VD_1_V_WAIT_FULL		0x0011
#define LP1_VD_1_DONE_FULL		0x0018
#define LP1_VD_1_Y_WAIT_PART		0x001A
#define LP1_VD_1_U_WAIT_PART		0x001D
#define LP1_VD_1_V_WAIT_PART		0x0020
#define LP1_VD_1_Y_TS_WAIT		0x0023
#define LP1_VD_1_U_TS_WAIT		0x0026
#define LP1_VD_1_V_TS_WAIT		0x0029
#define LP1_VD_1_LOAD_ADDR		0x0041

#define LP1_VD_2			0x004C
#define LP1_VD_2_READY			0x004E
#define LP1_VD_2_Y_WAIT_FULL		0x0053
#define LP1_VD_2_U_WAIT_FULL		0x0058
#define LP1_VD_2_V_WAIT_FULL		0x005D
#define LP1_VD_2_DONE_FULL		0x0064
#define LP1_VD_2_Y_WAIT_PART		0x0066
#define LP1_VD_2_U_WAIT_PART		0x0069
#define LP1_VD_2_V_WAIT_PART		0x006C
#define LP1_VD_2_Y_TS_WAIT		0x006F
#define LP1_VD_2_U_TS_WAIT		0x0072
#define LP1_VD_2_V_TS_WAIT		0x0075
#define LP1_VD_2_LOAD_ADDR		0x008D

#define LP1_VD_3			0x0098
#define LP1_VD_3_READY			0x009A
#define LP1_VD_3_Y_WAIT_FULL		0x009F
#define LP1_VD_3_U_WAIT_FULL		0x00A4
#define LP1_VD_3_V_WAIT_FULL		0x00A9
#define LP1_VD_3_DONE_FULL		0x00B0
#define LP1_VD_3_Y_WAIT_PART		0x00B2
#define LP1_VD_3_U_WAIT_PART		0x00B5
#define LP1_VD_3_V_WAIT_PART		0x00B8
#define LP1_VD_3_Y_TS_WAIT		0x00BB
#define LP1_VD_3_U_TS_WAIT		0x00BE
#define LP1_VD_3_V_TS_WAIT		0x00C1
#define LP1_VD_3_LOAD_ADDR		0x00D9

#define LP1_VD_4			0x00E4
#define LP1_VD_4_READY			0x00E6
#define LP1_VD_4_Y_WAIT_FULL		0x00EB
#define LP1_VD_4_U_WAIT_FULL		0x00F0
#define LP1_VD_4_V_WAIT_FULL		0x00F5
#define LP1_VD_4_DONE_FULL		0x00FC
#define LP1_VD_4_Y_WAIT_PART		0x00FE
#define LP1_VD_4_U_WAIT_PART		0x0101
#define LP1_VD_4_V_WAIT_PART		0x0104
#define LP1_VD_4_Y_TS_WAIT		0x0107
#define LP1_VD_4_U_TS_WAIT		0x010A
#define LP1_VD_4_V_TS_WAIT		0x010D
#define LP1_VD_4_LOAD_ADDR		0x0125

#define LP1_AD_1			0x0130
#define LP1_AD_1_READY			0x0132
#define LP1_AD_1_SP1_WAIT_FULL		0x0133
#define LP1_AD_1_SP2_WAIT_FULL		0x0136
#define LP1_AD_1_SP3_WAIT_FULL		0x0139
#define LP1_AD_1_SP4_WAIT_FULL		0x013C
#define LP1_AD_1_SP1_TS_WAIT		0x0146
#define LP1_AD_1_SP2_TS_WAIT		0x0149
#define LP1_AD_1_SP3_TS_WAIT		0x014C
#define LP1_AD_1_SP4_TS_WAIT		0x014F
#define LP1_AD_1_LOAD_ADDR		0x0166

#define LP1_AD_2			0x016D
#define LP1_AD_2_READY			0x016F
#define LP1_AD_2_SP1_WAIT_FULL		0x0170
#define LP1_AD_2_SP2_WAIT_FULL		0x0173
#define LP1_AD_2_SP3_WAIT_FULL		0x0176
#define LP1_AD_2_SP4_WAIT_FULL		0x0179
#define LP1_AD_2_SP1_TS_WAIT		0x0183
#define LP1_AD_2_SP2_TS_WAIT		0x0186
#define LP1_AD_2_SP3_TS_WAIT		0x0189
#define LP1_AD_2_SP4_TS_WAIT		0x018C
#define LP1_AD_2_LOAD_ADDR		0x01A3

#define LP1_AD_3			0x01AA
#define LP1_AD_3_READY			0x01AC
#define LP1_AD_3_SP1_WAIT_FULL		0x01AD
#define LP1_AD_3_SP2_WAIT_FULL		0x01B0
#define LP1_AD_3_SP3_WAIT_FULL		0x01B3
#define LP1_AD_3_SP4_WAIT_FULL		0x01B6
#define LP1_AD_3_SP1_TS_WAIT		0x01C0
#define LP1_AD_3_SP2_TS_WAIT		0x01C3
#define LP1_AD_3_SP3_TS_WAIT		0x01C6
#define LP1_AD_3_SP4_TS_WAIT		0x01C9
#define LP1_AD_3_LOAD_ADDR		0x01E0

#define LP1_AD_4			0x01E7
#define LP1_AD_4_READY			0x01EA
#define LP1_AD_4_SP1_WAIT_FULL		0x01EB
#define LP1_AD_4_SP2_WAIT_FULL		0x01EE
#define LP1_AD_4_SP3_WAIT_FULL		0x01F1
#define LP1_AD_4_SP4_WAIT_FULL		0x01F4
#define LP1_AD_4_SP1_TS_WAIT		0x01FE
#define LP1_AD_4_SP2_TS_WAIT		0x0201
#define LP1_AD_4_SP3_TS_WAIT		0x0204
#define LP1_AD_4_SP4_TS_WAIT		0x0207
#define LP1_AD_4_LOAD_ADDR		0x021E

#define LP1_AD_5			0x0225
#define LP1_AD_5_READY			0x0227
#define LP1_AD_5_SP1_WAIT_FULL		0x0228
#define LP1_AD_5_SP1_TS_WAIT		0x0231
#define LP1_AD_5_LOAD_ADDR		0x0248

#define LP1_AD_6			0x0252
#define LP1_AD_6_READY			0x0254
#define LP1_AD_6_SP1_WAIT_FULL		0x0255
#define LP1_AD_6_SP1_TS_WAIT		0x025E
#define LP1_AD_6_LOAD_ADDR		0x0275

#define LP1_AD_7			0x027C
#define LP1_AD_7_READY			0x027E
#define LP1_AD_7_SP1_WAIT_FULL		0x027F
#define LP1_AD_7_SP1_TS_WAIT		0x0288
#define LP1_AD_7_LOAD_ADDR		0x029F

#define LP1_AD_8			0x02A6
#define LP1_AD_8_READY			0x02A8
#define LP1_AD_8_SP1_WAIT_FULL		0x02A9
#define LP1_AD_8_SP1_TS_WAIT		0x02B2
#define LP1_AD_8_LOAD_ADDR		0x02C9
#define FINISH				0x02D0
#define LDM_IDLE_WAIT			0x02D2
#define SIGNAL_EVENT			0x02D4

#define DMA_BUFFERS			MAX_VIDEO_DSP_FRAMES
#define FIFO_COUNT			12
#define MAX_DEVICE_COUNT		8
#define NO_VBI				1

#define VDMA_COMMON_VARIABLE_START	0x00000320
#define VDMA_VIDEO_VARIABLE_START	VDMA_COMMON_VARIABLE_START + 32
#define VDMA_AUDIO_VARIABLE_START	0x000004E0
#define MINUS_1				(VDMA_COMMON_VARIABLE_START + 0)
#define MINUS_2				(VDMA_COMMON_VARIABLE_START + 1)
#define MINUS_6				(VDMA_COMMON_VARIABLE_START + 2)
#define MINUS_8				(VDMA_COMMON_VARIABLE_START + 3)
#define PLUS_8				(VDMA_COMMON_VARIABLE_START + 4)
#define PLUS_16				(VDMA_COMMON_VARIABLE_START + 5)
#define EVENT_ID_0			(VDMA_COMMON_VARIABLE_START + 6)
#define EVENT_ID_1			(VDMA_COMMON_VARIABLE_START + 7)
#define EVENT_ID_2			(VDMA_COMMON_VARIABLE_START + 8)
#define EVENT_ID_3			(VDMA_COMMON_VARIABLE_START + 9)
#define EVENT_ID_4			(VDMA_COMMON_VARIABLE_START + 10)
#define EVENT_ID_5			(VDMA_COMMON_VARIABLE_START + 11)
#define EVENT_ID_6			(VDMA_COMMON_VARIABLE_START + 12)
#define EVENT_ID_7			(VDMA_COMMON_VARIABLE_START + 13)
#define EVENT_ID_8			(VDMA_COMMON_VARIABLE_START + 14)
#define EVENT_ID_9			(VDMA_COMMON_VARIABLE_START + 15)
#define EVENT_ID_10			(VDMA_COMMON_VARIABLE_START + 16)
#define EVENT_ID_11			(VDMA_COMMON_VARIABLE_START + 17)
#define EVENT_ID_12			(VDMA_COMMON_VARIABLE_START + 18)
#define EVENT_ID_13			(VDMA_COMMON_VARIABLE_START + 19)
#define EVENT_ID_14			(VDMA_COMMON_VARIABLE_START + 20)
#define EVENT_ID_15			(VDMA_COMMON_VARIABLE_START + 21)
#define EVENTS_TO_FIRE			(VDMA_COMMON_VARIABLE_START + 22)
#define SIG_EVENT_CODE			(VDMA_COMMON_VARIABLE_START + 23)
#define ZERO				(VDMA_COMMON_VARIABLE_START + 24)
#define AUDIO_CHANNEL_SIZE		(VDMA_COMMON_VARIABLE_START + 25)

#define VD_1_XFER_CNT			(VDMA_VIDEO_VARIABLE_START + 0)
#define VD_1_DEST_XFER_CNT		(VD_1_XFER_CNT + 1)
#define VD_1_Y_XFER_SIZE_FULL		(VD_1_XFER_CNT + 2)
#define VD_1_UV_XFER_SIZE_FULL		(VD_1_XFER_CNT + 3)
#define VD_1_Y_XFER_CTRL_FULL		(VD_1_XFER_CNT + 4)
#define VD_1_U_XFER_CTRL_FULL		(VD_1_XFER_CNT + 5)
#define VD_1_V_XFER_CTRL_FULL		(VD_1_XFER_CNT + 6)
#define VD_1_Y_XFER_SIZE_PART		(VD_1_XFER_CNT + 7)
#define VD_1_UV_XFER_SIZE_PART		(VD_1_XFER_CNT + 8)
#define VD_1_Y_XFER_CTRL_PART		(VD_1_XFER_CNT + 9)
#define VD_1_U_XFER_CTRL_PART		(VD_1_XFER_CNT + 10)
#define VD_1_V_XFER_CTRL_PART		(VD_1_XFER_CNT + 11)
#define VD_1_Y_TS_XFER_CTRL		(VD_1_XFER_CNT + 12)
#define VD_1_U_TS_XFER_CTRL		(VD_1_XFER_CNT + 13)
#define VD_1_V_TS_XFER_CTRL		(VD_1_XFER_CNT + 14)
#define VD_1_FRAME_CNT			(VD_1_XFER_CNT + 15)
#define VD_1_STREAM_DISABLED		(VD_1_XFER_CNT + 16)
#define VD_1_LAST_FRAME_INDEX		(VD_1_XFER_CNT + 17)
#define VD_1_LAST_ADDR_PTR		(VD_1_XFER_CNT + 18)
#define VD_1_DEST_ADDR_PTR		(VD_1_XFER_CNT + 19)
#define VD_1_LAST_TS_ADDR_PTR		(VD_1_XFER_CNT + 20)
#define VD_1_DEST_TS_ADDR_PTR		(VD_1_XFER_CNT + 21)
#define VD_1_FRAME_XFER_CNT		(VD_1_XFER_CNT + 22)
#define VD_1_Y_DEST_ADDR_L		(VD_1_XFER_CNT + 24)
#define VD_1_Y_DEST_ADDR_H		(VD_1_XFER_CNT + 25)
#define VD_1_U_DEST_ADDR_L		(VD_1_XFER_CNT + 26)
#define VD_1_U_DEST_ADDR_H		(VD_1_XFER_CNT + 27)
#define VD_1_V_DEST_ADDR_L		(VD_1_XFER_CNT + 28)
#define VD_1_V_DEST_ADDR_H		(VD_1_XFER_CNT + 29)
#define VD_1_Y_TS_DEST_ADDR_L		(VD_1_XFER_CNT + 30)
#define VD_1_Y_TS_DEST_ADDR_H		(VD_1_XFER_CNT + 31)
#define VD_1_Y_DMA_ADDR_L		(VD_1_XFER_CNT + 32)
#define VD_1_Y_DMA_ADDR_H		(VD_1_XFER_CNT + 33)
#define VD_1_U_DMA_ADDR_L		(VD_1_XFER_CNT + 34)
#define VD_1_U_DMA_ADDR_H		(VD_1_XFER_CNT + 35)
#define VD_1_V_DMA_ADDR_L		(VD_1_XFER_CNT + 36)
#define VD_1_V_DMA_ADDR_H		(VD_1_XFER_CNT + 37)

// Skip 48 slots to accomodate the remaining 8 possible frames
#define VD_1_Y_TS_DMA_ADDR_L		(VD_1_V_DMA_ADDR_H + 49)
#define VD_1_Y_TS_DMA_ADDR_H		(VD_1_Y_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible time stamps
#define VD_2_XFER_CNT			(VD_1_Y_TS_DMA_ADDR_H + 17)
#define VD_2_DEST_XFER_CNT		(VD_2_XFER_CNT + 1)
#define VD_2_Y_XFER_SIZE_FULL		(VD_2_XFER_CNT + 2)
#define VD_2_UV_XFER_SIZE_FULL		(VD_2_XFER_CNT + 3)
#define VD_2_Y_XFER_CTRL_FULL		(VD_2_XFER_CNT + 4)
#define VD_2_U_XFER_CTRL_FULL		(VD_2_XFER_CNT + 5)
#define VD_2_V_XFER_CTRL_FULL		(VD_2_XFER_CNT + 6)
#define VD_2_Y_XFER_SIZE_PART		(VD_2_XFER_CNT + 7)
#define VD_2_UV_XFER_SIZE_PART		(VD_2_XFER_CNT + 8)
#define VD_2_Y_XFER_CTRL_PART		(VD_2_XFER_CNT + 9)
#define VD_2_U_XFER_CTRL_PART		(VD_2_XFER_CNT + 10)
#define VD_2_V_XFER_CTRL_PART		(VD_2_XFER_CNT + 11)
#define VD_2_Y_TS_XFER_CTRL		(VD_2_XFER_CNT + 12)
#define VD_2_U_TS_XFER_CTRL		(VD_2_XFER_CNT + 13)
#define VD_2_V_TS_XFER_CTRL		(VD_2_XFER_CNT + 14)
#define VD_2_FRAME_CNT			(VD_2_XFER_CNT + 15)
#define VD_2_STREAM_DISABLED		(VD_2_XFER_CNT + 16)
#define VD_2_LAST_FRAME_INDEX		(VD_2_XFER_CNT + 17)
#define VD_2_LAST_ADDR_PTR		(VD_2_XFER_CNT + 18)
#define VD_2_DEST_ADDR_PTR		(VD_2_XFER_CNT + 19)
#define VD_2_LAST_TS_ADDR_PTR		(VD_2_XFER_CNT + 20)
#define VD_2_DEST_TS_ADDR_PTR		(VD_2_XFER_CNT + 21)
#define VD_2_FRAME_XFER_CNT		(VD_2_XFER_CNT + 22)
#define VD_2_Y_DEST_ADDR_L		(VD_2_XFER_CNT + 24)
#define VD_2_Y_DEST_ADDR_H		(VD_2_XFER_CNT + 25)
#define VD_2_U_DEST_ADDR_L		(VD_2_XFER_CNT + 26)
#define VD_2_U_DEST_ADDR_H		(VD_2_XFER_CNT + 27)
#define VD_2_V_DEST_ADDR_L		(VD_2_XFER_CNT + 28)
#define VD_2_V_DEST_ADDR_H		(VD_2_XFER_CNT + 29)
#define VD_2_Y_TS_DEST_ADDR_L		(VD_2_XFER_CNT + 30)
#define VD_2_Y_TS_DEST_ADDR_H		(VD_2_XFER_CNT + 31)
#define VD_2_Y_DMA_ADDR_L		(VD_2_XFER_CNT + 32)
#define VD_2_Y_DMA_ADDR_H		(VD_2_XFER_CNT + 33)
#define VD_2_U_DMA_ADDR_L		(VD_2_XFER_CNT + 34)
#define VD_2_U_DMA_ADDR_H		(VD_2_XFER_CNT + 35)
#define VD_2_V_DMA_ADDR_L		(VD_2_XFER_CNT + 36)
#define VD_2_V_DMA_ADDR_H		(VD_2_XFER_CNT + 37)
// Skip 48 slots to accomodate the remaining 8 possible frames
#define VD_2_Y_TS_DMA_ADDR_L		(VD_2_V_DMA_ADDR_H + 49)
#define VD_2_Y_TS_DMA_ADDR_H		(VD_2_Y_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible time stamps
#define VD_3_XFER_CNT			(VD_2_Y_TS_DMA_ADDR_H + 17)
#define VD_3_DEST_XFER_CNT		(VD_3_XFER_CNT + 1)
#define VD_3_Y_XFER_SIZE_FULL		(VD_3_XFER_CNT + 2)
#define VD_3_UV_XFER_SIZE_FULL		(VD_3_XFER_CNT + 3)
#define VD_3_Y_XFER_CTRL_FULL		(VD_3_XFER_CNT + 4)
#define VD_3_U_XFER_CTRL_FULL		(VD_3_XFER_CNT + 5)
#define VD_3_V_XFER_CTRL_FULL		(VD_3_XFER_CNT + 6)
#define VD_3_Y_XFER_SIZE_PART		(VD_3_XFER_CNT + 7)
#define VD_3_UV_XFER_SIZE_PART		(VD_3_XFER_CNT + 8)
#define VD_3_Y_XFER_CTRL_PART		(VD_3_XFER_CNT + 9)
#define VD_3_U_XFER_CTRL_PART		(VD_3_XFER_CNT + 10)
#define VD_3_V_XFER_CTRL_PART		(VD_3_XFER_CNT + 11)
#define VD_3_Y_TS_XFER_CTRL		(VD_3_XFER_CNT + 12)
#define VD_3_U_TS_XFER_CTRL		(VD_3_XFER_CNT + 13)
#define VD_3_V_TS_XFER_CTRL		(VD_3_XFER_CNT + 14)
#define VD_3_FRAME_CNT			(VD_3_XFER_CNT + 15)
#define VD_3_STREAM_DISABLED		(VD_3_XFER_CNT + 16)
#define VD_3_LAST_FRAME_INDEX		(VD_3_XFER_CNT + 17)
#define VD_3_LAST_ADDR_PTR		(VD_3_XFER_CNT + 18)
#define VD_3_DEST_ADDR_PTR		(VD_3_XFER_CNT + 19)
#define VD_3_LAST_TS_ADDR_PTR		(VD_3_XFER_CNT + 20)
#define VD_3_DEST_TS_ADDR_PTR		(VD_3_XFER_CNT + 21)
#define VD_3_FRAME_XFER_CNT		(VD_3_XFER_CNT + 22)
#define VD_3_Y_DEST_ADDR_L		(VD_3_XFER_CNT + 24)
#define VD_3_Y_DEST_ADDR_H		(VD_3_XFER_CNT + 25)
#define VD_3_U_DEST_ADDR_L		(VD_3_XFER_CNT + 26)
#define VD_3_U_DEST_ADDR_H		(VD_3_XFER_CNT + 27)
#define VD_3_V_DEST_ADDR_L		(VD_3_XFER_CNT + 28)
#define VD_3_V_DEST_ADDR_H		(VD_3_XFER_CNT + 29)
#define VD_3_Y_TS_DEST_ADDR_L		(VD_3_XFER_CNT + 30)
#define VD_3_Y_TS_DEST_ADDR_H		(VD_3_XFER_CNT + 31)
#define VD_3_Y_DMA_ADDR_L		(VD_3_XFER_CNT + 32)
#define VD_3_Y_DMA_ADDR_H		(VD_3_XFER_CNT + 33)
#define VD_3_U_DMA_ADDR_L		(VD_3_XFER_CNT + 34)
#define VD_3_U_DMA_ADDR_H		(VD_3_XFER_CNT + 35)
#define VD_3_V_DMA_ADDR_L		(VD_3_XFER_CNT + 36)
#define VD_3_V_DMA_ADDR_H		(VD_3_XFER_CNT + 37)
// Skip 48 slots to accomodate the remaining 8 possible frames
#define VD_3_Y_TS_DMA_ADDR_L		(VD_3_V_DMA_ADDR_H + 49)
#define VD_3_Y_TS_DMA_ADDR_H		(VD_3_Y_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible time stamps
#define VD_4_XFER_CNT			(VD_3_Y_TS_DMA_ADDR_H + 17)
#define VD_4_DEST_XFER_CNT		(VD_4_XFER_CNT + 1)
#define VD_4_Y_XFER_SIZE_FULL		(VD_4_XFER_CNT + 2)
#define VD_4_UV_XFER_SIZE_FULL		(VD_4_XFER_CNT + 3)
#define VD_4_Y_XFER_CTRL_FULL		(VD_4_XFER_CNT + 4)
#define VD_4_U_XFER_CTRL_FULL		(VD_4_XFER_CNT + 5)
#define VD_4_V_XFER_CTRL_FULL		(VD_4_XFER_CNT + 6)
#define VD_4_Y_XFER_SIZE_PART		(VD_4_XFER_CNT + 7)
#define VD_4_UV_XFER_SIZE_PART		(VD_4_XFER_CNT + 8)
#define VD_4_Y_XFER_CTRL_PART		(VD_4_XFER_CNT + 9)
#define VD_4_U_XFER_CTRL_PART		(VD_4_XFER_CNT + 10)
#define VD_4_V_XFER_CTRL_PART		(VD_4_XFER_CNT + 11)
#define VD_4_Y_TS_XFER_CTRL		(VD_4_XFER_CNT + 12)
#define VD_4_U_TS_XFER_CTRL		(VD_4_XFER_CNT + 13)
#define VD_4_V_TS_XFER_CTRL		(VD_4_XFER_CNT + 14)
#define VD_4_FRAME_CNT			(VD_4_XFER_CNT + 15)
#define VD_4_STREAM_DISABLED		(VD_4_XFER_CNT + 16)
#define VD_4_LAST_FRAME_INDEX		(VD_4_XFER_CNT + 17)
#define VD_4_LAST_ADDR_PTR		(VD_4_XFER_CNT + 18)
#define VD_4_DEST_ADDR_PTR		(VD_4_XFER_CNT + 19)
#define VD_4_LAST_TS_ADDR_PTR		(VD_4_XFER_CNT + 20)
#define VD_4_DEST_TS_ADDR_PTR		(VD_4_XFER_CNT + 21)
#define VD_4_FRAME_XFER_CNT		(VD_4_XFER_CNT + 22)
#define VD_4_Y_DEST_ADDR_L		(VD_4_XFER_CNT + 24)
#define VD_4_Y_DEST_ADDR_H		(VD_4_XFER_CNT + 25)
#define VD_4_U_DEST_ADDR_L		(VD_4_XFER_CNT + 26)
#define VD_4_U_DEST_ADDR_H		(VD_4_XFER_CNT + 27)
#define VD_4_V_DEST_ADDR_L		(VD_4_XFER_CNT + 28)
#define VD_4_V_DEST_ADDR_H		(VD_4_XFER_CNT + 29)
#define VD_4_Y_TS_DEST_ADDR_L		(VD_4_XFER_CNT + 30)
#define VD_4_Y_TS_DEST_ADDR_H		(VD_4_XFER_CNT + 31)
#define VD_4_Y_DMA_ADDR_L		(VD_4_XFER_CNT + 32)
#define VD_4_Y_DMA_ADDR_H		(VD_4_XFER_CNT + 33)
#define VD_4_U_DMA_ADDR_L		(VD_4_XFER_CNT + 34)
#define VD_4_U_DMA_ADDR_H		(VD_4_XFER_CNT + 35)
#define VD_4_V_DMA_ADDR_L		(VD_4_XFER_CNT + 36)
#define VD_4_V_DMA_ADDR_H		(VD_4_XFER_CNT + 37)
// Skip 48 slots to accomodate the remaining 8 possible frames
#define VD_4_Y_TS_DMA_ADDR_L		(VD_4_V_DMA_ADDR_H + 49)
#define VD_4_Y_TS_DMA_ADDR_H		(VD_4_Y_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames

#define AD_1_XFER_CNT			(VDMA_AUDIO_VARIABLE_START)
#define AD_1_DEST_XFER_CNT		(AD_1_XFER_CNT + 1)
#define AD_1_SP1_XFER_STEP_SIZE		(AD_1_XFER_CNT + 2)
#define AD_1_SP1_XFER_CTRL		(AD_1_XFER_CNT + 3)
#define AD_1_SP2_XFER_CTRL		(AD_1_XFER_CNT + 4)
#define AD_1_SP3_XFER_CTRL		(AD_1_XFER_CNT + 5)
#define AD_1_SP4_XFER_CTRL		(AD_1_XFER_CNT + 6)
#define AD_1_SP1_TS_XFER_CTRL		(AD_1_XFER_CNT + 7)
#define AD_1_SP2_TS_XFER_CTRL		(AD_1_XFER_CNT + 8)
#define AD_1_SP3_TS_XFER_CTRL		(AD_1_XFER_CNT + 9)
#define AD_1_SP4_TS_XFER_CTRL		(AD_1_XFER_CNT + 10)
#define AD_1_FRAME_CNT			(AD_1_XFER_CNT + 11)
#define AD_1_STREAM_DISABLED		(AD_1_XFER_CNT + 12)
#define AD_1_LAST_FRAME_INDEX		(AD_1_XFER_CNT + 13)
#define AD_1_LAST_ADDR_PTR		(AD_1_XFER_CNT + 14)
#define AD_1_DEST_ADDR_PTR		(AD_1_XFER_CNT + 15)
#define AD_1_LAST_TS_ADDR_PTR		(AD_1_XFER_CNT + 16)
#define AD_1_DEST_TS_ADDR_PTR		(AD_1_XFER_CNT + 17)
#define AD_1_FRAME_XFER_CNT		(AD_1_XFER_CNT + 18)
#define AD_1_SP1_DEST_ADDR_L		(AD_1_XFER_CNT + 20)
#define AD_1_SP1_DEST_ADDR_H		(AD_1_XFER_CNT + 21)
#define AD_1_SP1_TS_DEST_ADDR_L		(AD_1_XFER_CNT + 22)
#define AD_1_SP1_TS_DEST_ADDR_H		(AD_1_XFER_CNT + 23)
#define AD_1_SP1_DMA_ADDR_L		(AD_1_XFER_CNT + 24)
#define AD_1_SP1_DMA_ADDR_H		(AD_1_XFER_CNT + 25)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_1_SP1_TS_DMA_ADDR_L		(AD_1_SP1_DMA_ADDR_H + 17)
#define AD_1_SP1_TS_DMA_ADDR_H		(AD_1_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_2_XFER_CNT			(AD_1_SP1_TS_DMA_ADDR_H + 17)
#define AD_2_DEST_XFER_CNT		(AD_2_XFER_CNT + 1)
#define AD_2_SP1_XFER_STEP_SIZE		(AD_2_XFER_CNT + 2)
#define AD_2_SP1_XFER_CTRL		(AD_2_XFER_CNT + 3)
#define AD_2_SP2_XFER_CTRL		(AD_2_XFER_CNT + 4)
#define AD_2_SP3_XFER_CTRL		(AD_2_XFER_CNT + 5)
#define AD_2_SP4_XFER_CTRL		(AD_2_XFER_CNT + 6)
#define AD_2_SP1_TS_XFER_CTRL		(AD_2_XFER_CNT + 7)
#define AD_2_SP2_TS_XFER_CTRL		(AD_2_XFER_CNT + 8)
#define AD_2_SP3_TS_XFER_CTRL		(AD_2_XFER_CNT + 9)
#define AD_2_SP4_TS_XFER_CTRL		(AD_2_XFER_CNT + 10)
#define AD_2_FRAME_CNT			(AD_2_XFER_CNT + 11)
#define AD_2_STREAM_DISABLED		(AD_2_XFER_CNT + 12)
#define AD_2_LAST_FRAME_INDEX		(AD_2_XFER_CNT + 13)
#define AD_2_LAST_ADDR_PTR		(AD_2_XFER_CNT + 14)
#define AD_2_DEST_ADDR_PTR		(AD_2_XFER_CNT + 15)
#define AD_2_LAST_TS_ADDR_PTR		(AD_2_XFER_CNT + 16)
#define AD_2_DEST_TS_ADDR_PTR		(AD_2_XFER_CNT + 17)
#define AD_2_FRAME_XFER_CNT		(AD_2_XFER_CNT + 18)
#define AD_2_SP1_DEST_ADDR_L		(AD_2_XFER_CNT + 20)
#define AD_2_SP1_DEST_ADDR_H		(AD_2_XFER_CNT + 21)
#define AD_2_SP1_TS_DEST_ADDR_L		(AD_2_XFER_CNT + 22)
#define AD_2_SP1_TS_DEST_ADDR_H		(AD_2_XFER_CNT + 23)
#define AD_2_SP1_DMA_ADDR_L		(AD_2_XFER_CNT + 24)
#define AD_2_SP1_DMA_ADDR_H		(AD_2_XFER_CNT + 25)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_2_SP1_TS_DMA_ADDR_L		(AD_2_SP1_DMA_ADDR_H + 17)
#define AD_2_SP1_TS_DMA_ADDR_H		(AD_2_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_3_XFER_CNT			(AD_2_SP1_TS_DMA_ADDR_H + 17)
#define AD_3_DEST_XFER_CNT		(AD_3_XFER_CNT + 1)
#define AD_3_SP1_XFER_STEP_SIZE		(AD_3_XFER_CNT + 2)
#define AD_3_SP1_XFER_CTRL		(AD_3_XFER_CNT + 3)
#define AD_3_SP2_XFER_CTRL		(AD_3_XFER_CNT + 4)
#define AD_3_SP3_XFER_CTRL		(AD_3_XFER_CNT + 5)
#define AD_3_SP4_XFER_CTRL		(AD_3_XFER_CNT + 6)
#define AD_3_SP1_TS_XFER_CTRL		(AD_3_XFER_CNT + 7)
#define AD_3_SP2_TS_XFER_CTRL		(AD_3_XFER_CNT + 8)
#define AD_3_SP3_TS_XFER_CTRL		(AD_3_XFER_CNT + 9)
#define AD_3_SP4_TS_XFER_CTRL		(AD_3_XFER_CNT + 10)
#define AD_3_FRAME_CNT			(AD_3_XFER_CNT + 11)
#define AD_3_STREAM_DISABLED		(AD_3_XFER_CNT + 12)
#define AD_3_LAST_FRAME_INDEX		(AD_3_XFER_CNT + 13)
#define AD_3_LAST_ADDR_PTR		(AD_3_XFER_CNT + 14)
#define AD_3_DEST_ADDR_PTR		(AD_3_XFER_CNT + 15)
#define AD_3_LAST_TS_ADDR_PTR		(AD_3_XFER_CNT + 16)
#define AD_3_DEST_TS_ADDR_PTR		(AD_3_XFER_CNT + 17)
#define AD_3_FRAME_XFER_CNT		(AD_3_XFER_CNT + 18)
#define AD_3_SP1_DEST_ADDR_L		(AD_3_XFER_CNT + 20)
#define AD_3_SP1_DEST_ADDR_H		(AD_3_XFER_CNT + 21)
#define AD_3_SP1_TS_DEST_ADDR_L		(AD_3_XFER_CNT + 22)
#define AD_3_SP1_TS_DEST_ADDR_H		(AD_3_XFER_CNT + 23)
#define AD_3_SP1_DMA_ADDR_L		(AD_3_XFER_CNT + 24)
#define AD_3_SP1_DMA_ADDR_H		(AD_3_XFER_CNT + 25)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_3_SP1_TS_DMA_ADDR_L		(AD_3_SP1_DMA_ADDR_H + 17)
#define AD_3_SP1_TS_DMA_ADDR_H		(AD_3_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_4_XFER_CNT			(AD_3_SP1_TS_DMA_ADDR_H + 17)
#define AD_4_DEST_XFER_CNT		(AD_4_XFER_CNT + 1)
#define AD_4_SP1_XFER_STEP_SIZE		(AD_4_XFER_CNT + 2)
#define AD_4_SP1_XFER_CTRL		(AD_4_XFER_CNT + 3)
#define AD_4_SP2_XFER_CTRL		(AD_4_XFER_CNT + 4)
#define AD_4_SP3_XFER_CTRL		(AD_4_XFER_CNT + 5)
#define AD_4_SP4_XFER_CTRL		(AD_4_XFER_CNT + 6)
#define AD_4_SP1_TS_XFER_CTRL		(AD_4_XFER_CNT + 7)
#define AD_4_SP2_TS_XFER_CTRL		(AD_4_XFER_CNT + 8)
#define AD_4_SP3_TS_XFER_CTRL		(AD_4_XFER_CNT + 9)
#define AD_4_SP4_TS_XFER_CTRL		(AD_4_XFER_CNT + 10)
#define AD_4_FRAME_CNT			(AD_4_XFER_CNT + 11)
#define AD_4_STREAM_DISABLED		(AD_4_XFER_CNT + 12)
#define AD_4_LAST_FRAME_INDEX		(AD_4_XFER_CNT + 13)
#define AD_4_LAST_ADDR_PTR		(AD_4_XFER_CNT + 14)
#define AD_4_DEST_ADDR_PTR		(AD_4_XFER_CNT + 15)
#define AD_4_LAST_TS_ADDR_PTR		(AD_4_XFER_CNT + 16)
#define AD_4_DEST_TS_ADDR_PTR		(AD_4_XFER_CNT + 17)
#define AD_4_FRAME_XFER_CNT		(AD_4_XFER_CNT + 18)
#define AD_4_SP1_DEST_ADDR_L		(AD_4_XFER_CNT + 20)
#define AD_4_SP1_DEST_ADDR_H		(AD_4_XFER_CNT + 21)
#define AD_4_SP1_TS_DEST_ADDR_L		(AD_4_XFER_CNT + 22)
#define AD_4_SP1_TS_DEST_ADDR_H		(AD_4_XFER_CNT + 23)
#define AD_4_SP1_DMA_ADDR_L		(AD_4_XFER_CNT + 24)
#define AD_4_SP1_DMA_ADDR_H		(AD_4_XFER_CNT + 25)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_4_SP1_TS_DMA_ADDR_L		(AD_4_SP1_DMA_ADDR_H + 17)
#define AD_4_SP1_TS_DMA_ADDR_H		(AD_4_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_5_XFER_CNT			(AD_4_SP1_TS_DMA_ADDR_H + 17)
#define AD_5_DEST_XFER_CNT		(AD_5_XFER_CNT + 1)
#define AD_5_SP1_XFER_STEP_SIZE		(AD_5_XFER_CNT + 2)
#define AD_5_SP1_XFER_CTRL		(AD_5_XFER_CNT + 3)
#define AD_5_SP1_TS_XFER_CTRL		(AD_5_XFER_CNT + 4)
#define AD_5_FRAME_CNT			(AD_5_XFER_CNT + 5)
#define AD_5_STREAM_DISABLED		(AD_5_XFER_CNT + 6)
#define AD_5_LAST_FRAME_INDEX		(AD_5_XFER_CNT + 7)
#define AD_5_LAST_ADDR_PTR		(AD_5_XFER_CNT + 8)
#define AD_5_DEST_ADDR_PTR		(AD_5_XFER_CNT + 9)
#define AD_5_LAST_TS_ADDR_PTR		(AD_5_XFER_CNT + 10)
#define AD_5_DEST_TS_ADDR_PTR		(AD_5_XFER_CNT + 11)
#define AD_5_FRAME_XFER_CNT		(AD_5_XFER_CNT + 12)
#define AD_5_SP1_DEST_ADDR_L		(AD_5_XFER_CNT + 14)
#define AD_5_SP1_DEST_ADDR_H		(AD_5_XFER_CNT + 15)
#define AD_5_SP1_TS_DEST_ADDR_L		(AD_5_XFER_CNT + 16)
#define AD_5_SP1_TS_DEST_ADDR_H		(AD_5_XFER_CNT + 17)
#define AD_5_SP1_DMA_ADDR_L		(AD_5_XFER_CNT + 18)
#define AD_5_SP1_DMA_ADDR_H		(AD_5_XFER_CNT + 19)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_5_SP1_TS_DMA_ADDR_L		(AD_5_SP1_DMA_ADDR_H + 17)
#define AD_5_SP1_TS_DMA_ADDR_H		(AD_5_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_6_XFER_CNT			(AD_5_SP1_TS_DMA_ADDR_H + 17)
#define AD_6_DEST_XFER_CNT		(AD_6_XFER_CNT + 1)
#define AD_6_SP1_XFER_STEP_SIZE		(AD_6_XFER_CNT + 2)
#define AD_6_SP1_XFER_CTRL		(AD_6_XFER_CNT + 3)
#define AD_6_SP1_TS_XFER_CTRL		(AD_6_XFER_CNT + 4)
#define AD_6_FRAME_CNT			(AD_6_XFER_CNT + 5)
#define AD_6_STREAM_DISABLED		(AD_6_XFER_CNT + 6)
#define AD_6_LAST_FRAME_INDEX		(AD_6_XFER_CNT + 7)
#define AD_6_LAST_ADDR_PTR		(AD_6_XFER_CNT + 8)
#define AD_6_DEST_ADDR_PTR		(AD_6_XFER_CNT + 9)
#define AD_6_LAST_TS_ADDR_PTR		(AD_6_XFER_CNT + 10)
#define AD_6_DEST_TS_ADDR_PTR		(AD_6_XFER_CNT + 11)
#define AD_6_FRAME_XFER_CNT		(AD_6_XFER_CNT + 12)
#define AD_6_SP1_DEST_ADDR_L		(AD_6_XFER_CNT + 14)
#define AD_6_SP1_DEST_ADDR_H		(AD_6_XFER_CNT + 15)
#define AD_6_SP1_TS_DEST_ADDR_L		(AD_6_XFER_CNT + 16)
#define AD_6_SP1_TS_DEST_ADDR_H		(AD_6_XFER_CNT + 17)
#define AD_6_SP1_DMA_ADDR_L		(AD_6_XFER_CNT + 18)
#define AD_6_SP1_DMA_ADDR_H		(AD_6_XFER_CNT + 19)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_6_SP1_TS_DMA_ADDR_L		(AD_6_SP1_DMA_ADDR_H + 17)
#define AD_6_SP1_TS_DMA_ADDR_H		(AD_6_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_7_XFER_CNT			(AD_6_SP1_TS_DMA_ADDR_H + 17)
#define AD_7_DEST_XFER_CNT		(AD_7_XFER_CNT + 1)
#define AD_7_SP1_XFER_STEP_SIZE		(AD_7_XFER_CNT + 2)
#define AD_7_SP1_XFER_CTRL		(AD_7_XFER_CNT + 3)
#define AD_7_SP1_TS_XFER_CTRL		(AD_7_XFER_CNT + 4)
#define AD_7_FRAME_CNT			(AD_7_XFER_CNT + 5)
#define AD_7_STREAM_DISABLED		(AD_7_XFER_CNT + 6)
#define AD_7_LAST_FRAME_INDEX		(AD_7_XFER_CNT + 7)
#define AD_7_LAST_ADDR_PTR		(AD_7_XFER_CNT + 8)
#define AD_7_DEST_ADDR_PTR		(AD_7_XFER_CNT + 9)
#define AD_7_LAST_TS_ADDR_PTR		(AD_7_XFER_CNT + 10)
#define AD_7_DEST_TS_ADDR_PTR		(AD_7_XFER_CNT + 11)
#define AD_7_FRAME_XFER_CNT		(AD_7_XFER_CNT + 12)
#define AD_7_SP1_DEST_ADDR_L		(AD_7_XFER_CNT + 14)
#define AD_7_SP1_DEST_ADDR_H		(AD_7_XFER_CNT + 15)
#define AD_7_SP1_TS_DEST_ADDR_L		(AD_7_XFER_CNT + 16)
#define AD_7_SP1_TS_DEST_ADDR_H		(AD_7_XFER_CNT + 17)
#define AD_7_SP1_DMA_ADDR_L		(AD_7_XFER_CNT + 18)
#define AD_7_SP1_DMA_ADDR_H		(AD_7_XFER_CNT + 19)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_7_SP1_TS_DMA_ADDR_L		(AD_7_SP1_DMA_ADDR_H + 17)
#define AD_7_SP1_TS_DMA_ADDR_H		(AD_7_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_8_XFER_CNT			(AD_7_SP1_TS_DMA_ADDR_H + 17)
#define AD_8_DEST_XFER_CNT		(AD_8_XFER_CNT + 1)
#define AD_8_SP1_XFER_STEP_SIZE		(AD_8_XFER_CNT + 2)
#define AD_8_SP1_XFER_CTRL		(AD_8_XFER_CNT + 3)
#define AD_8_SP1_TS_XFER_CTRL		(AD_8_XFER_CNT + 4)
#define AD_8_FRAME_CNT			(AD_8_XFER_CNT + 5)
#define AD_8_STREAM_DISABLED		(AD_8_XFER_CNT + 6)
#define AD_8_LAST_FRAME_INDEX		(AD_8_XFER_CNT + 7)
#define AD_8_LAST_ADDR_PTR		(AD_8_XFER_CNT + 8)
#define AD_8_DEST_ADDR_PTR		(AD_8_XFER_CNT + 9)
#define AD_8_LAST_TS_ADDR_PTR		(AD_8_XFER_CNT + 10)
#define AD_8_DEST_TS_ADDR_PTR		(AD_8_XFER_CNT + 11)
#define AD_8_FRAME_XFER_CNT		(AD_8_XFER_CNT + 12)
#define AD_8_SP1_DEST_ADDR_L		(AD_8_XFER_CNT + 14)
#define AD_8_SP1_DEST_ADDR_H		(AD_8_XFER_CNT + 15)
#define AD_8_SP1_TS_DEST_ADDR_L		(AD_8_XFER_CNT + 16)
#define AD_8_SP1_TS_DEST_ADDR_H		(AD_8_XFER_CNT + 17)
#define AD_8_SP1_DMA_ADDR_L		(AD_8_XFER_CNT + 18)
#define AD_8_SP1_DMA_ADDR_H		(AD_8_XFER_CNT + 19)
// Skip 16 slots to accomodate the remaining 8 possible frames
#define AD_8_SP1_TS_DMA_ADDR_L		(AD_8_SP1_DMA_ADDR_H + 17)
#define AD_8_SP1_TS_DMA_ADDR_H		(AD_8_SP1_TS_DMA_ADDR_L + 1)
// Skip 16 slots to accomodate the remaining 8 possible frames

/* DMA SEQUENCER PROGRAM */

#if 1
static u32 vc8x0_FlexDMAProgram[] = 
{
//START:
//MAIN:
/*	0x0000	*/	VDMA_LOAD_RA(VD_1_STREAM_DISABLED),										// Is this channel enabled?
/*	0x0001	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_2),									// No.  Check next FIFO then
// LP1_VD_1_READY
/*	0x0002	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_0, LP1_VD_2),					// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x0003	*/	VDMA_LOAD_RA(VD_1_DEST_XFER_CNT),										// Get the remaining XFER count
/*	0x0004	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_1_DONE_FULL),						// Branch if we are done with this frame
/*	0x0005	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the remaining XFER count
/*	0x0006	*/	VDMA_STORE_RA(VD_1_DEST_XFER_CNT),										// Save the count
// LP1_VD_1_Y_WAIT_FULL
/*	0x0007	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0008	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x0009	*/	VDMA_LOAD_XFER_CTL(0, VD_1_Y_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x000A	*/	VDMA_ADD_SYS_ADDR_I(VD_1_Y_XFER_SIZE_FULL),							// Bump the Y plane destination address
/*	0x000B	*/	VDMA_STORE_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),							// Save the new Y plane destination address
// LP1_VD_1_U_WAIT_FULL
/*	0x000C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x000D	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x000E	*/	VDMA_LOAD_XFER_CTL(0, VD_1_U_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x000F	*/	VDMA_ADD_SYS_ADDR_I(VD_1_UV_XFER_SIZE_FULL),							// Bump the U plane destination address
/*	0x0010	*/	VDMA_STORE_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),							// Save the new destination address
// LP1_VD_1_V_WAIT_FULL
/*	0x0011	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0012	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x0013	*/	VDMA_LOAD_XFER_CTL(0, VD_1_V_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x0014	*/	VDMA_ADD_SYS_ADDR_I(VD_1_UV_XFER_SIZE_FULL),							// Bump the V Plane destination address
/*	0x0015	*/	VDMA_STORE_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),							// Save the new V Plane destination address
///*	0x0016	*/	VDMA_LOAD_RA(VD_1_DEST_XFER_CNT),										// Get the remaining XFER count
///*	0x0017	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_2),										// Check next FIFO channel
/*	0x0016	*/	VDMA_JMP(VDMA_ALWAYS, 0, LP1_VD_2),										// Check next FIFO channel
/*	0x0017	*/	VDMA_NOP(),																		// Check next FIFO channel
// LP1_VD_1_DONE_FULL:	
/*	0x0018	*/	VDMA_LOAD_RA(VD_1_Y_XFER_SIZE_PART),									// Load partial transfer byte count
/*	0x0019	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_1_Y_TS_WAIT),						// Branch if we are done with this frame
// LP1_VD_1_Y_WAIT_PART
/*	0x001A	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x001B	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x001C	*/	VDMA_LOAD_XFER_CTL(0, VD_1_Y_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_1_U_WAIT_PART
/*	0x001D	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x001E	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x001F	*/	VDMA_LOAD_XFER_CTL(0, VD_1_U_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_1_V_WAIT_PART
/*	0x0020	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0021	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x0022	*/	VDMA_LOAD_XFER_CTL(0, VD_1_V_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_1_Y_TS_WAIT:
/*	0x0023	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0024	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_Y_TS_DEST_ADDR_L),						// Load the V Plane time stamp address
/*	0x0025	*/	VDMA_LOAD_XFER_CTL(0, VD_1_Y_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_1_U_TS_WAIT:
/*	0x0026	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0027	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),												// Load the U Plane time stamp address
/*	0x0028	*/	VDMA_LOAD_XFER_CTL(0, VD_1_U_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_1_V_TS_WAIT:
/*	0x0029	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x002A	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),												// Load the U Plane time stamp address
/*	0x002B	*/	VDMA_LOAD_XFER_CTL(0, VD_1_V_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x002C	*/	VDMA_LOAD_RA(VD_1_FRAME_XFER_CNT),										// Load the old last frame xfer'd index
/*	0x002D	*/	VDMA_STORE_RA(VD_1_LAST_FRAME_INDEX),									// Save the new last frame xfer'd index
/*	0x002E	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x002F	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_0),
/*	0x0030	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0031	*/	VDMA_LOAD_RA(VD_1_FRAME_XFER_CNT),										// Get the count of DMA frames remaining to xfer before wrap
/*	0x0032	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the count
/*	0x0033	*/	VDMA_STORE_RA(VD_1_FRAME_XFER_CNT),										// Save the new DMA frames to xfer
/*	0x0034	*/	VDMA_LOAD_RB(VD_1_DEST_TS_ADDR_PTR),									// Load the last time stamp desination address pointer
/*	0x0035	*/	VDMA_ADD_RB(MINUS_2),														// Decrement the count
/*	0x0036	*/	VDMA_STORE_RB(VD_1_DEST_TS_ADDR_PTR),									// Save the new time stamp desination pointer
/*	0x0037	*/	VDMA_LOAD_RB(VD_1_DEST_ADDR_PTR),										// Load the last desination address pointer
/*	0x0038	*/	VDMA_ADD_RB(MINUS_6),														// Decrement the count
/*	0x0039	*/	VDMA_STORE_RB(VD_1_DEST_ADDR_PTR),										// Save the new desination pointer
/*	0x003A	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_1_LOAD_ADDR),						// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x003B	*/	VDMA_LOAD_RB(VD_1_LAST_TS_ADDR_PTR),									// Load last DMA time stamp pointer for this FIFO
/*	0x003C	*/	VDMA_STORE_RB(VD_1_DEST_TS_ADDR_PTR),									// Save it as the current DMA time stamp pointer
/*	0x003D	*/	VDMA_LOAD_RB(VD_1_LAST_ADDR_PTR),										// Load last DMA frame pointer for this FIFO
/*	0x003E	*/	VDMA_STORE_RB(VD_1_DEST_ADDR_PTR),										// Save it as the current DMA frame pointer
/*	0x003F	*/	VDMA_LOAD_RA(VD_1_FRAME_CNT),												// Load DMA frames to xfer count
/*	0x0040	*/	VDMA_STORE_RA(VD_1_FRAME_XFER_CNT),										// Save it 
// LP1_VD_1_LOAD_ADDR
/*	0x0041	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane
/*	0x0042	*/	VDMA_STORE_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),							// Save the new destination address for the Y plane
/*	0x0043	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),											// Load the new destination address for the U plane
/*	0x0044	*/	VDMA_STORE_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),							// Save the new destination address for the U plane
/*	0x0045	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),											// Load the new destination address for the V plane
/*	0x0046	*/	VDMA_STORE_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),							// Save the new destination address for the V plane
/*	0x0047	*/	VDMA_LOAD_RB(VD_1_DEST_TS_ADDR_PTR),									// Load the last desination address pointer
/*	0x0048	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane time stamp
/*	0x0049	*/	VDMA_STORE_SYS_ADDR(0, VD_1_Y_TS_DEST_ADDR_L),						// Save the new destination address for the Y plane time stamp
/*	0x004A	*/	VDMA_LOAD_RA(VD_1_XFER_CNT),												// Load initial XFERS per DMA frame
/*	0x004B	*/	VDMA_STORE_RA(VD_1_DEST_XFER_CNT),										// Reset XFERS per DMA frame

// LP1_VD_2:
/*	0x004C	*/	VDMA_LOAD_RA(VD_2_STREAM_DISABLED),										// Is this channel enabled?
/*	0x004D	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_3),									// No.  Check next FIFO then
// LP1_VD_2_READY
/*	0x004E	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_2, LP1_VD_3),					// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x004F	*/	VDMA_LOAD_RA(VD_2_DEST_XFER_CNT),										// Get the remaining XFER count
/*	0x0050	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_2_DONE_FULL),						// Branch if we are done with this frame
/*	0x0051	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the remaining XFER count
/*	0x0052	*/	VDMA_STORE_RA(VD_2_DEST_XFER_CNT),										// Save the count
// LP1_VD_2_Y_WAIT_FULL
/*	0x0053	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0054	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x0055	*/	VDMA_LOAD_XFER_CTL(0, VD_2_Y_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x0056	*/	VDMA_ADD_SYS_ADDR_I(VD_2_Y_XFER_SIZE_FULL),							// Bump the Y plane destination address
/*	0x0057	*/	VDMA_STORE_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),							// Save the new Y plane destination address
// LP1_VD_2_U_WAIT_FULL
/*	0x0058	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0059	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x005A	*/	VDMA_LOAD_XFER_CTL(0, VD_2_U_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x005B	*/	VDMA_ADD_SYS_ADDR_I(VD_2_UV_XFER_SIZE_FULL),							// Bump the U plane destination address
/*	0x005C	*/	VDMA_STORE_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),							// Save the new destination address
// LP1_VD_2_V_WAIT_FULL
/*	0x005D	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x005E	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x005F	*/	VDMA_LOAD_XFER_CTL(0, VD_2_V_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x0060	*/	VDMA_ADD_SYS_ADDR_I(VD_2_UV_XFER_SIZE_FULL),							// Bump the V Plane destination address
/*	0x0061	*/	VDMA_STORE_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),							// Save the new V Plane destination address
///*	0x0062	*/	VDMA_LOAD_RA(VD_2_DEST_XFER_CNT),										// Get the remaining XFER count
///*	0x0063	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_3),										// Check next FIFO channel
/*	0x0062	*/	VDMA_JMP(VDMA_ALWAYS, 0, LP1_VD_3),										// Check next FIFO channel
/*	0x0063	*/	VDMA_NOP(),																		// Check next FIFO channel
// LP1_VD_2_DONE_FULL:	
/*	0x0064	*/	VDMA_LOAD_RA(VD_2_Y_XFER_SIZE_PART),									// Load partial transfer byte count
/*	0x0065	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_2_Y_TS_WAIT),						// Branch if we are done with this frame
// LP1_VD_2_Y_WAIT_PART
/*	0x0066	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0067	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x0068	*/	VDMA_LOAD_XFER_CTL(0, VD_2_Y_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_2_U_WAIT_PART
/*	0x0069	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x006A	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x006B	*/	VDMA_LOAD_XFER_CTL(0, VD_2_U_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_2_V_WAIT_PART
/*	0x006C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x006D	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x006E	*/	VDMA_LOAD_XFER_CTL(0, VD_2_V_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_2_Y_TS_WAIT:
/*	0x006F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0070	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_Y_TS_DEST_ADDR_L),						// Load the V Plane time stamp address
/*	0x0071	*/	VDMA_LOAD_XFER_CTL(0, VD_2_Y_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_2_U_TS_WAIT:
/*	0x0072	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0073	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),												// Load the U Plane time stamp address
/*	0x0074	*/	VDMA_LOAD_XFER_CTL(0, VD_2_U_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_2_V_TS_WAIT:
/*	0x0075	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0076	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),												// Load the U Plane time stamp address
/*	0x0077	*/	VDMA_LOAD_XFER_CTL(0, VD_2_V_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x0078	*/	VDMA_LOAD_RA(VD_2_FRAME_XFER_CNT),										// Load the old last frame xfer'd index
/*	0x0079	*/	VDMA_STORE_RA(VD_2_LAST_FRAME_INDEX),									// Save the new last frame xfer'd index
/*	0x007A	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x007B	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_2),
/*	0x007C	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x007D	*/	VDMA_LOAD_RA(VD_2_FRAME_XFER_CNT),										// Get the count of DMA frames remaining to xfer before wrap
/*	0x007E	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the count
/*	0x007F	*/	VDMA_STORE_RA(VD_2_FRAME_XFER_CNT),										// Save the new DMA frames to xfer
/*	0x0080	*/	VDMA_LOAD_RB(VD_2_DEST_TS_ADDR_PTR),									// Load the last time stamp desination address pointer
/*	0x0081	*/	VDMA_ADD_RB(MINUS_2),														// Decrement the count
/*	0x0082	*/	VDMA_STORE_RB(VD_2_DEST_TS_ADDR_PTR),									// Save the new time stamp desination pointer
/*	0x0083	*/	VDMA_LOAD_RB(VD_2_DEST_ADDR_PTR),										// Load the last desination address pointer
/*	0x0084	*/	VDMA_ADD_RB(MINUS_6),														// Decrement the count
/*	0x0085	*/	VDMA_STORE_RB(VD_2_DEST_ADDR_PTR),										// Save the new desination pointer
/*	0x0086	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_2_LOAD_ADDR),						// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x0087	*/	VDMA_LOAD_RB(VD_2_LAST_TS_ADDR_PTR),									// Load last DMA time stamp pointer for this FIFO
/*	0x0088	*/	VDMA_STORE_RB(VD_2_DEST_TS_ADDR_PTR),									// Save it as the current DMA time stamp pointer
/*	0x0089	*/	VDMA_LOAD_RB(VD_2_LAST_ADDR_PTR),										// Load last DMA frame pointer for this FIFO
/*	0x008A	*/	VDMA_STORE_RB(VD_2_DEST_ADDR_PTR),										// Save it as the current DMA frame pointer
/*	0x008B	*/	VDMA_LOAD_RA(VD_2_FRAME_CNT),												// Load DMA frames to xfer count
/*	0x008C	*/	VDMA_STORE_RA(VD_2_FRAME_XFER_CNT),										// Save it 
// LP1_VD_2_LOAD_ADDR
/*	0x008D	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane
/*	0x008E	*/	VDMA_STORE_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),							// Save the new destination address for the Y plane
/*	0x008F	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),											// Load the new destination address for the U plane
/*	0x0090	*/	VDMA_STORE_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),							// Save the new destination address for the U plane
/*	0x0091	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),											// Load the new destination address for the V plane
/*	0x0092	*/	VDMA_STORE_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),							// Save the new destination address for the V plane
/*	0x0093	*/	VDMA_LOAD_RB(VD_2_DEST_TS_ADDR_PTR),									// Load the last desination address pointer
/*	0x0094	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane time stamp
/*	0x0095	*/	VDMA_STORE_SYS_ADDR(0, VD_2_Y_TS_DEST_ADDR_L),						// Save the new destination address for the Y plane time stamp
/*	0x0096	*/	VDMA_LOAD_RA(VD_2_XFER_CNT),												// Load initial XFERS per DMA frame
/*	0x0097	*/	VDMA_STORE_RA(VD_2_DEST_XFER_CNT),										// Reset XFERS per DMA frame

//	LP1_VD_3
/*	0x0098	*/	VDMA_LOAD_RA(VD_3_STREAM_DISABLED),										// Is this channel enabled?
/*	0x0099	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_4),									// No.  Check next FIFO then
// LP1_VD_3_READY
/*	0x009A	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_4, LP1_VD_4),					// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x009B	*/	VDMA_LOAD_RA(VD_3_DEST_XFER_CNT),										// Get the remaining XFER count
/*	0x009C	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_3_DONE_FULL),						// Branch if we are done with this frame
/*	0x009D	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the remaining XFER count
/*	0x009E	*/	VDMA_STORE_RA(VD_3_DEST_XFER_CNT),										// Save the count
// LP1_VD_3_Y_WAIT_FULL
/*	0x009F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00A0	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x00A1	*/	VDMA_LOAD_XFER_CTL(0, VD_3_Y_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x00A2	*/	VDMA_ADD_SYS_ADDR_I(VD_3_Y_XFER_SIZE_FULL),							// Bump the Y plane destination address
/*	0x00A3	*/	VDMA_STORE_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),							// Save the new Y plane destination address
// LP1_VD_3_U_WAIT_FULL
/*	0x00A4	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00A5	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x00A6	*/	VDMA_LOAD_XFER_CTL(0, VD_3_U_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x00A7	*/	VDMA_ADD_SYS_ADDR_I(VD_3_UV_XFER_SIZE_FULL),							// Bump the U plane destination address
/*	0x00A8	*/	VDMA_STORE_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),							// Save the new destination address
// LP1_VD_3_V_WAIT_FULL
/*	0x00A9	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00AA	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x00AB	*/	VDMA_LOAD_XFER_CTL(0, VD_3_V_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x00AC	*/	VDMA_ADD_SYS_ADDR_I(VD_3_UV_XFER_SIZE_FULL),							// Bump the V Plane destination address
/*	0x00AD	*/	VDMA_STORE_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),							// Save the new V Plane destination address
///*	0x00AE	*/	VDMA_LOAD_RA(VD_3_DEST_XFER_CNT),										// Get the remaining XFER count
///*	0x00AF	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_4),									// Check next FIFO channel
/*	0x00AE	*/	VDMA_JMP(VDMA_ALWAYS, 0, LP1_VD_4),										// Check next FIFO channel
/*	0x00AF	*/	VDMA_NOP(),																		// Check next FIFO channel
// LP1_VD_3_DONE_FULL:	
/*	0x00B0	*/	VDMA_LOAD_RA(VD_3_Y_XFER_SIZE_PART),									// Load partial transfer byte count
/*	0x00B1	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_3_Y_TS_WAIT),						// Branch if we are done with this frame
// LP1_VD_3_Y_WAIT_PART
/*	0x00B2	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00B3	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x00B4	*/	VDMA_LOAD_XFER_CTL(0, VD_3_Y_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_3_U_WAIT_PART
/*	0x00B5	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00B6	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x00B7	*/	VDMA_LOAD_XFER_CTL(0, VD_3_U_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_3_V_WAIT_PART
/*	0x00B8	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00B9	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x00BA	*/	VDMA_LOAD_XFER_CTL(0, VD_3_V_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_3_Y_TS_WAIT:
/*	0x00BB	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x00BC	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_Y_TS_DEST_ADDR_L),						// Load the V Plane time stamp address
/*	0x00BD	*/	VDMA_LOAD_XFER_CTL(0, VD_3_Y_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_3_U_TS_WAIT:
/*	0x00BE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x00BF	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),												// Load the U Plane time stamp address
/*	0x00C0	*/	VDMA_LOAD_XFER_CTL(0, VD_3_U_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_3_V_TS_WAIT:
/*	0x00C1	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x00C2	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),												// Load the U Plane time stamp address
/*	0x00C3	*/	VDMA_LOAD_XFER_CTL(0, VD_3_V_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x00C4	*/	VDMA_LOAD_RA(VD_3_FRAME_XFER_CNT),										// Load the old last frame xfer'd index
/*	0x00C5	*/	VDMA_STORE_RA(VD_3_LAST_FRAME_INDEX),									// Save the new last frame xfer'd index
/*	0x00C6	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x00C7	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_4),
/*	0x00C8	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x00C9	*/	VDMA_LOAD_RA(VD_3_FRAME_XFER_CNT),										// Get the count of DMA frames remaining to xfer before wrap
/*	0x00CA	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the count
/*	0x00CB	*/	VDMA_STORE_RA(VD_3_FRAME_XFER_CNT),										// Save the new DMA frames to xfer
/*	0x00CC	*/	VDMA_LOAD_RB(VD_3_DEST_TS_ADDR_PTR),									// Load the last time stamp desination address pointer
/*	0x00CD	*/	VDMA_ADD_RB(MINUS_2),														// Decrement the count
/*	0x00CE	*/	VDMA_STORE_RB(VD_3_DEST_TS_ADDR_PTR),									// Save the new time stamp desination pointer
/*	0x00CF	*/	VDMA_LOAD_RB(VD_3_DEST_ADDR_PTR),										// Load the last desination address pointer
/*	0x00D0	*/	VDMA_ADD_RB(MINUS_6),														// Decrement the count
/*	0x00D1	*/	VDMA_STORE_RB(VD_3_DEST_ADDR_PTR),										// Save the new desination pointer
/*	0x00D2	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_3_LOAD_ADDR),						// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x00D3	*/	VDMA_LOAD_RB(VD_3_LAST_TS_ADDR_PTR),									// Load last DMA time stamp pointer for this FIFO
/*	0x00D4	*/	VDMA_STORE_RB(VD_3_DEST_TS_ADDR_PTR),									// Save it as the current DMA time stamp pointer
/*	0x00D5	*/	VDMA_LOAD_RB(VD_3_LAST_ADDR_PTR),										// Load last DMA frame pointer for this FIFO
/*	0x00D6	*/	VDMA_STORE_RB(VD_3_DEST_ADDR_PTR),										// Save it as the current DMA frame pointer
/*	0x00D7	*/	VDMA_LOAD_RA(VD_3_FRAME_CNT),												// Load DMA frames to xfer count
/*	0x00D8	*/	VDMA_STORE_RA(VD_3_FRAME_XFER_CNT),										// Save it 
// LP1_VD_3_LOAD_ADDR
/*	0x00D9	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane
/*	0x00DA	*/	VDMA_STORE_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),							// Save the new destination address for the Y plane
/*	0x00DB	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),											// Load the new destination address for the U plane
/*	0x00DC	*/	VDMA_STORE_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),							// Save the new destination address for the U plane
/*	0x00DD	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),											// Load the new destination address for the V plane
/*	0x00DE	*/	VDMA_STORE_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),							// Save the new destination address for the V plane
/*	0x00DF	*/	VDMA_LOAD_RB(VD_3_DEST_TS_ADDR_PTR),									// Load the last desination address pointer
/*	0x00E0	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane time stamp
/*	0x00E1	*/	VDMA_STORE_SYS_ADDR(0, VD_3_Y_TS_DEST_ADDR_L),						// Save the new destination address for the Y plane time stamp
/*	0x00E2	*/	VDMA_LOAD_RA(VD_3_XFER_CNT),												// Load initial XFERS per DMA frame
/*	0x00E3	*/	VDMA_STORE_RA(VD_3_DEST_XFER_CNT),										// Reset XFERS per DMA frame

//	LP1_VD_4
/*	0x00E4	*/	VDMA_LOAD_RA(VD_4_STREAM_DISABLED),										// Is this channel enabled?
/*	0x00E5	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_1),									// No.  Check next FIFO then
// LP1_VD_4_READY
/*	0x00E6	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_6, LP1_AD_1),					// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x00E7	*/	VDMA_LOAD_RA(VD_4_DEST_XFER_CNT),										// Get the remaining XFER count
/*	0x00E8	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_4_DONE_FULL),						// Branch if we are done with this frame
/*	0x00E9	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the remaining XFER count
/*	0x00EA	*/	VDMA_STORE_RA(VD_4_DEST_XFER_CNT),										// Save the count
// LP1_VD_4_Y_WAIT_FULL
/*	0x00EB	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00EC	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x00ED	*/	VDMA_LOAD_XFER_CTL(0, VD_4_Y_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x00EE	*/	VDMA_ADD_SYS_ADDR_I(VD_4_Y_XFER_SIZE_FULL),							// Bump the Y plane destination address
/*	0x00EF	*/	VDMA_STORE_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),							// Save the new Y plane destination address
// LP1_VD_4_U_WAIT_FULL
/*	0x00F0	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00F1	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x00F2	*/	VDMA_LOAD_XFER_CTL(0, VD_4_U_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x00F3	*/	VDMA_ADD_SYS_ADDR_I(VD_4_UV_XFER_SIZE_FULL),							// Bump the U plane destination address
/*	0x00F4	*/	VDMA_STORE_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),							// Save the new destination address
// LP1_VD_4_V_WAIT_FULL
/*	0x00F5	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00F6	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x00F7	*/	VDMA_LOAD_XFER_CTL(0, VD_4_V_XFER_CTRL_FULL),						// Transfer this portion of the frame
/*	0x00F8	*/	VDMA_ADD_SYS_ADDR_I(VD_4_UV_XFER_SIZE_FULL),							// Bump the V Plane destination address
/*	0x00F9	*/	VDMA_STORE_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),							// Save the new V Plane destination address
///*	0x00FA	*/	VDMA_LOAD_RA(VD_4_DEST_XFER_CNT),										// Get the remaining XFER count
///*	0x00FB	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_1),										// Check next FIFO channel
/*	0x00FA	*/	VDMA_JMP(VDMA_ALWAYS, 0, LP1_AD_1),										// Check next FIFO channel
/*	0x00FB	*/	VDMA_NOP(),																		// Check next FIFO channel
// LP1_VD_4_DONE_FULL:	
/*	0x00FC	*/	VDMA_LOAD_RA(VD_4_Y_XFER_SIZE_PART),									// Load partial transfer byte count
/*	0x00FD	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_4_Y_TS_WAIT),						// Branch if we are done with this frame
// LP1_VD_4_Y_WAIT_PART
/*	0x00FE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00FF	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),							// Load the Y plane destination addresso
/*	0x0100	*/	VDMA_LOAD_XFER_CTL(0, VD_4_Y_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_4_U_WAIT_PART
/*	0x0101	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0102	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),							// Load the U plane destination address
/*	0x0103	*/	VDMA_LOAD_XFER_CTL(0, VD_4_U_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_4_V_WAIT_PART
/*	0x0104	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0105	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),							// Load the V Plane destination address
/*	0x0106	*/	VDMA_LOAD_XFER_CTL(0, VD_4_V_XFER_CTRL_PART),						// Transfer this portion of the frame
// LP1_VD_4_Y_TS_WAIT:
/*	0x0107	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0108	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_Y_TS_DEST_ADDR_L),						// Load the V Plane time stamp address
/*	0x0109	*/	VDMA_LOAD_XFER_CTL(0, VD_4_Y_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_4_U_TS_WAIT:
/*	0x010A	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x010B	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),												// Load the U Plane time stamp address
/*	0x010C	*/	VDMA_LOAD_XFER_CTL(0, VD_4_U_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_VD_4_V_TS_WAIT:
/*	0x010D	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x010E	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),												// Load the U Plane time stamp address
/*	0x010F	*/	VDMA_LOAD_XFER_CTL(0, VD_4_V_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x0110	*/	VDMA_LOAD_RA(VD_4_FRAME_XFER_CNT),										// Load the old last frame xfer'd index
/*	0x0111	*/	VDMA_STORE_RA(VD_4_LAST_FRAME_INDEX),									// Save the new last frame xfer'd index
/*	0x0112	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0113	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_6),
/*	0x0114	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0115	*/	VDMA_LOAD_RA(VD_4_FRAME_XFER_CNT),										// Get the count of DMA frames remaining to xfer before wrap
/*	0x0116	*/	VDMA_ADD_RA(MINUS_1),														// Decrement the count
/*	0x0117	*/	VDMA_STORE_RA(VD_4_FRAME_XFER_CNT),										// Save the new DMA frames to xfer
/*	0x0118	*/	VDMA_LOAD_RB(VD_4_DEST_TS_ADDR_PTR),									// Load the last time stamp desination address pointer
/*	0x0119	*/	VDMA_ADD_RB(MINUS_2),														// Decrement the count
/*	0x011A	*/	VDMA_STORE_RB(VD_4_DEST_TS_ADDR_PTR),									// Save the new time stamp desination pointer
/*	0x011B	*/	VDMA_LOAD_RB(VD_4_DEST_ADDR_PTR),										// Load the last desination address pointer
/*	0x011C	*/	VDMA_ADD_RB(MINUS_6),														// Decrement the count
/*	0x011D	*/	VDMA_STORE_RB(VD_4_DEST_ADDR_PTR),										// Save the new desination pointer
/*	0x011E	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_4_LOAD_ADDR),						// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x011F	*/	VDMA_LOAD_RB(VD_4_LAST_TS_ADDR_PTR),									// Load last DMA time stamp pointer for this FIFO
/*	0x0120	*/	VDMA_STORE_RB(VD_4_DEST_TS_ADDR_PTR),									// Save it as the current DMA time stamp pointer
/*	0x0121	*/	VDMA_LOAD_RB(VD_4_LAST_ADDR_PTR),										// Load last DMA frame pointer for this FIFO
/*	0x0122	*/	VDMA_STORE_RB(VD_4_DEST_ADDR_PTR),										// Save it as the current DMA frame pointer
/*	0x0123	*/	VDMA_LOAD_RA(VD_4_FRAME_CNT),												// Load DMA frames to xfer count
/*	0x0124	*/	VDMA_STORE_RA(VD_4_FRAME_XFER_CNT),										// Save it 
// LP1_VD_4_LOAD_ADDR
/*	0x0125	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane
/*	0x0126	*/	VDMA_STORE_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),							// Save the new destination address for the Y plane
/*	0x0127	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),											// Load the new destination address for the U plane
/*	0x0128	*/	VDMA_STORE_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),							// Save the new destination address for the U plane
/*	0x0129	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),											// Load the new destination address for the V plane
/*	0x012A	*/	VDMA_STORE_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),							// Save the new destination address for the V plane
/*	0x012B	*/	VDMA_LOAD_RB(VD_4_DEST_TS_ADDR_PTR),									// Load the last desination address pointer
/*	0x012C	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),											// Load the new destination address for the Y plane time stamp
/*	0x012D	*/	VDMA_STORE_SYS_ADDR(0, VD_4_Y_TS_DEST_ADDR_L),						// Save the new destination address for the Y plane time stamp
/*	0x012E	*/	VDMA_LOAD_RA(VD_4_XFER_CNT),												// Load initial XFERS per DMA frame
/*	0x012F	*/	VDMA_STORE_RA(VD_4_DEST_XFER_CNT),										// Reset XFERS per DMA frame

//	LP1_AD_1
/*	0x0130	*/	VDMA_LOAD_RA(AD_1_STREAM_DISABLED),										// Is this channel enabled?
/*	0x0131	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_2),										// No.  See if it is ready then then
// LP1_AD_1_READY
/*	0x0132	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_1, LP1_AD_2),						// Check next FIFO if this one is not ready
// LP1_AD_1_SP1_WAIT_FULL
/*	0x0133	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0134	*/	VDMA_LOAD_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x0135	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP1_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_1_SP2_WAIT_FULL
/*	0x0136	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP2_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0137	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 2
/*	0x0138	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP2_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_1_SP3_WAIT_FULL
/*	0x0139	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP3_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x013A	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 3
/*	0x013B	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP3_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_1_SP4_WAIT_FULL
/*	0x013C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP4_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x013D	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 4
/*	0x013E	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP4_XFER_CTRL),								// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x013F	*/	VDMA_LOAD_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x0140	*/	VDMA_ADD_SYS_ADDR_I(AD_1_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x0141	*/	VDMA_STORE_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
/*	0x0142	*/	VDMA_LOAD_RA(AD_1_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x0143	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x0144	*/	VDMA_STORE_RA(AD_1_DEST_XFER_CNT),											// Save the count
/*	0x0145	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_2),										// Branch if we are not done with this frame
// LP1_AD_1_SP1_TS_WAIT:
/*	0x0146	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0147	*/	VDMA_LOAD_SYS_ADDR(0, AD_1_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x0148	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_1_SP2_TS_WAIT:
/*	0x0149	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP2_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x014A	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 2 time stamp address
/*	0x014B	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP2_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_1_SP3_TS_WAIT:
/*	0x014C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP3_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x014D	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 3 time stamp address
/*	0x014E	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP3_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_1_SP4_TS_WAIT:
/*	0x014F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP4_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0150	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 4 time stamp address
/*	0x0151	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP4_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x0152	*/	VDMA_LOAD_RA(AD_1_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x0153	*/	VDMA_STORE_RA(AD_1_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x0154	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0155	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_1),
/*	0x0156	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0157	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x0158	*/	VDMA_STORE_RA(AD_1_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x0159	*/	VDMA_LOAD_RB(AD_1_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x015A	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x015B	*/	VDMA_STORE_RB(AD_1_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x015C	*/	VDMA_LOAD_RB(AD_1_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x015D	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x015E	*/	VDMA_STORE_RB(AD_1_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x015F	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_1_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x0160	*/	VDMA_LOAD_RB(AD_1_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x0161	*/	VDMA_STORE_RB(AD_1_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x0162	*/	VDMA_LOAD_RB(AD_1_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x0163	*/	VDMA_STORE_RB(AD_1_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x0164	*/	VDMA_LOAD_RA(AD_1_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x0165	*/	VDMA_STORE_RA(AD_1_FRAME_XFER_CNT),											// Save it 
// LP1_AD_1_LOAD_ADDR
/*	0x0166	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x0167	*/	VDMA_STORE_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x0168	*/	VDMA_LOAD_RB(AD_1_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x0169	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x016A	*/	VDMA_STORE_SYS_ADDR(0, AD_1_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x016B	*/	VDMA_LOAD_RA(AD_1_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x016C	*/	VDMA_STORE_RA(AD_1_DEST_XFER_CNT),											// Reset XFERS per DMA frame

//	LP1_AD_2
/*	0x016D	*/	VDMA_LOAD_RA(AD_2_STREAM_DISABLED),											// Is this channel enabled?
/*	0x016E	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_3),									// Yes.  See if it is ready then then
// LP1_AD_2_READY
/*	0x016F	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_3, LP1_AD_3),						// Check next FIFO if this one is not ready
// LP1_AD_2_SP1_WAIT_FULL
/*	0x0170	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0171	*/	VDMA_LOAD_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x0172	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP1_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_2_SP2_WAIT_FULL
/*	0x0173	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP2_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0174	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 2
/*	0x0175	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP2_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_2_SP3_WAIT_FULL
/*	0x0176	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP3_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0177	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 3
/*	0x0178	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP3_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_2_SP4_WAIT_FULL
/*	0x0179	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP4_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x017A	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 4
/*	0x017B	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP4_XFER_CTRL),								// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x017C	*/	VDMA_LOAD_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x017D	*/	VDMA_ADD_SYS_ADDR_I(AD_2_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x017E	*/	VDMA_STORE_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
/*	0x017F	*/	VDMA_LOAD_RA(AD_2_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x0180	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x0181	*/	VDMA_STORE_RA(AD_2_DEST_XFER_CNT),											// Save the count
/*	0x0182	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_3),										// Branch if we are not done with this frame
// LP1_AD_2_SP1_TS_WAIT:
/*	0x0183	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0184	*/	VDMA_LOAD_SYS_ADDR(0, AD_2_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x0185	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_2_SP2_TS_WAIT:
/*	0x0186	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP2_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0187	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 2 time stamp address
/*	0x0188	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP2_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_2_SP3_TS_WAIT:
/*	0x0189	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP3_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x018A	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 3 time stamp address
/*	0x018B	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP3_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_2_SP4_TS_WAIT:
/*	0x018C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP4_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x018D	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 4 time stamp address
/*	0x018E	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP4_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x018F	*/	VDMA_LOAD_RA(AD_2_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x0190	*/	VDMA_STORE_RA(AD_2_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x0191	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0192	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_3),
/*	0x0193	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0194	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x0195	*/	VDMA_STORE_RA(AD_2_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x0196	*/	VDMA_LOAD_RB(AD_2_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x0197	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x0198	*/	VDMA_STORE_RB(AD_2_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x0199	*/	VDMA_LOAD_RB(AD_2_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x019A	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x019B	*/	VDMA_STORE_RB(AD_2_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x019C	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_2_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x019D	*/	VDMA_LOAD_RB(AD_2_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x019E	*/	VDMA_STORE_RB(AD_2_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x019F	*/	VDMA_LOAD_RB(AD_2_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x01A0	*/	VDMA_STORE_RB(AD_2_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x01A1	*/	VDMA_LOAD_RA(AD_2_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x01A2	*/	VDMA_STORE_RA(AD_2_FRAME_XFER_CNT),											// Save it 
// LP1_AD_2_LOAD_ADDR
/*	0x01A3	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x01A4	*/	VDMA_STORE_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x01A5	*/	VDMA_LOAD_RB(AD_2_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x01A6	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x01A7	*/	VDMA_STORE_SYS_ADDR(0, AD_2_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x01A8	*/	VDMA_LOAD_RA(AD_2_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x01A9	*/	VDMA_STORE_RA(AD_2_DEST_XFER_CNT),											// Reset XFERS per DMA frame

// LP1_AD_3
/*	0x01AA	*/	VDMA_LOAD_RA(AD_3_STREAM_DISABLED),											// Is this channel enabled?
/*	0x01AB	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_4),										// Yes.  See if it is ready then then
// LP1_AD_3_READY
/*	0x01AC	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_5, LP1_AD_4),						// Check next FIFO if this one is not ready
// LP1_AD_3_SP1_WAIT_FULL
/*	0x01AD	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01AE	*/	VDMA_LOAD_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x01AF	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP1_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_3_SP2_WAIT_FULL
/*	0x01B0	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP2_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01B1	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 2
/*	0x01B2	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP2_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_3_SP3_WAIT_FULL
/*	0x01B3	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP3_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01B4	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 3
/*	0x01B5	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP3_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_3_SP4_WAIT_FULL
/*	0x01B6	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP4_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01B7	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 4
/*	0x01B8	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP4_XFER_CTRL),								// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x01B9	*/	VDMA_LOAD_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x01BA	*/	VDMA_ADD_SYS_ADDR_I(AD_3_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x01BB	*/	VDMA_STORE_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
/*	0x01BC	*/	VDMA_LOAD_RA(AD_3_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x01BD	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x01BE	*/	VDMA_STORE_RA(AD_3_DEST_XFER_CNT),											// Save the count
/*	0x01BF	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_4),										// Branch if we are not done with this frame
// LP1_AD_3_SP1_TS_WAIT:
/*	0x01C0	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x01C1	*/	VDMA_LOAD_SYS_ADDR(0, AD_3_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x01C2	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_3_SP2_TS_WAIT:
/*	0x01C3	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP2_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x01C4	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 2 time stamp address
/*	0x01C5	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP2_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_3_SP3_TS_WAIT:
/*	0x01C6	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP3_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x01C7	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 3 time stamp address
/*	0x01C8	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP3_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_3_SP4_TS_WAIT:
/*	0x01C9	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP4_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x01CA	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 4 time stamp address
/*	0x01CB	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP4_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x01CC	*/	VDMA_LOAD_RA(AD_3_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x01CD	*/	VDMA_STORE_RA(AD_3_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x01CE	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x01CF	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_5),
/*	0x01D0	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x01D1	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x01D2	*/	VDMA_STORE_RA(AD_3_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x01D3	*/	VDMA_LOAD_RB(AD_3_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x01D4	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x01D5	*/	VDMA_STORE_RB(AD_3_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x01D6	*/	VDMA_LOAD_RB(AD_3_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x01D7	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x01D8	*/	VDMA_STORE_RB(AD_3_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x01D9	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_3_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x01DA	*/	VDMA_LOAD_RB(AD_3_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x01DB	*/	VDMA_STORE_RB(AD_3_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x01DC	*/	VDMA_LOAD_RB(AD_3_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x01DD	*/	VDMA_STORE_RB(AD_3_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x01DE	*/	VDMA_LOAD_RA(AD_3_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x01DF	*/	VDMA_STORE_RA(AD_3_FRAME_XFER_CNT),											// Save it 
// LP1_AD_3_LOAD_ADDR
/*	0x01E0	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x01E1	*/	VDMA_STORE_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x01E2	*/	VDMA_LOAD_RB(AD_3_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x01E3	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x01E4	*/	VDMA_STORE_SYS_ADDR(0, AD_3_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x01E5	*/	VDMA_LOAD_RA(AD_3_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x01E6	*/	VDMA_STORE_RA(AD_3_DEST_XFER_CNT),											// Reset XFERS per DMA frame

// LP1_AD_4
/*	0x01E7	*/	VDMA_LOAD_RA(AD_4_STREAM_DISABLED),											// Is this channel enabled?
/*	0x01E8	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_5),										// Yes.  See if it is ready then then
/*	0x01E9	*/	VDMA_NOP(),																			// Check next FIFO if this one is not ready
// LP1_AD_4_READY
/*	0x01EA	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_7, LP1_AD_5),						// Check next FIFO if this one is not ready
// LP1_AD_4_SP1_WAIT_FULL
/*	0x01EB	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01EC	*/	VDMA_LOAD_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x01ED	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP1_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_4_SP2_WAIT_FULL
/*	0x01EE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP2_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01EF	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 2
/*	0x01F0	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP2_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_4_SP3_WAIT_FULL
/*	0x01F1	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP3_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01F2	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 3
/*	0x01F3	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP3_XFER_CTRL),								// Transfer this portion of the frame
// LP1_AD_4_SP4_WAIT_FULL
/*	0x01F4	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP4_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x01F5	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),									// Bump the destination address to stereo pair 4
/*	0x01F6	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP4_XFER_CTRL),								// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x01F7	*/	VDMA_LOAD_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x01F8	*/	VDMA_ADD_SYS_ADDR_I(AD_4_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x01F9	*/	VDMA_STORE_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
/*	0x01FA	*/	VDMA_LOAD_RA(AD_4_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x01FB	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x01FC	*/	VDMA_STORE_RA(AD_4_DEST_XFER_CNT),											// Save the count
/*	0x01FD	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_5),										// Branch if we are not done with this frame
// LP1_AD_4_SP1_TS_WAIT:
/*	0x01FE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x01FF	*/	VDMA_LOAD_SYS_ADDR(0, AD_4_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x0200	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_4_SP2_TS_WAIT:
/*	0x0201	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP2_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0202	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 2 time stamp address
/*	0x0203	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP2_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_4_SP3_TS_WAIT:
/*	0x0204	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP3_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0205	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 3 time stamp address
/*	0x0206	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP3_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
// LP1_AD_4_SP4_TS_WAIT:
/*	0x0207	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP4_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0208	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),													// Load the audio channel 4 time stamp address
/*	0x0209	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP4_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x020A	*/	VDMA_LOAD_RA(AD_4_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x020B	*/	VDMA_STORE_RA(AD_4_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x020C	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x020D	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_7),
/*	0x020E	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x020F	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x0210	*/	VDMA_STORE_RA(AD_4_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x0211	*/	VDMA_LOAD_RB(AD_4_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x0212	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x0213	*/	VDMA_STORE_RB(AD_4_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x0214	*/	VDMA_LOAD_RB(AD_4_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x0215	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x0216	*/	VDMA_STORE_RB(AD_4_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x0217	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_4_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x0218	*/	VDMA_LOAD_RB(AD_4_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x0219	*/	VDMA_STORE_RB(AD_4_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x021A	*/	VDMA_LOAD_RB(AD_4_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x021B	*/	VDMA_STORE_RB(AD_4_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x021C	*/	VDMA_LOAD_RA(AD_4_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x021D	*/	VDMA_STORE_RA(AD_4_FRAME_XFER_CNT),											// Save it 
// LP1_AD_4_LOAD_ADDR
/*	0x021E	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x021F	*/	VDMA_STORE_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x0220	*/	VDMA_LOAD_RB(AD_4_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x0221	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x0222	*/	VDMA_STORE_SYS_ADDR(0, AD_4_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x0223	*/	VDMA_LOAD_RA(AD_4_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x0224	*/	VDMA_STORE_RA(AD_4_DEST_XFER_CNT),											// Reset XFERS per DMA frame

// LP1_AD_5
/*	0x0225	*/	VDMA_LOAD_RA(AD_5_STREAM_DISABLED),											// Is this channel enabled?
/*	0x0226	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_6),										// Yes.  See if it is ready then then
// LP1_AD_5_READY
/*	0x0227	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_8, LP1_AD_6),						// Check next FIFO if this one is not ready
// LP1_AD_5_SP1_WAIT_FULL
/*	0x0228	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_5_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0229	*/	VDMA_LOAD_SYS_ADDR(0, AD_5_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x022A	*/	VDMA_LOAD_XFER_CTL(0, AD_5_SP1_XFER_CTRL),								// Transfer this portion of the frame
/*	0x022B	*/	VDMA_ADD_SYS_ADDR_I(AD_5_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x022C	*/	VDMA_STORE_SYS_ADDR(0, AD_5_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
// Decrement the remining XFER count and set next destination address
/*	0x022D	*/	VDMA_LOAD_RA(AD_5_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x022E	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x022F	*/	VDMA_STORE_RA(AD_5_DEST_XFER_CNT),											// Save the count
/*	0x0230	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_6),										// Branch if we are done with this frame
// LP1_AD_5_SP1_TS_WAIT:
/*	0x0231	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_5_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0232	*/	VDMA_LOAD_SYS_ADDR(0, AD_5_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x0233	*/	VDMA_LOAD_XFER_CTL(0, AD_5_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x0234	*/	VDMA_LOAD_RA(AD_5_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x0235	*/	VDMA_STORE_RA(AD_5_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x0236	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0237	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_11),
/*	0x0238	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0239	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x023A	*/	VDMA_STORE_RA(AD_5_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x023B	*/	VDMA_LOAD_RB(AD_5_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x023C	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x023D	*/	VDMA_STORE_RB(AD_5_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x023E	*/	VDMA_LOAD_RB(AD_5_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x023F	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x0240	*/	VDMA_STORE_RB(AD_5_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x0241	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_5_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x0242	*/	VDMA_LOAD_RB(AD_5_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x0243	*/	VDMA_STORE_RB(AD_5_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x0244	*/	VDMA_LOAD_RB(AD_5_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x0245	*/	VDMA_STORE_RB(AD_5_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x0246	*/	VDMA_LOAD_RA(AD_5_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x0247	*/	VDMA_STORE_RA(AD_5_FRAME_XFER_CNT),											// Save it 
// LP1_AD_5_LOAD_ADDR
/*	0x0248	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x0249	*/	VDMA_STORE_SYS_ADDR(0, AD_5_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x024A	*/	VDMA_LOAD_RB(AD_5_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x024B	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x024C	*/	VDMA_STORE_SYS_ADDR(0, AD_5_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x024D	*/	VDMA_LOAD_RA(AD_5_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x024E	*/	VDMA_STORE_RA(AD_5_DEST_XFER_CNT),											// Reset XFERS per DMA frame
/*	0x024F	*/	VDMA_NOP(),
/*	0x0250	*/	VDMA_NOP(),
/*	0x0251	*/	VDMA_NOP(),

// LP1_AD_6
/*	0x0252	*/	VDMA_LOAD_RA(AD_6_STREAM_DISABLED),											// Is this channel enabled?
/*	0x0253	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_7),										// Yes.  See if it is ready then then
// LP1_AD_6_READY
/*	0x0254	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_9, LP1_AD_7),						// Check next FIFO if this one is not ready
// LP1_AD_6_SP1_WAIT_FULL
/*	0x0255	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_6_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0256	*/	VDMA_LOAD_SYS_ADDR(0, AD_6_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x0257	*/	VDMA_LOAD_XFER_CTL(0, AD_6_SP1_XFER_CTRL),								// Transfer this portion of the frame
/*	0x0258	*/	VDMA_ADD_SYS_ADDR_I(AD_6_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x0259	*/	VDMA_STORE_SYS_ADDR(0, AD_6_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
// DecreEent the remining XFER count and set next destination address
/*	0x025A	*/	VDMA_LOAD_RA(AD_6_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x025B	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x025C	*/	VDMA_STORE_RA(AD_6_DEST_XFER_CNT),											// Save the count
/*	0x025D	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_7),										// Branch if we are done with this frame
// LP1_AD_6_SP1_TS_WAIT:
/*	0x025E	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_6_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x025F	*/	VDMA_LOAD_SYS_ADDR(0, AD_6_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x0260	*/	VDMA_LOAD_XFER_CTL(0, AD_6_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x0261	*/	VDMA_LOAD_RA(AD_6_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x0262	*/	VDMA_STORE_RA(AD_6_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x0263	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0264	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_12),
/*	0x0265	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0266	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x0267	*/	VDMA_STORE_RA(AD_6_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x0268	*/	VDMA_LOAD_RB(AD_6_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x0269	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x026A	*/	VDMA_STORE_RB(AD_6_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x026B	*/	VDMA_LOAD_RB(AD_6_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x026C	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x026D	*/	VDMA_STORE_RB(AD_6_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x026E	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_6_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x026F	*/	VDMA_LOAD_RB(AD_6_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x0270	*/	VDMA_STORE_RB(AD_6_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x0271	*/	VDMA_LOAD_RB(AD_6_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x0272	*/	VDMA_STORE_RB(AD_6_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x0273	*/	VDMA_LOAD_RA(AD_6_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x0274	*/	VDMA_STORE_RA(AD_6_FRAME_XFER_CNT),											// Save it 
// LP1_AD_6_LOAD_ADDR
/*	0x0275	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x0276	*/	VDMA_STORE_SYS_ADDR(0, AD_6_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x0277	*/	VDMA_LOAD_RB(AD_6_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x0278	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x0279	*/	VDMA_STORE_SYS_ADDR(0, AD_6_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x027A	*/	VDMA_LOAD_RA(AD_6_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x027B	*/	VDMA_STORE_RA(AD_6_DEST_XFER_CNT),											// Reset XFERS per DMA frame

// LP1_AD_7
/*	0x027C	*/	VDMA_LOAD_RA(AD_7_STREAM_DISABLED),											// Is this channel enabled?
/*	0x027D	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_8),										// Yes.  See if it is ready then then
// LP1_AD_7_READY
/*	0x027E	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_10, LP1_AD_8),						// Check next FIFO if this one is not ready
// LP1_AD_7_SP1_WAIT_FULL
/*	0x027F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_7_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0280	*/	VDMA_LOAD_SYS_ADDR(0, AD_7_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x0281	*/	VDMA_LOAD_XFER_CTL(0, AD_7_SP1_XFER_CTRL),								// Transfer this portion of the frame
/*	0x0282	*/	VDMA_ADD_SYS_ADDR_I(AD_7_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x0283	*/	VDMA_STORE_SYS_ADDR(0, AD_7_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
// Decrement the remining XFER count and set next destination address
/*	0x0284	*/	VDMA_LOAD_RA(AD_7_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x0285	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x0286	*/	VDMA_STORE_RA(AD_7_DEST_XFER_CNT),											// Save the count
/*	0x0287	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_8),										// Branch if we are done with this frame
// LP1_AD_7_SP1_TS_WAIT:
/*	0x0288	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_7_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x0289	*/	VDMA_LOAD_SYS_ADDR(0, AD_7_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x028A	*/	VDMA_LOAD_XFER_CTL(0, AD_7_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x028B	*/	VDMA_LOAD_RA(AD_7_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x028C	*/	VDMA_STORE_RA(AD_7_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x028D	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x028E	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_13),
/*	0x028F	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x0290	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x0291	*/	VDMA_STORE_RA(AD_7_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x0292	*/	VDMA_LOAD_RB(AD_7_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x0293	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x0294	*/	VDMA_STORE_RB(AD_7_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x0295	*/	VDMA_LOAD_RB(AD_7_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x0296	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x0297	*/	VDMA_STORE_RB(AD_7_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x0298	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_7_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around	to the start
/*	0x0299	*/	VDMA_LOAD_RB(AD_7_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x029A	*/	VDMA_STORE_RB(AD_7_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x029B	*/	VDMA_LOAD_RB(AD_7_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x029C	*/	VDMA_STORE_RB(AD_7_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x029D	*/	VDMA_LOAD_RA(AD_7_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x029E	*/	VDMA_STORE_RA(AD_7_FRAME_XFER_CNT),											// Save it 
// LP1_AD_7_LOAD_ADDR
/*	0x029F	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x02A0	*/	VDMA_STORE_SYS_ADDR(0, AD_7_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x02A1	*/	VDMA_LOAD_RB(AD_7_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x02A2	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x02A3	*/	VDMA_STORE_SYS_ADDR(0, AD_7_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x02A4	*/	VDMA_LOAD_RA(AD_7_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x02A5	*/	VDMA_STORE_RA(AD_7_DEST_XFER_CNT),											// Reset XFERS per DMA frame

// LP1_AD_8
/*	0x02A6	*/	VDMA_LOAD_RA(AD_8_STREAM_DISABLED),											// Is this channel enabled?
/*	0x02A7	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, FINISH),											// Yes.  See if it is ready then then
// LP1_AD_8_READY
/*	0x02A8	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_11, FINISH),						// Check next FIFO if this one is not ready
// LP1_AD_8_SP1_WAIT_FULL
/*	0x02A9	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_8_SP1_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x02AA	*/	VDMA_LOAD_SYS_ADDR(0, AD_8_SP1_DEST_ADDR_L),								// Load the destination address of stereo pair 1
/*	0x02AB	*/	VDMA_LOAD_XFER_CTL(0, AD_8_SP1_XFER_CTRL),								// Transfer this portion of the frame
/*	0x02AC	*/	VDMA_ADD_SYS_ADDR_I(AD_8_SP1_XFER_STEP_SIZE),							// Bump the stereo pair destination address
/*	0x02AD	*/	VDMA_STORE_SYS_ADDR(0, AD_8_SP1_DEST_ADDR_L),							// Save the new stereo pair destination address
// Decrement the remining XFER count and set next destination address
/*	0x02AE	*/	VDMA_LOAD_RA(AD_8_DEST_XFER_CNT),											// Get the remaining XFER count
/*	0x02AF	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the remaining XFER count
/*	0x02B0	*/	VDMA_STORE_RA(AD_8_DEST_XFER_CNT),											// Save the count
/*	0x02B1	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, FINISH),											// Branch if we are done with this frame
// LP1_AD_8_SP1_TS_WAIT:
/*	0x02B2	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_8_SP1_TS_WAIT),		// Do not issue xfer command when LDM queue is full
/*	0x02B3	*/	VDMA_LOAD_SYS_ADDR(0, AD_8_SP1_TS_DEST_ADDR_L),							// Load the audio channel 1 time stamp address
/*	0x02B4	*/	VDMA_LOAD_XFER_CTL(0, AD_8_SP1_TS_XFER_CTRL),							// Transfer the time stamp portion of the frame
/*	0x02B5	*/	VDMA_LOAD_RA(AD_8_FRAME_XFER_CNT),											// Load the old last frame xfer'd index
/*	0x02B6	*/	VDMA_STORE_RA(AD_8_LAST_FRAME_INDEX),										// Save the new last frame xfer'd index
/*	0x02B7	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x02B8	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_14),
/*	0x02B9	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
// Setup to start DMA'ing into the next video frame
/*	0x02BA	*/	VDMA_ADD_RA(MINUS_1),															// Decrement the count
/*	0x02BB	*/	VDMA_STORE_RA(AD_8_FRAME_XFER_CNT),											// Save the new DMA frames to xfer
/*	0x02BC	*/	VDMA_LOAD_RB(AD_8_DEST_TS_ADDR_PTR),										// Load the last time stamp desination address pointer
/*	0x02BD	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x02BE	*/	VDMA_STORE_RB(AD_8_DEST_TS_ADDR_PTR),										// Save the new time stamp desination pointer
/*	0x02BF	*/	VDMA_LOAD_RB(AD_8_DEST_ADDR_PTR),											// Load the last desination address pointer
/*	0x02C0	*/	VDMA_ADD_RB(MINUS_2),															// Decrement the count
/*	0x02C1	*/	VDMA_STORE_RB(AD_8_DEST_ADDR_PTR),											// Save the new desination pointer
/*	0x02C2	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_8_LOAD_ADDR),							// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x02C3	*/	VDMA_LOAD_RB(AD_8_LAST_TS_ADDR_PTR),										// Load last DMA time stamp pointer for this FIFO
/*	0x02C4	*/	VDMA_STORE_RB(AD_8_DEST_TS_ADDR_PTR),										// Save it as the current DMA time stamp pointer
/*	0x02C5	*/	VDMA_LOAD_RB(AD_8_LAST_ADDR_PTR),											// Load last DMA frame pointer for this FIFO
/*	0x02C6	*/	VDMA_STORE_RB(AD_8_DEST_ADDR_PTR),											// Save it as the current DMA frame pointer
/*	0x02C7	*/	VDMA_LOAD_RA(AD_8_FRAME_CNT),													// Load DMA frames to xfer count
/*	0x02C8	*/	VDMA_STORE_RA(AD_8_FRAME_XFER_CNT),											// Save it 
// LP1_AD_8_LOAD_ADDR
/*	0x02C9	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for stereo pair 1
/*	0x02CA	*/	VDMA_STORE_SYS_ADDR(0, AD_8_SP1_DEST_ADDR_L),							// Save the new destination address for stereo pair 1
/*	0x02CB	*/	VDMA_LOAD_RB(AD_8_DEST_TS_ADDR_PTR),										// Load the last desination address pointer
/*	0x02CC	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),												// Load the new destination address for the stereo pair 1 time stamp
/*	0x02CD	*/	VDMA_STORE_SYS_ADDR(0, AD_8_SP1_TS_DEST_ADDR_L),						// Save the new destination address for the stereo pair 1 time stamp
/*	0x02CE	*/	VDMA_LOAD_RA(AD_8_XFER_CNT),													// Load initial XFERS per DMA frame
/*	0x02CF	*/	VDMA_STORE_RA(AD_8_DEST_XFER_CNT),											// Reset XFERS per DMA frame
// FINISH
/*	0x02D0	*/	VDMA_LOAD_RA(SIGNAL_EVENT),
/*	0x02D1	*/	VDMA_JMP(VDMA_RA_EQZ, 0, MAIN),
// LDM_IDLE_WAIT
/*	0x02D2	*/	VDMA_JMP(VDMA_EXT_COND_LO, LDM_IDLE, LDM_IDLE_WAIT),
/* 0x02D3	*/	VDMA_NOP(),
// SIGNAL_EVENT
/* 0x02D4	*/	VDMA_SIG_EVENT(0, 1, 0),
/* 0x02D5	*/	VDMA_NOP(),
/*	0x02D6	*/	VDMA_LOAD_SYS_ADDR(0, SIG_EVENT_CODE),
/*	0x02D7	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/* 0x02D8	*/	VDMA_NOP(),
/* 0x02D9	*/	VDMA_NOP(),
/*	0x02DA	*/	VDMA_JMP(VDMA_ALWAYS, 0, MAIN),
};

#else
static u32 vc8x0_FlexDMAProgram[] = 
{
//START:
//MAIN:
/*	0x0000	*/	VDMA_LOAD_RA(VD_1_STREAM_DISABLED),				// Is this channel enabled?
/*	0x0001	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_2),				// No.  Check next FIFO then
// LP1_VD_1_READY
/*	0x0002	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_0, LP1_VD_2),		// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x0003	*/	VDMA_LOAD_RA(VD_1_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0004	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_1_DONE_FULL),			// Branch if we are done with this frame
/*	0x0005	*/	VDMA_ADD_RA(MINUS_1),		// Decrement the remaining XFER count
/*	0x0006	*/	VDMA_STORE_RA(VD_1_DEST_XFER_CNT),				// Save the count
// LP1_VD_1_Y_WAIT_FULL
/*	0x0007	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0008	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x0009	*/	VDMA_LOAD_XFER_CTL(0, VD_1_Y_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x000A	*/	VDMA_ADD_SYS_ADDR_I(VD_1_Y_XFER_SIZE_FULL),			// Bump the Y plane destination address
/*	0x000B	*/	VDMA_STORE_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),			// Save the new Y plane destination address
// LP1_VD_1_U_WAIT_FULL
/*	0x000C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x000D	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x000E	*/	VDMA_LOAD_XFER_CTL(0, VD_1_U_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x000F	*/	VDMA_ADD_SYS_ADDR_I(VD_1_UV_XFER_SIZE_FULL),			// Bump the U plane destination address
/*	0x0010	*/	VDMA_STORE_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),			// Save the new destination address
// LP1_VD_1_V_WAIT_FULL
/*	0x0011	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0012	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x0013	*/	VDMA_LOAD_XFER_CTL(0, VD_1_V_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x0014	*/	VDMA_ADD_SYS_ADDR_I(VD_1_UV_XFER_SIZE_FULL),			// Bump the V Plane destination address
/*	0x0015	*/	VDMA_STORE_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),			// Save the new V Plane destination address
/*	0x0016	*/	VDMA_LOAD_RA(VD_1_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0017	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_2),				// Check next FIFO channel
// LP1_VD_1_DONE_FULL:	
/*	0x0018	*/	VDMA_LOAD_RA(VD_1_Y_XFER_SIZE_PART),				// Load partial transfer byte count
/*	0x0019	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_1_Y_TS_WAIT),			// Branch if we are done with this frame
// LP1_VD_1_Y_WAIT_PART
/*	0x001A	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x001B	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x001C	*/	VDMA_LOAD_XFER_CTL(0, VD_1_Y_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_1_U_WAIT_PART
/*	0x001D	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x001E	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x001F	*/	VDMA_LOAD_XFER_CTL(0, VD_1_U_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_1_V_WAIT_PART
/*	0x0020	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0021	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x0022	*/	VDMA_LOAD_XFER_CTL(0, VD_1_V_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_1_Y_TS_WAIT:
/*	0x0023	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0024	*/	VDMA_LOAD_SYS_ADDR(0, VD_1_Y_TS_DEST_ADDR_L),			// Load the V Plane time stamp address
/*	0x0025	*/	VDMA_LOAD_XFER_CTL(0, VD_1_Y_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_1_U_TS_WAIT:
/*	0x0026	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0027	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),// Load the U Plane time stamp address
/*	0x0028	*/	VDMA_LOAD_XFER_CTL(0, VD_1_U_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_1_V_TS_WAIT:
/*	0x0029	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_1_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x002A	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),// Load the U Plane time stamp address
/*	0x002B	*/	VDMA_LOAD_XFER_CTL(0, VD_1_V_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x002C	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x002D	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_0),
/*	0x002E	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x002F	*/	VDMA_LOAD_RA(VD_1_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x0030	*/	VDMA_STORE_RA(VD_1_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0031	*/	VDMA_LOAD_RA(VD_1_FRAME_XFER_CNT),				// Get the count of DMA frames remaining to xfer before wrap
/*	0x0032	*/	VDMA_ADD_RA(MINUS_1),		// Decrement the count
/*	0x0033	*/	VDMA_STORE_RA(VD_1_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0034	*/	VDMA_LOAD_RB(VD_1_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0035	*/	VDMA_ADD_RB(MINUS_2),		// Decrement the count
/*	0x0036	*/	VDMA_STORE_RB(VD_1_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x0037	*/	VDMA_LOAD_RB(VD_1_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x0038	*/	VDMA_ADD_RB(MINUS_6),		// Decrement the count
/*	0x0039	*/	VDMA_STORE_RB(VD_1_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x003A	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_1_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x003B	*/	VDMA_LOAD_RB(VD_1_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x003C	*/	VDMA_STORE_RB(VD_1_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x003D	*/	VDMA_LOAD_RB(VD_1_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x003E	*/	VDMA_STORE_RB(VD_1_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x003F	*/	VDMA_LOAD_RA(VD_1_FRAME_CNT),// Load DMA frames to xfer count
/*	0x0040	*/	VDMA_STORE_RA(VD_1_FRAME_XFER_CNT),				// Save it 
// LP1_VD_1_LOAD_ADDR
/*	0x0041	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane
/*	0x0042	*/	VDMA_STORE_SYS_ADDR(0, VD_1_Y_DEST_ADDR_L),			// Save the new destination address for the Y plane
/*	0x0043	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),					// Load the new destination address for the U plane
/*	0x0044	*/	VDMA_STORE_SYS_ADDR(0, VD_1_U_DEST_ADDR_L),			// Save the new destination address for the U plane
/*	0x0045	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),					// Load the new destination address for the V plane
/*	0x0046	*/	VDMA_STORE_SYS_ADDR(0, VD_1_V_DEST_ADDR_L),			// Save the new destination address for the V plane
/*	0x0047	*/	VDMA_LOAD_RB(VD_1_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x0048	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane time stamp
/*	0x0049	*/	VDMA_STORE_SYS_ADDR(0, VD_1_Y_TS_DEST_ADDR_L),			// Save the new destination address for the Y plane time stamp
/*	0x004A	*/	VDMA_LOAD_RA(VD_1_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x004B	*/	VDMA_STORE_RA(VD_1_DEST_XFER_CNT),				// Reset XFERS per DMA frame

// LP1_VD_2:
/*	0x004C	*/	VDMA_LOAD_RA(VD_2_STREAM_DISABLED),				// Is this channel enabled?
/*	0x004D	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_3),				// No.  Check next FIFO then
// LP1_VD_2_READY
/*	0x004E	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_2, LP1_VD_3),		// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x004F	*/	VDMA_LOAD_RA(VD_2_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0050	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_2_DONE_FULL),			// Branch if we are done with this frame
/*	0x0051	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x0052	*/	VDMA_STORE_RA(VD_2_DEST_XFER_CNT),				// Save the count
// LP1_VD_2_Y_WAIT_FULL
/*	0x0053	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0054	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x0055	*/	VDMA_LOAD_XFER_CTL(0, VD_2_Y_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x0056	*/	VDMA_ADD_SYS_ADDR_I(VD_2_Y_XFER_SIZE_FULL),			// Bump the Y plane destination address
/*	0x0057	*/	VDMA_STORE_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),			// Save the new Y plane destination address
// LP1_VD_2_U_WAIT_FULL
/*	0x0058	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x0059	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x005A	*/	VDMA_LOAD_XFER_CTL(0, VD_2_U_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x005B	*/	VDMA_ADD_SYS_ADDR_I(VD_2_UV_XFER_SIZE_FULL),			// Bump the U plane destination address
/*	0x005C	*/	VDMA_STORE_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),			// Save the new destination address
// LP1_VD_2_V_WAIT_FULL
/*	0x005D	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x005E	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x005F	*/	VDMA_LOAD_XFER_CTL(0, VD_2_V_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x0060	*/	VDMA_ADD_SYS_ADDR_I(VD_2_UV_XFER_SIZE_FULL),			// Bump the V Plane destination address
/*	0x0061	*/	VDMA_STORE_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),			// Save the new V Plane destination address
/*	0x0062	*/	VDMA_LOAD_RA(VD_2_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0063	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_3),				// Check next FIFO channel
// LP1_VD_2_DONE_FULL:	
/*	0x0064	*/	VDMA_LOAD_RA(VD_2_Y_XFER_SIZE_PART),				// Load partial transfer byte count
/*	0x0065	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_2_Y_TS_WAIT),			// Branch if we are done with this frame
// LP1_VD_2_Y_WAIT_PART
/*	0x0066	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0067	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x0068	*/	VDMA_LOAD_XFER_CTL(0, VD_2_Y_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_2_U_WAIT_PART
/*	0x0069	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x006A	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x006B	*/	VDMA_LOAD_XFER_CTL(0, VD_2_U_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_2_V_WAIT_PART
/*	0x006C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x006D	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x006E	*/	VDMA_LOAD_XFER_CTL(0, VD_2_V_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_2_Y_TS_WAIT:
/*	0x006F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0070	*/	VDMA_LOAD_SYS_ADDR(0, VD_2_Y_TS_DEST_ADDR_L),			// Load the V Plane time stamp address
/*	0x0071	*/	VDMA_LOAD_XFER_CTL(0, VD_2_Y_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_2_U_TS_WAIT:
/*	0x0072	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0073	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),					// Load the U Plane time stamp address
/*	0x0074	*/	VDMA_LOAD_XFER_CTL(0, VD_2_U_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_2_V_TS_WAIT:
/*	0x0075	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_2_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0076	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the U Plane time stamp address
/*	0x0077	*/	VDMA_LOAD_XFER_CTL(0, VD_2_V_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x0078	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0079	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_2),
/*	0x007A	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x007B	*/	VDMA_LOAD_RA(VD_2_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x007C	*/	VDMA_STORE_RA(VD_2_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x007D	*/	VDMA_LOAD_RA(VD_2_FRAME_XFER_CNT),				// Get the count of DMA frames remaining to xfer before wrap
/*	0x007E	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x007F	*/	VDMA_STORE_RA(VD_2_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0080	*/	VDMA_LOAD_RB(VD_2_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0081	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0082	*/	VDMA_STORE_RB(VD_2_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x0083	*/	VDMA_LOAD_RB(VD_2_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x0084	*/	VDMA_ADD_RB(MINUS_6),						// Decrement the count
/*	0x0085	*/	VDMA_STORE_RB(VD_2_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x0086	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_2_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x0087	*/	VDMA_LOAD_RB(VD_2_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x0088	*/	VDMA_STORE_RB(VD_2_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x0089	*/	VDMA_LOAD_RB(VD_2_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x008A	*/	VDMA_STORE_RB(VD_2_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x008B	*/	VDMA_LOAD_RA(VD_2_FRAME_CNT),// Load DMA frames to xfer count
/*	0x008C	*/	VDMA_STORE_RA(VD_2_FRAME_XFER_CNT),				// Save it 
// LP1_VD_2_LOAD_ADDR
/*	0x008D	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane
/*	0x008E	*/	VDMA_STORE_SYS_ADDR(0, VD_2_Y_DEST_ADDR_L),			// Save the new destination address for the Y plane
/*	0x008F	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),					// Load the new destination address for the U plane
/*	0x0090	*/	VDMA_STORE_SYS_ADDR(0, VD_2_U_DEST_ADDR_L),			// Save the new destination address for the U plane
/*	0x0091	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),					// Load the new destination address for the V plane
/*	0x0092	*/	VDMA_STORE_SYS_ADDR(0, VD_2_V_DEST_ADDR_L),			// Save the new destination address for the V plane
/*	0x0093	*/	VDMA_LOAD_RB(VD_2_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x0094	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane time stamp
/*	0x0095	*/	VDMA_STORE_SYS_ADDR(0, VD_2_Y_TS_DEST_ADDR_L),			// Save the new destination address for the Y plane time stamp
/*	0x0096	*/	VDMA_LOAD_RA(VD_2_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x0097	*/	VDMA_STORE_RA(VD_2_DEST_XFER_CNT),				// Reset XFERS per DMA frame

//	LP1_VD_3
/*	0x0098	*/	VDMA_LOAD_RA(VD_3_STREAM_DISABLED),				// Is this channel enabled?
/*	0x0099	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_4),				// No.  Check next FIFO then
// LP1_VD_3_READY
/*	0x009A	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_4, LP1_VD_4),		// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x009B	*/	VDMA_LOAD_RA(VD_3_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x009C	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_3_DONE_FULL),			// Branch if we are done with this frame
/*	0x009D	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x009E	*/	VDMA_STORE_RA(VD_3_DEST_XFER_CNT),				// Save the count
// LP1_VD_3_Y_WAIT_FULL
/*	0x009F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00A0	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x00A1	*/	VDMA_LOAD_XFER_CTL(0, VD_3_Y_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x00A2	*/	VDMA_ADD_SYS_ADDR_I(VD_3_Y_XFER_SIZE_FULL),			// Bump the Y plane destination address
/*	0x00A3	*/	VDMA_STORE_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),			// Save the new Y plane destination address
// LP1_VD_3_U_WAIT_FULL
/*	0x00A4	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00A5	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x00A6	*/	VDMA_LOAD_XFER_CTL(0, VD_3_U_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x00A7	*/	VDMA_ADD_SYS_ADDR_I(VD_3_UV_XFER_SIZE_FULL),			// Bump the U plane destination address
/*	0x00A8	*/	VDMA_STORE_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),			// Save the new destination address
// LP1_VD_3_V_WAIT_FULL
/*	0x00A9	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00AA	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x00AB	*/	VDMA_LOAD_XFER_CTL(0, VD_3_V_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x00AC	*/	VDMA_ADD_SYS_ADDR_I(VD_3_UV_XFER_SIZE_FULL),			// Bump the V Plane destination address
/*	0x00AD	*/	VDMA_STORE_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),			// Save the new V Plane destination address
/*	0x00AE	*/	VDMA_LOAD_RA(VD_3_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x00AF	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_4),				// Check next FIFO channel
// LP1_VD_3_DONE_FULL:	
/*	0x00B0	*/	VDMA_LOAD_RA(VD_3_Y_XFER_SIZE_PART),				// Load partial transfer byte count
/*	0x00B1	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_3_Y_TS_WAIT),			// Branch if we are done with this frame
// LP1_VD_3_Y_WAIT_PART
/*	0x00B2	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00B3	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x00B4	*/	VDMA_LOAD_XFER_CTL(0, VD_3_Y_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_3_U_WAIT_PART
/*	0x00B5	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00B6	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x00B7	*/	VDMA_LOAD_XFER_CTL(0, VD_3_U_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_3_V_WAIT_PART
/*	0x00B8	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00B9	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x00BA	*/	VDMA_LOAD_XFER_CTL(0, VD_3_V_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_3_Y_TS_WAIT:
/*	0x00BB	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x00BC	*/	VDMA_LOAD_SYS_ADDR(0, VD_3_Y_TS_DEST_ADDR_L),			// Load the V Plane time stamp address
/*	0x00BD	*/	VDMA_LOAD_XFER_CTL(0, VD_3_Y_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_3_U_TS_WAIT:
/*	0x00BE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x00BF	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),					// Load the U Plane time stamp address
/*	0x00C0	*/	VDMA_LOAD_XFER_CTL(0, VD_3_U_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_3_V_TS_WAIT:
/*	0x00C1	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_3_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x00C2	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the U Plane time stamp address
/*	0x00C3	*/	VDMA_LOAD_XFER_CTL(0, VD_3_V_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x00C4	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x00C5	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_4),
/*	0x00C6	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x00C7	*/	VDMA_LOAD_RA(VD_3_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x00C8	*/	VDMA_STORE_RA(VD_3_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x00C9	*/	VDMA_LOAD_RA(VD_3_FRAME_XFER_CNT),				// Get the count of DMA frames remaining to xfer before wrap
/*	0x00CA	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x00CB	*/	VDMA_STORE_RA(VD_3_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x00CC	*/	VDMA_LOAD_RB(VD_3_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x00CD	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x00CE	*/	VDMA_STORE_RB(VD_3_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x00CF	*/	VDMA_LOAD_RB(VD_3_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x00D0	*/	VDMA_ADD_RB(MINUS_6),						// Decrement the count
/*	0x00D1	*/	VDMA_STORE_RB(VD_3_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x00D2	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_3_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x00D3	*/	VDMA_LOAD_RB(VD_3_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x00D4	*/	VDMA_STORE_RB(VD_3_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x00D5	*/	VDMA_LOAD_RB(VD_3_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x00D6	*/	VDMA_STORE_RB(VD_3_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x00D7	*/	VDMA_LOAD_RA(VD_3_FRAME_CNT),// Load DMA frames to xfer count
/*	0x00D8	*/	VDMA_STORE_RA(VD_3_FRAME_XFER_CNT),				// Save it 
// LP1_VD_3_LOAD_ADDR
/*	0x00D9	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane
/*	0x00DA	*/	VDMA_STORE_SYS_ADDR(0, VD_3_Y_DEST_ADDR_L),			// Save the new destination address for the Y plane
/*	0x00DB	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),					// Load the new destination address for the U plane
/*	0x00DC	*/	VDMA_STORE_SYS_ADDR(0, VD_3_U_DEST_ADDR_L),			// Save the new destination address for the U plane
/*	0x00DD	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),					// Load the new destination address for the V plane
/*	0x00DE	*/	VDMA_STORE_SYS_ADDR(0, VD_3_V_DEST_ADDR_L),			// Save the new destination address for the V plane
/*	0x00DF	*/	VDMA_LOAD_RB(VD_3_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x00E0	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane time stamp
/*	0x00E1	*/	VDMA_STORE_SYS_ADDR(0, VD_3_Y_TS_DEST_ADDR_L),			// Save the new destination address for the Y plane time stamp
/*	0x00E2	*/	VDMA_LOAD_RA(VD_3_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x00E3	*/	VDMA_STORE_RA(VD_3_DEST_XFER_CNT),				// Reset XFERS per DMA frame

//	LP1_VD_4
/*	0x0098	*/	VDMA_LOAD_RA(VD_4_STREAM_DISABLED),				// Is this channel enabled?
/*	0x00E5	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_1),				// No.  Check next FIFO then
// LP1_VD_4_READY
/*	0x00E6	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_6, LP1_AD_1),		// Check next FIFO if this one is not ready
// Setup to do a DMA xfer from the first video fifo
/*	0x00E7	*/	VDMA_LOAD_RA(VD_4_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x00E8	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_4_DONE_FULL),			// Branch if we are done with this frame
/*	0x00E9	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x00EA	*/	VDMA_STORE_RA(VD_4_DEST_XFER_CNT),				// Save the count
// LP1_VD_4_Y_WAIT_FULL
/*	0x00EB	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_Y_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00EC	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x00ED	*/	VDMA_LOAD_XFER_CTL(0, VD_4_Y_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x00EE	*/	VDMA_ADD_SYS_ADDR_I(VD_4_Y_XFER_SIZE_FULL),			// Bump the Y plane destination address
/*	0x00EF	*/	VDMA_STORE_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),			// Save the new Y plane destination address
// LP1_VD_4_U_WAIT_FULL
/*	0x00F0	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_U_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00F1	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x00F2	*/	VDMA_LOAD_XFER_CTL(0, VD_4_U_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x00F3	*/	VDMA_ADD_SYS_ADDR_I(VD_4_UV_XFER_SIZE_FULL),			// Bump the U plane destination address
/*	0x00F4	*/	VDMA_STORE_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),			// Save the new destination address
// LP1_VD_4_V_WAIT_FULL
/*	0x00F5	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_V_WAIT_FULL),	// Do not issue xfer command when LDM queue is full
/*	0x00F6	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x00F7	*/	VDMA_LOAD_XFER_CTL(0, VD_4_V_XFER_CTRL_FULL),			// Transfer this portion of the frame
/*	0x00F8	*/	VDMA_ADD_SYS_ADDR_I(VD_4_UV_XFER_SIZE_FULL),			// Bump the V Plane destination address
/*	0x00F9	*/	VDMA_STORE_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),			// Save the new V Plane destination address
/*	0x00FA	*/	VDMA_LOAD_RA(VD_4_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x00FB	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_1),				// Check next FIFO channel
// LP1_VD_4_DONE_FULL:	
/*	0x00FC	*/	VDMA_LOAD_RA(VD_4_Y_XFER_SIZE_PART),				// Load partial transfer byte count
/*	0x00FD	*/	VDMA_JMP(VDMA_RA_EQZ, 0, LP1_VD_4_Y_TS_WAIT),			// Branch if we are done with this frame
// LP1_VD_4_Y_WAIT_PART
/*	0x00FE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_Y_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x00FF	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),			// Load the Y plane destination addresso
/*	0x0100	*/	VDMA_LOAD_XFER_CTL(0, VD_4_Y_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_4_U_WAIT_PART
/*	0x0101	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_U_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0102	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),			// Load the U plane destination address
/*	0x0103	*/	VDMA_LOAD_XFER_CTL(0, VD_4_U_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_4_V_WAIT_PART
/*	0x0104	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_V_WAIT_PART),	// Do not issue xfer command when LDM queue is full
/*	0x0105	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),			// Load the V Plane destination address
/*	0x0106	*/	VDMA_LOAD_XFER_CTL(0, VD_4_V_XFER_CTRL_PART),			// Transfer this portion of the frame
// LP1_VD_4_Y_TS_WAIT:
/*	0x0107	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_Y_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0108	*/	VDMA_LOAD_SYS_ADDR(0, VD_4_Y_TS_DEST_ADDR_L),			// Load the V Plane time stamp address
/*	0x0109	*/	VDMA_LOAD_XFER_CTL(0, VD_4_Y_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_4_U_TS_WAIT:
/*	0x010A	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_U_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x010B	*/	VDMA_ADD_SYS_ADDR_I(PLUS_16),					// Load the U Plane time stamp address
/*	0x010C	*/	VDMA_LOAD_XFER_CTL(0, VD_4_U_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_VD_4_V_TS_WAIT:
/*	0x010D	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_VD_4_V_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x010E	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the U Plane time stamp address
/*	0x010F	*/	VDMA_LOAD_XFER_CTL(0, VD_4_V_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x0110	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0111	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_6),
/*	0x0112	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0113	*/	VDMA_LOAD_RA(VD_4_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x0114	*/	VDMA_STORE_RA(VD_4_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0115	*/	VDMA_LOAD_RA(VD_4_FRAME_XFER_CNT),				// Get the count of DMA frames remaining to xfer before wrap
/*	0x0116	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x0117	*/	VDMA_STORE_RA(VD_4_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0118	*/	VDMA_LOAD_RB(VD_4_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0119	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x011A	*/	VDMA_STORE_RB(VD_4_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x011B	*/	VDMA_LOAD_RB(VD_4_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x011C	*/	VDMA_ADD_RB(MINUS_6),						// Decrement the count
/*	0x011D	*/	VDMA_STORE_RB(VD_4_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x011E	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_VD_4_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x011F	*/	VDMA_LOAD_RB(VD_4_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x0120	*/	VDMA_STORE_RB(VD_4_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x0121	*/	VDMA_LOAD_RB(VD_4_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x0122	*/	VDMA_STORE_RB(VD_4_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x0123	*/	VDMA_LOAD_RA(VD_4_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x0124	*/	VDMA_STORE_RA(VD_4_FRAME_XFER_CNT),				// Save it 
// LP1_VD_4_LOAD_ADDR
/*	0x0125	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane
/*	0x0126	*/	VDMA_STORE_SYS_ADDR(0, VD_4_Y_DEST_ADDR_L),			// Save the new destination address for the Y plane
/*	0x0127	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 2),					// Load the new destination address for the U plane
/*	0x0128	*/	VDMA_STORE_SYS_ADDR(0, VD_4_U_DEST_ADDR_L),			// Save the new destination address for the U plane
/*	0x0129	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 4),					// Load the new destination address for the V plane
/*	0x012A	*/	VDMA_STORE_SYS_ADDR(0, VD_4_V_DEST_ADDR_L),			// Save the new destination address for the V plane
/*	0x012B	*/	VDMA_LOAD_RB(VD_4_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x012C	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the Y plane time stamp
/*	0x012D	*/	VDMA_STORE_SYS_ADDR(0, VD_4_Y_TS_DEST_ADDR_L),			// Save the new destination address for the Y plane time stamp
/*	0x012E	*/	VDMA_LOAD_RA(VD_4_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x012F	*/	VDMA_STORE_RA(VD_4_DEST_XFER_CNT),				// Reset XFERS per DMA frame

//	LP1_AD_1
/*	0x0130	*/	VDMA_LOAD_RA(AD_1_STREAM_DISABLED),				// Is this channel enabled?
/*	0x0131	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_2),				// No.  See if it is ready then then
// LP1_AD_1_READY
/*	0x0132	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_1, LP1_AD_2),		// Check next FIFO if this one is not ready
// LP1_AD_1_SP1_WAIT_FULL
/*	0x0133	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0134	*/	VDMA_LOAD_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x0135	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP1_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_1_SP2_WAIT_FULL
/*	0x0136	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP2_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0137	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 2
/*	0x0138	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP2_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_1_SP3_WAIT_FULL
/*	0x0139	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP3_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x013A	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 3
/*	0x013B	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP3_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_1_SP4_WAIT_FULL
/*	0x013C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP4_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x013D	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 4
/*	0x013E	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP4_XFER_CTRL),			// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x013F	*/	VDMA_LOAD_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x0140	*/	VDMA_ADD_SYS_ADDR_I(AD_1_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x0141	*/	VDMA_STORE_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
/*	0x0142	*/	VDMA_LOAD_RA(AD_1_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0143	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x0144	*/	VDMA_STORE_RA(AD_1_DEST_XFER_CNT),				// Save the count
/*	0x0145	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_2),				// Branch if we are not done with this frame
// LP1_AD_1_SP1_TS_WAIT:
/*	0x0146	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0147	*/	VDMA_LOAD_SYS_ADDR(0, AD_1_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x0148	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_1_SP2_TS_WAIT:
/*	0x0149	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP2_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x014A	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 2 time stamp address
/*	0x014B	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP2_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_1_SP3_TS_WAIT:
/*	0x014C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP3_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x014D	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 3 time stamp address
/*	0x014E	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP3_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_1_SP4_TS_WAIT:
/*	0x014F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_1_SP4_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0150	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 4 time stamp address
/*	0x0151	*/	VDMA_LOAD_XFER_CTL(0, AD_1_SP4_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x0152	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0153	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_1),
/*	0x0154	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0155	*/	VDMA_LOAD_RA(AD_1_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x0156	*/	VDMA_STORE_RA(AD_1_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0157	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x0158	*/	VDMA_STORE_RA(AD_1_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0159	*/	VDMA_LOAD_RB(AD_1_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x015A	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x015B	*/	VDMA_STORE_RB(AD_1_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x015C	*/	VDMA_LOAD_RB(AD_1_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x015D	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x015E	*/	VDMA_STORE_RB(AD_1_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x015F	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_1_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x0160	*/	VDMA_LOAD_RB(AD_1_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x0161	*/	VDMA_STORE_RB(AD_1_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x0162	*/	VDMA_LOAD_RB(AD_1_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x0163	*/	VDMA_STORE_RB(AD_1_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x0164	*/	VDMA_LOAD_RA(AD_1_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x0165	*/	VDMA_STORE_RA(AD_1_FRAME_XFER_CNT),				// Save it 
// LP1_AD_1_LOAD_ADDR
/*	0x0166	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x0167	*/	VDMA_STORE_SYS_ADDR(0, AD_1_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x0168	*/	VDMA_LOAD_RB(AD_1_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x0169	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x016A	*/	VDMA_STORE_SYS_ADDR(0, AD_1_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x016B	*/	VDMA_LOAD_RA(AD_1_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x016C	*/	VDMA_STORE_RA(AD_1_DEST_XFER_CNT),				// Reset XFERS per DMA frame

//	LP1_AD_2
/*	0x016D	*/	VDMA_LOAD_RA(AD_2_STREAM_DISABLED),				// Is this channel enabled?
/*	0x016E	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_3),				// Yes.  See if it is ready then then
// LP1_AD_2_READY
/*	0x016F	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_3, LP1_AD_3),		// Check next FIFO if this one is not ready
// LP1_AD_2_SP1_WAIT_FULL
/*	0x0170	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0171	*/	VDMA_LOAD_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x0172	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP1_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_2_SP2_WAIT_FULL
/*	0x0173	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP2_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0174	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 2
/*	0x0175	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP2_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_2_SP3_WAIT_FULL
/*	0x0176	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP3_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0177	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 3
/*	0x0178	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP3_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_2_SP4_WAIT_FULL
/*	0x0179	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP4_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x017A	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 4
/*	0x017B	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP4_XFER_CTRL),			// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x017C	*/	VDMA_LOAD_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x017D	*/	VDMA_ADD_SYS_ADDR_I(AD_2_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x017E	*/	VDMA_STORE_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
/*	0x017F	*/	VDMA_LOAD_RA(AD_2_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0180	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x0181	*/	VDMA_STORE_RA(AD_2_DEST_XFER_CNT),				// Save the count
/*	0x0182	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_3),				// Branch if we are not done with this frame
// LP1_AD_2_SP1_TS_WAIT:
/*	0x0183	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0184	*/	VDMA_LOAD_SYS_ADDR(0, AD_2_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x0185	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_2_SP2_TS_WAIT:
/*	0x0186	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP2_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0187	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 2 time stamp address
/*	0x0188	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP2_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_2_SP3_TS_WAIT:
/*	0x0189	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP3_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x018A	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 3 time stamp address
/*	0x018B	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP3_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_2_SP4_TS_WAIT:
/*	0x018C	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_2_SP4_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x018D	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 4 time stamp address
/*	0x018E	*/	VDMA_LOAD_XFER_CTL(0, AD_2_SP4_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x018F	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0190	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_3),
/*	0x0191	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0192	*/	VDMA_LOAD_RA(AD_2_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x0193	*/	VDMA_STORE_RA(AD_2_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0194	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x0195	*/	VDMA_STORE_RA(AD_2_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0196	*/	VDMA_LOAD_RB(AD_2_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0197	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0198	*/	VDMA_STORE_RB(AD_2_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x0199	*/	VDMA_LOAD_RB(AD_2_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x019A	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x019B	*/	VDMA_STORE_RB(AD_2_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x019C	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_2_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x019D	*/	VDMA_LOAD_RB(AD_2_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x019E	*/	VDMA_STORE_RB(AD_2_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x019F	*/	VDMA_LOAD_RB(AD_2_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x01A0	*/	VDMA_STORE_RB(AD_2_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x01A1	*/	VDMA_LOAD_RA(AD_2_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x01A2	*/	VDMA_STORE_RA(AD_2_FRAME_XFER_CNT),				// Save it 
// LP1_AD_2_LOAD_ADDR
/*	0x01A3	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x01A4	*/	VDMA_STORE_SYS_ADDR(0, AD_2_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x01A5	*/	VDMA_LOAD_RB(AD_2_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x01A6	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x01A7	*/	VDMA_STORE_SYS_ADDR(0, AD_2_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x01A8	*/	VDMA_LOAD_RA(AD_2_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x01A9	*/	VDMA_STORE_RA(AD_2_DEST_XFER_CNT),				// Reset XFERS per DMA frame

// LP1_AD_3
/*	0x01AA	*/	VDMA_LOAD_RA(AD_3_STREAM_DISABLED),				// Is this channel enabled?
/*	0x01AB	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_4),				// Yes.  See if it is ready then then
// LP1_AD_3_READY
/*	0x01AC	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_5, LP1_AD_4),		// Check next FIFO if this one is not ready
// LP1_AD_3_SP1_WAIT_FULL
/*	0x01AD	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01AE	*/	VDMA_LOAD_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x01AF	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP1_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_3_SP2_WAIT_FULL
/*	0x01B0	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP2_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01B1	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 2
/*	0x01B2	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP2_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_3_SP3_WAIT_FULL
/*	0x01B3	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP3_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01B4	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 3
/*	0x01B5	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP3_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_3_SP4_WAIT_FULL
/*	0x01B6	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP4_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01B7	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 4
/*	0x01B8	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP4_XFER_CTRL),			// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x01B9	*/	VDMA_LOAD_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x01BA	*/	VDMA_ADD_SYS_ADDR_I(AD_3_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x01BB	*/	VDMA_STORE_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
/*	0x01BC	*/	VDMA_LOAD_RA(AD_3_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x01BD	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x01BE	*/	VDMA_STORE_RA(AD_3_DEST_XFER_CNT),				// Save the count
/*	0x01BF	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_4),				// Branch if we are not done with this frame
// LP1_AD_3_SP1_TS_WAIT:
/*	0x01C0	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x01C1	*/	VDMA_LOAD_SYS_ADDR(0, AD_3_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x01C2	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_3_SP2_TS_WAIT:
/*	0x01C3	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP2_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x01C4	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 2 time stamp address
/*	0x01C5	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP2_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_3_SP3_TS_WAIT:
/*	0x01C6	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP3_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x01C7	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 3 time stamp address
/*	0x01C8	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP3_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_3_SP4_TS_WAIT:
/*	0x01C9	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_3_SP4_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x01CA	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 4 time stamp address
/*	0x01CB	*/	VDMA_LOAD_XFER_CTL(0, AD_3_SP4_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x01CC	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x01CD	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_5),
/*	0x01CE	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x01CF	*/	VDMA_LOAD_RA(AD_3_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x01D0	*/	VDMA_STORE_RA(AD_3_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x01D1	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x01D2	*/	VDMA_STORE_RA(AD_3_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x01D3	*/	VDMA_LOAD_RB(AD_3_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x01D4	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x01D5	*/	VDMA_STORE_RB(AD_3_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x01D6	*/	VDMA_LOAD_RB(AD_3_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x01D7	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x01D8	*/	VDMA_STORE_RB(AD_3_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x01D9	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_3_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x01DA	*/	VDMA_LOAD_RB(AD_3_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x01DB	*/	VDMA_STORE_RB(AD_3_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x01DC	*/	VDMA_LOAD_RB(AD_3_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x01DD	*/	VDMA_STORE_RB(AD_3_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x01DE	*/	VDMA_LOAD_RA(AD_3_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x01DF	*/	VDMA_STORE_RA(AD_3_FRAME_XFER_CNT),				// Save it 
// LP1_AD_3_LOAD_ADDR
/*	0x01E0	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x01E1	*/	VDMA_STORE_SYS_ADDR(0, AD_3_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x01E2	*/	VDMA_LOAD_RB(AD_3_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x01E3	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x01E4	*/	VDMA_STORE_SYS_ADDR(0, AD_3_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x01E5	*/	VDMA_LOAD_RA(AD_3_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x01E6	*/	VDMA_STORE_RA(AD_3_DEST_XFER_CNT),				// Reset XFERS per DMA frame

// LP1_AD_4
/*	0x01E7	*/	VDMA_LOAD_RA(AD_4_STREAM_DISABLED),				// Is this channel enabled?
/*	0x01E8	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_5),				// Yes.  See if it is ready then then
/*	0x01E9	*/	VDMA_JMP(VDMA_ALWAYS, 0, LP1_AD_5),				// Check next FIFO if this one is not ready
// LP1_AD_4_READY
/*	0x01EA	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_7, LP1_AD_5),		// Check next FIFO if this one is not ready
// LP1_AD_4_SP1_WAIT_FULL
/*	0x01EB	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01EC	*/	VDMA_LOAD_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x01ED	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP1_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_4_SP2_WAIT_FULL
/*	0x01EE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP2_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01EF	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 2
/*	0x01F0	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP2_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_4_SP3_WAIT_FULL
/*	0x01F1	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP3_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01F2	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 3
/*	0x01F3	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP3_XFER_CTRL),			// Transfer this portion of the frame
// LP1_AD_4_SP4_WAIT_FULL
/*	0x01F4	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP4_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x01F5	*/	VDMA_ADD_SYS_ADDR_I(AUDIO_CHANNEL_SIZE),			// Bump the destination address to stereo pair 4
/*	0x01F6	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP4_XFER_CTRL),			// Transfer this portion of the frame
// Decrement the remining XFER count and set next destination address
/*	0x01F7	*/	VDMA_LOAD_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x01F8	*/	VDMA_ADD_SYS_ADDR_I(AD_4_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x01F9	*/	VDMA_STORE_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
/*	0x01FA	*/	VDMA_LOAD_RA(AD_4_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x01FB	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x01FC	*/	VDMA_STORE_RA(AD_4_DEST_XFER_CNT),				// Save the count
/*	0x01FD	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_5),				// Branch if we are not done with this frame
// LP1_AD_4_SP1_TS_WAIT:
/*	0x01FE	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x01FF	*/	VDMA_LOAD_SYS_ADDR(0, AD_4_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x0200	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_4_SP2_TS_WAIT:
/*	0x0201	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP2_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0202	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 2 time stamp address
/*	0x0203	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP2_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_4_SP3_TS_WAIT:
/*	0x0204	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP3_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0205	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 3 time stamp address
/*	0x0206	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP3_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
// LP1_AD_4_SP4_TS_WAIT:
/*	0x0207	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_4_SP4_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0208	*/	VDMA_ADD_SYS_ADDR_I(PLUS_8),					// Load the audio channel 4 time stamp address
/*	0x0209	*/	VDMA_LOAD_XFER_CTL(0, AD_4_SP4_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x020A	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x020B	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_7),
/*	0x020C	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x020D	*/	VDMA_LOAD_RA(AD_4_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x020E	*/	VDMA_STORE_RA(AD_4_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x020F	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x0210	*/	VDMA_STORE_RA(AD_4_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0211	*/	VDMA_LOAD_RB(AD_4_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0212	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0213	*/	VDMA_STORE_RB(AD_4_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x0214	*/	VDMA_LOAD_RB(AD_4_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x0215	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0216	*/	VDMA_STORE_RB(AD_4_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x0217	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_4_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x0218	*/	VDMA_LOAD_RB(AD_4_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x0219	*/	VDMA_STORE_RB(AD_4_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x021A	*/	VDMA_LOAD_RB(AD_4_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x021B	*/	VDMA_STORE_RB(AD_4_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x021C	*/	VDMA_LOAD_RA(AD_4_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x021D	*/	VDMA_STORE_RA(AD_4_FRAME_XFER_CNT),				// Save it 
// LP1_AD_4_LOAD_ADDR
/*	0x021E	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x021F	*/	VDMA_STORE_SYS_ADDR(0, AD_4_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x0220	*/	VDMA_LOAD_RB(AD_4_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x0221	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x0222	*/	VDMA_STORE_SYS_ADDR(0, AD_4_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x0223	*/	VDMA_LOAD_RA(AD_4_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x0224	*/	VDMA_STORE_RA(AD_4_DEST_XFER_CNT),				// Reset XFERS per DMA frame

// LP1_AD_5
/*	0x0225	*/	VDMA_LOAD_RA(AD_5_STREAM_DISABLED),				// Is this channel enabled?
/*	0x0226	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_6),				// Yes.  See if it is ready then then
// LP1_AD_5_READY
/*	0x0227	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_8, LP1_AD_6),		// Check next FIFO if this one is not ready
// LP1_AD_5_SP1_WAIT_FULL
/*	0x0228	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_5_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0229	*/	VDMA_LOAD_SYS_ADDR(0, AD_5_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x022A	*/	VDMA_LOAD_XFER_CTL(0, AD_5_SP1_XFER_CTRL),			// Transfer this portion of the frame
/*	0x022B	*/	VDMA_ADD_SYS_ADDR_I(AD_5_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x022C	*/	VDMA_STORE_SYS_ADDR(0, AD_5_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
// Decrement the remining XFER count and set next destination address
/*	0x022D	*/	VDMA_LOAD_RA(AD_5_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x022E	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x022F	*/	VDMA_STORE_RA(AD_5_DEST_XFER_CNT),				// Save the count
/*	0x0230	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_6),				// Branch if we are done with this frame
// LP1_AD_5_SP1_TS_WAIT:
/*	0x0231	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_5_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0232	*/	VDMA_LOAD_SYS_ADDR(0, AD_5_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x0233	*/	VDMA_LOAD_XFER_CTL(0, AD_5_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x0234	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0235	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_11),
/*	0x0236	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0237	*/	VDMA_LOAD_RA(AD_5_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x0238	*/	VDMA_STORE_RA(AD_5_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0239	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x023A	*/	VDMA_STORE_RA(AD_5_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x023B	*/	VDMA_LOAD_RB(AD_5_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x023C	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x023D	*/	VDMA_STORE_RB(AD_5_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x023E	*/	VDMA_LOAD_RB(AD_5_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x023F	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0240	*/	VDMA_STORE_RB(AD_5_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x0241	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_5_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x0242	*/	VDMA_LOAD_RB(AD_5_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x0243	*/	VDMA_STORE_RB(AD_5_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x0244	*/	VDMA_LOAD_RB(AD_5_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x0245	*/	VDMA_STORE_RB(AD_5_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x0246	*/	VDMA_LOAD_RA(AD_5_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x0247	*/	VDMA_STORE_RA(AD_5_FRAME_XFER_CNT),				// Save it 
// LP1_AD_5_LOAD_ADDR
/*	0x0248	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x0249	*/	VDMA_STORE_SYS_ADDR(0, AD_5_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x024A	*/	VDMA_LOAD_RB(AD_5_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x024B	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x024C	*/	VDMA_STORE_SYS_ADDR(0, AD_5_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x024D	*/	VDMA_LOAD_RA(AD_5_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x024E	*/	VDMA_STORE_RA(AD_5_DEST_XFER_CNT),				// Reset XFERS per DMA frame
/*	0x024F	*/	VDMA_NOP(),
/*	0x0250	*/	VDMA_NOP(),
/*	0x0251	*/	VDMA_NOP(),

// LP1_AD_6
/*	0x0252	*/	VDMA_LOAD_RA(AD_6_STREAM_DISABLED),				// Is this channel enabled?
/*	0x0253	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_7),				// Yes.  See if it is ready then then
// LP1_AD_6_READY
/*	0x0254	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_9, LP1_AD_7),		// Check next FIFO if this one is not ready
// LP1_AD_6_SP1_WAIT_FULL
/*	0x0255	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_6_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0256	*/	VDMA_LOAD_SYS_ADDR(0, AD_6_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x0257	*/	VDMA_LOAD_XFER_CTL(0, AD_6_SP1_XFER_CTRL),			// Transfer this portion of the frame
/*	0x0258	*/	VDMA_ADD_SYS_ADDR_I(AD_6_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x0259	*/	VDMA_STORE_SYS_ADDR(0, AD_6_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
// DecreEent the remining XFER count and set next destination address
/*	0x025A	*/	VDMA_LOAD_RA(AD_6_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x025B	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x025C	*/	VDMA_STORE_RA(AD_6_DEST_XFER_CNT),				// Save the count
/*	0x025D	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_7),				// Branch if we are done with this frame
// LP1_AD_6_SP1_TS_WAIT:
/*	0x025E	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_6_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x025F	*/	VDMA_LOAD_SYS_ADDR(0, AD_6_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x0260	*/	VDMA_LOAD_XFER_CTL(0, AD_6_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x0261	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0262	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_12),
/*	0x0263	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x0264	*/	VDMA_LOAD_RA(AD_6_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x0265	*/	VDMA_STORE_RA(AD_6_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0266	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x0267	*/	VDMA_STORE_RA(AD_6_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0268	*/	VDMA_LOAD_RB(AD_6_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0269	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x026A	*/	VDMA_STORE_RB(AD_6_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x026B	*/	VDMA_LOAD_RB(AD_6_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x026C	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x026D	*/	VDMA_STORE_RB(AD_6_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x026E	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_6_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around 	to the start
/*	0x026F	*/	VDMA_LOAD_RB(AD_6_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x0270	*/	VDMA_STORE_RB(AD_6_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x0271	*/	VDMA_LOAD_RB(AD_6_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x0272	*/	VDMA_STORE_RB(AD_6_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x0273	*/	VDMA_LOAD_RA(AD_6_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x0274	*/	VDMA_STORE_RA(AD_6_FRAME_XFER_CNT),				// Save it 
// LP1_AD_6_LOAD_ADDR
/*	0x0275	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x0276	*/	VDMA_STORE_SYS_ADDR(0, AD_6_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x0277	*/	VDMA_LOAD_RB(AD_6_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x0278	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x0279	*/	VDMA_STORE_SYS_ADDR(0, AD_6_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x027A	*/	VDMA_LOAD_RA(AD_6_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x027B	*/	VDMA_STORE_RA(AD_6_DEST_XFER_CNT),				// Reset XFERS per DMA frame

// LP1_AD_7
/*	0x027C	*/	VDMA_LOAD_RA(AD_7_STREAM_DISABLED),				// Is this channel enabled?
/*	0x027D	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_8),				// Yes.  See if it is ready then then
// LP1_AD_7_READY
/*	0x027E	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_10, LP1_AD_8),		// Check next FIFO if this one is not ready
// LP1_AD_7_SP1_WAIT_FULL
/*	0x027F	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_7_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x0280	*/	VDMA_LOAD_SYS_ADDR(0, AD_7_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x0281	*/	VDMA_LOAD_XFER_CTL(0, AD_7_SP1_XFER_CTRL),			// Transfer this portion of the frame
/*	0x0282	*/	VDMA_ADD_SYS_ADDR_I(AD_7_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x0283	*/	VDMA_STORE_SYS_ADDR(0, AD_7_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
// Decrement the remining XFER count and set next destination address
/*	0x0284	*/	VDMA_LOAD_RA(AD_7_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x0285	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x0286	*/	VDMA_STORE_RA(AD_7_DEST_XFER_CNT),				// Save the count
/*	0x0287	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_8),				// Branch if we are done with this frame
// LP1_AD_7_SP1_TS_WAIT:
/*	0x0288	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_7_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x0289	*/	VDMA_LOAD_SYS_ADDR(0, AD_7_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x028A	*/	VDMA_LOAD_XFER_CTL(0, AD_7_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x028B	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x028C	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_13),
/*	0x028D	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x028E	*/	VDMA_LOAD_RA(AD_7_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x028F	*/	VDMA_STORE_RA(AD_7_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x0290	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x0291	*/	VDMA_STORE_RA(AD_7_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x0292	*/	VDMA_LOAD_RB(AD_7_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x0293	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0294	*/	VDMA_STORE_RB(AD_7_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x0295	*/	VDMA_LOAD_RB(AD_7_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x0296	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x0297	*/	VDMA_STORE_RB(AD_7_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x0298	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_7_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around	to the start
/*	0x0299	*/	VDMA_LOAD_RB(AD_7_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x029A	*/	VDMA_STORE_RB(AD_7_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x029B	*/	VDMA_LOAD_RB(AD_7_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x029C	*/	VDMA_STORE_RB(AD_7_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x029D	*/	VDMA_LOAD_RA(AD_7_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x029E	*/	VDMA_STORE_RA(AD_7_FRAME_XFER_CNT),				// Save it 
// LP1_AD_7_LOAD_ADDR
/*	0x029F	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x02A0	*/	VDMA_STORE_SYS_ADDR(0, AD_7_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x02A1	*/	VDMA_LOAD_RB(AD_7_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x02A2	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x02A3	*/	VDMA_STORE_SYS_ADDR(0, AD_7_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x02A4	*/	VDMA_LOAD_RA(AD_7_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x02A5	*/	VDMA_STORE_RA(AD_7_DEST_XFER_CNT),				// Reset XFERS per DMA frame

// LP1_AD_8
/*	0x02A6	*/	VDMA_LOAD_RA(AD_8_STREAM_DISABLED),				// Is this channel enabled?
/*	0x02A7	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, FINISH),				// Yes.  See if it is ready then then
// LP1_AD_8_READY
/*	0x02A8	*/	VDMA_JMP(VDMA_EXT_COND_LO, EXT_COND_11, FINISH),		// Check next FIFO if this one is not ready
// LP1_AD_8_SP1_WAIT_FULL
/*	0x02A9	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_8_SP1_WAIT_FULL),// Do not issue xfer command when LDM queue is full
/*	0x02AA	*/	VDMA_LOAD_SYS_ADDR(0, AD_8_SP1_DEST_ADDR_L),			// Load the destination address of stereo pair 1
/*	0x02AB	*/	VDMA_LOAD_XFER_CTL(0, AD_8_SP1_XFER_CTRL),			// Transfer this portion of the frame
/*	0x02AC	*/	VDMA_ADD_SYS_ADDR_I(AD_8_SP1_XFER_STEP_SIZE),			// Bump the stereo pair destination address
/*	0x02AD	*/	VDMA_STORE_SYS_ADDR(0, AD_8_SP1_DEST_ADDR_L),			// Save the new stereo pair destination address
// Decrement the remining XFER count and set next destination address
/*	0x02AE	*/	VDMA_LOAD_RA(AD_8_DEST_XFER_CNT),				// Get the remaining XFER count
/*	0x02AF	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the remaining XFER count
/*	0x02B0	*/	VDMA_STORE_RA(AD_8_DEST_XFER_CNT),				// Save the count
/*	0x02B1	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, FINISH),				// Branch if we are done with this frame
// LP1_AD_8_SP1_TS_WAIT:
/*	0x02B2	*/	VDMA_JMP(VDMA_LDM_CMD_QUEUE_FULL_HI, 0, LP1_AD_8_SP1_TS_WAIT),	// Do not issue xfer command when LDM queue is full
/*	0x02B3	*/	VDMA_LOAD_SYS_ADDR(0, AD_8_SP1_TS_DEST_ADDR_L),			// Load the audio channel 1 time stamp address
/*	0x02B4	*/	VDMA_LOAD_XFER_CTL(0, AD_8_SP1_TS_XFER_CTRL),			// Transfer the time stamp portion of the frame
/*	0x02B5	*/	VDMA_LOAD_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x02B6	*/	VDMA_ADD_SYS_ADDR_I(EVENT_ID_14),
/*	0x02B7	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/*	0x02B8	*/	VDMA_LOAD_RA(AD_8_FRAME_XFER_CNT),				// Load the old last frame xfer'd index
/*	0x02B9	*/	VDMA_STORE_RA(AD_8_LAST_FRAME_INDEX),				// Save the new last frame xfer'd index
// Setup to start DMA'ing into the next video frame
/*	0x02BA	*/	VDMA_ADD_RA(MINUS_1),						// Decrement the count
/*	0x02BB	*/	VDMA_STORE_RA(AD_8_FRAME_XFER_CNT),				// Save the new DMA frames to xfer
/*	0x02BC	*/	VDMA_LOAD_RB(AD_8_DEST_TS_ADDR_PTR),				// Load the last time stamp desination address pointer
/*	0x02BD	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x02BE	*/	VDMA_STORE_RB(AD_8_DEST_TS_ADDR_PTR),				// Save the new time stamp desination pointer
/*	0x02BF	*/	VDMA_LOAD_RB(AD_8_DEST_ADDR_PTR),				// Load the last desination address pointer
/*	0x02C0	*/	VDMA_ADD_RB(MINUS_2),						// Decrement the count
/*	0x02C1	*/	VDMA_STORE_RB(AD_8_DEST_ADDR_PTR),				// Save the new desination pointer
/*	0x02C2	*/	VDMA_JMP(VDMA_RA_NEQZ, 0, LP1_AD_8_LOAD_ADDR),			// Branch if we don't need to wrap yet
// At this point we have xfer'd video to all of our DMA buffers.  Wrap around to the start
/*	0x02C3	*/	VDMA_LOAD_RB(AD_8_LAST_TS_ADDR_PTR),				// Load last DMA time stamp pointer for this FIFO
/*	0x02C4	*/	VDMA_STORE_RB(AD_8_DEST_TS_ADDR_PTR),				// Save it as the current DMA time stamp pointer
/*	0x02C5	*/	VDMA_LOAD_RB(AD_8_LAST_ADDR_PTR),				// Load last DMA frame pointer for this FIFO
/*	0x02C6	*/	VDMA_STORE_RB(AD_8_DEST_ADDR_PTR),				// Save it as the current DMA frame pointer
/*	0x02C7	*/	VDMA_LOAD_RA(AD_8_FRAME_CNT),					// Load DMA frames to xfer count
/*	0x02C8	*/	VDMA_STORE_RA(AD_8_FRAME_XFER_CNT),				// Save it 
// LP1_AD_8_LOAD_ADDR
/*	0x02C9	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for stereo pair 1
/*	0x02CA	*/	VDMA_STORE_SYS_ADDR(0, AD_8_SP1_DEST_ADDR_L),			// Save the new destination address for stereo pair 1
/*	0x02CB	*/	VDMA_LOAD_RB(AD_8_DEST_TS_ADDR_PTR),				// Load the last desination address pointer
/*	0x02CC	*/	VDMA_LOAD_SYS_ADDR(VDMA_RB, 0),					// Load the new destination address for the stereo pair 1 time stamp
/*	0x02CD	*/	VDMA_STORE_SYS_ADDR(0, AD_8_SP1_TS_DEST_ADDR_L),		// Save the new destination address for the stereo pair 1 time stamp
/*	0x02CE	*/	VDMA_LOAD_RA(AD_8_XFER_CNT),					// Load initial XFERS per DMA frame
/*	0x02CF	*/	VDMA_STORE_RA(AD_8_DEST_XFER_CNT),				// Reset XFERS per DMA frame
// FINISH
/*	0x02D0	*/	VDMA_LOAD_RA(SIGNAL_EVENT),
/*	0x02D1	*/	VDMA_JMP(VDMA_RA_EQZ, 0, MAIN),
// LDM_IDLE_WAIT
/*	0x02D2	*/	VDMA_JMP(VDMA_EXT_COND_LO, LDM_IDLE, LDM_IDLE_WAIT),
/* 0x02D3	*/	VDMA_NOP(),
// SIGNAL_EVENT
/* 0x02D4	*/	VDMA_SIG_EVENT(0, 1, 0),
/* 0x02D5	*/	VDMA_NOP(),
/*	0x02D6	*/	VDMA_LOAD_SYS_ADDR(0, SIG_EVENT_CODE),
/*	0x02D7	*/	VDMA_STORE_SYS_ADDR(0, SIGNAL_EVENT),
/* 0x02D8	*/	VDMA_NOP(),
/* 0x02D9	*/	VDMA_NOP(),
/*	0x02DA	*/	VDMA_JMP(VDMA_ALWAYS, 0, MAIN),
};
#endif

void vc8x0_dma_dump_channel_registers_audio(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;

	dprintk(1, "m_nXferCnt        = %x(%08x)\n", channel->regs.m_nXferCnt, vc_read32(channel->regs.m_nXferCnt));
	dprintk(1, "m_nDestXferCnt    = %x(%08x)\n", channel->regs.m_nDestXferCnt, vc_read32(channel->regs.m_nDestXferCnt));
	dprintk(1, "m_nSP1XferStepSize= %x(%08x)\n", channel->regs.m_nSP1XferStepSize, vc_read32(channel->regs.m_nSP1XferStepSize));
	dprintk(1, "m_nSP1XferCtrl    = %x(%08x)\n", channel->regs.m_nSP1XferCtrl, vc_read32(channel->regs.m_nSP1XferCtrl));
	dprintk(1, "m_nSP2XferCtrl    = %x(%08x)\n", channel->regs.m_nSP2XferCtrl, vc_read32(channel->regs.m_nSP2XferCtrl));
	dprintk(1, "m_nSP3XferCtrl    = %x(%08x)\n", channel->regs.m_nSP3XferCtrl, vc_read32(channel->regs.m_nSP3XferCtrl));
	dprintk(1, "m_nSP4XferCtrl    = %x(%08x)\n", channel->regs.m_nSP4XferCtrl, vc_read32(channel->regs.m_nSP4XferCtrl));
	dprintk(1, "m_nSP1TSXferCtrl  = %x(%08x)\n", channel->regs.m_nSP1TSXferCtrl, vc_read32(channel->regs.m_nSP1TSXferCtrl));
	dprintk(1, "m_nSP2TSXferCtrl  = %x(%08x)\n", channel->regs.m_nSP2TSXferCtrl, vc_read32(channel->regs.m_nSP2TSXferCtrl));
	dprintk(1, "m_nSP3TSXferCtrl  = %x(%08x)\n", channel->regs.m_nSP3TSXferCtrl, vc_read32(channel->regs.m_nSP3TSXferCtrl));
	dprintk(1, "m_nSP4TSXferCtrl  = %x(%08x)\n", channel->regs.m_nSP4TSXferCtrl, vc_read32(channel->regs.m_nSP4TSXferCtrl));
	dprintk(1, "m_nFrameCnt       = %x(%08x)\n", channel->regs.m_nFrameCnt, vc_read32(channel->regs.m_nFrameCnt));
	dprintk(1, "m_nStreamDisabled = %x(%08x)\n", channel->regs.m_nStreamDisabled, vc_read32(channel->regs.m_nStreamDisabled));
	dprintk(1, "m_nLastFrameIndex = %x(%08x)\n", channel->regs.m_nLastFrameIndex, vc_read32(channel->regs.m_nLastFrameIndex));
	dprintk(1, "m_nLastAddrPtr    = %x(%08x)\n", channel->regs.m_nLastAddrPtr, vc_read32(channel->regs.m_nLastAddrPtr));
	dprintk(1, "m_nDestAddrPtr    = %x(%08x)\n", channel->regs.m_nDestAddrPtr, vc_read32(channel->regs.m_nDestAddrPtr));
	dprintk(1, "m_nLastTSAddrPtr  = %x(%08x)\n", channel->regs.m_nLastTSAddrPtr, vc_read32(channel->regs.m_nLastTSAddrPtr));
	dprintk(1, "m_nDestTSAddrPtr  = %x(%08x)\n", channel->regs.m_nDestTSAddrPtr, vc_read32(channel->regs.m_nDestTSAddrPtr));
	dprintk(1, "m_nFrameXferCnt   = %x(%08x)\n", channel->regs.m_nFrameXferCnt, vc_read32(channel->regs.m_nFrameXferCnt));
	dprintk(1, "m_nSP1DestAddrL   = %x(%08x)\n", channel->regs.m_nSP1DestAddrL, vc_read32(channel->regs.m_nSP1DestAddrL));
	dprintk(1, "m_nSP1DestAddrH   = %x(%08x)\n", channel->regs.m_nSP1DestAddrH, vc_read32(channel->regs.m_nSP1DestAddrH));
	dprintk(1, "m_nSP1TSDestAddrL = %x(%08x)\n", channel->regs.m_nSP1TSDestAddrL, vc_read32(channel->regs.m_nSP1TSDestAddrL));
	dprintk(1, "m_nSP1TSDestAddrH = %x(%08x)\n", channel->regs.m_nSP1TSDestAddrH, vc_read32(channel->regs.m_nSP1TSDestAddrH));
	dprintk(1, "m_nSP1DmaAddrL    = %x(%08x)\n", channel->regs.m_nSP1DmaAddrL, vc_read32(channel->regs.m_nSP1DmaAddrL));
	dprintk(1, "m_nSP1DmaAddrH    = %x(%08x)\n", channel->regs.m_nSP1DmaAddrH, vc_read32(channel->regs.m_nSP1DmaAddrH));
	dprintk(1, "m_nSP1TSDmaAddrL  = %x(%08x)\n", channel->regs.m_nSP1TSDmaAddrL, vc_read32(channel->regs.m_nSP1TSDmaAddrL));
	dprintk(1, "m_nSP1TSDmaAddrH  = %x(%08x)\n", channel->regs.m_nSP1TSDmaAddrH, vc_read32(channel->regs.m_nSP1TSDmaAddrH));
	dprintk(1, "m_nFifoGo         = %x(%08x)\n", channel->regs.m_nFifoGo, vc_read32(channel->regs.m_nFifoGo));
	dprintk(1, "m_nMisc           = %x(%08x)\n", channel->regs.m_nMisc, vc_read32(channel->regs.m_nMisc));

	dprintk(1, "OFFSETS\n");
	dprintk(1, "m_nXferCnt        = %x\n", channel->offsets.m_nXferCnt);
	dprintk(1, "m_nDestXferCnt    = %x\n", channel->offsets.m_nDestXferCnt);
	dprintk(1, "m_nSP1XferStepSize= %x\n", channel->offsets.m_nSP1XferStepSize);
	dprintk(1, "m_nSP1XferCtrl    = %x\n", channel->offsets.m_nSP1XferCtrl);
	dprintk(1, "m_nSP2XferCtrl    = %x\n", channel->offsets.m_nSP2XferCtrl);
	dprintk(1, "m_nSP3XferCtrl    = %x\n", channel->offsets.m_nSP3XferCtrl);
	dprintk(1, "m_nSP4XferCtrl    = %x\n", channel->offsets.m_nSP4XferCtrl);
	dprintk(1, "m_nFrameCnt       = %x\n", channel->offsets.m_nFrameCnt);
	dprintk(1, "m_nStreamDisabled = %x\n", channel->offsets.m_nStreamDisabled);
	dprintk(1, "m_nLastFrameIndex = %x\n", channel->offsets.m_nLastFrameIndex);
	dprintk(1, "m_nLastAddrPtr    = %x\n", channel->offsets.m_nLastAddrPtr);
	dprintk(1, "m_nDestAddrPtr    = %x\n", channel->offsets.m_nDestAddrPtr);
	dprintk(1, "m_nLastTSAddrPtr  = %x\n", channel->offsets.m_nLastTSAddrPtr);
	dprintk(1, "m_nDestTSAddrPtr  = %x\n", channel->offsets.m_nDestTSAddrPtr);
	dprintk(1, "m_nFrameXferCnt   = %x\n", channel->offsets.m_nFrameXferCnt);
	dprintk(1, "m_nSP1TSXferCtrl  = %x\n", channel->offsets.m_nSP1TSXferCtrl);
	dprintk(1, "m_nSP2TSXferCtrl  = %x\n", channel->offsets.m_nSP2TSXferCtrl);
	dprintk(1, "m_nSP3TSXferCtrl  = %x\n", channel->offsets.m_nSP3TSXferCtrl);
	dprintk(1, "m_nSP4TSXferCtrl  = %x\n", channel->offsets.m_nSP4TSXferCtrl);
	dprintk(1, "m_nSP1DestAddrL   = %x\n", channel->offsets.m_nSP1DestAddrL);
	dprintk(1, "m_nSP1DestAddrH   = %x\n", channel->offsets.m_nSP1DestAddrH);
	dprintk(1, "m_nSP1TSDestAddrL = %x\n", channel->offsets.m_nSP1TSDestAddrL);
	dprintk(1, "m_nSP1TSDestAddrH = %x\n", channel->offsets.m_nSP1TSDestAddrH);
	dprintk(1, "m_nSP1DmaAddrL    = %x\n", channel->offsets.m_nSP1DmaAddrL);
	dprintk(1, "m_nSP1DmaAddrH    = %x\n", channel->offsets.m_nSP1DmaAddrH);
	dprintk(1, "m_nSP1TSDmaAddrL  = %x\n", channel->offsets.m_nSP1TSDmaAddrL);
	dprintk(1, "m_nSP1TSDmaAddrH  = %x\n", channel->offsets.m_nSP1TSDmaAddrH);
	dprintk(1, "m_nFifoGo         = %x\n", channel->offsets.m_nFifoGo);
	dprintk(1, "m_nMisc           = %x\n", channel->offsets.m_nMisc);

	dprintk(1, "DEFAULTS\n");
	dprintk(1, "m_nXferCnt        = %x\n", channel->defaults.m_nXferCnt);
	dprintk(1, "m_nDestXferCnt    = %x\n", channel->defaults.m_nDestXferCnt);
	dprintk(1, "m_nFrameCnt       = %x\n", channel->defaults.m_nFrameCnt);
	dprintk(1, "m_nStreamDisabled = %x\n", channel->defaults.m_nStreamDisabled);
	dprintk(1, "m_nLastFrameIndex = %x\n", channel->defaults.m_nLastFrameIndex);
	dprintk(1, "m_nLastAddrPtr    = %x\n", channel->defaults.m_nLastAddrPtr);
	dprintk(1, "m_nDestAddrPtr    = %x\n", channel->defaults.m_nDestAddrPtr);
	dprintk(1, "m_nLastTSAddrPtr  = %x\n", channel->defaults.m_nLastTSAddrPtr);
	dprintk(1, "m_nDestTSAddrPtr  = %x\n", channel->defaults.m_nDestTSAddrPtr);
	dprintk(1, "m_nFrameXferCnt   = %x\n", channel->defaults.m_nFrameXferCnt);
	dprintk(1, "m_nSP1XferStepSize= %x\n", channel->defaults.m_nSP1XferStepSize);
	dprintk(1, "m_nSP1XferCtrl    = %x\n", channel->defaults.m_nSP1XferCtrl);
	dprintk(1, "m_nSP2XferCtrl    = %x\n", channel->defaults.m_nSP2XferCtrl);
	dprintk(1, "m_nSP3XferCtrl    = %x\n", channel->defaults.m_nSP3XferCtrl);
	dprintk(1, "m_nSP4XferCtrl    = %x\n", channel->defaults.m_nSP4XferCtrl);
	dprintk(1, "m_nSP1TSXferCtrl  = %x\n", channel->defaults.m_nSP1TSXferCtrl);
	dprintk(1, "m_nSP2TSXferCtrl  = %x\n", channel->defaults.m_nSP2TSXferCtrl);
	dprintk(1, "m_nSP3TSXferCtrl  = %x\n", channel->defaults.m_nSP3TSXferCtrl);
	dprintk(1, "m_nSP4TSXferCtrl  = %x\n", channel->defaults.m_nSP4TSXferCtrl);
	dprintk(1, "m_nSP1DestAddrL   = %x\n", channel->defaults.m_nSP1DestAddrL);
	dprintk(1, "m_nSP1DestAddrH   = %x\n", channel->defaults.m_nSP1DestAddrH);
	dprintk(1, "m_nSP1TSDestAddrL = %x\n", channel->defaults.m_nSP1TSDestAddrL);
	dprintk(1, "m_nSP1TSDestAddrH = %x\n", channel->defaults.m_nSP1TSDestAddrH);
	dprintk(1, "m_nSP1DmaAddrL    = %x\n", channel->defaults.m_nSP1DmaAddrL);
	dprintk(1, "m_nSP1DmaAddrH    = %x\n", channel->defaults.m_nSP1DmaAddrH);
	dprintk(1, "m_nSP1TSDmaAddrL  = %x\n", channel->defaults.m_nSP1TSDmaAddrL);
	dprintk(1, "m_nSP1TSDmaAddrH  = %x\n", channel->defaults.m_nSP1TSDmaAddrH);
	dprintk(1, "m_nFifoGo         = %x\n", channel->defaults.m_nFifoGo);
	dprintk(1, "m_nMisc           = %x\n", channel->defaults.m_nMisc);
}

void vc8x0_dma_dump_channel_registers_video(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;

	dprintk(1, "m_nXferCnt        = %x(%08x)\n", channel->regs.m_nXferCnt, vc_read32(channel->regs.m_nXferCnt));
	dprintk(1, "m_nDestXferCnt    = %x(%08x)\n", channel->regs.m_nDestXferCnt, vc_read32(channel->regs.m_nDestXferCnt));
	dprintk(1, "m_nYXferSizeFull  = %x(%08x)\n", channel->regs.m_nYXferSizeFull, vc_read32(channel->regs.m_nYXferSizeFull));
	dprintk(1, "m_nUVXferSizeFull = %x(%08x)\n", channel->regs.m_nUVXferSizeFull, vc_read32(channel->regs.m_nUVXferSizeFull));
	dprintk(1, "m_nYXferCtrlFull  = %x(%08x)\n", channel->regs.m_nYXferCtrlFull, vc_read32(channel->regs.m_nYXferCtrlFull));
	dprintk(1, "m_nUXferCtrlFull  = %x(%08x)\n", channel->regs.m_nUXferCtrlFull, vc_read32(channel->regs.m_nUXferCtrlFull));
	dprintk(1, "m_nVXferCtrlFull  = %x(%08x)\n", channel->regs.m_nVXferCtrlFull, vc_read32(channel->regs.m_nVXferCtrlFull));
	dprintk(1, "m_nYXferSizePart  = %x(%08x)\n", channel->regs.m_nYXferSizePart, vc_read32(channel->regs.m_nYXferSizePart));
	dprintk(1, "m_nUVXferSizePart = %x(%08x)\n", channel->regs.m_nUVXferSizePart, vc_read32(channel->regs.m_nUVXferSizePart));
	dprintk(1, "m_nYXferCtrlPart  = %x(%08x)\n", channel->regs.m_nYXferCtrlPart, vc_read32(channel->regs.m_nYXferCtrlPart));
	dprintk(1, "m_nUXferCtrlPart  = %x(%08x)\n", channel->regs.m_nUXferCtrlPart, vc_read32(channel->regs.m_nUXferCtrlPart));
	dprintk(1, "m_nVXferCtrlPart  = %x(%08x)\n", channel->regs.m_nVXferCtrlPart, vc_read32(channel->regs.m_nVXferCtrlPart));
	dprintk(1, "m_nYTSXferCtrl    = %x(%08x)\n", channel->regs.m_nYTSXferCtrl, vc_read32(channel->regs.m_nYTSXferCtrl));
	dprintk(1, "m_nUTSXferCtrl    = %x(%08x)\n", channel->regs.m_nUTSXferCtrl, vc_read32(channel->regs.m_nUTSXferCtrl));
	dprintk(1, "m_nVTSXferCtrl    = %x(%08x)\n", channel->regs.m_nVTSXferCtrl, vc_read32(channel->regs.m_nVTSXferCtrl));
	dprintk(1, "m_nFrameCnt       = %x(%08x)\n", channel->regs.m_nFrameCnt, vc_read32(channel->regs.m_nFrameCnt));
	dprintk(1, "m_nStreamDisabled = %x(%08x)\n", channel->regs.m_nStreamDisabled, vc_read32(channel->regs.m_nStreamDisabled));
	dprintk(1, "m_nLastFrameIndex = %x(%08x)\n", channel->regs.m_nLastFrameIndex, vc_read32(channel->regs.m_nLastFrameIndex));
	dprintk(1, "m_nLastAddrPtr    = %x(%08x)\n", channel->regs.m_nLastAddrPtr, vc_read32(channel->regs.m_nLastAddrPtr));
	dprintk(1, "m_nDestAddrPtr    = %x(%08x)\n", channel->regs.m_nDestAddrPtr, vc_read32(channel->regs.m_nDestAddrPtr));
	dprintk(1, "m_nLastTSAddrPtr  = %x(%08x)\n", channel->regs.m_nLastTSAddrPtr, vc_read32(channel->regs.m_nLastTSAddrPtr));
	dprintk(1, "m_nDestTSAddrPtr  = %x(%08x)\n", channel->regs.m_nDestTSAddrPtr, vc_read32(channel->regs.m_nDestTSAddrPtr));
	dprintk(1, "m_nFrameXferCnt   = %x(%08x)\n", channel->regs.m_nFrameXferCnt, vc_read32(channel->regs.m_nFrameXferCnt));
	dprintk(1, "m_nYDestAddrL     = %x(%08x)\n", channel->regs.m_nYDestAddrL, vc_read32(channel->regs.m_nYDestAddrL));
	dprintk(1, "m_nYDestAddrH     = %x(%08x)\n", channel->regs.m_nYDestAddrH, vc_read32(channel->regs.m_nYDestAddrH));
	dprintk(1, "m_nUDestAddrL     = %x(%08x)\n", channel->regs.m_nUDestAddrL, vc_read32(channel->regs.m_nUDestAddrL));
	dprintk(1, "m_nUDestAddrH     = %x(%08x)\n", channel->regs.m_nUDestAddrH, vc_read32(channel->regs.m_nUDestAddrH));
	dprintk(1, "m_nVDestAddrL     = %x(%08x)\n", channel->regs.m_nVDestAddrL, vc_read32(channel->regs.m_nVDestAddrL));
	dprintk(1, "m_nVDestAddrH     = %x(%08x)\n", channel->regs.m_nVDestAddrH, vc_read32(channel->regs.m_nVDestAddrH));
	dprintk(1, "m_nYTSDestAddrL   = %x(%08x)\n", channel->regs.m_nYTSDestAddrL, vc_read32(channel->regs.m_nYTSDestAddrL));
	dprintk(1, "m_nYTSDestAddrH   = %x(%08x)\n", channel->regs.m_nYTSDestAddrH, vc_read32(channel->regs.m_nYTSDestAddrH));
	dprintk(1, "m_nYDmaAddrL      = %x(%08x)\n", channel->regs.m_nYDmaAddrL, vc_read32(channel->regs.m_nYDmaAddrL));
	dprintk(1, "m_nYDmaAddrH      = %x(%08x)\n", channel->regs.m_nYDmaAddrH, vc_read32(channel->regs.m_nYDmaAddrH));
	dprintk(1, "m_nUDmaAddrL      = %x(%08x)\n", channel->regs.m_nUDmaAddrL, vc_read32(channel->regs.m_nUDmaAddrL));
	dprintk(1, "m_nUDmaAddrH      = %x(%08x)\n", channel->regs.m_nUDmaAddrH, vc_read32(channel->regs.m_nUDmaAddrH));
	dprintk(1, "m_nVDmaAddrL      = %x(%08x)\n", channel->regs.m_nVDmaAddrL, vc_read32(channel->regs.m_nVDmaAddrL));
	dprintk(1, "m_nVDmaAddrH      = %x(%08x)\n", channel->regs.m_nVDmaAddrH, vc_read32(channel->regs.m_nVDmaAddrH));
	dprintk(1, "m_nYTSDmaAddrL    = %x(%08x)\n", channel->regs.m_nYTSDmaAddrL, vc_read32(channel->regs.m_nYTSDmaAddrL));
	dprintk(1, "m_nYTSDmaAddrH    = %x(%08x)\n", channel->regs.m_nYTSDmaAddrH, vc_read32(channel->regs.m_nYTSDmaAddrH));
	dprintk(1, "m_nFifoGo         = %x(%08x)\n", channel->regs.m_nFifoGo, vc_read32(channel->regs.m_nFifoGo));
	dprintk(1, "m_nMisc           = %x(%08x)\n", channel->regs.m_nMisc, vc_read32(channel->regs.m_nMisc));

	dprintk(1, "DEFAULTS\n");
	dprintk(1, "m_nXferCnt        = %x\n", channel->defaults.m_nXferCnt);
	dprintk(1, "m_nDestXferCnt    = %x\n", channel->defaults.m_nDestXferCnt);
	dprintk(1, "m_nYXferSizeFull  = %x\n", channel->defaults.m_nYXferSizeFull);
	dprintk(1, "m_nUVXferSizeFull = %x\n", channel->defaults.m_nUVXferSizeFull);
	dprintk(1, "m_nYXferCtrlFull  = %x\n", channel->defaults.m_nYXferCtrlFull);
	dprintk(1, "m_nUXferCtrlFull  = %x\n", channel->defaults.m_nUXferCtrlFull);
	dprintk(1, "m_nVXferCtrlFull  = %x\n", channel->defaults.m_nVXferCtrlFull);
	dprintk(1, "m_nYXferSizePart  = %x\n", channel->defaults.m_nYXferSizePart);
	dprintk(1, "m_nUVXferSizePart = %x\n", channel->defaults.m_nUVXferSizePart);
	dprintk(1, "m_nYXferCtrlPart  = %x\n", channel->defaults.m_nYXferCtrlPart);
	dprintk(1, "m_nUXferCtrlPart  = %x\n", channel->defaults.m_nUXferCtrlPart);
	dprintk(1, "m_nVXferCtrlPart  = %x\n", channel->defaults.m_nVXferCtrlPart);
	dprintk(1, "m_nYTSXferCtrl    = %x\n", channel->defaults.m_nYTSXferCtrl);
	dprintk(1, "m_nUTSXferCtrl    = %x\n", channel->defaults.m_nUTSXferCtrl);
	dprintk(1, "m_nVTSXferCtrl    = %x\n", channel->defaults.m_nVTSXferCtrl);
	dprintk(1, "m_nFrameCnt       = %x\n", channel->defaults.m_nFrameCnt);
	dprintk(1, "m_nStreamDisabled = %x\n", channel->defaults.m_nStreamDisabled);
	dprintk(1, "m_nLastFrameIndex = %x\n", channel->defaults.m_nLastFrameIndex);
	dprintk(1, "m_nLastAddrPtr    = %x\n", channel->defaults.m_nLastAddrPtr);
	dprintk(1, "m_nDestAddrPtr    = %x\n", channel->defaults.m_nDestAddrPtr);
	dprintk(1, "m_nLastTSAddrPtr  = %x\n", channel->defaults.m_nLastTSAddrPtr);
	dprintk(1, "m_nDestTSAddrPtr  = %x\n", channel->defaults.m_nDestTSAddrPtr);
	dprintk(1, "m_nFrameXferCnt   = %x\n", channel->defaults.m_nFrameXferCnt);
	dprintk(1, "m_nYDestAddrL     = %x\n", channel->defaults.m_nYDestAddrL);
	dprintk(1, "m_nYDestAddrH     = %x\n", channel->defaults.m_nYDestAddrH);
	dprintk(1, "m_nUDestAddrL     = %x\n", channel->defaults.m_nUDestAddrL);
	dprintk(1, "m_nUDestAddrH     = %x\n", channel->defaults.m_nUDestAddrH);
	dprintk(1, "m_nVDestAddrL     = %x\n", channel->defaults.m_nVDestAddrL);
	dprintk(1, "m_nVDestAddrH     = %x\n", channel->defaults.m_nVDestAddrH);
	dprintk(1, "m_nYTSDestAddrL   = %x\n", channel->defaults.m_nYTSDestAddrL);
	dprintk(1, "m_nYTSDestAddrH   = %x\n", channel->defaults.m_nYTSDestAddrH);
	dprintk(1, "m_nYDmaAddrL      = %x\n", channel->defaults.m_nYDmaAddrL);
	dprintk(1, "m_nYDmaAddrH      = %x\n", channel->defaults.m_nYDmaAddrH);
	dprintk(1, "m_nUDmaAddrL      = %x\n", channel->defaults.m_nUDmaAddrL);
	dprintk(1, "m_nUDmaAddrH      = %x\n", channel->defaults.m_nUDmaAddrH);
	dprintk(1, "m_nVDmaAddrL      = %x\n", channel->defaults.m_nVDmaAddrL);
	dprintk(1, "m_nVDmaAddrH      = %x\n", channel->defaults.m_nVDmaAddrH);
	dprintk(1, "m_nYTSDmaAddrL    = %x\n", channel->defaults.m_nYTSDmaAddrL);
	dprintk(1, "m_nYTSDmaAddrH    = %x\n", channel->defaults.m_nYTSDmaAddrH);
	dprintk(1, "m_nFifoGo         = %x\n", channel->defaults.m_nFifoGo);
	dprintk(1, "m_nMisc           = %x\n", channel->defaults.m_nMisc);
}

void vc8x0_dma_dump_channel_registers(struct vc8x0_dma_channel *channel)
{
	if (channel->type == TYPE_VIDEO)
		vc8x0_dma_dump_channel_registers_video(channel);
	else
	if (channel->type == TYPE_AUDIO)
		vc8x0_dma_dump_channel_registers_audio(channel);
}

/*
 * Configure up the MMIO register addresses for each video channel.
 * The always between 0x4000 and 0x5fff in BAR0. I've verified these
 * values in a spreadsheet based on the defines above. The values are
 * also properly scaled so that a plain old vc_write32(reg, value),
 * will >> 2 correctly inside the vc_write32 macpro.
 */
static void vc8x0_dma_configure_channel_registers_video(struct vc8x0_dma_channel *channel)
{
	u32 offset = (channel->nr - 1) * (VD_2_XFER_CNT - VD_1_XFER_CNT);
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* Flex DMA registers ready for use with vc_write32() */
	channel->regs.m_nXferCnt = (VD_1_XFER_CNT + offset) << 2;
	channel->regs.m_nDestXferCnt = (VD_1_DEST_XFER_CNT + offset) << 2;
	channel->regs.m_nYXferSizeFull = (VD_1_Y_XFER_SIZE_FULL + offset) << 2;
	channel->regs.m_nUVXferSizeFull = (VD_1_UV_XFER_SIZE_FULL + offset) << 2;
	channel->regs.m_nYXferCtrlFull = (VD_1_Y_XFER_CTRL_FULL + offset) << 2;
	channel->regs.m_nUXferCtrlFull = (VD_1_U_XFER_CTRL_FULL + offset) << 2;
	channel->regs.m_nVXferCtrlFull = (VD_1_V_XFER_CTRL_FULL + offset) << 2;
	channel->regs.m_nYXferSizePart = (VD_1_Y_XFER_SIZE_PART + offset) << 2;
	channel->regs.m_nUVXferSizePart = (VD_1_UV_XFER_SIZE_PART + offset) << 2;
	channel->regs.m_nYXferCtrlPart = (VD_1_Y_XFER_CTRL_PART + offset) << 2;
	channel->regs.m_nUXferCtrlPart = (VD_1_U_XFER_CTRL_PART + offset) << 2;
	channel->regs.m_nVXferCtrlPart = (VD_1_V_XFER_CTRL_PART + offset) << 2;
	channel->regs.m_nYTSXferCtrl = (VD_1_Y_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nUTSXferCtrl = (VD_1_U_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nVTSXferCtrl = (VD_1_V_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nFrameCnt = (VD_1_FRAME_CNT + offset) << 2;
	channel->regs.m_nStreamDisabled = (VD_1_STREAM_DISABLED + offset) << 2;
	channel->regs.m_nLastFrameIndex = (VD_1_LAST_FRAME_INDEX + offset) << 2;
	channel->regs.m_nLastAddrPtr = (VD_1_LAST_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestAddrPtr = (VD_1_DEST_ADDR_PTR + offset) << 2;
	channel->regs.m_nLastTSAddrPtr = (VD_1_LAST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestTSAddrPtr = (VD_1_DEST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nFrameXferCnt = (VD_1_FRAME_XFER_CNT + offset) << 2;
	channel->regs.m_nYDestAddrL = (VD_1_Y_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nYDestAddrH = (VD_1_Y_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nUDestAddrL = (VD_1_U_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nUDestAddrH = (VD_1_U_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nVDestAddrL = (VD_1_V_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nVDestAddrH = (VD_1_V_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nYTSDestAddrL = (VD_1_Y_TS_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nYTSDestAddrH = (VD_1_Y_TS_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nYDmaAddrL = (VD_1_Y_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nYDmaAddrH = (VD_1_Y_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nUDmaAddrL = (VD_1_U_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nUDmaAddrH = (VD_1_U_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nVDmaAddrL = (VD_1_V_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nVDmaAddrH = (VD_1_V_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nYTSDmaAddrL = (VD_1_Y_TS_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nYTSDmaAddrH = (VD_1_Y_TS_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nFifoGo = (channel->nr * 0x100) + 0x28;
	channel->regs.m_nMisc = (channel->nr * 0x100) + 0x30;

	/* Flex DMA registers ready for use by other flex registers, and have
	 * different offset calculations. */
	offset = (channel->nr - 1) * (VD_2_XFER_CNT - VD_1_XFER_CNT);
	channel->offsets.m_nXferCnt = (VD_1_XFER_CNT + offset);
	channel->offsets.m_nDestXferCnt = (VD_1_DEST_XFER_CNT + offset);
	channel->offsets.m_nYXferSizeFull = (VD_1_Y_XFER_SIZE_FULL + offset);
	channel->offsets.m_nUVXferSizeFull = (VD_1_UV_XFER_SIZE_FULL + offset);
	channel->offsets.m_nYXferCtrlFull = (VD_1_Y_XFER_CTRL_FULL + offset);
	channel->offsets.m_nUXferCtrlFull = (VD_1_U_XFER_CTRL_FULL + offset);
	channel->offsets.m_nVXferCtrlFull = (VD_1_V_XFER_CTRL_FULL + offset);
	channel->offsets.m_nYXferSizePart = (VD_1_Y_XFER_SIZE_PART + offset);
	channel->offsets.m_nUVXferSizePart = (VD_1_UV_XFER_SIZE_PART + offset);
	channel->offsets.m_nYXferCtrlPart = (VD_1_Y_XFER_CTRL_PART + offset);
	channel->offsets.m_nUXferCtrlPart = (VD_1_U_XFER_CTRL_PART + offset);
	channel->offsets.m_nVXferCtrlPart = (VD_1_V_XFER_CTRL_PART + offset);
	channel->offsets.m_nYTSXferCtrl = (VD_1_Y_TS_XFER_CTRL + offset);
	channel->offsets.m_nUTSXferCtrl = (VD_1_U_TS_XFER_CTRL + offset);
	channel->offsets.m_nVTSXferCtrl = (VD_1_V_TS_XFER_CTRL + offset);
	channel->offsets.m_nFrameCnt = (VD_1_FRAME_CNT + offset);
	channel->offsets.m_nStreamDisabled = (VD_1_STREAM_DISABLED + offset);
	channel->offsets.m_nLastFrameIndex = (VD_1_LAST_FRAME_INDEX + offset);
	channel->offsets.m_nLastAddrPtr = (VD_1_LAST_ADDR_PTR + offset);
	channel->offsets.m_nDestAddrPtr = (VD_1_DEST_ADDR_PTR + offset);
	channel->offsets.m_nLastTSAddrPtr = (VD_1_LAST_TS_ADDR_PTR + offset);
	channel->offsets.m_nDestTSAddrPtr = (VD_1_DEST_TS_ADDR_PTR + offset);
	channel->offsets.m_nFrameXferCnt = (VD_1_FRAME_XFER_CNT + offset);
	channel->offsets.m_nYDestAddrL = (VD_1_Y_DEST_ADDR_L + offset);
	channel->offsets.m_nYDestAddrH = (VD_1_Y_DEST_ADDR_H + offset);
	channel->offsets.m_nUDestAddrL = (VD_1_U_DEST_ADDR_L + offset);
	channel->offsets.m_nUDestAddrH = (VD_1_U_DEST_ADDR_H + offset);
	channel->offsets.m_nVDestAddrL = (VD_1_V_DEST_ADDR_L + offset);
	channel->offsets.m_nVDestAddrH = (VD_1_V_DEST_ADDR_H + offset);
	channel->offsets.m_nYTSDestAddrL = (VD_1_Y_TS_DEST_ADDR_L + offset);
	channel->offsets.m_nYTSDestAddrH = (VD_1_Y_TS_DEST_ADDR_H + offset);
	channel->offsets.m_nYDmaAddrL = (VD_1_Y_DMA_ADDR_L + offset);
	channel->offsets.m_nYDmaAddrH = (VD_1_Y_DMA_ADDR_H + offset);
	channel->offsets.m_nUDmaAddrL = (VD_1_U_DMA_ADDR_L + offset);
	channel->offsets.m_nUDmaAddrH = (VD_1_U_DMA_ADDR_H + offset);
	channel->offsets.m_nVDmaAddrL = (VD_1_V_DMA_ADDR_L + offset);
	channel->offsets.m_nVDmaAddrH = (VD_1_V_DMA_ADDR_H + offset);
	channel->offsets.m_nYTSDmaAddrL = (VD_1_Y_TS_DMA_ADDR_L + offset);
	channel->offsets.m_nYTSDmaAddrH = (VD_1_Y_TS_DMA_ADDR_H + offset);

	/* Finally, a holding place for some default register values */
	memset(&channel->defaults, 0, sizeof(channel->defaults));

	/* Settings for 0820 */
	channel->defaults.m_nUXferCtrlFull = 0x00080000;
	channel->defaults.m_nUXferCtrlPart = 0x00080000;
	channel->defaults.m_nUTSXferCtrl   = 0x00080008;
	channel->defaults.m_nVXferCtrlFull = 0x000C0000;
	channel->defaults.m_nVXferCtrlPart = 0x000C0000;
	channel->defaults.m_nVTSXferCtrl   = 0x000C0008;

	if (channel->nr == 1) {
		channel->defaults.m_nYXferCtrlFull = 0x00040000;
		channel->defaults.m_nYXferCtrlPart = 0x00040000;
		channel->defaults.m_nYTSXferCtrl = 0x00040010;
	} else
	if (channel->nr == 2) {
		channel->defaults.m_nYXferCtrlFull = 0x00200000;
		channel->defaults.m_nYXferCtrlPart = 0x00200000;
		channel->defaults.m_nYTSXferCtrl = 0x00200010;
		channel->defaults.m_nUXferCtrlFull = 0x00240000;
		channel->defaults.m_nUXferCtrlPart = 0x00240000;
		channel->defaults.m_nUTSXferCtrl = 0x00240008;
		channel->defaults.m_nVXferCtrlFull = 0x00280000;
		channel->defaults.m_nVXferCtrlPart = 0x00280000;
		channel->defaults.m_nVTSXferCtrl = 0x00280008;
	} else
	if (channel->nr == 3) {
		channel->defaults.m_nYXferCtrlFull = 0x003c0000;
		channel->defaults.m_nYXferCtrlPart = 0x003c0000;
		channel->defaults.m_nYTSXferCtrl = 0x003c0010;
		channel->defaults.m_nUXferCtrlFull = 0x00400000;
		channel->defaults.m_nUXferCtrlPart = 0x00400000;
		channel->defaults.m_nUTSXferCtrl = 0x00400008;
		channel->defaults.m_nVXferCtrlFull = 0x00440000;
		channel->defaults.m_nVXferCtrlPart = 0x00440000;
		channel->defaults.m_nVTSXferCtrl = 0x00440008;
	} else
	if (channel->nr == 4) {
		channel->defaults.m_nYXferCtrlFull = 0x00580000;
		channel->defaults.m_nYXferCtrlPart = 0x00580000;
		channel->defaults.m_nYTSXferCtrl = 0x00580010;
		channel->defaults.m_nUXferCtrlFull = 0x005c0000;
		channel->defaults.m_nUXferCtrlPart = 0x005c0000;
		channel->defaults.m_nUTSXferCtrl = 0x005c0008;
		channel->defaults.m_nVXferCtrlFull = 0x00600000;
		channel->defaults.m_nVXferCtrlPart = 0x00600000;
		channel->defaults.m_nVTSXferCtrl = 0x00600008;
	} else
	if (channel->nr == 9) {
	} else
		BUG();

}

static void vc8x0_dma_configure_channel9_registers_audio(struct vc8x0_dma_channel *channel)
{
	u32 offset = (channel->nr - 9) * (AD_2_XFER_CNT - AD_1_XFER_CNT);
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* TODO: Hardcoded assumptions about unbalcned audio here */
	/* TODO / BUG : Hardcoded */
	offset = 0; /* 56c0 */
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* Flex DMA registers ready for use with vc_write32() */
	channel->regs.m_nXferCnt = (AD_5_XFER_CNT + offset) << 2;
	channel->regs.m_nDestXferCnt = (AD_5_DEST_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1XferStepSize = (AD_5_SP1_XFER_STEP_SIZE + offset) << 2;
	channel->regs.m_nSP1XferCtrl = (AD_5_SP1_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP1TSXferCtrl = (AD_5_SP1_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nFrameCnt = (AD_5_FRAME_CNT + offset) << 2;
	channel->regs.m_nStreamDisabled = (AD_5_STREAM_DISABLED + offset) << 2;
	channel->regs.m_nLastFrameIndex = (AD_5_LAST_FRAME_INDEX + offset) << 2;
	channel->regs.m_nLastAddrPtr = (AD_5_LAST_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestAddrPtr = (AD_5_DEST_ADDR_PTR + offset) << 2;
	channel->regs.m_nLastTSAddrPtr = (AD_5_LAST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestTSAddrPtr = (AD_5_DEST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nFrameXferCnt = (AD_5_FRAME_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1DestAddrL = (AD_5_SP1_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DestAddrH = (AD_5_SP1_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDestAddrL = (AD_5_SP1_TS_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDestAddrH = (AD_5_SP1_TS_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1DmaAddrL = (AD_5_SP1_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DmaAddrH = (AD_5_SP1_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrL = (AD_5_SP1_TS_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrH = (AD_5_SP1_TS_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nFifoGo = 0x928; /* TODO, fix me */
	channel->regs.m_nMisc = 0x930; /* TODO, fix me */

	/* Flex DMA registers ready for use by other flex registers, and have
	 * different offset calculations. */
	/* TODO: Large assumptions about offsets here */
	offset = 0; /* 56c0 */
	channel->offsets.m_nXferCnt = (AD_5_XFER_CNT + offset);
	channel->offsets.m_nDestXferCnt = (AD_5_DEST_XFER_CNT + offset);
	channel->offsets.m_nSP1XferStepSize = (AD_5_SP1_XFER_STEP_SIZE + offset);
	channel->offsets.m_nSP1XferCtrl = (AD_5_SP1_XFER_CTRL + offset);
	channel->offsets.m_nSP1TSXferCtrl = (AD_5_SP1_TS_XFER_CTRL + offset);
	channel->offsets.m_nFrameCnt = (AD_5_FRAME_CNT + offset);
	channel->offsets.m_nStreamDisabled = (AD_5_STREAM_DISABLED + offset);
	channel->offsets.m_nLastFrameIndex = (AD_5_LAST_FRAME_INDEX + offset);
	channel->offsets.m_nLastAddrPtr = (AD_5_LAST_ADDR_PTR + offset);
	channel->offsets.m_nDestAddrPtr = (AD_5_DEST_ADDR_PTR + offset);
	channel->offsets.m_nLastTSAddrPtr = (AD_5_LAST_TS_ADDR_PTR + offset);
	channel->offsets.m_nDestTSAddrPtr = (AD_5_DEST_TS_ADDR_PTR + offset);
	channel->offsets.m_nFrameXferCnt = (AD_5_FRAME_XFER_CNT + offset);
	channel->offsets.m_nSP1DestAddrL = (AD_5_SP1_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1DestAddrH = (AD_5_SP1_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1TSDestAddrL = (AD_5_SP1_TS_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1TSDestAddrH = (AD_5_SP1_TS_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1DmaAddrL = (AD_5_SP1_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1DmaAddrH = (AD_1_SP1_DMA_ADDR_H + offset);
	channel->offsets.m_nSP1TSDmaAddrL = (AD_5_SP1_TS_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1TSDmaAddrH = (AD_5_SP1_TS_DMA_ADDR_H + offset);

	/* Finally, a holding place for some default register values */
	memset(&channel->defaults, 0, sizeof(channel->defaults));

	if ((channel->nr == 5) || (channel->nr == 9)) {
		/* Unbalanced Audio */
		channel->defaults.m_nSP1XferCtrl   = 0x00740000;
		channel->defaults.m_nSP1TSXferCtrl = 0x00740008;
	} else
	if (channel->nr == 10) {
		channel->defaults.m_nSP1XferCtrl = 0x002C0000;
		channel->defaults.m_nSP1TSXferCtrl = 0x002C0008;
		channel->defaults.m_nSP2XferCtrl = 0x00300000;
		channel->defaults.m_nSP2TSXferCtrl = 0x00300008;
		channel->defaults.m_nSP3XferCtrl = 0x00340000;
		channel->defaults.m_nSP3TSXferCtrl = 0x00340008;
		channel->defaults.m_nSP4XferCtrl = 0x00380000;
		channel->defaults.m_nSP4TSXferCtrl = 0x00380008;
	} else
		BUG();

}

static void vc8x0_dma_configure_channel5_registers_audio(struct vc8x0_dma_channel *channel)
{
	u32 offset = (channel->nr - 5) * (AD_2_XFER_CNT - AD_1_XFER_CNT);
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* TODO: Hardcoded assumptions about unbalcned audio here */
	/* TODO / BUG : Hardcoded */
	offset = 0; /* 5380 */
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* Flex DMA registers ready for use with vc_write32() */
	channel->regs.m_nXferCnt = (AD_1_XFER_CNT + offset) << 2;
	channel->regs.m_nDestXferCnt = (AD_1_DEST_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1XferStepSize = (AD_1_SP1_XFER_STEP_SIZE + offset) << 2;
	channel->regs.m_nSP1XferCtrl = (AD_1_SP1_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2XferCtrl = (AD_1_SP2_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3XferCtrl = (AD_1_SP3_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4XferCtrl = (AD_1_SP4_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP1TSXferCtrl = (AD_1_SP1_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2TSXferCtrl = (AD_1_SP2_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3TSXferCtrl = (AD_1_SP3_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4TSXferCtrl = (AD_1_SP4_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nFrameCnt = (AD_1_FRAME_CNT + offset) << 2;
	channel->regs.m_nStreamDisabled = (AD_1_STREAM_DISABLED + offset) << 2;
	channel->regs.m_nLastFrameIndex = (AD_1_LAST_FRAME_INDEX + offset) << 2;
	channel->regs.m_nLastAddrPtr = (AD_1_LAST_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestAddrPtr = (AD_1_DEST_ADDR_PTR + offset) << 2;
	channel->regs.m_nLastTSAddrPtr = (AD_1_LAST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestTSAddrPtr = (AD_1_DEST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nFrameXferCnt = (AD_1_FRAME_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1DestAddrL = (AD_1_SP1_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DestAddrH = (AD_1_SP1_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDestAddrL = (AD_1_SP1_TS_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDestAddrH = (AD_1_SP1_TS_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1DmaAddrL = (AD_1_SP1_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DmaAddrH = (AD_1_SP1_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrL = (AD_1_SP1_TS_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrH = (AD_1_SP1_TS_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nFifoGo = 0x928; /* TODO, fix me */
	channel->regs.m_nMisc = 0x930; /* TODO, fix me */

	/* Flex DMA registers ready for use by other flex registers, and have
	 * different offset calculations. */
	/* TODO: Large assumptions about offsets here */
	offset = 0; /* 5380 */
	channel->offsets.m_nXferCnt = (AD_1_XFER_CNT + offset);
	channel->offsets.m_nDestXferCnt = (AD_1_DEST_XFER_CNT + offset);
	channel->offsets.m_nSP1XferStepSize = (AD_1_SP1_XFER_STEP_SIZE + offset);
	channel->offsets.m_nSP1XferCtrl = (AD_1_SP1_XFER_CTRL + offset);
	channel->offsets.m_nSP2XferCtrl = (AD_1_SP2_XFER_CTRL + offset);
	channel->offsets.m_nSP3XferCtrl = (AD_1_SP3_XFER_CTRL + offset);
	channel->offsets.m_nSP4XferCtrl = (AD_1_SP4_XFER_CTRL + offset);
	channel->offsets.m_nSP1TSXferCtrl = (AD_1_SP1_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP2TSXferCtrl = (AD_1_SP2_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP3TSXferCtrl = (AD_1_SP3_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP4TSXferCtrl = (AD_1_SP4_TS_XFER_CTRL + offset);
	channel->offsets.m_nFrameCnt = (AD_1_FRAME_CNT + offset);
	channel->offsets.m_nStreamDisabled = (AD_1_STREAM_DISABLED + offset);
	channel->offsets.m_nLastFrameIndex = (AD_1_LAST_FRAME_INDEX + offset);
	channel->offsets.m_nLastAddrPtr = (AD_1_LAST_ADDR_PTR + offset);
	channel->offsets.m_nDestAddrPtr = (AD_1_DEST_ADDR_PTR + offset);
	channel->offsets.m_nLastTSAddrPtr = (AD_1_LAST_TS_ADDR_PTR + offset);
	channel->offsets.m_nDestTSAddrPtr = (AD_1_DEST_TS_ADDR_PTR + offset);
	channel->offsets.m_nFrameXferCnt = (AD_1_FRAME_XFER_CNT + offset);
	channel->offsets.m_nSP1DestAddrL = (AD_1_SP1_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1DestAddrH = (AD_1_SP1_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1TSDestAddrL = (AD_1_SP1_TS_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1TSDestAddrH = (AD_1_SP1_TS_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1DmaAddrL = (AD_1_SP1_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1DmaAddrH = (AD_1_SP1_DMA_ADDR_H + offset);
	channel->offsets.m_nSP1TSDmaAddrL = (AD_1_SP1_TS_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1TSDmaAddrH = (AD_1_SP1_TS_DMA_ADDR_H + offset);

	/* Finally, a holding place for some default register values */
	memset(&channel->defaults, 0, sizeof(channel->defaults));

	if (channel->nr == 5) {
		channel->defaults.m_nSP1XferCtrl   = 0x00100040;
		channel->defaults.m_nSP2XferCtrl   = 0x00140040;
		channel->defaults.m_nSP3XferCtrl   = 0x00180040;
		channel->defaults.m_nSP4XferCtrl   = 0x001c0040;
		channel->defaults.m_nSP1TSXferCtrl = 0x00100008;
		channel->defaults.m_nSP2TSXferCtrl = 0x00140008;
		channel->defaults.m_nSP3TSXferCtrl = 0x00180008;
		channel->defaults.m_nSP4TSXferCtrl = 0x001c0008;
	} else
		BUG();

}

static void vc8x0_dma_configure_channel6_registers_audio(struct vc8x0_dma_channel *channel)
{
	u32 offset;

	/* TODO: Hardcoded assumptions about unbalcned audio here */
	/* TODO / BUG : Hardcoded */
	offset = 0; /* 5380 */
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* Flex DMA registers ready for use with vc_write32() */
	channel->regs.m_nXferCnt = (AD_2_XFER_CNT + offset) << 2;
	channel->regs.m_nDestXferCnt = (AD_2_DEST_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1XferStepSize = (AD_2_SP1_XFER_STEP_SIZE + offset) << 2;
	channel->regs.m_nSP1XferCtrl = (AD_2_SP1_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2XferCtrl = (AD_2_SP2_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3XferCtrl = (AD_2_SP3_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4XferCtrl = (AD_2_SP4_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP1TSXferCtrl = (AD_2_SP1_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2TSXferCtrl = (AD_2_SP2_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3TSXferCtrl = (AD_2_SP3_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4TSXferCtrl = (AD_2_SP4_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nFrameCnt = (AD_2_FRAME_CNT + offset) << 2;
	channel->regs.m_nStreamDisabled = (AD_2_STREAM_DISABLED + offset) << 2;
	channel->regs.m_nLastFrameIndex = (AD_2_LAST_FRAME_INDEX + offset) << 2;
	channel->regs.m_nLastAddrPtr = (AD_2_LAST_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestAddrPtr = (AD_2_DEST_ADDR_PTR + offset) << 2;
	channel->regs.m_nLastTSAddrPtr = (AD_2_LAST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestTSAddrPtr = (AD_2_DEST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nFrameXferCnt = (AD_2_FRAME_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1DestAddrL = (AD_2_SP1_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DestAddrH = (AD_2_SP1_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDestAddrL = (AD_2_SP1_TS_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDestAddrH = (AD_2_SP1_TS_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1DmaAddrL = (AD_2_SP1_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DmaAddrH = (AD_2_SP1_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrL = (AD_2_SP1_TS_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrH = (AD_2_SP1_TS_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nFifoGo = 0x928; /* TODO, fix me */
	channel->regs.m_nMisc = 0x930; /* TODO, fix me */

	/* Flex DMA registers ready for use by other flex registers, and have
	 * different offset calculations. */
	/* TODO: Large assumptions about offsets here */
	offset = 0; /* 5380 */
	channel->offsets.m_nXferCnt = (AD_2_XFER_CNT + offset);
	channel->offsets.m_nDestXferCnt = (AD_2_DEST_XFER_CNT + offset);
	channel->offsets.m_nSP1XferStepSize = (AD_2_SP1_XFER_STEP_SIZE + offset);
	channel->offsets.m_nSP1XferCtrl = (AD_2_SP1_XFER_CTRL + offset);
	channel->offsets.m_nSP2XferCtrl = (AD_2_SP2_XFER_CTRL + offset);
	channel->offsets.m_nSP3XferCtrl = (AD_2_SP3_XFER_CTRL + offset);
	channel->offsets.m_nSP4XferCtrl = (AD_2_SP4_XFER_CTRL + offset);
	channel->offsets.m_nSP1TSXferCtrl = (AD_2_SP1_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP2TSXferCtrl = (AD_2_SP2_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP3TSXferCtrl = (AD_2_SP3_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP4TSXferCtrl = (AD_2_SP4_TS_XFER_CTRL + offset);
	channel->offsets.m_nFrameCnt = (AD_2_FRAME_CNT + offset);
	channel->offsets.m_nStreamDisabled = (AD_2_STREAM_DISABLED + offset);
	channel->offsets.m_nLastFrameIndex = (AD_2_LAST_FRAME_INDEX + offset);
	channel->offsets.m_nLastAddrPtr = (AD_2_LAST_ADDR_PTR + offset);
	channel->offsets.m_nDestAddrPtr = (AD_2_DEST_ADDR_PTR + offset);
	channel->offsets.m_nLastTSAddrPtr = (AD_2_LAST_TS_ADDR_PTR + offset);
	channel->offsets.m_nDestTSAddrPtr = (AD_2_DEST_TS_ADDR_PTR + offset);
	channel->offsets.m_nFrameXferCnt = (AD_2_FRAME_XFER_CNT + offset);
	channel->offsets.m_nSP1DestAddrL = (AD_2_SP1_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1DestAddrH = (AD_2_SP1_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1TSDestAddrL = (AD_2_SP1_TS_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1TSDestAddrH = (AD_2_SP1_TS_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1DmaAddrL = (AD_2_SP1_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1DmaAddrH = (AD_2_SP1_DMA_ADDR_H + offset);
	channel->offsets.m_nSP1TSDmaAddrL = (AD_2_SP1_TS_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1TSDmaAddrH = (AD_2_SP1_TS_DMA_ADDR_H + offset);

	/* Finally, a holding place for some default register values */
	memset(&channel->defaults, 0, sizeof(channel->defaults));

	if (channel->nr == 6) {
		channel->defaults.m_nSP1XferCtrl   = 0x002c0000;
		channel->defaults.m_nSP1TSXferCtrl = 0x002c0008;
		channel->defaults.m_nSP2XferCtrl   = 0x00300000;
		channel->defaults.m_nSP2TSXferCtrl = 0x00300008;
		channel->defaults.m_nSP3XferCtrl   = 0x00340000;
		channel->defaults.m_nSP3TSXferCtrl = 0x00340008;
		channel->defaults.m_nSP4XferCtrl   = 0x00380000;
		channel->defaults.m_nSP4TSXferCtrl = 0x00380008;
	} else
		BUG();

}

static void vc8x0_dma_configure_channel7_registers_audio(struct vc8x0_dma_channel *channel)
{
	u32 offset;

	/* TODO: Hardcoded assumptions about unbalcned audio here */
	/* TODO / BUG : Hardcoded */
	offset = 0; /* 5380 */
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* Flex DMA registers ready for use with vc_write32() */
	channel->regs.m_nXferCnt = (AD_3_XFER_CNT + offset) << 2;
	channel->regs.m_nDestXferCnt = (AD_3_DEST_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1XferStepSize = (AD_3_SP1_XFER_STEP_SIZE + offset) << 2;
	channel->regs.m_nSP1XferCtrl = (AD_3_SP1_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2XferCtrl = (AD_3_SP2_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3XferCtrl = (AD_3_SP3_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4XferCtrl = (AD_3_SP4_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP1TSXferCtrl = (AD_3_SP1_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2TSXferCtrl = (AD_3_SP2_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3TSXferCtrl = (AD_3_SP3_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4TSXferCtrl = (AD_3_SP4_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nFrameCnt = (AD_3_FRAME_CNT + offset) << 2;
	channel->regs.m_nStreamDisabled = (AD_3_STREAM_DISABLED + offset) << 2;
	channel->regs.m_nLastFrameIndex = (AD_3_LAST_FRAME_INDEX + offset) << 2;
	channel->regs.m_nLastAddrPtr = (AD_3_LAST_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestAddrPtr = (AD_3_DEST_ADDR_PTR + offset) << 2;
	channel->regs.m_nLastTSAddrPtr = (AD_3_LAST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestTSAddrPtr = (AD_3_DEST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nFrameXferCnt = (AD_3_FRAME_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1DestAddrL = (AD_3_SP1_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DestAddrH = (AD_3_SP1_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDestAddrL = (AD_3_SP1_TS_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDestAddrH = (AD_3_SP1_TS_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1DmaAddrL = (AD_3_SP1_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DmaAddrH = (AD_3_SP1_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrL = (AD_3_SP1_TS_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrH = (AD_3_SP1_TS_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nFifoGo = 0x928; /* TODO, fix me */
	channel->regs.m_nMisc = 0x930; /* TODO, fix me */

	/* Flex DMA registers ready for use by other flex registers, and have
	 * different offset calculations. */
	/* TODO: Large assumptions about offsets here */
	offset = 0; /* 5380 */
	channel->offsets.m_nXferCnt = (AD_3_XFER_CNT + offset);
	channel->offsets.m_nDestXferCnt = (AD_3_DEST_XFER_CNT + offset);
	channel->offsets.m_nSP1XferStepSize = (AD_3_SP1_XFER_STEP_SIZE + offset);
	channel->offsets.m_nSP1XferCtrl = (AD_3_SP1_XFER_CTRL + offset);
	channel->offsets.m_nSP2XferCtrl = (AD_3_SP2_XFER_CTRL + offset);
	channel->offsets.m_nSP3XferCtrl = (AD_3_SP3_XFER_CTRL + offset);
	channel->offsets.m_nSP4XferCtrl = (AD_3_SP4_XFER_CTRL + offset);
	channel->offsets.m_nSP1TSXferCtrl = (AD_3_SP1_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP2TSXferCtrl = (AD_3_SP2_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP3TSXferCtrl = (AD_3_SP3_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP4TSXferCtrl = (AD_3_SP4_TS_XFER_CTRL + offset);
	channel->offsets.m_nFrameCnt = (AD_3_FRAME_CNT + offset);
	channel->offsets.m_nStreamDisabled = (AD_3_STREAM_DISABLED + offset);
	channel->offsets.m_nLastFrameIndex = (AD_3_LAST_FRAME_INDEX + offset);
	channel->offsets.m_nLastAddrPtr = (AD_3_LAST_ADDR_PTR + offset);
	channel->offsets.m_nDestAddrPtr = (AD_3_DEST_ADDR_PTR + offset);
	channel->offsets.m_nLastTSAddrPtr = (AD_3_LAST_TS_ADDR_PTR + offset);
	channel->offsets.m_nDestTSAddrPtr = (AD_3_DEST_TS_ADDR_PTR + offset);
	channel->offsets.m_nFrameXferCnt = (AD_3_FRAME_XFER_CNT + offset);
	channel->offsets.m_nSP1DestAddrL = (AD_3_SP1_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1DestAddrH = (AD_3_SP1_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1TSDestAddrL = (AD_3_SP1_TS_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1TSDestAddrH = (AD_3_SP1_TS_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1DmaAddrL = (AD_3_SP1_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1DmaAddrH = (AD_3_SP1_DMA_ADDR_H + offset);
	channel->offsets.m_nSP1TSDmaAddrL = (AD_3_SP1_TS_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1TSDmaAddrH = (AD_3_SP1_TS_DMA_ADDR_H + offset);

	/* Finally, a holding place for some default register values */
	memset(&channel->defaults, 0, sizeof(channel->defaults));

	if (channel->nr == 7) {
		channel->defaults.m_nSP1XferCtrl   = 0x00480000;
		channel->defaults.m_nSP1TSXferCtrl = 0x00480008;
		channel->defaults.m_nSP2XferCtrl   = 0x004c0000;
		channel->defaults.m_nSP2TSXferCtrl = 0x004c0008;
		channel->defaults.m_nSP3XferCtrl   = 0x00500000;
		channel->defaults.m_nSP3TSXferCtrl = 0x00500008;
		channel->defaults.m_nSP4XferCtrl   = 0x00540000;
		channel->defaults.m_nSP4TSXferCtrl = 0x00540008;
	} else
		BUG();

}

static void vc8x0_dma_configure_channel8_registers_audio(struct vc8x0_dma_channel *channel)
{
	u32 offset;

	/* TODO: Hardcoded assumptions about unbalcned audio here */
	/* TODO / BUG : Hardcoded */
	offset = 0; /* 5380 */
	offset |= (DPRAM_INSTRUCTION >> 2); 

	/* Flex DMA registers ready for use with vc_write32() */
	channel->regs.m_nXferCnt = (AD_4_XFER_CNT + offset) << 2;
	channel->regs.m_nDestXferCnt = (AD_4_DEST_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1XferStepSize = (AD_4_SP1_XFER_STEP_SIZE + offset) << 2;
	channel->regs.m_nSP1XferCtrl = (AD_4_SP1_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2XferCtrl = (AD_4_SP2_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3XferCtrl = (AD_4_SP3_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4XferCtrl = (AD_4_SP4_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP1TSXferCtrl = (AD_4_SP1_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP2TSXferCtrl = (AD_4_SP2_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP3TSXferCtrl = (AD_4_SP3_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nSP4TSXferCtrl = (AD_4_SP4_TS_XFER_CTRL + offset) << 2;
	channel->regs.m_nFrameCnt = (AD_4_FRAME_CNT + offset) << 2;
	channel->regs.m_nStreamDisabled = (AD_4_STREAM_DISABLED + offset) << 2;
	channel->regs.m_nLastFrameIndex = (AD_4_LAST_FRAME_INDEX + offset) << 2;
	channel->regs.m_nLastAddrPtr = (AD_4_LAST_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestAddrPtr = (AD_4_DEST_ADDR_PTR + offset) << 2;
	channel->regs.m_nLastTSAddrPtr = (AD_4_LAST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nDestTSAddrPtr = (AD_4_DEST_TS_ADDR_PTR + offset) << 2;
	channel->regs.m_nFrameXferCnt = (AD_4_FRAME_XFER_CNT + offset) << 2;
	channel->regs.m_nSP1DestAddrL = (AD_4_SP1_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DestAddrH = (AD_4_SP1_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDestAddrL = (AD_4_SP1_TS_DEST_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDestAddrH = (AD_4_SP1_TS_DEST_ADDR_H + offset) << 2;
	channel->regs.m_nSP1DmaAddrL = (AD_4_SP1_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1DmaAddrH = (AD_4_SP1_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrL = (AD_4_SP1_TS_DMA_ADDR_L + offset) << 2;
	channel->regs.m_nSP1TSDmaAddrH = (AD_4_SP1_TS_DMA_ADDR_H + offset) << 2;
	channel->regs.m_nFifoGo = 0x928; /* TODO, fix me */
	channel->regs.m_nMisc = 0x930; /* TODO, fix me */

	/* Flex DMA registers ready for use by other flex registers, and have
	 * different offset calculations. */
	/* TODO: Large assumptions about offsets here */
	offset = 0; /* 5380 */
	channel->offsets.m_nXferCnt = (AD_4_XFER_CNT + offset);
	channel->offsets.m_nDestXferCnt = (AD_4_DEST_XFER_CNT + offset);
	channel->offsets.m_nSP1XferStepSize = (AD_4_SP1_XFER_STEP_SIZE + offset);
	channel->offsets.m_nSP1XferCtrl = (AD_4_SP1_XFER_CTRL + offset);
	channel->offsets.m_nSP2XferCtrl = (AD_4_SP2_XFER_CTRL + offset);
	channel->offsets.m_nSP3XferCtrl = (AD_4_SP3_XFER_CTRL + offset);
	channel->offsets.m_nSP4XferCtrl = (AD_4_SP4_XFER_CTRL + offset);
	channel->offsets.m_nSP1TSXferCtrl = (AD_4_SP1_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP2TSXferCtrl = (AD_4_SP2_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP3TSXferCtrl = (AD_4_SP3_TS_XFER_CTRL + offset);
	channel->offsets.m_nSP4TSXferCtrl = (AD_4_SP4_TS_XFER_CTRL + offset);
	channel->offsets.m_nFrameCnt = (AD_4_FRAME_CNT + offset);
	channel->offsets.m_nStreamDisabled = (AD_4_STREAM_DISABLED + offset);
	channel->offsets.m_nLastFrameIndex = (AD_4_LAST_FRAME_INDEX + offset);
	channel->offsets.m_nLastAddrPtr = (AD_4_LAST_ADDR_PTR + offset);
	channel->offsets.m_nDestAddrPtr = (AD_4_DEST_ADDR_PTR + offset);
	channel->offsets.m_nLastTSAddrPtr = (AD_4_LAST_TS_ADDR_PTR + offset);
	channel->offsets.m_nDestTSAddrPtr = (AD_4_DEST_TS_ADDR_PTR + offset);
	channel->offsets.m_nFrameXferCnt = (AD_4_FRAME_XFER_CNT + offset);
	channel->offsets.m_nSP1DestAddrL = (AD_4_SP1_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1DestAddrH = (AD_4_SP1_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1TSDestAddrL = (AD_4_SP1_TS_DEST_ADDR_L + offset);
	channel->offsets.m_nSP1TSDestAddrH = (AD_4_SP1_TS_DEST_ADDR_H + offset);
	channel->offsets.m_nSP1DmaAddrL = (AD_4_SP1_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1DmaAddrH = (AD_4_SP1_DMA_ADDR_H + offset);
	channel->offsets.m_nSP1TSDmaAddrL = (AD_4_SP1_TS_DMA_ADDR_L + offset);
	channel->offsets.m_nSP1TSDmaAddrH = (AD_4_SP1_TS_DMA_ADDR_H + offset);

	/* Finally, a holding place for some default register values */
	memset(&channel->defaults, 0, sizeof(channel->defaults));

	if (channel->nr == 8) {
		channel->defaults.m_nSP1XferCtrl   = 0x00640000;
		channel->defaults.m_nSP1TSXferCtrl = 0x00640008;
		channel->defaults.m_nSP2XferCtrl   = 0x00680000;
		channel->defaults.m_nSP2TSXferCtrl = 0x00680008;
		channel->defaults.m_nSP3XferCtrl   = 0x006c0000;
		channel->defaults.m_nSP3TSXferCtrl = 0x006c0008;
		channel->defaults.m_nSP4XferCtrl   = 0x00700000;
		channel->defaults.m_nSP4TSXferCtrl = 0x00700008;
	} else
		BUG();

}

void vc8x0_dma_configure_channel_registers(struct vc8x0_dma_channel *channel)
{
	if (channel->type == TYPE_VIDEO)
		vc8x0_dma_configure_channel_registers_video(channel);
	else
	if (channel->type == TYPE_AUDIO) {
		if (IS_820E(channel->dev))
			vc8x0_dma_configure_channel9_registers_audio(channel);
		if (IS_845E(channel->dev)) {
			if (channel->nr == 5)
				vc8x0_dma_configure_channel5_registers_audio(channel);
			if (channel->nr == 6)
				vc8x0_dma_configure_channel6_registers_audio(channel);
			if (channel->nr == 7)
				vc8x0_dma_configure_channel7_registers_audio(channel);
			if (channel->nr == 8)
				vc8x0_dma_configure_channel8_registers_audio(channel);
		}
	}
}

int vc8x0_dma_loadsequencer(struct vc8x0_dev *dev)
{
	struct vc8x0_dma_channel *channel;

	u32 i, v, reg;
	int ret = 0, flexlen = sizeof(vc8x0_FlexDMAProgram) / sizeof(u32);
	dprintk(1, "%s()\n", __func__);

	dprintk(1, "%s() DMA Sequencer program contains %d instructions\n", __func__, flexlen);

	do {
		/* Load the FLEX program into the PCIe bridge */
		for (i = 0, reg = DPRAM_INSTRUCTION; i < flexlen; i++, reg += 4) {
			vc_write32(reg, vc8x0_FlexDMAProgram[ i ]);
		}

		/* Verify the FLEX program */
		for (i = 0, reg = DPRAM_INSTRUCTION; i < flexlen; i++, reg += 4) {
			v = vc_read32(reg);
			if (v != vc8x0_FlexDMAProgram[ i ]) {
				printk(KERN_ERR "%s() %x != %x\n", __func__, v, vc8x0_FlexDMAProgram[ i ]);
				ret = -EINVAL;
				break;
			}
		}

		vc_write32(DPRAM_INSTRUCTION + (MINUS_1 << 2), -1);
		vc_write32(DPRAM_INSTRUCTION + (MINUS_2 << 2), -2);
		vc_write32(DPRAM_INSTRUCTION + (MINUS_6 << 2), -6);
		vc_write32(DPRAM_INSTRUCTION + (MINUS_8 << 2), -8);
		vc_write32(DPRAM_INSTRUCTION + (PLUS_8  << 2),  8);
		vc_write32(DPRAM_INSTRUCTION + (PLUS_16 << 2), 16);
		
		/* Ensure the video streams are disabled */
		for (i = 0; i < MAX_DMA_CHANNELS; i++) {
			channel = &dev->dma_channel[ i ];
			if (IS_CHANNEL_CONFIGURED(channel))
				vc_write32(channel->regs.m_nStreamDisabled, 1);
		}

		vc_write32(DPRAM_INSTRUCTION + (EVENTS_TO_FIRE << 2), 0);
		vc_write32(DPRAM_INSTRUCTION + (SIG_EVENT_CODE << 2), VDMA_SIG_EVENT(0, 1, 0));
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_0 << 2), VDMA_EVENT_0);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_1 << 2), VDMA_EVENT_1);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_2 << 2), VDMA_EVENT_2);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_3 << 2), VDMA_EVENT_3);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_4 << 2), VDMA_EVENT_4);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_5 << 2), VDMA_EVENT_5);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_6 << 2), VDMA_EVENT_6);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_7 << 2), VDMA_EVENT_7);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_8 << 2), VDMA_EVENT_8);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_9 << 2), VDMA_EVENT_9);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_10 << 2), VDMA_EVENT_10);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_11 << 2), VDMA_EVENT_11);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_12 << 2), VDMA_EVENT_12);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_13 << 2), VDMA_EVENT_13);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_14 << 2), VDMA_EVENT_14);
		vc_write32(DPRAM_INSTRUCTION + (EVENT_ID_15 << 2), VDMA_EVENT_15);
		vc_write32(DPRAM_INSTRUCTION + (ZERO << 2), 0);
		vc_write32(DPRAM_INSTRUCTION + (AUDIO_CHANNEL_SIZE << 2), 1024 * 16);

	} while (0);

	if (ret == 0)
		dprintk(1, "%s() DMA Sequencer loaded successfully\n", __func__);
	else
		printk(KERN_ERR "%s() DMA Sequencer failed at step %d\n", __func__, i);

	return ret;
}

/* Initialize the DMA risc processor back to the first instruction */
/* Start the RISC engine */
int vc8x0_dma_sequencer_start(struct vc8x0_dev *dev)
{
	dprintk(1, "%s()\n", __func__);

	if (dev->risc_running)
		BUG();

	dev->risc_running = 1;

	/* Reset the RISC instruction pointers */
	vc_write32(VDMA_DPTR, 0);

	/* Start the RISC processor */
	vc_write32(VDMA_CSR, VDMA_CSR_IPR);

	/* Start GN4124 Interrupts */
	gn_write(GN4124_GPIO_INT_ENABLE, 0x00000001);

	return 0;
}

/* Stop the RISC engine */
int vc8x0_dma_sequencer_stop(struct vc8x0_dev *dev)
{
	dprintk(1, "%s()\n", __func__);

	if (dev->risc_running == 0)
		return 0;

	dev->risc_running = 0;

	/* Stop GN4124 Interrupts */
	gn_write(GN4124_GPIO_INT_DISABLE, 0x00000001);

	/* Stop the RISC processor */
	vc_write32(VDMA_CSR, VDMA_CSR_HALT);
	vc_write32(VDMA_EVENT_CLR, VDMA_EVENT_ALL);
	return 0;
}

int vc8x0_dma_configure_video(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_dma_buffer *buf, *tsbuf = 0;
	struct list_head *p, *q, *x, *y;
	struct vc8x0_format *fmt = channel->fmt;
	u32 v, xfersize = 0, addr, tsaddr, offset;
	unsigned long flags;
	u32 nVDMAAddrOffset = 0;
	u32 nUDMAAddrOffset = 0;
	u32 nBlockXferCount = 0;
	u32 vbi_total_lines = 0;

	dprintk(1, "%s(nr=%d)\n", __func__, channel->nr);

	if (fmt == 0) {
		printk(KERN_ERR "%s() fmt is unset\n", __func__);
		return -EINVAL;
	}

	channel->m_nDMABuffers = MAX_VIDEO_DSP_FRAMES;
	channel->m_nBlockSize = 512;

	if (channel->vbi_enabled)
		vbi_total_lines = fmt->vbi_total_lines;

	/* Configure the complete and total block size for each DMA transferred buffer */
	vc_write32(channel->regs.m_nYXferSizeFull, channel->m_nBlockSize); // 512
	vc_write32(channel->regs.m_nUVXferSizeFull, channel->m_nBlockSize / 2);
	xfersize = (((fmt->width * 2) * (fmt->height + vbi_total_lines)) / channel->m_nBlockSize) / 2;
	dprintk(1, "%s() calculated xfersize = 0x%x\n", __func__, xfersize);

	vc_write32(channel->regs.m_nXferCnt, xfersize);
	vc_write32(channel->regs.m_nDestXferCnt, xfersize);
	channel->defaults.m_nXferCnt = xfersize;
	channel->defaults.m_nDestXferCnt = xfersize;

	dprintk(1, "%s(), Y xfer size full = %d, UV xfer size full = %d, Xfer cnt = %d",
		__func__,
		channel->m_nBlockSize,
		channel->m_nBlockSize / 2,
		xfersize);

	/* Set the Y, U, & V plane transfer control values */
	v = channel->defaults.m_nYXferCtrlFull | (channel->m_nBlockSize & 0x00000FFF);
	vc_write32(channel->regs.m_nYXferCtrlFull, v);

	v = channel->defaults.m_nUXferCtrlFull | ((channel->m_nBlockSize / 2) & 0x00000FFF);
	vc_write32(channel->regs.m_nUXferCtrlFull, v);

	v = channel->defaults.m_nVXferCtrlFull | ((channel->m_nBlockSize / 2) & 0x00000FFF);
	vc_write32(channel->regs.m_nVXferCtrlFull, v);


	/* Compute the remaining Y pixels */
	v = (fmt->width * (fmt->height + vbi_total_lines)) % channel->m_nBlockSize;
	vc_write32(channel->regs.m_nYXferSizePart, v);

	/* Set the Y, U, & V plane transfer control values */
	/* BUG? */
	vc_write32(channel->regs.m_nYXferCtrlPart, channel->defaults.m_nYXferCtrlPart | v);
	dprintk(1, "Y xfer size PART = %d, UV xfer size PART = %d\n", v, v / 2);

	/* Compute the remaining U & V pixels */
	v /= 2;
	vc_write32(channel->regs.m_nUVXferSizePart, v & 0x00000FFF);
	vc_write32(channel->regs.m_nUXferCtrlPart, channel->defaults.m_nUXferCtrlPart | (v & 0x00000fff));
	vc_write32(channel->regs.m_nVXferCtrlPart, channel->defaults.m_nVXferCtrlPart | (v & 0x00000fff));

	/* Set the Y, U, & V timestamp transfer control values */
	/* ? defauls or offsets here? */
	vc_write32(channel->regs.m_nYTSXferCtrl, channel->defaults.m_nYTSXferCtrl);
	vc_write32(channel->regs.m_nUTSXferCtrl, channel->defaults.m_nUTSXferCtrl);
	vc_write32(channel->regs.m_nVTSXferCtrl, channel->defaults.m_nVTSXferCtrl);

	/* Set the # of DMA frames for this FIFO */
	vc_write32(channel->regs.m_nFrameCnt, channel->m_nDMABuffers);
	vc_write32(channel->regs.m_nFrameXferCnt, channel->m_nDMABuffers);
	vc_write32(channel->regs.m_nLastFrameIndex, channel->m_nDMABuffers);
	/* m_arrLastFrameIndexId[nFIFOIndex] = m_arrVideoSequencerVariableValues[nSubDeviceIndex].Video.m_nLastFrameIndex; */
	channel->m_arrLastFrameIndexId = channel->regs.m_nLastFrameIndex;

	/* Set the dma info for video channel 0 */
	nBlockXferCount = (fmt->width * (fmt->height + vbi_total_lines)) / channel->m_nBlockSize;
	nBlockXferCount += ((fmt->width * (fmt->height + vbi_total_lines)) % channel->m_nBlockSize) ? 1 : 0;
	nUDMAAddrOffset = nBlockXferCount * channel->m_nBlockSize;

	nBlockXferCount = ((fmt->width * (fmt->height + vbi_total_lines)) / 2) / channel->m_nBlockSize;
	nBlockXferCount += (((fmt->width * (fmt->height + vbi_total_lines)) / 2) % channel->m_nBlockSize) ? 1 : 0;
	nVDMAAddrOffset = nUDMAAddrOffset + (nBlockXferCount * channel->m_nBlockSize);

	/* We only ever have 1 timestamp buffer in the list, grab it for references purposes */
	/* We'll configure the video channel to refernece various parts of this buffer */
	spin_lock_irqsave(&channel->dma_timestamp_buffers_free_lock, flags);
	list_for_each_safe(x, y, &channel->dma_timestamp_buffers_free) {
		tsbuf = list_entry(x, struct vc8x0_dma_buffer, list);

		/* Configure the buffer plane pointers and sizes */
		tsbuf->dma_y_plane = tsbuf->cpu_dma;
		tsbuf->cpu_y_plane = tsbuf->cpu;
		tsbuf->cpu_y_plane_size = 0;

		break;
	}
	spin_unlock_irqrestore(&channel->dma_timestamp_buffers_free_lock, flags);
	BUG_ON(tsbuf == 0);

	vc_write32(channel->regs.m_nYTSDestAddrL, tsbuf->dma_y_plane); /* Physical address */
	vc_write32(channel->regs.m_nYTSDestAddrH, 0); /* Bits 32-63 */
	channel->defaults.m_nYTSDestAddrL = tsbuf->dma_y_plane;
	channel->defaults.m_nYTSDestAddrH = 0;

	spin_lock_irqsave(&channel->dma_buffers_free_lock, flags);
	list_for_each_safe(p, q, &channel->dma_buffers_free) {

		buf = list_entry(p, struct vc8x0_dma_buffer, list);
		BUG_ON(buf->nr >= MAX_VIDEO_DSP_FRAMES);

		/* Configure the buffer plane pointers and sizes */
		buf->dma_y_plane = buf->cpu_dma;
		buf->cpu_y_plane = buf->cpu;
		buf->cpu_y_plane_size = 0;

		buf->dma_u_plane = buf->cpu_dma + nUDMAAddrOffset;
		//buf->dma_u_plane = buf->cpu_dma;
		buf->cpu_u_plane = buf->cpu + nUDMAAddrOffset;
		buf->cpu_u_plane_size = 0;

		buf->dma_v_plane = buf->cpu_dma + nVDMAAddrOffset;
		//buf->dma_v_plane = buf->cpu_dma;
		buf->cpu_v_plane = buf->cpu + nVDMAAddrOffset;
		buf->cpu_v_plane_size = 0;

		/* Calculate this buffer FLEX address and program the flex engine */
		/* Each buffer 0-8 is given 6 DWORD for Y_addr_L, Y_addr_h, U... v... */
		addr = channel->regs.m_nYDmaAddrL;
		addr += (buf->nr * (6 * sizeof(u32)));
		offset = channel->offsets.m_nYDmaAddrL;
		offset += (buf->nr * 6);

		/* Each Timestamp buffer is given 2 dwords, so is 2 * u32 size */
		tsaddr = channel->regs.m_nYTSDmaAddrL;
		tsaddr += (buf->nr * (2 * sizeof(u32)));
		vc_write32(tsaddr, tsbuf->dma_y_plane + (buf->nr * 0x20));

		/* Write the Y/U/V DMA addresses */
		vc_write32(addr + (0 * sizeof(u32)), buf->dma_y_plane); /* PCI Physical ADDR 0-31bits */
		vc_write32(addr + (1 * sizeof(u32)), 0); /* Bits 32-63 */
		vc_write32(addr + (2 * sizeof(u32)), buf->dma_u_plane); /* PCI Physical ADDR 0-31bits */
		vc_write32(addr + (3 * sizeof(u32)), 0); /* Bits 32-63 */
		vc_write32(addr + (4 * sizeof(u32)), buf->dma_v_plane); /* PCI Physical ADDR 0-31bits */
		vc_write32(addr + (5 * sizeof(u32)), 0); /* Bits 32-63 */

		/* Ensure our last buffer is references by the FLEX as the first place to put
		 * video content */
		if (buf->nr == (MAX_VIDEO_DSP_FRAMES - 1)) {
			vc_write32(channel->regs.m_nYDestAddrL, buf->dma_y_plane);
			vc_write32(channel->regs.m_nYDestAddrH, 0);
			vc_write32(channel->regs.m_nUDestAddrL, buf->dma_u_plane);
			vc_write32(channel->regs.m_nUDestAddrH, 0);
			vc_write32(channel->regs.m_nVDestAddrL, buf->dma_v_plane);
			vc_write32(channel->regs.m_nVDestAddrH, 0);
			/* ... and remember the above values as defaults when we restart */
			channel->defaults.m_nYDestAddrL = buf->dma_y_plane;
			channel->defaults.m_nYDestAddrH = 0;
			channel->defaults.m_nUDestAddrL = buf->dma_u_plane;
			channel->defaults.m_nUDestAddrH = 0;
			channel->defaults.m_nVDestAddrL = buf->dma_v_plane;
			channel->defaults.m_nVDestAddrH = 0;

			/* Write the Y/U/V DMA addresses */
			offset = channel->offsets.m_nYDmaAddrL;
			offset += (buf->nr * 6);

			/* Set the last DMA address pointer for this FIFO */
			vc_write32(channel->regs.m_nLastAddrPtr, offset);
			vc_write32(channel->regs.m_nDestAddrPtr, offset);
			/* ... and remember the above values as defaults when we restart */
			channel->defaults.m_nLastAddrPtr = offset;
			channel->defaults.m_nDestAddrPtr = offset;

			/* Configure the TS buffers */
			offset = channel->offsets.m_nYTSDmaAddrL;
			offset += (buf->nr * 2);
			vc_write32(channel->regs.m_nLastTSAddrPtr, offset);
			vc_write32(channel->regs.m_nDestTSAddrPtr, offset);
			/* ... and remember the above values as defaults when we restart */
			channel->defaults.m_nLastTSAddrPtr = offset;
			channel->defaults.m_nDestTSAddrPtr = offset;
		}
	}
	spin_unlock_irqrestore(&channel->dma_buffers_free_lock, flags);

	/* Indicate this channel is in-active now */
	vc_write32(channel->regs.m_nStreamDisabled, 1);

	return 0;
}

int vc8x0_dma_configure_audio(struct vc8x0_dma_channel *channel)
{
/* The audio DMA frame sizes are based on a maximum of 1920 left and right audio samples per 10ms frame
 * multiplied by the number of stereo pairs the device is capable of delivering.  Due to DMA sequencer issues
 * with spanning 4k address boundries each channel is up sized to 16K and the different transfer segments
 * are aligned on 1K boundries.  This does waste space but it is an absolute requirement.
 */
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_dma_buffer *buf, *tsbuf = 0;
	struct list_head *p, *q, *x, *y;
	u32 nXferSize = 0;
	u32 nXferCount = 0;
	u32 rate = 48000;
	u32 offset, addr;
	unsigned long flags;

	dprintk(1, "%s(nr=%d)\n", __func__, channel->nr);

	channel->m_nDMABuffers = MAX_AUDIO_DSP_FRAMES;
	channel->m_nBlockSize = 64;

//	int nChannelCount = 2;
//	int nChannelCount2 = (nChannelCount < 5) ? nChannelCount : 4;

	switch (rate) {
        case 48000:
		/* Thats 16 32bit left and 16 32bit right samples */
                nXferSize = channel->m_nBlockSize;
                nXferCount = 480 / (channel->m_nBlockSize / 8);
                break;
        case 44100:
		/* Thats 9 32bit left and 9 32bit right samples */
                nXferSize = (7 * 8);
                nXferCount = 441 / 7;
                break;
        case 32000:
		/* Thats 10 32bit left and 10 32bit right samples */
                nXferSize = channel->m_nBlockSize;
                nXferCount = 320 / (channel->m_nBlockSize / 8);
                break;
        case 96000:
		/* Thats 20 32bit left and 20 32bit right samples */
                nXferSize = channel->m_nBlockSize;
                nXferCount = 960 / (channel->m_nBlockSize / 8);
                break;
        case 192000:
		/* Thats 20 32bit left and 20 32bit right samples */
                nXferSize = channel->m_nBlockSize;
                nXferCount = 1920 / (channel->m_nBlockSize / 8);
                break;
        }

	vc_write32(channel->regs.m_nSP1XferStepSize, nXferSize);
	vc_write32(channel->regs.m_nXferCnt, nXferCount);
	vc_write32(channel->regs.m_nDestXferCnt, nXferCount);
	channel->defaults.m_nXferCnt = nXferCount;
	channel->defaults.m_nDestXferCnt = nXferCount;

	/* Set the transfer control values for stereo pair one */
	vc_write32(channel->regs.m_nSP1XferCtrl, channel->defaults.m_nSP1XferCtrl | (nXferSize & 0x00000FFF));

	/* Set the time stamp transfer control values */
        vc_write32(channel->regs.m_nSP1TSXferCtrl, channel->defaults.m_nSP1TSXferCtrl);

	/* Set the transfer control values for stereo pairs two thru four */
	vc_write32(channel->regs.m_nSP2XferCtrl, channel->defaults.m_nSP2XferCtrl | (nXferSize & 0x00000FFF));
	vc_write32(channel->regs.m_nSP3XferCtrl, channel->defaults.m_nSP3XferCtrl | (nXferSize & 0x00000FFF));
	vc_write32(channel->regs.m_nSP4XferCtrl, channel->defaults.m_nSP4XferCtrl | (nXferSize & 0x00000FFF));

	/* Set the time stamp transfer control values for steroe pairs two thru four */
        vc_write32(channel->regs.m_nSP2TSXferCtrl, channel->defaults.m_nSP2TSXferCtrl);
        vc_write32(channel->regs.m_nSP3TSXferCtrl, channel->defaults.m_nSP3TSXferCtrl);
        vc_write32(channel->regs.m_nSP4TSXferCtrl, channel->defaults.m_nSP4TSXferCtrl);

	/* Set the # of DMA frames for this FIFO */ 
	vc_write32(channel->regs.m_nFrameCnt, channel->m_nDMABuffers);
	vc_write32(channel->regs.m_nFrameXferCnt, channel->m_nDMABuffers);
	vc_write32(channel->regs.m_nLastFrameIndex, channel->m_nDMABuffers);
	/* m_arrLastFrameIndexId[nFIFOIndex] = m_arrVideoSequencerVariableValues[nSubDeviceIndex].Video.m_nLastFrameIndex; */
	channel->m_arrLastFrameIndexId = channel->defaults.m_nLastFrameIndex;

	/* We only ever have 1 timestamp buffer in the list, grab it for references purposes */
	/* We'll configure the video channel to refernece various parts of this buffer */
	spin_lock_irqsave(&channel->dma_timestamp_buffers_free_lock, flags);
	list_for_each_safe(x, y, &channel->dma_timestamp_buffers_free) {
		tsbuf = list_entry(x, struct vc8x0_dma_buffer, list);

		/* Configure the buffer plane pointers and sizes */
		tsbuf->dma_y_plane = tsbuf->cpu_dma;
		tsbuf->cpu_y_plane = tsbuf->cpu;
		tsbuf->cpu_y_plane_size = 0;

		break;
	}
	spin_unlock_irqrestore(&channel->dma_timestamp_buffers_free_lock, flags);
	BUG_ON(tsbuf == 0);

	/* Set the dma info for audio stereo pairs 1 thru 4 for this device */
	spin_lock_irqsave(&channel->dma_buffers_free_lock, flags);
	list_for_each_safe(p, q, &channel->dma_buffers_free) {

		buf = list_entry(p, struct vc8x0_dma_buffer, list);
		BUG_ON(buf->nr >= MAX_VIDEO_DSP_FRAMES);

		/* Configure the buffer plane pointers and sizes */
		buf->dma_y_plane = buf->cpu_dma;
		buf->cpu_y_plane = buf->cpu;
		buf->cpu_y_plane_size = 0;
		buf->cpu_u_plane = buf->cpu;
		buf->cpu_u_plane_size = 0;
		buf->cpu_v_plane = buf->cpu;
		buf->cpu_v_plane_size = 0;

		addr = channel->regs.m_nSP1DmaAddrL + (buf->nr * (2 * sizeof(u32)));
		offset = channel->offsets.m_nSP1DmaAddrL + (buf->nr * 2);

		vc_write32(addr + (0 * sizeof(u32)), buf->cpu_dma); /* PCI Physical ADDR 0-31bits */
		vc_write32(addr + (1 * sizeof(u32)), 0); /* Bits 32-63 */

		vc_write32(channel->regs.m_nSP1DestAddrL, buf->cpu_dma);
		vc_write32(channel->regs.m_nSP1DestAddrH, 0);
		channel->defaults.m_nSP1DestAddrL = buf->cpu_dma;
		channel->defaults.m_nSP1DestAddrH = 0;

		vc_write32(channel->regs.m_nLastAddrPtr, offset);
		vc_write32(channel->regs.m_nDestAddrPtr, offset);
		channel->defaults.m_nLastAddrPtr = offset;
		channel->defaults.m_nDestAddrPtr = offset;

		/* TS related */
		addr = channel->regs.m_nSP1TSDmaAddrL + (buf->nr * (2 * sizeof(u32)));
		offset = channel->offsets.m_nSP1TSDmaAddrL + (buf->nr * 2);
		vc_write32(addr + (0 * sizeof(u32)), tsbuf->dma_y_plane + (channel->nr * 1024));
		vc_write32(addr + (1 * sizeof(u32)), 0);

		/* Set the starting dma info for audio stereo pairs for this device */
		vc_write32(channel->regs.m_nSP1TSDestAddrL, tsbuf->dma_y_plane + (channel->nr * 1024) + (buf->nr * 0x20));
		vc_write32(channel->regs.m_nSP1TSDestAddrH, 0);
		channel->defaults.m_nSP1TSDestAddrL = tsbuf->dma_y_plane + (channel->nr * 1024) + (buf->nr * 0x20);
		channel->defaults.m_nSP1TSDestAddrH = 0;

		/* Set the last DMA address pointer for this FIFO */
		vc_write32(channel->regs.m_nLastTSAddrPtr, offset);
		vc_write32(channel->regs.m_nDestTSAddrPtr, offset);
		channel->defaults.m_nLastTSAddrPtr = offset;
		channel->defaults.m_nDestTSAddrPtr = offset;

	}
	spin_unlock_irqrestore(&channel->dma_buffers_free_lock, flags);
	
	/* Indicate this channel is in-active now */
	vc_write32(channel->regs.m_nStreamDisabled, 1);

	return 0;
}

int vc8x0_dma_configure(struct vc8x0_dma_channel *channel)
{
	if (channel->type == TYPE_VIDEO)
		return vc8x0_dma_configure_video(channel);
	if (channel->type == TYPE_AUDIO)
		return vc8x0_dma_configure_audio(channel);

	return vc8x0_dma_configure_video(channel);
}

int vc8x0_dma_reset_video(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;
	struct vc8x0_format *fmt = channel->fmt;
	u32 xfersize, vbi_total_lines = 0;

	if (channel->vbi_enabled)
		vbi_total_lines = fmt->vbi_total_lines;

	/* Stop the channel */
	vc_write32(channel->regs.m_nStreamDisabled, 1); /* Video flex off */
	vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Video fifo Off */
	msleep(250);
	xfersize = (((fmt->width * 2) * (fmt->height + vbi_total_lines)) / channel->m_nBlockSize) / 2;
	dprintk(1, "%s() calculated xfersize = 0x%x\n", __func__, xfersize);
	channel->defaults.m_nXferCnt = xfersize;
	channel->defaults.m_nDestXferCnt = xfersize;

	/* Reset the Flex values */
	vc_write32(channel->regs.m_nXferCnt, channel->defaults.m_nXferCnt);
	vc_write32(channel->regs.m_nDestXferCnt, channel->defaults.m_nXferCnt);
	vc_write32(channel->regs.m_nLastFrameIndex, MAX_VIDEO_DSP_FRAMES);
	vc_write32(channel->regs.m_nFrameXferCnt, MAX_VIDEO_DSP_FRAMES);

	vc_write32(channel->regs.m_nLastAddrPtr, channel->defaults.m_nLastAddrPtr);
	vc_write32(channel->regs.m_nDestAddrPtr, channel->defaults.m_nDestAddrPtr);
	vc_write32(channel->regs.m_nLastTSAddrPtr, channel->defaults.m_nLastTSAddrPtr);
	vc_write32(channel->regs.m_nDestTSAddrPtr, channel->defaults.m_nDestTSAddrPtr);
	vc_write32(channel->regs.m_nYDestAddrL, channel->defaults.m_nYDestAddrL);
	vc_write32(channel->regs.m_nYDestAddrH, channel->defaults.m_nYDestAddrH);
	vc_write32(channel->regs.m_nUDestAddrL, channel->defaults.m_nUDestAddrL);
	vc_write32(channel->regs.m_nUDestAddrH, channel->defaults.m_nUDestAddrH);
	vc_write32(channel->regs.m_nVDestAddrL, channel->defaults.m_nVDestAddrL);
	vc_write32(channel->regs.m_nVDestAddrH, channel->defaults.m_nVDestAddrH);
	vc_write32(channel->regs.m_nYTSDestAddrL, channel->defaults.m_nYTSDestAddrL);
	vc_write32(channel->regs.m_nYTSDestAddrH, channel->defaults.m_nYTSDestAddrH);

	/* Start the channel */
	vc_write32(channel->regs.m_nStreamDisabled, 0); /* Video flex on */
	vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Video fifo on */

	return 0;
}

int vc8x0_dma_reset_audio(struct vc8x0_dma_channel *channel)
{
	struct vc8x0_dev *dev = channel->dev;

	/* Stop the channel */
	vc_write32(channel->regs.m_nStreamDisabled, 1); /* Audio flex off */
	vc_write32(channel->regs.m_nFifoGo, 0x00000000); /* Audio fifo off */
	msleep(250);

	/* Reset the Flex values */
	vc_write32(channel->regs.m_nXferCnt, channel->defaults.m_nXferCnt);
	vc_write32(channel->regs.m_nDestXferCnt, channel->defaults.m_nXferCnt);
	vc_write32(channel->regs.m_nLastFrameIndex, MAX_AUDIO_DSP_FRAMES);
	vc_write32(channel->regs.m_nFrameXferCnt, MAX_AUDIO_DSP_FRAMES);

	vc_write32(channel->regs.m_nLastAddrPtr, channel->defaults.m_nLastAddrPtr);
	vc_write32(channel->regs.m_nDestAddrPtr, channel->defaults.m_nDestAddrPtr);
	vc_write32(channel->regs.m_nLastTSAddrPtr, channel->defaults.m_nLastTSAddrPtr);
	vc_write32(channel->regs.m_nDestTSAddrPtr, channel->defaults.m_nDestTSAddrPtr);
	vc_write32(channel->regs.m_nSP1DestAddrL, channel->defaults.m_nSP1DestAddrL);
	vc_write32(channel->regs.m_nSP1DestAddrH, channel->defaults.m_nSP1DestAddrH);
	vc_write32(channel->regs.m_nSP1TSDestAddrL, channel->defaults.m_nSP1TSDestAddrL);
	vc_write32(channel->regs.m_nSP1TSDestAddrH, channel->defaults.m_nSP1TSDestAddrH);

	/* Start the channel */
	vc_write32(channel->regs.m_nStreamDisabled, 0); /* Audio flex on */
	vc_write32(channel->regs.m_nFifoGo, 0x00000001); /* Audio fifo on */

	return 0;
}

int vc8x0_dma_reset(struct vc8x0_dma_channel *channel)
{
	if (channel->type == TYPE_VIDEO)
		return vc8x0_dma_reset_video(channel);
	if (channel->type == TYPE_AUDIO)
		return vc8x0_dma_reset_audio(channel);

	return -EINVAL;
}

